package pl.slawas.scjdbc4jpa.config;

import java.io.Serializable;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.persistence.FlushModeType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import org.apache.commons.lang.StringUtils;

import pl.slawas.paging.Page;
import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.security.SecurePasswordIdentity;

/**
 * 
 * JdbcConfig - klasa zarządzająca parametrami uruchomienia JDBC kontenera.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
@XmlRootElement(name = "jdbc-settings")
@XmlType(propOrder = { "jndiName", "jnpDisableDiscovery", "providerUrl",
		"urlPkgPrefixes", "initialContextFactory", "dbDialect", "dbUser",
		"dbPassword", "batchSize", "batchCommitSize", "pageResultSize",
		"maxPages", "flushModeType", "declaredEntities" })
public class JdbcConfig implements Serializable {

	private static final long serialVersionUID = 1302128333610279616L;

	/** domyślna lokalizacja pliku z konfiguracją entity manager-a */
	public static String DEFAULT_CONFIG_URL = "/META-INF/scjdbc4jpa-config-jdbc.xml";

	/**
	 * Na razie robie tak, ze trzeba zadeklarowac encje w pliku konfiguracyjnym,
	 * później zastanowimy się nad tym jak wyszukiwać encje.
	 */
	@XmlElementWrapper(name = "entities")
	@XmlElement(name = "entity")
	private Collection<JdbcConfigEntityMetadata> declaredEntities;

	@XmlElement(name = "batch-commit-size")
	private Integer batchCommitSize;

	@XmlElement(name = "batch-size")
	private Integer batchSize;

	@XmlElement(name = "jndi-jnp-disable-discovery")
	private String jnpDisableDiscovery;

	@XmlElement(name = "jndi-provider-url")
	private String providerUrl;

	@XmlElement(name = "jndi-url-pkg-prefixes")
	private String urlPkgPrefixes;

	@XmlElement(name = "jndi-initial-context-factory")
	private String initialContextFactory;

	@XmlElement(name = "ds-jndi-name")
	private String jndiName;

	@XmlElement(name = "page-result-size")
	private Integer pageResultSize = Page.DEFAULT_PAGE_SIZE;

	@XmlElement(name = "max-pages")
	private Integer maxPages = Page.DEFAULT_MAX_PAGES;

	@XmlElement(name = "flush-mode-type")
	private FlushModeType flushModeType = FlushModeType.COMMIT;

	@XmlElement(name = "db-dialect")
	private Dialect dbDialect;

	@XmlElement(name = "db-user")
	private String dbUser;

	@XmlElement(name = "db-password")
	private String dbPassword;

	/**
	 * @return the batchCommitSize
	 */
	public Integer getBatchCommitSize() {
		return batchCommitSize;
	}

	/**
	 * @return the batchSize
	 */
	public Integer getBatchSize() {
		return batchSize;
	}

	/**
	 * @return the jnpDisableDiscovery
	 */
	public String getJnpDisableDiscovery() {
		return jnpDisableDiscovery;
	}

	/**
	 * @return the providerUrl
	 */
	public String getProviderUrl() {
		return providerUrl;
	}

	/**
	 * @return the urlPkgPrefixes
	 */
	public String getUrlPkgPrefixes() {
		return urlPkgPrefixes;
	}

	/**
	 * @return the initialContextFactory
	 */
	public String getInitialContextFactory() {
		return initialContextFactory;
	}

	/**
	 * @return the jndiName
	 */
	public String getJndiName() {
		return jndiName;
	}

	/**
	 * @return the declaredEntities
	 */
	public Collection<JdbcConfigEntityMetadata> getDeclaredEntities() {
		return declaredEntities;
	}

	/**
	 * @return the dbDialect
	 */
	public Dialect getDbDialect() {
		return dbDialect;
	}

	/**
	 * @return the pageResultSize
	 */
	public Integer getPageResultSize() {
		return pageResultSize;
	}

	/**
	 * @return the flushModeType
	 */
	public FlushModeType getFlushModeType() {
		return flushModeType;
	}

	/**
	 * @return the dbUser
	 */
	public String getDbUser() {
		return dbUser;
	}

	/**
	 * @return the dbPassword
	 * @throws IllegalBlockSizeException
	 * @throws BadPaddingException
	 * @throws NoSuchAlgorithmException
	 * @throws NoSuchPaddingException
	 * @throws InvalidKeyException
	 */
	public String getDbPassword() throws InvalidKeyException,
			NoSuchPaddingException, NoSuchAlgorithmException,
			BadPaddingException, IllegalBlockSizeException {

		return StringUtils.isNotBlank(dbPassword) ? SecurePasswordIdentity
				.decode(dbPassword) : dbPassword;
	}

	/**
	 * @return the serialversionuid
	 */
	public static long getSerialversionuid() {
		return serialVersionUID;
	}

	/**
	 * @return the dEFAULT_CONFIG_URL
	 */
	public static String getDEFAULT_CONFIG_URL() {
		return DEFAULT_CONFIG_URL;
	}

	/**
	 * @return the maxPages
	 */
	public Integer getMaxPages() {
		return maxPages;
	}

}
