package pl.slawas.scjdbc4jpa.core;

import java.io.Serializable;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Hashtable;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.FactoryBean;
import org.springframework.beans.factory.InitializingBean;

import pl.slawas.helper.jndi.JndiParameters;
import pl.slawas.helper.jndi.JndiParametersUtils;
import pl.slawas.scjdbc4jpa.security.SecurePasswordIdentity;

/**
 * 
 * _CommonAbstractEntityManagerFactory ogolna abstrakcja
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
@SuppressWarnings("serial")
public abstract class ScEntityManagerFactorySupport implements
		Serializable, ScDefaults, JndiParameters, FactoryBean,
		InitializingBean, DisposableBean {

	final transient protected Logger logger = LoggerFactory
			.getLogger(getClass());

	protected int batchCommitSize = DefaultBatchCommitSize;

	protected int batchSize = -1;

	/**
	 * JNDI Data Source Context Name
	 */
	protected String dsJndiName = DetaultJndiName;

	protected String dialectName = DefaultDialect;

	protected String jnpDisableDiscovery;

	protected String providerUrl;

	protected String urlPkgPrefixes;

	protected String initialContextFactory;

	private String dbUser;

	private boolean passwordIsEncoded = true;

	private String dbPassword;

	/**
	 * Flaga odpowiedzialna za usunięcie konfliktu wywołania dwóch metod:
	 * {@link #loadDataSourceFromJndi()} oraz {@link #setDataSource(DataSource)}
	 * , które to różnymi sposobami próbują ustawić to samo źródło danych.
	 */
	protected boolean settedDataSource = false;

	public void setJnpDisableDiscovery(String jnpDiscovery) {
		this.jnpDisableDiscovery = jnpDiscovery;
	}

	public void setProviderUrl(String providerUrl) {
		this.providerUrl = providerUrl;
	}

	public void setUrlPkgPrefixes(String urlPkgPrefixes) {
		this.urlPkgPrefixes = urlPkgPrefixes;
	}

	public void setInitialContextFactory(String initialContextFactory) {
		this.initialContextFactory = initialContextFactory;
	}

	/**
	 * @return the jnpDisableDiscovery
	 */
	public String getJnpDisableDiscovery() {
		return JndiParametersUtils.getJnpDisableDiscovery(jnpDisableDiscovery);
	}

	/**
	 * @return the providerUrl
	 */
	public String getProviderUrl() {
		return JndiParametersUtils.getProviderUrl(providerUrl);
	}

	/**
	 * @return the urlPkgPrefixes
	 */
	public String getUrlPkgPrefixes() {
		return JndiParametersUtils.getUrlPkgPrefixes(urlPkgPrefixes);
	}

	/**
	 * @return the initialContextFactory
	 */
	public String getInitialContextFactory() {
		return JndiParametersUtils
				.getInitialContextFactory(initialContextFactory);
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public Hashtable getJndiEnv() {
		Hashtable env = new Hashtable();
		env.put(Context.INITIAL_CONTEXT_FACTORY, getInitialContextFactory());
		env.put(Context.PROVIDER_URL, getProviderUrl());
		env.put(JndiParameters.JNP_DISABLEDISCOVERY, getJnpDisableDiscovery());
		env.put(Context.URL_PKG_PREFIXES, getUrlPkgPrefixes());
		return env;
	}

	public String getDialectName() {
		return dialectName;
	}

	public void setDialectName(String dialectName) {
		this.dialectName = dialectName;
	}

	public int getBatchCommitSize() {
		int batchSize = getBatchSize();
		if (this.dialectName.equals(DefaultDialect)) {
			this.batchCommitSize = batchSize;
		} else {
			this.batchCommitSize = (batchCommitSize >= batchSize ? batchCommitSize
					: batchSize);
		}
		return this.batchCommitSize;
	}

	public void setBatchCommitSize(int batchCommitSize) {
		this.batchCommitSize = batchCommitSize;
	}

	public int getBatchSize() {
		if (this.batchSize < 0) {
			this.batchSize = DefaultBatchSize;
		}
		return this.batchSize;
	}

	/**
	 * @param batchSize
	 *            the batchSize to set
	 */
	public void setBatchSize(int batchSize) {
		this.batchSize = batchSize;
	}

	/**
	 * @return the dsJndiName
	 */
	public String getDsJndiName() {
		return dsJndiName;
	}

	/**
	 * @param dsJndiName
	 *            the dsJndiName to set
	 */
	public void setDsJndiName(String dsJndiName) {
		this.dsJndiName = dsJndiName;
	}

	/**
	 * Metoda pobierania zrodla danych z kontekstu JNDI
	 */
	public void loadDataSourceFromJndi() {

		if (!settedDataSource) {

			if (getDsJndiName() == null)
				throw new IllegalArgumentException(
						"Nie udalo sie ustalic zrodla danych. "
								+ "Nazwa JNDI zrodla danych nie moze byc pusta");

			logger.info("Using dsJndiName : " + getDsJndiName());
			DataSource dataSource = null;
			try {
				InitialContext ctx = new InitialContext(this.getJndiEnv());
				dataSource = (DataSource) ctx.lookup(getDsJndiName());
			} catch (NamingException e) {
				dataSource = null;
				logger.error("", e);
			} catch (Exception e) {
				dataSource = null;
				logger.error("", e);
			}
			if (dataSource != null) {
				this.setDataSource(dataSource);
			} else {
				throw new IllegalArgumentException(
						"Nie udalo sie ustalic zrodla danych");
			}
		}
	}

	/**
	 * Metoda ustawiania źródła danych. Metoda dodatkowo ustawia flagę
	 * {@link #settedDataSource}.
	 * 
	 * @param dataSource
	 *            źródlo danych
	 */
	public void setDataSource(DataSource dataSource) {
		if (!settedDataSource) {
			setterOfDataSourceMethodImplementation(dataSource);
			this.settedDataSource = true;
		} else {
			logger.warn("Parameter 'dataSource' is setted by piervous value. Use only one of 'dataSource' and 'dsJndiName' parameters. ");
		}
	}

	public boolean isSingleton() {
		return true;
	}

	public void afterPropertiesSet() throws Exception {
		loadDataSourceFromJndi();
	}

	/**
	 * Metoda właściwa ustawiająca źródło danych wymagana do ustawienia źródła
	 * za pomocą metody {@link #setDataSource(DataSource)}, która dodatkowo
	 * ustawia flagę {@link #settedDataSource}.
	 * 
	 * @param dataSource
	 */
	protected abstract void setterOfDataSourceMethodImplementation(
			DataSource dataSource);

	/**
	 * Pobranie źródła danych
	 * 
	 * @return źródło danych
	 */
	public abstract DataSource getDataSource();

	/**
	 * Pobieranie użytkownika bazy danych, potrzebnego do inicjalizacji
	 * połączenia pobieranego ze źródła danych.
	 * 
	 * @return użytkownik bazy danych
	 */
	public String getDbUser() {
		return dbUser;
	}

	/**
	 * Ustawianie użytkownika bazy danych, potrzebnego do inicjalizacji
	 * połączenia pobieranego ze źródła danych.
	 * 
	 * @param dbUser
	 *            użytkownik bazy danych
	 */
	public void setDbUser(String dbUser) {
		this.dbUser = dbUser;
	}

	/**
	 * Pobierania hasła użytkownika bazy danych, potrzebnego do inicjalizacji
	 * połączenia pobieranego ze źródła danych.
	 * 
	 * @return niezakodowane hasło użytkownika bazy danych
	 * @throws IllegalBlockSizeException
	 * @throws BadPaddingException
	 * @throws NoSuchAlgorithmException
	 * @throws NoSuchPaddingException
	 * @throws InvalidKeyException
	 */
	public String getDbPassword() throws InvalidKeyException,
			NoSuchPaddingException, NoSuchAlgorithmException,
			BadPaddingException, IllegalBlockSizeException {
		return (this.passwordIsEncoded && StringUtils.isNotBlank(dbPassword) ? SecurePasswordIdentity
				.decode(dbPassword) : dbPassword);
	}

	/**
	 * Ustawianie niezakodowanego hasła. Metoda nie zalecana, ponieważ wartość
	 * przechowywana jest w obiekcie, a wiadomo, że można je podejżeć różnymi
	 * metodami.
	 * 
	 * @param dbPassword
	 *            niezakodowane (pisane otwartym tekstem)
	 */
	@Deprecated
	public void setDbPassword(String dbPassword) {
		this.passwordIsEncoded = false;
		this.dbPassword = dbPassword;
	}

	/**
	 * Ustawianie zakodowanego hasła.
	 * 
	 * @see SecurePasswordIdentity
	 * 
	 * @param dbPassword
	 *            zakodowane hasło
	 */
	public void setEncodedDbPassword(String dbPassword) {
		this.passwordIsEncoded = true;
		this.dbPassword = dbPassword;
	}
}
