package pl.slawas.scjdbc4jpa.core;

import java.sql.Savepoint;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.transaction.CannotCreateTransactionException;
import org.springframework.transaction.NestedTransactionNotSupportedException;
import org.springframework.transaction.SavepointManager;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionSystemException;
import org.springframework.transaction.TransactionUsageException;
import org.springframework.transaction.support.SmartTransactionObject;

public abstract class ScTransactionObjectSupport implements SavepointManager,
		SmartTransactionObject {

	final protected Log log = LogFactory.getLog(getClass());

	private ScEntityManagerHolder entityHolder;

	private Integer previousIsolationLevel;

	private boolean savepointAllowed = false;

	public void setEntityHolder(ScEntityManagerHolder entityHolder) {
		this.entityHolder = entityHolder;
	}

	public ScEntityManagerHolder getEntityHolder() {
		return this.entityHolder;
	}

	public boolean hasEntityHolder() {
		return (this.entityHolder != null);
	}

	public void setPreviousIsolationLevel(Integer previousIsolationLevel) {
		this.previousIsolationLevel = previousIsolationLevel;
	}

	public Integer getPreviousIsolationLevel() {
		return this.previousIsolationLevel;
	}

	public void setSavepointAllowed(boolean savepointAllowed) {
		this.savepointAllowed = savepointAllowed;
	}

	public boolean isSavepointAllowed() {
		return this.savepointAllowed;
	}

	// ---------------------------------------------------------------------
	// Implementation of SavepointManager
	// ---------------------------------------------------------------------

	/**
	 * This implementation creates a JDBC 3.0 Savepoint and returns it.
	 * 
	 * @see java.sql.Connection#setSavepoint
	 */
	public Object createSavepoint() throws TransactionException {
		ScEntityManagerHolder entHolder = getEntityHolderForSavepoint();
		try {
			if (!entHolder.supportsSavepoints()) {
				throw new NestedTransactionNotSupportedException(
						"Cannot create a nested transaction because savepoints are not supported by your JDBC driver");
			}
		} catch (Throwable ex) {
			throw new NestedTransactionNotSupportedException(
					"Cannot create a nested transaction because your JDBC driver is not a JDBC 3.0 driver",
					ex);
		}
		try {
			return entHolder.createSavepoint();
		} catch (Throwable ex) {
			throw new CannotCreateTransactionException(
					"Could not create JDBC savepoint", ex);
		}
	}

	/**
	 * This implementation rolls back to the given JDBC 3.0 Savepoint.
	 * 
	 * @see java.sql.Connection#rollback(java.sql.Savepoint)
	 */
	public void rollbackToSavepoint(Object savepoint)
			throws TransactionException {
		try {
			getEntityHolderForSavepoint().getConnection().rollback(
					(Savepoint) savepoint);
		} catch (Throwable ex) {
			throw new TransactionSystemException(
					"Could not roll back to JDBC savepoint", ex);
		}
	}

	/**
	 * This implementation releases the given JDBC 3.0 Savepoint.
	 * 
	 * @see java.sql.Connection#releaseSavepoint
	 */
	public void releaseSavepoint(Object savepoint) throws TransactionException {
		try {
			getEntityHolderForSavepoint().getConnection().releaseSavepoint(
					(Savepoint) savepoint);
		} catch (Throwable ex) {
			log.debug("Could not explicitly release JDBC savepoint", ex);
		}
	}

	protected ScEntityManagerHolder getEntityHolderForSavepoint()
			throws TransactionException {
		if (!isSavepointAllowed()) {
			throw new NestedTransactionNotSupportedException(
					"Transaction manager does not allow nested transactions");
		}
		if (!hasEntityHolder()) {
			throw new TransactionUsageException(
					"Cannot create nested transaction if not exposing a JDBC transaction");
		}
		return getEntityHolder();
	}

}
