package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Vector;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import javax.persistence.FlushModeType;
import javax.persistence.LockModeType;
import javax.persistence.Query;

import org.apache.commons.lang.StringUtils;
import org.hibernate.cfg.Configuration;
import org.hibernate.engine.Mapping;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.entities.IDUtils;
import pl.slawas.entities.IDUtilsErrorException;
import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.paging.PagingParamsReadOnly;
import pl.slawas.paging.ResultMessage;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.core.ScEntityManager;
import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntityCacheUtils;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntityFirstLevelCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntityFirstLevelCacheManager.FirstLevelCacheObject;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntitySecondLevelCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheFirstLevelManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.dml.DmlQuerySupport;
import pl.slawas.scjdbc4jpa.jdbc.dml.LockRow;
import pl.slawas.scjdbc4jpa.jdbc.dml._IDmlQuery;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.CreateSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.EntityNotJetSavedIdDbException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ExecuteSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.IdIsNotEmbeddableException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityManagerException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectIsNotEntityException;
import pl.slawas.scjdbc4jpa.jdbc.generators.ValueGenerator;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityParserLevel;
import pl.slawas.scjdbc4jpa.jdbc.where.BindParams;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;
import pl.slawas.scjdbc4jpa.jpa.mapping.Table;

/**
 * 
 * JDBCEntityManager - obiekt entity manager-a
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.2 $
 * 
 */
public class JDBCEntityManager implements ScEntityManager,
		_IFisrtLevelCacheManager, Serializable {

	private static final long serialVersionUID = 6049437632156464706L;

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCEntityManager.class);

	/* OBIEKTY PRYWATNE wykonawcze - start */
	/** fabryka, ktora utworzyla entity manager-a */
	private final JDBCEntityManagerFactory emf;

	/** połączenie do bazy danych */
	private final Connection conn;

	/** lista polecen DML, gromadzonych az do metody {@link #flush()} */
	private final Vector<DmlOperation> dmlCommands = new Vector<DmlOperation>();

	/** obiekt transakcji */
	private JDBCTransaction transaction;

	/** obiekt kesza pierwszego poziomu */
	private final EntityFirstLevelCacheManager firstLevelCache;

	private final QueryCacheFirstLevelManager localQueryCacheManager;

	/* OBIEKTY PRYWATNE wykonawcze - koniec */

	/* PARAMETRY - start */
	/** poziom parsowania encji */
	private JDBCEntityParserLevel level = JDBCEntityParserLevel.THIRD;

	/** mod uruchomienia operacji {@link #flush()} */
	private FlushModeType flushModeType = FlushModeType.COMMIT;

	/**
	 * rozmiar paczki operacji DML po której musi nastąpić commit, na razie brak
	 * zastosoawnia, może trzeba sie zastanowic nad savepointami w metodzie
	 * {@link #flush()}?
	 */
	private Integer batchCommitSize = JDBCEntityManagerFactory.DefaultBatchCommitSize;

	/**
	 * rozmiar paczki przesylanych zbiorow operacji DML
	 * 
	 * @see #flush()
	 */
	private Integer batchSize = JDBCEntityManagerFactory.DefaultBatchSize;

	private final PagingParams defaultPagingParams;

	private final String name;

	/* PARAMETRY - koniec */

	JDBCEntityManager(PagingParams pagingParams, String dbUser,
			String dbPassword, JDBCEntityManagerFactory emf,
			JDBCEntityParserLevel level) {
		String user = emf.getDbUser();
		String password = null;
		try {
			password = emf.getDbPassword();
		} catch (InvalidKeyException e) {
			throw new JDBCEntityManagerException(e);
		} catch (NoSuchPaddingException e) {
			throw new JDBCEntityManagerException(e);
		} catch (NoSuchAlgorithmException e) {
			throw new JDBCEntityManagerException(e);
		} catch (BadPaddingException e) {
			throw new JDBCEntityManagerException(e);
		} catch (IllegalBlockSizeException e) {
			throw new JDBCEntityManagerException(e);
		}
		if (StringUtils.isNotBlank(dbUser)
				&& StringUtils.isNotBlank(dbPassword)) {
			user = dbUser;
			password = dbPassword;
		}
		if (StringUtils.isNotBlank(user) && StringUtils.isNotBlank(password)) {
			try {
				this.conn = emf.getDataSource().getConnection(user, password);
				this.conn.setAutoCommit(false);
			} catch (SQLException e) {
				throw new JDBCEntityManagerException(e);
			}
		} else {
			try {
				this.conn = emf.getDataSource().getConnection();
				this.conn.setAutoCommit(false);
			} catch (SQLException e) {
				throw new JDBCEntityManagerException(e);
			}
		}
		long seqValue = emf.getEntityManagerSequenceNextValue();
		this.name = "JDBCEM.FirstLevelCache." + Long.toString(seqValue);
		this.firstLevelCache = new EntityFirstLevelCacheManager(this.name);
		String localQueryCacheRegionName = "JDBCQR.FirstLevelCache."
				+ Long.toString(seqValue);
		this.localQueryCacheManager = new QueryCacheFirstLevelManager(
				localQueryCacheRegionName);
		this.emf = emf;
		this.transaction = new JDBCTransaction(this);
		this.batchCommitSize = emf.getBatchCommitSize();
		this.batchSize = emf.getBatchSize();
		this.level = level;
		if (pagingParams == null) {
			pagingParams = LoadJdbcConfiguration.getPagingParams(null);
		}
		pagingParams.setPage(pagingParams.getFirstPage());
		this.defaultPagingParams = new PagingParamsReadOnly(pagingParams);
	}

	public JDBCEntityManager(PagingParams pagingParams,
			JDBCEntityManagerFactory emf) {
		this(pagingParams, null, null, emf, JDBCEntityParserLevel.THIRD);
	}

	public JDBCEntityManager(PagingParams pagingParams,
			JDBCEntityManagerFactory emf, JDBCEntityParserLevel level) {
		this(pagingParams, null, null, emf, level);
	}

	/**
	 * Sprawdzenie czy entity manager jest otwarty. Jeżeli nie zrzucony jet
	 * wyjątek informujący o tym. (nie może być prywatna ze wzgledu na obiekt
	 * transakcji)
	 * 
	 * @throws JDBCEntityManagerException
	 *             błąd, gdy manager nie jest otwarty
	 */
	void checkEntityManagerIsOpen() throws JDBCEntityManagerException {
		if (!this.isOpen()) {
			throw new JDBCEntityManagerException(
					"Manager encji nie jest otwarty.");
		}
	}

	/**
	 * Sprawdzenie czy entity manager jest otwarty, a jeżeli tak to czy ma
	 * aktywną transackcję.
	 * 
	 * @throws JDBCEntityManagerException
	 *             błąd, gdy manager nie jest otwarty, albo gdy transakcja nie
	 *             jest aktywna.
	 */
	void checkTransactionIsActive() throws JDBCEntityManagerException {
		checkEntityManagerIsOpen();
		if (!getTransaction().isActive()) {
			throw new JDBCEntityManagerException(
					"Transakcja nie jest aktywna. Nie mozna dokonywac modyfikacji danych.");
		}
	}

	/**
	 * zwraca wartość {@link #batchCommitSize}
	 */
	public int getBatchCommitSize() {
		return this.batchCommitSize;
	}

	/**
	 * zwraca wartość {@link #batchSize}
	 */
	public int getBatchSize() {
		return this.batchSize;
	}

	/**
	 * @param batchCommitSize
	 *            the {@link #batchCommitSize} to set
	 */
	void setBatchCommitSize(Integer batchCommitSize) {
		this.batchCommitSize = batchCommitSize;
	}

	/**
	 * @param batchSize
	 *            the {@link #batchSize} to set
	 */
	void setBatchSize(Integer batchSize) {
		this.batchSize = batchSize;
	}

	/**
	 * @return the firstPageDefinition defincja pierwszej strony
	 */
	public Page getFirstPageDefinition() {
		return this.defaultPagingParams.getFirstPage();
	}

	/**
	 * @return the level
	 */
	public JDBCEntityParserLevel getLevel() {
		return level;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	@Deprecated
	public Configuration getConfiguration() {
		return emf.getConfiguration();
	}

	public Connection getConnection() throws SQLException {
		checkEntityManagerIsOpen();

		return this.conn;
	}

	public EntityManager getJpaEntityManager() {
		checkEntityManagerIsOpen();

		return this;
	}

	@Deprecated
	public Mapping getMapping() {
		return emf.getMapping();
	}

	public boolean isAllowedConfiguration() {
		return false;
	}

	public boolean isAllowedMapping() {
		return false;
	}

	public void release() {
		this.close();
	}

	public void setTable(Table table) {
		throw new UnsupportedOperationException(
				"Entity manager nie wspiera obslugi obiektow pl.slawas.scjdbc4jpa.jpa.mapping.Table.");
	}

	public void clear() {
		if (this.firstLevelCache.cacheIsAllowed())
			try {
				this.firstLevelCache.getCache().clear();
			} catch (ObjectCacheErrorException ignore) {
			}
		dmlCommands.clear();
	}

	public void close() {
		this.clear();
		try {
			if (this.isOpen()) {
				if (transaction != null && transaction.isActive()) {
					transaction.rollback();
					transaction = null;
				}
				this.conn.close();
			}

			this.firstLevelCache.close();
			this.localQueryCacheManager.close();
		} catch (SQLException ignore) {
		}
	}

	public boolean contains(Object entity) {

		Object primaryKey = null;
		try {
			primaryKey = IDUtils.getObjectId(entity);
		} catch (IDUtilsErrorException ignore) {
		}
		if (primaryKey != null) {
			EntityObjectWithEntityId entityInfo = getFromFirstLevelCache(
					entity.getClass(), primaryKey);
			return (entityInfo.getCachedObject() != null);
		}
		return false;
	}

	/**
	 * Utworzenie klauzuli zapytania, poprzez analizę paramerów warunków
	 * zapytania oraz warunków sortowania.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param conditions
	 *            warunki zapytania w postaci macierzy
	 * @param sort
	 *            warunki sortowania
	 * @return przygotowanie zapytanie
	 */
	public JDBCQuery createQuery(Class<?> entityClass,
			ConditionClauseSupport[] conditions, SortClause sort) {
		JDBCQuery query = createQuery(entityClass);
		if (sort != null) {
			query.setSortClause(sort);
		}
		if (conditions != null && conditions.length != 0) {
			for (ConditionClauseSupport con : conditions) {
				query.addCondition(con);
			}
		}
		return query;
	}

	/**
	 * Utworzenie klauzuli zapytania, poprzez analizę paramerów warunków
	 * zapytania oraz warunków sortowania.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param conditions
	 *            warunki zapytania w postaci wektora
	 * @param sort
	 *            warunki sortowania
	 * @return przygotowanie zapytanie
	 */
	public JDBCQuery createQuery(Class<?> entityClass,
			Vector<ConditionClauseSupport> conditions, SortClause sort) {
		JDBCQuery query = createQuery(entityClass);
		if (sort != null) {
			query.setSortClause(sort);
		}
		if (conditions != null && !conditions.isEmpty()) {
			for (ConditionClauseSupport con : conditions) {
				query.addCondition(con);
			}
		}
		return query;
	}

	/**
	 * Pobieranie metadanych danej encji.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @return metadane encji
	 */
	public JDBCEntityMetaData getMetaData(Class<?> entityClass) {
		String entityClassName = entityClass.getName();
		JDBCEntityMetaData metaData = this.emf.getMetaData(entityClassName);
		if (metaData == null) {
			throw new JDBCEntityManagerException("Klasa " + entityClassName
					+ " nie jest zarejestrowana klasa encji.");
		}
		return metaData;
	}

	/**
	 * Pobieranie opisu tabeli dla encji
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @return opis tabeli danej encji
	 */
	public JDBCTableDesc getTableDefinition(Class<?> entityClass) {
		return getMetaData(entityClass).getTableDefinition(level);
	}

	/**
	 * Utworzenie podstwowej klauzuli zapytania, bez warunków wyszukiwania oraz
	 * warunków sortowania.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @return przygotowanie zapytanie
	 */
	public JDBCQuery createQuery(Class<?> entityClass) {
		checkEntityManagerIsOpen();

		try {
			JDBCEntityMetaData metaData = getMetaData(entityClass);
			JDBCTableDesc table = metaData.getTableDefinition(level);
			JDBCQuery findAllQuery = metaData.getFindAllQuery(level);
			QueryCacheManager queryCacheManager = (this.transaction.isActive() ? this.localQueryCacheManager
					: metaData.getOverQueryCacheManager());
			return (new JDBCQuery(table, this.defaultPagingParams.copy(),
					queryCacheManager, findAllQuery.getQueryClause()
							.getBaseClause())).setEm(this);
		} catch (ObjectIsNotEntityException e) {
			throw new JDBCEntityManagerException(e);
		} catch (JDBCEntityParserException e) {
			throw new JDBCEntityManagerException(e);
		} catch (IdIsNotEmbeddableException e) {
			throw new JDBCEntityManagerException(e);
		} catch (CreateSQLQueryException e) {
			throw new JDBCEntityManagerException(e);
		}
	}

	/**
	 * Pobieranie informacji o encji z kesz-a
	 * 
	 * @param <T>
	 * @param entityClass
	 * @param primaryKey
	 * @return pomocniczy obiekt z informacjami o encji
	 */
	@SuppressWarnings("unchecked")
	<T> EntityObjectWithEntityId getFromFirstLevelCache(Class<T> entityClass,
			Object primaryKey) {

		EntityObjectWithEntityId obj;
		T entity = null;
		String entityId = null;

		String entityClassName = entityClass.getName();
		EntityFirstLevelCacheManager.FirstLevelCacheObject cachedObject = null;
		if (this.firstLevelCache.cacheIsAllowed()) {
			entityId = EntityCacheUtils
					.getEntityId(primaryKey, entityClassName);
			cachedObject = this.firstLevelCache
					.getFirstLevelCacheObject(entityId);
			try {
				if (EntityCacheUtils.cachedObjectCanBeUsed(cachedObject)) {
					if (logger.isTraceEnabled()) {
						logger.trace("[{}] Pobieram obiekt z kesz-a",
								new Object[] { getName(), entityId });
					}
				} else if (cachedObject != null
						&& cachedObject.getObjectCacheStatus().equals(
								CacheUsage.TO_REMOVE)) {
					this.firstLevelCache.removeFirstLevelCacheObject(entityId,
							false);
					cachedObject = null;
				}
			} catch (EntityNotJetSavedIdDbException e) {
				if (logger.isWarnEnabled()) {
					String message = "Pobieram nie zatwierdzony w bazie danych obiekt";
					if (!cachedObject.isFlushedAfterDML()) {
						message = "Pobieram nie wyslany do bazy danych obiekt";
					}
					logger.warn("[{}] {} z kesz-a", new Object[] { getName(),
							message });
				}
			}
		}
		if (cachedObject != null) {
			entity = (T) cachedObject.getEntity();
		}
		obj = new EntityObjectWithEntityId(entity, entityId);
		obj.setCachedObject(cachedObject);
		return obj;
	}

	@SuppressWarnings("unchecked")
	public <T> T find(Class<T> entityClass, Object primaryKey) {
		checkEntityManagerIsOpen();

		EntityObjectWithEntityId entityInfo = getFromFirstLevelCache(
				entityClass, primaryKey);
		if (entityInfo.getEntityObject() != null
				&& !entityInfo.getCachedObject().isFlushedAfterDML()) {
			return (T) entityInfo.getEntityObject();
		}
		/**
		 * jeżeli tu jestem to znaczy, że obiekt nie jest dostępny w cache, albo
		 * wymaga odświeżenia
		 */
		return loadResultForPrimaryKeyFromDb(entityClass, primaryKey,
				entityInfo.getCachedObject(), entityInfo.getEntityId());
	}

	/**
	 * Metoda wspierająca odczytanie wartości encji bazy danych i ewentualnie
	 * umieszczenie go w keszu pierwszego poziomu.
	 * 
	 * @param <T>
	 * @param entityClass
	 *            klasa encji
	 * @param primaryKey
	 *            obiekt klucza głownego encji
	 * @param cachedObject
	 *            obiekt przechowywany w keszu
	 * @param entityId
	 *            identyfikator obiektu przechowywanego w keszu
	 *            {@link #getEntityId(Object, String)})
	 * @return gotowa do uzycia encja
	 */
	@SuppressWarnings("unchecked")
	private <T> T loadResultForPrimaryKeyFromDb(Class<T> entityClass,
			Object primaryKey, FirstLevelCacheObject cachedObject,
			String entityId) {

		JDBCEntityMetaData metaData = getMetaData(entityClass);
		JDBCQuery query = metaData.getFindByIdQuery(level);
		JDBCTableDesc table = metaData.getTableDefinition(level);
		BindParams bindParams = WhereClause.createBindParams(1,
				table.getPrimaryKey(), primaryKey);
		SelectQuery sQuery = query.getQuery();
		sQuery.getWhereClause().setBindParams(bindParams.getBindParams());
		QueryCacheManager queryCacheManager = (this.transaction.isActive() ? this.localQueryCacheManager
				: metaData.getUniqueQueryCacheManager());
		query.setQueryCacheManager(queryCacheManager);
		query.setExecuted(false);
		PagedResult qR = query.setEm(this).execute();
		if (logger.isTraceEnabled()) {
			logger.trace("[{}] loadResultForPrimaryKeyFromDb(): Komunikat {}",
					new Object[] { getName(), qR.getMessage() });
		}
		if (!qR.getMessage().equals(ResultMessage.NO_DATA_FOUND)) {
			ResultRow row = qR.getUniqueResult();
			if (row.isEntityAvailable()) {
				try {
					T newEntityValue = (T) row.getEntity();
					if (this.firstLevelCache.cacheIsAllowed()
							&& entityId != null) {
						this.firstLevelCache
								.refreshEntityInFirstLevelCache(
										getName(),
										cachedObject,
										entityId,
										qR.getQueryInfo(),
										metaData,
										newEntityValue,
										(this.transaction.isActive() ? CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE
												: CacheUsage.TO_USE));
					}
					return newEntityValue;
				} catch (InstantiationException e) {
					throw new JDBCEntityManagerException(e);
				} catch (IllegalAccessException e) {
					throw new JDBCEntityManagerException(e);
				} catch (NoSuchMethodException e) {
					throw new JDBCEntityManagerException(e);
				} catch (InvocationTargetException e) {
					throw new JDBCEntityManagerException(e);
				} catch (SQLException e) {
					throw new JDBCEntityManagerException(e);
				}
			} else {
				throw new JDBCEntityManagerException(
						"Zla definicja zapytania. Obiekt encji nie jest dostępny");
			}
		}
		return null;
	}

	/**
	 * Pobieranie wierszy rezydujących w bazie danych na określonej sronie.
	 * Wynik zapytania <b>NIE</b> jest keszowany oraz <b>NIE</b> podlega
	 * restrykcjom maksymalnej liczby stron wyników (zobacz
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}).
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param page
	 *            strona wynikow
	 * @param conditions
	 *            warunki zapytania
	 * @param sort
	 *            klauzula sortowania
	 * @return wynik zapytania
	 */
	public PagedResult dump(Class<?> entityClass, Page page,
			ConditionClauseSupport[] conditions, SortClause sort) {
		checkEntityManagerIsOpen();

		JDBCQuery query = null;
		query = createQuery(entityClass, conditions, sort);
		query.setPage((page == null ? this.defaultPagingParams.getFirstPage()
				: page));
		PagedResult pagedResult = query.setEm(this).createDump();
		return pagedResult;
	}

	/**
	 * Pobieranie wierszy rezydujących w bazie danych na określonej sronie.
	 * Wynik zapytania jest keszowany (w zależności od ustawień parametru
	 * {@link #cacheIsAllowed()}) oraz podlega restrykcjom maksymalnej liczby
	 * stron wyników (zobacz {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}).
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param page
	 *            strona wynikow
	 * @param conditions
	 *            warunki zapytania
	 * @param sort
	 *            klauzula sortowania
	 * @return wynik zapytania
	 */
	public PagedResult find(Class<?> entityClass, Page page,
			ConditionClauseSupport[] conditions, SortClause sort) {
		checkEntityManagerIsOpen();

		JDBCQuery query = null;
		query = createQuery(entityClass, conditions, sort);
		query.setPage((page == null ? this.defaultPagingParams.getFirstPage()
				: page));
		PagedResult pagedResult = query.setEm(this).execute();
		return pagedResult;
	}

	public void flush() {
		checkTransactionIsActive();
		int updateCounts = 0;
		logger.trace("[{}] flush()", getName());
		boolean clearLocalQueryCacheManager = false;
		for (DmlOperation command : this.dmlCommands) {
			try {
				_IDmlQuery query = command.query;
				logger.trace(
						"[{}] Wysylam {} dla {} encji {}...",
						new Object[] {
								getName(),
								query.getType(),
								command.params.size(),
								command.metaData.getTableDefinition(
										JDBCEntityParserLevel.FIRST)
										.getEntityName() });
				query.prepareStatement(conn);
				if (command.params.size() > 1) {
					boolean isOracleStatement = query.isOracleStatement();
					if (isOracleStatement) {
						query.setBatchSize(this.batchSize);
					}
					int i = 0;
					for (Object param : command.params) {
						query.addBatch(param);
						if (!isOracleStatement && i > this.batchSize) {
							i = 0;
							updateCounts = updateCounts + query.executeBatch();
						}
						i++;
					}
					updateCounts = updateCounts + query.executeBatch();
					QueryCacheManager cacheToClear;
					switch (query.getType()) {
					case INSERT:
						cacheToClear = command.metaData
								.getOverQueryCacheManager();
						if (cacheToClear.cacheIsAllowed()) {
							this.transaction.addCacheToClear(cacheToClear);
						}
						break;
					default:
						cacheToClear = command.metaData
								.getOverQueryCacheManager();
						if (cacheToClear.cacheIsAllowed()) {
							this.transaction.addCacheToClear(cacheToClear);
						}
						cacheToClear = command.metaData
								.getUniqueQueryCacheManager();
						if (cacheToClear.cacheIsAllowed()) {
							this.transaction.addCacheToClear(cacheToClear);
						}
						break;
					}
					clearLocalQueryCacheManager = true;
					cacheToClear = this.localQueryCacheManager;
					if (cacheToClear.cacheIsAllowed()) {
						this.transaction.addCacheToClear(cacheToClear);
					}
				} else {
					EntityObjectWithEntityId entityObject = command.params
							.get(0);
					updateCounts = updateCounts
							+ query.executeUpdate(entityObject
									.getEntityObject());
					switch (query.getType()) {
					case INSERT:
						QueryCacheManager cacheToClear = command.metaData
								.getOverQueryCacheManager();
						if (cacheToClear.cacheIsAllowed()) {
							this.transaction.addCacheToClear(cacheToClear);
						}
						clearLocalQueryCacheManager = true;
					case UPDATE:
						if (StringUtils.isNotBlank(entityObject.getEntityId())) {
							this.transaction
									.addChangedObject(entityObject.entityId);
							if (entityObject.getCachedObject() != null)
								entityObject.getCachedObject()
										.setIsFlushedAfterDML();
						}
						break;
					case DELETE:
						if (StringUtils.isNotBlank(entityObject.getEntityId())
								&& entityObject.getCachedObject() != null) {
							EntitySecondLevelCacheManager sc = entityObject
									.getCachedObject()
									.getSecondLevelCacheManager();
							this.transaction.addRemovedObject(
									entityObject.entityId, sc);
							this.firstLevelCache.removeFirstLevelCacheObject(
									entityObject.getEntityId(), false);
						}
						break;
					default:
						break;
					}
				}
			} catch (ExecuteSQLQueryException e) {
				this.transaction.setRollbackOnly();
				throw new JDBCEntityManagerException(e);
			} catch (SQLException e) {
				this.transaction.setRollbackOnly();
				throw new JDBCEntityManagerException(e);
			}
		}

		if (clearLocalQueryCacheManager
				&& this.localQueryCacheManager.cacheIsAllowed()) {
			logger.trace(
					"[{}] Czyszcze lokalny kesz zapytan bo bylo wstawianie...",
					new Object[] { getName(), updateCounts });
			this.localQueryCacheManager.clear();
		}
		logger.trace("[{}] Zmienilem {} wierszy", new Object[] { getName(),
				updateCounts });
		this.transaction.sumUpdateCounts(updateCounts);
		this.dmlCommands.clear();
	}

	public FlushModeType getFlushMode() {
		return this.flushModeType;
	}

	@SuppressWarnings("unchecked")
	public <T> T getReference(Class<T> entityClass, Object primaryKey) {
		checkEntityManagerIsOpen();

		if (this.firstLevelCache.cacheIsAllowed()) {
			String entityId = EntityCacheUtils.getEntityId(primaryKey,
					entityClass.getName());
			FirstLevelCacheObject cachedObject = this.firstLevelCache
					.getFirstLevelCacheObject(entityId);
			if (EntityCacheUtils.cachedObjectCanBeUsed(cachedObject)) {
				return (T) cachedObject.getEntity();
			} else if (cachedObject != null
					&& cachedObject.getObjectCacheStatus().equals(
							CacheUsage.TO_REMOVE)) {
				this.firstLevelCache.removeFirstLevelCacheObject(entityId,
						false);
				cachedObject = null;
			}
		}
		return null;
	}

	public EntityTransaction getTransaction() {
		checkEntityManagerIsOpen();

		return this.transaction;
	}

	public boolean isOpen() {
		try {
			return (!this.conn.isClosed());
		} catch (SQLException e) {
			return false;
		}
	}

	public void joinTransaction() {
		throw new UnsupportedOperationException(
				"Entity manager nie wspiera podlaczenia sie do transakcji JTA.");
	}

	public <T> T merge(T entity) {
		checkTransactionIsActive();

		JDBCEntityMetaData metaData = getMetaData(entity.getClass());

		EntityObjectWithEntityId entityObjectWithEntityId = createEntityObjectWithEntityId(entity);
		FirstLevelCacheObject firstLevelCacheObject = null;
		if (entityObjectWithEntityId.getEntityId() != null) {
			firstLevelCacheObject = this.firstLevelCache
					.getFirstLevelCacheObject(entityObjectWithEntityId
							.getEntityId());
			if (firstLevelCacheObject != null) {
				this.firstLevelCache.removeFirstLevelCacheObject(
						entityObjectWithEntityId.getEntityId(), false);
			}
			firstLevelCacheObject = this.firstLevelCache
					.addEntity2FirstLevelCache(getName(),
							entityObjectWithEntityId.getEntityId(), metaData,
							entity, CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE);
			firstLevelCacheObject.setIsNotFlushedAfterDML();
		}
		entityObjectWithEntityId.setCachedObject(firstLevelCacheObject);

		DmlQuerySupport query = metaData.getUpdateClause();
		this.dmlCommands.add(new DmlOperation(metaData, query,
				entityObjectWithEntityId));
		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}

		return entity;
	}

	public void persist(Object entity) {
		checkTransactionIsActive();

		JDBCEntityMetaData metaData = getMetaData(entity.getClass());
		/* obsluga generacji klucza głownego - START */
		try {
			Object primaryKey = IDUtils.getObjectId(entity);
			if (null == primaryKey
					&& ValueGenerator.valueGeneratorsAreAvalaible(getDialect())) {
				ValueGenerator vg = metaData.getValueGenerator();
				if (vg.primaryKeyCanBeGenerated()) {
					primaryKey = vg.getNext();
				}
				if (primaryKey != null) {
					IDUtils.setObjectId(entity, primaryKey);
				}
			}
		} catch (IDUtilsErrorException warn) {
			if (logger.isTraceEnabled())
				logger.warn(
						"Blad podczas obslugi wartosci klucza glownego dla encji {}: {}",
						new Object[] { entity.getClass().getName(),
								warn.getMessage() });
		}
		/* obsluga generacji klucza głownego - KONIEC */

		EntityObjectWithEntityId entityObjectWithEntityId = createEntityObjectWithEntityId(entity);
		FirstLevelCacheObject firstLevelCacheObject = null;
		if (entityObjectWithEntityId.getEntityId() != null) {
			firstLevelCacheObject = this.firstLevelCache
					.addEntity2FirstLevelCache(getName(),
							entityObjectWithEntityId.getEntityId(), metaData,
							entity, CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE);
			firstLevelCacheObject.setIsNotFlushedAfterDML();
		}
		entityObjectWithEntityId.setCachedObject(firstLevelCacheObject);

		DmlQuerySupport query = metaData.getInsertClause();
		this.dmlCommands.add(new DmlOperation(metaData, query,
				entityObjectWithEntityId));
		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}
	}

	public void remove(Object entity) {
		checkTransactionIsActive();

		JDBCEntityMetaData metaData = getMetaData(entity.getClass());

		EntityObjectWithEntityId entityObjectWithEntityId = createEntityObjectWithEntityId(entity);
		FirstLevelCacheObject firstLevelCacheObject = null;
		if (entityObjectWithEntityId.getEntityId() != null) {
			firstLevelCacheObject = this.firstLevelCache
					.getFirstLevelCacheObject(entityObjectWithEntityId
							.getEntityId());
			if (firstLevelCacheObject == null) {
				firstLevelCacheObject = this.firstLevelCache
						.addEntity2FirstLevelCache(getName(),
								entityObjectWithEntityId.getEntityId(),
								metaData, entity,
								CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE);
			}
			firstLevelCacheObject.setIsNotFlushedAfterDML();
		}
		entityObjectWithEntityId.setCachedObject(firstLevelCacheObject);

		DmlQuerySupport query = metaData.getDeleteClause();
		this.dmlCommands.add(new DmlOperation(metaData, query,
				entityObjectWithEntityId));

		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}
	}

	private EntityObjectWithEntityId createEntityObjectWithEntityId(
			Object entity) {
		String entityId = null;
		if (this.firstLevelCache.cacheIsAllowed()) {
			try {
				Object primaryKey = null;
				primaryKey = IDUtils.getObjectId(entity);
				if (primaryKey != null) {
					entityId = EntityCacheUtils.getEntityId(primaryKey, entity
							.getClass().getName());
				}
			} catch (IDUtilsErrorException ignore) {
			}
		}
		return new EntityObjectWithEntityId(entity, entityId);
	}

	public void lock(Object entity, LockModeType lockMode) {
		checkEntityManagerIsOpen();
		if (lockMode != null) {
			// TODO Uzupełnić o implementację argumentu LockModeType
			logger.warn(
					"Obiekt w tej implementacji nie wspiera argumentu LockModeType ({}).",
					lockMode);
		}
		if (getTransaction().isActive()) {
			JDBCEntityMetaData metaData = getMetaData(entity.getClass());
			LockRow lock = metaData.getLockClause();
			try {
				lock.prepareStatement(conn);
				lock.executeUpdate(entity);
			} catch (SQLException e) {
				throw new JDBCEntityManagerException(e);
			} catch (ExecuteSQLQueryException e) {
				throw new JDBCEntityManagerException(e);
			}
		} else {
			throw new UnsupportedOperationException(
					"Nie mozna blokowac do aktualizacji wiersza encji gdy "
							+ "nie zostala uaktywniona transakcja");
		}

	}

	public void refresh(Object entity) {
		checkEntityManagerIsOpen();

		String entityClassName = entity.getClass().getName();
		Object primaryKey;
		try {
			primaryKey = IDUtils.getObjectId(entity);
			FirstLevelCacheObject cachedObject = null;
			String entityId = null;
			if (this.firstLevelCache.cacheIsAllowed()) {
				entityId = EntityCacheUtils.getEntityId(primaryKey,
						entityClassName);
				cachedObject = this.firstLevelCache
						.getFirstLevelCacheObject(entityId);
			}
			loadResultForPrimaryKeyFromDb(entity.getClass(), primaryKey,
					cachedObject, entityId);
		} catch (IDUtilsErrorException e1) {
			throw new JDBCEntityManagerException(
					"Blad pobrania wartosci primary key.", e1);
		}
	}

	/**
	 * Metoda wspierająca masowe zmiany danych w encjach.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font> Metoda aktualizuje tylko informacje o
	 * encjach, o których sama wie na podstawie podręcznego kesza, i tylko dla
	 * nich aktualizuje stan kesza zapytań.
	 * </p>
	 * 
	 * @param <T>
	 * @param entities
	 * @return
	 */
	public <T> int merge(List<T> entities) {
		checkTransactionIsActive();

		if (entities == null || entities.isEmpty())
			throw new IllegalArgumentException(
					"Lista aktualizowanych encji nie moze byc pusta");

		T firstEntity = entities.get(0);
		JDBCEntityMetaData metaData = getMetaData(firstEntity.getClass());
		DmlQuerySupport query = metaData.getUpdateClause();
		DmlOperation operation = new DmlOperation(metaData, query);
		for (T entity : entities) {
			operation.addParam(new EntityObjectWithEntityId(entity, null));
		}
		this.dmlCommands.add(operation);
		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}
		// TODO dokończyc implementacje metody dla masowych aktualizacji encji
		return 0;
	}

	/**
	 * Metoda wspierająca masowe dodawanie danych.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font> Metoda nie aktualizuje żadnych
	 * informacji w o encjach, nie czyści kesza zapytań.
	 * </p>
	 * 
	 * @param <T>
	 * @param entities
	 * @return
	 */
	public <T> int persist(List<T> entities) {
		checkTransactionIsActive();

		if (entities == null || entities.isEmpty())
			throw new IllegalArgumentException(
					"Lista wstawianych encji nie moze byc pusta");

		T firstEntity = entities.get(0);
		JDBCEntityMetaData metaData = getMetaData(firstEntity.getClass());
		DmlQuerySupport query = metaData.getInsertClause();
		DmlOperation operation = new DmlOperation(metaData, query);
		for (T entity : entities) {
			operation.addParam(new EntityObjectWithEntityId(entity, null));
		}
		this.dmlCommands.add(operation);
		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}
		// TODO dokończyc implementacje metody dla masowego wstawiania encji
		return 0;
	}

	public <T> int remove(List<T> entities) {
		checkTransactionIsActive();

		if (entities == null || entities.isEmpty())
			throw new IllegalArgumentException(
					"Lista usuwanych encji nie moze byc pusta");

		T firstEntity = entities.get(0);
		JDBCEntityMetaData metaData = getMetaData(firstEntity.getClass());
		DmlQuerySupport query = metaData.getDeleteClause();
		DmlOperation operation = new DmlOperation(metaData, query);
		for (T entity : entities) {
			operation.addParam(new EntityObjectWithEntityId(entity, null));
		}
		this.dmlCommands.add(operation);
		if (this.flushModeType.equals(FlushModeType.AUTO)) {
			flush();
		}
		// TODO dokończyc implementacje metody dla masowego usuwania encji
		return 0;

	}

	public void setFlushMode(FlushModeType flushMode) {
		this.flushModeType = flushMode;
	}

	public Dialect getDialect() {
		return this.emf.getDbDialect();
	}

	@Override
	protected void finalize() throws Throwable {
		try {
			this.close();
		} finally {
			super.finalize();
		}
	}

	/**
	 * 
	 * DmlOperation klasa pomocnjcza z operacja modyfikacji danych
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.2 $
	 * 
	 */
	private class DmlOperation implements Serializable {

		private static final long serialVersionUID = 7030879279120317286L;

		final JDBCEntityMetaData metaData;

		final _IDmlQuery query;

		final Vector<EntityObjectWithEntityId> params;

		/**
		 * 
		 * @param metaData
		 * @param query
		 * @param param
		 */
		DmlOperation(JDBCEntityMetaData metaData, _IDmlQuery query,
				EntityObjectWithEntityId param) {
			super();
			this.metaData = metaData;
			this.query = query;
			this.params = new Vector<EntityObjectWithEntityId>();
			this.params.add(param);
		}

		/**
		 * 
		 * @param metaData
		 * @param query
		 */
		DmlOperation(JDBCEntityMetaData metaData, _IDmlQuery query) {
			super();
			this.metaData = metaData;
			this.query = query;
			this.params = new Vector<EntityObjectWithEntityId>();
		}

		void addParam(EntityObjectWithEntityId param) {
			this.params.add(param);
		}

	}

	/**
	 * 
	 * EntityObjectWithEntityId
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.2 $
	 * 
	 */
	class EntityObjectWithEntityId implements Serializable {

		private static final long serialVersionUID = -3282713197106844112L;

		private final Object entityObject;

		private final String entityId;

		private FirstLevelCacheObject cachedObject = null;

		/**
		 * @param entityObject
		 * @param entityId
		 */
		EntityObjectWithEntityId(Object entityObject, String entityId) {
			super();
			this.entityObject = entityObject;
			this.entityId = entityId;
		}

		/**
		 * @return the entityObject
		 */
		Object getEntityObject() {
			return entityObject;
		}

		/**
		 * @return the entityId
		 */
		String getEntityId() {
			return entityId;
		}

		/**
		 * @return the cachedObject
		 */
		FirstLevelCacheObject getCachedObject() {
			return cachedObject;
		}

		/**
		 * @param cachedObject
		 *            the cachedObject to set
		 */
		void setCachedObject(FirstLevelCacheObject cachedObject) {
			this.cachedObject = cachedObject;
		}

	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@Deprecated
	public Query createNamedQuery(String name) {
		checkEntityManagerIsOpen();
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@Deprecated
	public Query createNativeQuery(String sqlString) {
		checkEntityManagerIsOpen();
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@SuppressWarnings("rawtypes")
	@Deprecated
	public Query createNativeQuery(String sqlString, Class resultClass) {
		checkEntityManagerIsOpen();
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@Deprecated
	public Query createNativeQuery(String sqlString, String resultSetMapping) {
		checkEntityManagerIsOpen();
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@Deprecated
	public Query createQuery(String qlString) {
		checkEntityManagerIsOpen();
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	/**
	 * TODO Metoda jeszcze nie zaimplementowana
	 */
	@Deprecated
	public Object getDelegate() {
		throw new UnsupportedOperationException(
				"Metoda jeszcze nie zaimplementowana");
	}

	public boolean disableFirstLevelCache() {
		this.firstLevelCache.setCacheUsage(CacheUsage.NONE);
		this.localQueryCacheManager.setCacheUsage(CacheUsage.NONE);
		return (!this.firstLevelCache.cacheIsAllowed() || !this.localQueryCacheManager
				.cacheIsAllowed());
	}

	public boolean enableFirstLevelCache() {
		this.firstLevelCache.setCacheUsage(CacheUsage.TO_USE);
		this.localQueryCacheManager.setCacheUsage(CacheUsage.TO_USE);
		return (this.firstLevelCache.cacheIsAllowed() && this.localQueryCacheManager
				.cacheIsAllowed());
	}

	public boolean firstLevelCacheIsAllowed() {
		return (this.firstLevelCache.cacheIsAllowed() && this.localQueryCacheManager
				.cacheIsAllowed());
	}

	/**
	 * @return the pagingParams
	 */
	public PagingParams getDefaultPagingParams() {
		return defaultPagingParams;
	}

	/**
	 * @return the {@link #firstLevelCache}
	 */
	EntityFirstLevelCacheManager getFirstLevelCache() {
		return firstLevelCache;
	}

}
