package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.paging.PagingParamsReadOnly;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntitySecondLevelCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheType;
import pl.slawas.scjdbc4jpa.jdbc.dml.Delete;
import pl.slawas.scjdbc4jpa.jdbc.dml.Insert;
import pl.slawas.scjdbc4jpa.jdbc.dml.LockRow;
import pl.slawas.scjdbc4jpa.jdbc.dml.Update;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.CreateSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.IdIsNotEmbeddableException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectIsNotEntityException;
import pl.slawas.scjdbc4jpa.jdbc.generators.ValueGenerator;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityParserLevel;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCPersistenceEntityParser;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;

/**
 * 
 * JDBCEntityMetaData - obiekt metadanych przechowywany na liscie encji entity
 * manager-a
 * 
 * @see JDBCQuery
 * @see ResultRowMapper
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
public class JDBCEntityMetaData implements Serializable {

	private static final long serialVersionUID = -9176254511465791871L;

	private ConcurrentMap<JDBCEntityParserLevel, JDBCTableDesc> tableDefinitions = new ConcurrentHashMap<JDBCEntityParserLevel, JDBCTableDesc>(
			JDBCEntityParserLevel.values().length);

	private ConcurrentMap<JDBCEntityParserLevel, JDBCQuery> findAllQueries = new ConcurrentHashMap<JDBCEntityParserLevel, JDBCQuery>(
			JDBCEntityParserLevel.values().length);

	private ConcurrentMap<JDBCEntityParserLevel, JDBCQuery> findByIdQueries = new ConcurrentHashMap<JDBCEntityParserLevel, JDBCQuery>(
			JDBCEntityParserLevel.values().length);

	private final ValueGenerator valueGenerator;

	private final Insert insertClause;

	private final Update updateClause;

	private final Delete deleteClause;

	private final LockRow lockClause;

	private final EntitySecondLevelCacheManager entitySecondLevelCacheManager;

	private final QueryCacheManager uniqueQueryCacheManager;

	private final QueryCacheManager overQueryCacheManager;

	private final CacheUsage cacheUsage;

	private final CacheUsage queryCacheUsage;

	/**
	 * Konstruktor metadanych związanych z daną encją. Konstruktor inicjalizuje
	 * postać szeregu predefiniowanych zapytań, takie jak zapytanie o wszystkie
	 * elemenyty i wyszukiwanie po kluczu glownuym obiektu. Zadeklarowane użycie
	 * kesza oraz nazwa regionu przechowywania keszowanych danych będą
	 * zastosowane do tych właśnie zapytań i nie będą one mogły zostać
	 * zmienione. Ewentualne użycie metod {@link #setCacheRegionName(String)}
	 * oraz {@link #setCacheUsage(CacheUsage)}, mogą zostać wykorzystane do
	 * zmiany parametrów dla innych zapytań (później korzystac z getter-ów
	 * podczas ich budowy), ale trzeba być z tym bardzo ostrożnym, by czasem
	 * tych samych zapytań nie trzymać w różnych regionach.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param defaultCacheUsage
	 *            domyślne użycie kesz-a, jeżeli jest {@code null}, to zostanie
	 *            ustawiona wartość domyślna na {@link CacheUsage#TO_USE}. Jako,
	 *            że jest to inicjalizacjia obiektu, to parametr powinien
	 *            przyjmować jedną z dwóch wartości: {@link CacheUsage#TO_USE}
	 *            oraz {@link CacheUsage#NONE}
	 * @param customCacheRegionName
	 *            ogólnie podczas tworzenia definicji tabeli domyślnie jako
	 *            nazwa regionu kesza barana jest nazwa encji (zoabzcz
	 *            {@link JDBCTableDesc#getCacheRegionName()}). Pramater zmienia
	 *            ten stan i jeżeli nie jest {@code null} to dla tworzonej
	 *            definicji tabeli zostaje przypisany nowa nazwa regionu.
	 * @param queryCacheUsage
	 *            czy ma być używany kesz dla zapytań typu
	 *            {@link QueryCacheType#OVER} (opcjonalne, jezeli bedzię
	 *            {@code null}, to zostanie użyta wartość zdefiniowana w
	 *            parametrze 'defaultCacheUsage'. Jako, że jest to
	 *            inicjalizacjia obiektu, to parametr powinien przyjmować jedną
	 *            z dwóch wartości: {@link CacheUsage#TO_USE} oraz
	 *            {@link CacheUsage#NONE}
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws ObjectIsNotEntityException
	 * @throws JDBCEntityParserException
	 * @throws IdIsNotEmbeddableException
	 * @throws CreateSQLQueryException
	 */
	public JDBCEntityMetaData(Dialect dialect, Class<?> entityClass,
			PagingParams pagingParams, CacheUsage defaultCacheUsage,
			String customCacheRegionName, CacheUsage queryCacheUsage)
			throws SecurityException, NoSuchMethodException,
			ObjectIsNotEntityException, JDBCEntityParserException,
			IdIsNotEmbeddableException, CreateSQLQueryException {

		/** ustawiam na pierwszą stronę wyników */
		pagingParams.setPage(pagingParams.getFirstPage());

		this.cacheUsage = (defaultCacheUsage == null ? CacheUsage.TO_USE
				: defaultCacheUsage);
		this.queryCacheUsage = (queryCacheUsage == null ? this.cacheUsage
				: queryCacheUsage);

		JDBCTableDesc tableDefinition = JDBCPersistenceEntityParser.getTable(
				entityClass, JDBCEntityParserLevel.FIRST);
		tableDefinition.setCacheRegionName(customCacheRegionName);

		this.valueGenerator = new ValueGenerator(dialect, entityClass);
		this.insertClause = new Insert(tableDefinition);
		this.updateClause = new Update(tableDefinition);
		this.deleteClause = new Delete(tableDefinition);
		this.lockClause = new LockRow(tableDefinition);
		this.entitySecondLevelCacheManager = new EntitySecondLevelCacheManager(
				this.cacheUsage, tableDefinition);
		this.uniqueQueryCacheManager = new QueryCacheManager(this.cacheUsage,
				tableDefinition, QueryCacheType.UNIQUE);
		this.overQueryCacheManager = new QueryCacheManager(
				this.queryCacheUsage, tableDefinition, QueryCacheType.OVER);

		initMataData(JDBCEntityParserLevel.FIRST, entityClass, pagingParams,
				customCacheRegionName, tableDefinition);
		initMataData(JDBCEntityParserLevel.SECOND, entityClass, pagingParams,
				customCacheRegionName, null);
		initMataData(JDBCEntityParserLevel.THIRD, entityClass, pagingParams,
				customCacheRegionName, null);

	}

	/**
	 * Inicjalizacja struktur metadanych takich jak lista definicji tablicy (w
	 * zaleznosci od poziomu zagnieżdżenia), zapytan o wszystkie wiersze i
	 * zapytań z warunkiem po kluczu głównym
	 * 
	 * @param level
	 *            poziom zagnieżdżenia
	 * @param entityClass
	 *            klasa encji
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param customCacheRegionName
	 *            ogólnie podczas tworzenia definicji tabeli domyślnie jako
	 *            nazwa regionu kesza barana jest nazwa encji (zoabzcz
	 *            {@link JDBCTableDesc#getCacheRegionName()}). Pramater zmienia
	 *            ten stan i jeżeli nie jest {@code null} to dla tworzonej
	 *            definicji tabeli zostaje przypisany nowa nazwa regionu.
	 * @param tableDefinition
	 *            opcjonalnie przygotowana juz definicja tablicy.
	 * @return opis tablicy na zadanym poziomie zagnieżdżenia
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws ObjectIsNotEntityException
	 * @throws JDBCEntityParserException
	 * @throws IdIsNotEmbeddableException
	 * @throws CreateSQLQueryException
	 */
	private JDBCTableDesc initMataData(JDBCEntityParserLevel level,
			Class<?> entityClass, PagingParams pagingParams,
			String customCacheRegionName, JDBCTableDesc tableDefinition)
			throws SecurityException, NoSuchMethodException,
			ObjectIsNotEntityException, JDBCEntityParserException,
			IdIsNotEmbeddableException, CreateSQLQueryException {

		/* utworznie definicji tablicy zwiazanej z dana encją */
		if (tableDefinition == null
				|| !tableDefinition.getLevel().equals(level)) {
			tableDefinition = JDBCPersistenceEntityParser.getTable(entityClass,
					level);
			tableDefinition.setCacheRegionName(customCacheRegionName);
		}
		tableDefinitions.put(level, tableDefinition);

		/* utworznie zapytania o wszystkie wiersze */
		JDBCQuery findAllquery = new JDBCQuery(tableDefinition,
				new PagingParamsReadOnly(pagingParams),
				this.overQueryCacheManager);
		findAllquery.create();
		findAllQueries.put(level, findAllquery);
		String baseQueryClause = findAllquery.getQueryClause().getBaseClause();

		/* utworznie zapytania o wszystkie o podanym identyfikatorze */
		JDBCQuery findByIdquery = new JDBCQuery(tableDefinition,
				new PagingParamsReadOnly(pagingParams),
				this.uniqueQueryCacheManager, baseQueryClause);
		WhereClause where = new WhereClause(tableDefinition.getPrimaryKey());
		for (ConditionClauseSupport c : where.getConditionsArray()) {
			findByIdquery.addCondition(c);
		}
		findByIdquery.create();
		findByIdQueries.put(level, findByIdquery);

		return tableDefinition;
	}

	/**
	 * @param level
	 *            poziom zagnieżdzenia definicji {@link JDBCEntityParserLevel}
	 * @return definicja tabeli
	 */
	public JDBCTableDesc getTableDefinition(JDBCEntityParserLevel level) {
		return tableDefinitions.get(level);
	}

	/**
	 * @param level
	 *            poziom zagnieżdzenia definicji {@link JDBCEntityParserLevel}
	 * @return zapytanie o wszystkie wiersze
	 */
	public JDBCQuery getFindAllQuery(JDBCEntityParserLevel level) {
		return findAllQueries.get(level);
	}

	/**
	 * @param level
	 *            poziom zagnieżdzenia definicji {@link JDBCEntityParserLevel}
	 * @return zapytanie o wiersz o podanym identyfikatorze (wartościa klucza
	 *         głównego)
	 */
	public JDBCQuery getFindByIdQuery(JDBCEntityParserLevel level) {
		return findByIdQueries.get(level);
	}

	/**
	 * @return the cacheRegionName
	 */
	public String getCacheRegionName() {
		return this.getTableDefinition(JDBCEntityParserLevel.FIRST)
				.getCacheRegionName();
	}

	/**
	 * @return the cacheUsage
	 */
	public CacheUsage getCacheUsage() {
		return cacheUsage;
	}

	/**
	 * @return the insertClause
	 */
	public Insert getInsertClause() {
		return insertClause;
	}

	/**
	 * @return the updateClause
	 */
	public Update getUpdateClause() {
		return updateClause;
	}

	/**
	 * @return the deleteClause
	 */
	public Delete getDeleteClause() {
		return deleteClause;
	}

	/**
	 * @return the lockClause
	 */
	public LockRow getLockClause() {
		return lockClause;
	}

	@Override
	public String toString() {
		return "JDBCEntityMetaData [cacheRegionName=" + getCacheRegionName()
				+ ", cacheUsage=" + cacheUsage + ", tableDefinition="
				+ this.getTableDefinition(JDBCEntityParserLevel.THIRD) + "]";
	}

	/**
	 * @return the entitySecondLevelCacheManager
	 */
	public EntitySecondLevelCacheManager getEntitySecondLevelCacheManager() {
		return entitySecondLevelCacheManager;
	}

	/**
	 * @return the uniqueQueryCacheManager
	 */
	public QueryCacheManager getUniqueQueryCacheManager() {
		return uniqueQueryCacheManager;
	}

	/**
	 * @return the overQueryCacheManager
	 */
	public QueryCacheManager getOverQueryCacheManager() {
		return overQueryCacheManager;
	}

	/**
	 * @return the queryCacheUsage
	 */
	public CacheUsage getQueryCacheUsage() {
		return queryCacheUsage;
	}

	/**
	 * @return the valueGenerator
	 */
	public ValueGenerator getValueGenerator() {
		return valueGenerator;
	}

}
