package pl.slawas.scjdbc4jpa.jdbc;

import java.lang.reflect.InvocationTargetException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.persistence.FlushModeType;
import javax.persistence.TemporalType;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.entities.IDUtils;
import pl.slawas.entities.IDUtilsErrorException;
import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.paging.ResultMessage;
import pl.slawas.paging._IPagedQuery;
import pl.slawas.scjdbc4jpa.config.JdbcConfig;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.CreateSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ExecuteSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.IdIsNotEmbeddableException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCQueryExecuteException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectIsNotEntityException;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSimple;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionOperator;

/**
 * 
 * JDBCQuery stronicowane zapytanie do bazy danych.
 * 
 * <p>
 * Przykład tworzenia na podstawie tworzenia metadanych
 * {@link JDBCEntityMetaData}
 * </p>
 * 
 * <pre>
 * // utworznie definicji tablicy zwiazanej z dana encją
 * JDBCTableDesc tableDefinition = JDBCPersistenceEntityParser.getTable(
 * 		entityClass, level);
 * tableDefinitions.put(level, tableDefinition);
 * 
 * // utworznie zapytania o wszystkie wiersze przy
 * // użyciu konstruktora {@link #JDBCQuery(JDBCTableDesc, String, CacheUsage)}}
 * JDBCQuery findAllquery = new JDBCQuery(tableDefinition, cacheRegionName,
 * 		cacheUsage);
 * findAllquery.create();
 * findAllQueries.put(level, findAllquery);
 * 
 * // pobranie bazowej klauzuli zapytania
 * String baseQueryClause = findAllquery.getQueryClause().getBaseClause();
 * baseQueryClauses.put(level, baseQueryClause);
 * 
 * // utworznie zapytania o wszystkie o podanym identyfikatorze z bazową klauzulą
 * // użyciu konstruktora {@link #JDBCQuery(JDBCTableDesc, String, CacheUsage,
 * // String)}}
 * JDBCQuery findByIdquery = new JDBCQuery(tableDefinition, cacheRegionName,
 * 		cacheUsage, baseQueryClause);
 * WhereClause where = new WhereClause(tableDefinition.getPrimaryKey());
 * for (ConditionClause c : where.getConditionsArray()) {
 * 	findByIdquery.addCondition(c);
 * }
 * findByIdQueries.put(level, findByIdquery);
 * </pre>
 * 
 * <p>
 * Przykład tworzenia na zapytania w implementacji metody
 * {@link JDBCEntityManager#createQuery(Class)}
 * </p>
 * 
 * <pre>
 * public JDBCQuery createQuery(Class&lt;?&gt; entityClass) {
 * 	checkEntityManagerIsOpen();
 * 
 * 	String entityClassName = entityClass.getName();
 * 	JDBCEntityMetaData metaData = this.emf.getMetaData(entityClassName);
 * 	JDBCTableDesc table = metaData.getTableDefinition(level);
 * 	String baseQueryClause = metaData.getBaseQueryClause(level);
 * 	return (new JDBCQuery(table, metaData.getCacheRegionName(),
 * 			metaData.getCacheUsage(), baseQueryClause)).setEm(this);
 * }
 * 
 * </pre>
 * 
 * @see ResultRowMapper
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.10 $
 * 
 */
public class JDBCQuery extends SelectQuery implements javax.persistence.Query,
		_IPagedQuery<ResultRow>, _IFisrtLevelCacheManager {

	private static final long serialVersionUID = -8155029600607865920L;

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCQuery.class);

	private JDBCEntityManager em;

	private final CacheUsage defaultCacheUsage;

	private boolean isCreated = false;

	private boolean isExecuted = false;

	private ResultRowMapper resultRowMapper;

	private final Vector<ConditionClauseSupport> conditionClauses = new Vector<ConditionClauseSupport>();

	@SuppressWarnings("unused")
	private FlushModeType flushModeType = FlushModeType.AUTO;

	/**
	 * Konstruktor zapytania. Użycie tego konstruktora spowoduje utworzenie
	 * "całkowicie" od nowa klauzuli zapytania (poprzez parsowanie opisu
	 * tablicy). Można wykorzystać ten konstruktor do stworzenia klauzuli
	 * "bazowej" (używamy metody {@link #getQueryClause()} do pobrania
	 * "klauzuli zapytania" a nastepnie ( {@link QueryClause#getBaseClause()} do
	 * pobrania kluauzuli "bazowej").
	 * 
	 * @see SelectQuery#SelectQuery(QueryCacheManager, JDBCTableDesc, Integer)
	 * 
	 * @param table
	 *            opis tabeli
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param queryCacheManager
	 *            manager kesza
	 */
	public JDBCQuery(JDBCTableDesc table, PagingParams pagingParams,
			QueryCacheManager queryCacheManager) {
		super(queryCacheManager, table, pagingParams);
		this.defaultCacheUsage = (!this.cacheIsAllowed() ? CacheUsage.NONE
				: queryCacheManager.getCacheUsage());
	}

	/**
	 * Konstruktor zapytania. Użycie tego konstruktora wspiera podniesienie
	 * wydajności tworzenia zapytań poprzez pominięcie tworzenia klauzuli
	 * bazowej, która może być czasochłonna z punktu widzenia tworzenia
	 * zapytania.
	 * 
	 * @see SelectQuery#SelectQuery(QueryCacheManager, JDBCTableDesc, Integer)
	 * 
	 * @param table
	 *            definicja tablicy
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param queryCacheManager
	 *            manager kesza
	 * @param baseClause
	 *            klauzula bazowa, która może być gdzieś przechowywana, bo
	 *            przecież sie nie zmienia.
	 */
	public JDBCQuery(JDBCTableDesc table, PagingParams pagingParams,
			QueryCacheManager queryCacheManager, String baseClause) {
		super(queryCacheManager, table, pagingParams);
		this.queryClause = new QueryClause(baseClause);
		this.defaultCacheUsage = (!this.cacheIsAllowed() ? CacheUsage.NONE
				: queryCacheManager.getCacheUsage());
	}

	/**
	 * Konstruktor zapytania. Użycie tego konstruktora spowoduje utworzenie
	 * "całkowicie" od nowa klauzuli zapytania (poprzez parsowanie opisu
	 * tablicy). <b>Jedynym sensownym użyciem tego konstruktora jest budowa
	 * zapytań natywnych.</b> w przciwnym wypadku zaleca sie użycie
	 * konstruktorów:
	 * {@link #JDBCQuery(JDBCTableDesc, String, QueryCacheManager)} oraz
	 * {@link #JDBCQuery(JDBCTableDesc, String, QueryCacheManager, String)}
	 * 
	 * @see SelectQuery#SelectQuery(QueryCacheManager, JDBCTableDesc, Integer)
	 * 
	 * @param table
	 *            definicja tablicy
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param queryCacheManager
	 *            manager kesza
	 * @param sort
	 *            klauzula sortowania
	 * @param groupBy
	 *            klauzula grupowania
	 */
	public JDBCQuery(JDBCTableDesc table, PagingParams pagingParams,
			QueryCacheManager queryCacheManager, SortClause sort,
			GroupByClause groupBy) {
		super(queryCacheManager, table, pagingParams);
		this.defaultCacheUsage = (!this.cacheIsAllowed() ? CacheUsage.NONE
				: queryCacheManager.getCacheUsage());
		this.setGroupByClause(groupBy);
		this.setSortClause(sort);
	}

	/**
	 * Implementacja nie zakłada wykonywania operacji zmiany zawartości wierszy.
	 */
	@Deprecated
	public int executeUpdate() {
		throw new UnsupportedOperationException(
				"Obiekt nie wspiera operacji modyfikacji danych");
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony (metodą
	 * {@link #setPage(Page)}) albo dla domyślnej strony pierwszej (jeżeli
	 * strona nie została ustawiona). Wynik zapytania podlega restrykcjom liczby
	 * zwracanych wierszy (zobacz {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * ) i w zależności od parametru {@link JDBCQuery#defaultCacheUsage} może
	 * być keszowany.
	 * 
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult execute() {
		logger.trace("[execute()] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return execute(this.pagingParams.getPage());
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony (metodą
	 * {@link #setPage(Page)}) albo dla domyślnej strony pierwszej (jeżeli
	 * strona nie została ustawiona). Wynik zapytania podlega restrykcjom liczby
	 * zwracanych wierszy (zobacz {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * ) i w zależności od parametru 'cacheUsage' może być keszowany.
	 * 
	 * 
	 * @param cacheUsage
	 *            parametr defininiujacy użycie kesz-a
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult execute(CacheUsage cacheUsage) {
		logger.trace("[execute(CacheUsage)] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return execute(this.pagingParams.getPage(), cacheUsage);
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony. Wynik
	 * zapytania podlega restrykcjom liczby zwracanych wierszy (zobacz
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}) i w zależności od
	 * parametru 'cacheUsage' może być keszowany.
	 * 
	 * @param page
	 *            strona wyniku jaka ma być ustawiona w stronicowanym rezultacie
	 * @param cacheUsage
	 *            parametr defininiujacy użycie kesz-a
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult execute(Page page, CacheUsage cacheUsage) {
		this.pagingParams.setPage(page);
		logger.trace("[execute(Page, CacheUsage)] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return execute(this.pagingParams, ResultType.RESTRICTED, cacheUsage);
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony. Wynik
	 * zapytania podlega restrykcjom liczby zwracanych wierszy (zobacz
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}) i w zależności od
	 * parametru {@link JDBCQuery#defaultCacheUsage} może być keszowany.
	 * 
	 * @param page
	 *            strona wyniku jaka ma być ustawiona w stronicowanym rezultacie
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult execute(Page page) {
		this.pagingParams.setPage(page);
		logger.trace("[execute(Page)] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return this.execute(this.pagingParams, ResultType.RESTRICTED,
				this.defaultCacheUsage);
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony (metodą
	 * {@link #setPage(Page)}) albo dla domyślnej strony pierwszej (jeżeli
	 * strona nie została ustawiona). Wynik zapytania <b>NIE</b> podlega
	 * restrykcjom liczby zwracanych wierszy (zobacz
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES} ) oraz nie może być
	 * keszowany. Metoda na cele eksportu danych.
	 * 
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult createDump() {
		logger.trace("[createDump()] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return createDump(this.pagingParams.getPage());
	}

	/**
	 * Wykonanie zapytania i zwrócenie wyniku dla ustawionej strony. Wynik
	 * zapytania <b>NIE</b> podlega restrykcjom liczby zwracanych wierszy
	 * (zobacz {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}) oraz nie może być
	 * keszowany. Metoda na cele eksportu danych.
	 * 
	 * @param page
	 *            strona wyniku jaka ma być ustawiona w stronicowanym rezultacie
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult createDump(Page page) {
		this.pagingParams.setPage(page);
		logger.trace("[createDump(Page)] Parametry stronicowania: {}",
				this.pagingParams.toString());
		return execute(this.pagingParams, ResultType.UNLIMITED, CacheUsage.NONE);
	}

	/**
	 * Metoda właściwa wykonania zapytania na bazie dnaych
	 * 
	 * @param pagingParams
	 *            parametry stronicowania z ustawioną odpowiednio stroną
	 * @param resultType
	 *            typ rezultatu (definicja używania/nie używania restrykcji dla
	 *            wyniku - zobacz {@link ResultType})
	 * @param cacheUsage
	 *            sposób użycia kesz-a
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult execute(PagingParams pagingParams,
			ResultType resultType, CacheUsage cacheUsage) {
		try {
			if (em == null)
				throw new IllegalArgumentException(
						"Brak ustwionego entity manager-a");

			/* sprawdzam czy zapytanie zostalo utworzone */
			if (!this.isCreated) {
				this.create();
			}
			/*
			 * jezeli nie zostala zdefiniowana strona, to uzywam domyslnych jej
			 * parametrów.
			 */
			if (pagingParams != null) {
				if (!this.pagingParams.isReadOnly()
						&& !this.pagingParams.equals(pagingParams)) {
					logger.trace("Zmieniam parametry stronicowania z {} na {}",
							new Object[] { this.pagingParams.toString(),
									pagingParams.toString() });
					setPagingParams(pagingParams);
				} else if (logger.isTraceEnabled()
						&& this.pagingParams.isReadOnly()) {
					logger.trace(
							"Parametry stronicowania sa tylko do odczytu: {}",
							this.pagingParams.toString());
				}
			}
			/* ustwaiam identyfikator zapytania */
			String queryId = super.getQueryId();
			if (this.resultRowMapper == null || !isExecuted) {
				/*
				 * wynik zapytania nie jest ustwiony, lub zapytanie nie zostalo
				 * jeszcze wykonane
				 */
				logger.trace("Wykonuje zapytanie, bo nie mam resultRowMapper-a, albo zapytanie nie zostalo jeszcze wykonane...");
				this.resultRowMapper = super.executeQuery(em.getConnection(),
						cacheUsage, queryId, this.pagingParams.getOffset() + 1,
						null, resultType);
				isExecuted = true;
			}

			QueryResult queryResult = this.resultRowMapper.prepareQueryResult(
					createQueryCacheInfo(this.defaultCacheUsage, queryId),
					this.pagingParams.getPage().getFirstRowNumber(),
					this.pagingParams.getPage().getLastRowNumber());
			PagedResult pagedResult = new PagedResult(
					this.resultRowMapper.getPagingParams(), queryResult, this);
			this.pagingParams.setPage(pagedResult.getCurrentPageInfo());
			logger.trace("KESZ WLACZONY? {}", this.firstLevelCacheIsAllowed());
			if (!pagedResult.getMessage().equals(ResultMessage.NO_DATA_FOUND)
					&& resultType.equals(ResultType.RESTRICTED)
					&& this.firstLevelCacheIsAllowed())

				for (ResultRow row : pagedResult.getResult()) {
					if (!row.isEntityAvailable()) {
						if (logger.isDebugEnabled()) {
							logger.warn("Obiekt encji nie jest dostepny");
						}
						break;
					}

					Object primaryKey = null;
					try {
						primaryKey = IDUtils.getObjectId(row.getEntity());
						if (primaryKey != null) {
							JDBCEntityManager.EntityObjectWithEntityId entityInfo = em
									.getFromFirstLevelCache(
											this.table.getEntityClass(),
											primaryKey);
							JDBCEntityMetaData metaData = em
									.getMetaData(this.table.getEntityClass());
							em.getFirstLevelCache()
									.refreshEntityInFirstLevelCache(
											em.getName(),
											entityInfo.getCachedObject(),
											entityInfo.getEntityId(),
											pagedResult.getQueryInfo(),
											metaData,
											row.getEntity(),
											(em.getTransaction().isActive() ? CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE
													: CacheUsage.TO_USE));
						}
					} catch (IDUtilsErrorException ignore) {
					} catch (InstantiationException e) {
						if (logger.isTraceEnabled()) {
							logger.warn(e.getMessage());
						}
					} catch (IllegalAccessException e) {
						if (logger.isTraceEnabled()) {
							logger.warn(e.getMessage());
						}
					} catch (NoSuchMethodException e) {
						if (logger.isTraceEnabled()) {
							logger.warn(e.getMessage());
						}
					} catch (InvocationTargetException e) {
						if (logger.isTraceEnabled()) {
							logger.warn(e.getMessage());
						}
					} catch (SQLException e) {
						if (logger.isTraceEnabled()) {
							logger.warn(e.getMessage());
						}
					}
				}

			return pagedResult;
		} catch (SQLException e) {
			throw new JDBCQueryExecuteException(
					"Blad pobrania listy wynikow zapytania", e);
		} catch (ExecuteSQLQueryException e) {
			throw new JDBCQueryExecuteException(
					"Blad pobrania listy wynikow zapytania", e);
		} catch (ObjectIsNotEntityException e) {
			throw new JDBCQueryExecuteException("Blad utworzenia zapytania", e);
		} catch (JDBCEntityParserException e) {
			throw new JDBCQueryExecuteException("Blad utworzenia zapytania", e);
		} catch (IdIsNotEmbeddableException e) {
			throw new JDBCQueryExecuteException("Blad utworzenia zapytania", e);
		} catch (CreateSQLQueryException e) {
			throw new JDBCQueryExecuteException("Blad utworzenia zapytania", e);
		}
	}

	/**
	 * Używaj metody {@link #getPagedResult()} albo
	 * {@link #getPagedResult(Page)} .
	 * <p>
	 * Zaleca się nie używanie tej metody ze wzgledu na utratę informacji o
	 * stronicowaniu wyniku. Maksymalna liczba wierszy zwracana przez tę metode
	 * jest ograniczona do rozmiaru strony oraz zalezna od numeru strony,
	 * zdefiniowanej metodą {@link JDBCQuery#setPage(Page)}.
	 * </p>
	 */
	@Deprecated
	public List<ResultRow> getResultList() {
		PagedResult pagedResult = execute();
		if (pagedResult != null && pagedResult.getResultSize() != 0) {
			List<ResultRow> resultList = new ArrayList<ResultRow>();
			Collection<ResultRow> rows = pagedResult.getResult();
			for (ResultRow row : rows) {
				resultList.add(row);
			}
			return resultList;
		}
		return null;
	}

	public ResultRow getSingleResult() {
		PagedResult pagedResult = execute();
		if (pagedResult != null && pagedResult.getResultSize() != 0) {
			return pagedResult.getResult().iterator().next();
		}
		return null;
	}

	/**
	 * Metoda ustwiająca pierwszy numer wiersza (rozpoczynając od wartości
	 * {@code 1}) w wyniku jaki ma być zwrócony (ustawia offset parametrów
	 * stronicowania - zobacz {@link PagingParams #setOffset(int)}).
	 */
	public JDBCQuery setFirstResult(int arg0) {
		this.pagingParams.setOffset(arg0);
		return this;
	}

	public JDBCQuery setFlushMode(FlushModeType arg0) {
		this.flushModeType = arg0;
		return this;
	}

	public JDBCQuery setHint(String arg0, Object arg1) {
		if (arg1 != null) {
			logger.warn("Obiekt nie wspiera argumentu Object ({}).", arg1);
		}
		super.setComment(arg0);
		return this;
	}

	/**
	 * Metoda ustawiająca maksymalną możliwą liczbę wierszy w rezultacie
	 */
	public JDBCQuery setMaxResults(int arg0) {
		this.pagingParams.setMaxCount(arg0);
		return this;
	}

	public JDBCQuery setParameter(String arg0, Object arg1) {
		JDBCColumnDesc column = this.table.getColumn(arg0);
		ConditionClauseSupport condition = new ConditionClauseSimple(
				ConditionOperator.AND, column, arg1, true);
		this.conditionClauses.add(condition);
		return this;
	}

	/**
	 * Metoda nie wspierana przez dany obiekt. Można używać tyko metody
	 * {@link #setParameter(String, Object)}.
	 */
	@Deprecated
	public JDBCQuery setParameter(int arg0, Object arg1) {
		throw new UnsupportedOperationException(
				"Metoda nie wspierana przez dany obiekt");
	}

	/**
	 * Uzywaj metody {@link #setParameter(String, Object)}.
	 * <p>
	 * Implementacja metoda nie wspiera wszystkich argumentów.
	 * </p>
	 */
	@Deprecated
	public JDBCQuery setParameter(String arg0, Date arg1, TemporalType arg2) {
		return setParameter(arg0, arg1);
	}

	/**
	 * Uzywaj metody {@link #setParameter(String, Object)}.
	 * <p>
	 * Implementacja metoda nie wspiera wszystkich argumentów.
	 * </p>
	 */
	@Deprecated
	public JDBCQuery setParameter(String arg0, Calendar arg1, TemporalType arg2) {
		if (arg2 != null) {
			logger.warn("Obiekt nie wspiera argumentu TemporalType ({}). "
					+ "Uzyta zostanie metoda setParameter(int, Object)", arg2);
		}
		return setParameter(arg0, arg1);
	}

	/**
	 * Metoda nie wspierana przez dany obiekt. Można używać tyko metody
	 * {@link #setParameter(String, Object)}.
	 */
	@Deprecated
	public JDBCQuery setParameter(int arg0, Date arg1, TemporalType arg2) {
		throw new UnsupportedOperationException(
				"Metoda nie wspierana przez dany obiekt");
	}

	/**
	 * Metoda nie wspierana przez dany obiekt. Można używać tyko metody
	 * {@link #setParameter(String, Object)}.
	 */
	@Deprecated
	public JDBCQuery setParameter(int arg0, Calendar arg1, TemporalType arg2) {
		throw new UnsupportedOperationException(
				"Metoda nie wspierana przez dany obiekt");
	}

	/**
	 * Pobranie podstawowego obiektu zapytania.
	 * 
	 * @return obiekt zapytania JDBC
	 */
	public SelectQuery getQuery() {
		return this;
	}

	/**
	 * @return the isCreated
	 */
	public boolean isCreated() {
		return isCreated;
	}

	/**
	 * Usunięcie warunków zapytania. Aby metoda miała skuteczność po jej
	 * uruchomieniu nalzezy ponownie utworzyć zapytanie wywolując metodę
	 * {@link #createQuery()}. Metoda zmienia status zapytania ustawiają
	 * {@link #isCreated} na wartość {@code false}.
	 * 
	 * @return tworzone zapytanie
	 */
	public JDBCQuery clearConditions() {
		this.conditionClauses.clear();
		clearQueryStatus();
		return this;
	}

	/**
	 * Dodawanie zaawansowanej klauzuli wyszukiwania
	 * 
	 * @param condition
	 *            warunke wyszukiwania
	 * @return tworzone zapytanie
	 */
	public JDBCQuery addCondition(ConditionClauseSupport condition) {
		this.conditionClauses.add(condition);
		return this;
	}

	/**
	 * @return the em
	 */
	public JDBCEntityManager getEm() {
		return em;
	}

	/**
	 * @param em
	 *            the em to set
	 * @return zapytanie
	 */
	public JDBCQuery setEm(JDBCEntityManager em) {
		this.em = em;
		return this;
	}

	public JDBCQuery setGroupByClause(GroupByClause groupBy) {
		super.setGroupByClause(groupBy);
		clearQueryStatus();
		return this;
	}

	public JDBCQuery addFunction(_IFunction function)
			throws CreateSQLQueryException {
		super.addFunction(function);
		clearQueryStatus();
		return this;
	}

	/**
	 * Ustawienie flag braku utworzenia zapytania oraz wykonania zapytania.
	 * Czyszczenie jest potrzebne gdy po utworzeniu zapytania nastąpi dołożenie
	 * klauzul sortowania, grupowania, albo dodatkowych warunków zapytania, albo
	 * następi jakakolwiek inna modyfikacja klauzuli zapytania.
	 * 
	 * @see JDBCQuery#setSortClause(SortClause)
	 * @see JDBCQuery#setGroupByClause(GroupByClause)
	 * @see JDBCQuery#addCondition(ConditionClauseSupport)
	 * @see JDBCQuery#addFunction(pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction)
	 */
	private void clearQueryStatus() {
		this.isCreated = false;
		this.isExecuted = false;
	}

	public JDBCQuery setSortClause(SortClause sort) {
		super.setSortClause(sort);
		clearQueryStatus();
		return this;
	}

	/**
	 * Zwraca wynik zapytania na stronie zdefiniowanej metodą
	 * {@link #setPage(Page)}. Jeżeli strona nie jest ustawiona zostanie
	 * zwrócony wynik strony o numerze {@link JdbcConfig#DEFAULT_PAGE_NR}. Mozna
	 * tez uzywać metody {@link #setFirstResult(int)} do ustwiwania numeru
	 * strony ale takie rozwiązanie nie jest zalecane.
	 * 
	 * @return zwraca stronicowany wynik
	 */
	public PagedResult getPagedResult() {
		return execute();
	}

	/**
	 * Zwraca stronicowany wynik zapytania na zadanej stronie
	 * 
	 * @param page
	 *            strona wyników
	 * @return winik zapytania na danej stronie
	 */
	public PagedResult getPagedResult(Page page) {
		return execute(page);
	}

	/**
	 * @param page
	 *            the page to set
	 */
	public void setPage(Page page) {
		logger.trace("Ustawiam strone wynikow: {}", page.toString());
		if (!this.pagingParams.setPage(page)) {
			logger.warn("Nie udane ustawienie strony o parametrach {}",
					page.toString());
		}
	}

	@Override
	public SelectQuery clearFunctions() throws CreateSQLQueryException {
		super.clearFunctions();
		clearQueryStatus();
		return this;
	}

	@Override
	public SelectQuery create() throws ObjectIsNotEntityException,
			JDBCEntityParserException, IdIsNotEmbeddableException,
			CreateSQLQueryException {
		if (conditionClauses != null && !conditionClauses.isEmpty()) {
			super.setWhereClause(new WhereClause(conditionClauses));
		} else {
			super.setWhereClause(null);
		}
		super.create();
		this.isCreated = true;
		return this;
	}

	/**
	 * Użycie tej metody zatraca całą ideę stronicowanych wyników
	 */
	@Override
	@Deprecated
	public QueryResult execute(Connection conn, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		return super.execute(conn, cacheUsage);
	}

	/**
	 * Użycie tej metody zatraca całą ideę stronicowanych wyników
	 */
	@Override
	@Deprecated
	public QueryResult execute(Connection conn, Integer startPosition,
			Integer endPosition, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		return super.execute(conn, startPosition, endPosition, cacheUsage);
	}

	@Override
	public SelectQuery removeFunction(_IFunction function)
			throws CreateSQLQueryException {
		super.removeFunction(function);
		clearQueryStatus();
		return this;

	}

	@Override
	public SelectQuery removeFunction(String functionKey)
			throws CreateSQLQueryException {
		super.removeFunction(functionKey);
		clearQueryStatus();
		return this;
	}

	@Override
	public SelectQuery setAfterSelectClause(SelectAfterClause afterSelectClause) {
		super.setAfterSelectClause(afterSelectClause);
		clearQueryStatus();
		return this;
	}

	@Override
	public SelectQuery setColumnsToUse(Set<String> columnsToUseList) {
		super.setColumnsToUse(columnsToUseList);
		clearQueryStatus();
		return this;
	}

	@Override
	public SelectQuery setColumnsToUse(String[] columnsToUse) {
		super.setColumnsToUse(columnsToUse);
		clearQueryStatus();
		return this;
	}

	@Override
	public void setGenerateEntityObject(boolean generateEntityObject) {
		super.setGenerateEntityObject(generateEntityObject);
		clearQueryStatus();
	}

	@Override
	public void setUseFunctionsOnly(boolean useFunctionsOnly) {
		super.setUseFunctionsOnly(useFunctionsOnly);
		clearQueryStatus();
	}

	/**
	 * Metoda nie wspierana przez dany obiekt. Można używać tyko metody
	 * dodawania poszczególnych warunków zapytania.
	 * {@link #addCondition(ConditionClauseSupport)}.
	 */
	@Override
	@Deprecated
	public SelectQuery setWhereClause(WhereClause where) {
		throw new UnsupportedOperationException(
				"Metoda nie wspierana przez dany obiekt");
	}

	/**
	 * Użycie tej metody zatraca całą ideę stronicowanych wyników
	 */
	@Override
	@Deprecated
	public ResultRow uniqeExecute(Connection conn, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		return super.uniqeExecute(conn, cacheUsage);
	}

	/**
	 * @param isExecuted
	 *            the isExecuted to set
	 */
	public void setExecuted(boolean isExecuted) {
		this.isExecuted = isExecuted;
	}

	/**
	 * @return the {@link #page}
	 */
	public Page getPage() {
		return this.pagingParams.getPage();
	}

	public boolean disableFirstLevelCache() {
		return this.em.disableFirstLevelCache();
	}

	public boolean enableFirstLevelCache() {
		return this.em.enableFirstLevelCache();
	}

	public boolean firstLevelCacheIsAllowed() {
		return this.em.firstLevelCacheIsAllowed();
	}

}