package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import javax.persistence.EntityTransaction;
import javax.persistence.FlushModeType;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.cache.managers.EntitySecondLevelCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCTransactionException;

/**
 * 
 * JDBCTransaction - implementacja obiektu transakcji
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.1 $
 * 
 */
public class JDBCTransaction implements Serializable, EntityTransaction {

	private static final long serialVersionUID = -6279092815803428517L;

	private final JDBCEntityManager em;

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCTransaction.class);

	private int summaryUpdateCounts = 0;

	private boolean isActive = false;

	private boolean isRollbackOnly = false;

	private Set<QueryCacheManager> cachesToClear = new HashSet<QueryCacheManager>();

	private Set<String> changedObjects = new HashSet<String>();

	private Hashtable<String, EntitySecondLevelCacheManager> removedObjects = new Hashtable<String, EntitySecondLevelCacheManager>();

	JDBCTransaction(JDBCEntityManager em) {
		this.em = em;
	}

	/**
	 * Sprawdzenie czy transakcja jest aktywna.
	 */
	private void checkTransactionIsActive() {
		if (!isActive()) {
			throw new JDBCTransactionException("Transakcja nie jest aktywna");
		}
	}

	private int setChangedObjectAsCommited() {
		int i = 0;
		if (!this.changedObjects.isEmpty()) {
			for (String entityId : changedObjects) {
				this.em.getFirstLevelCache().setToRefreshFirstLevelCacheObject(
						entityId);
				i++;
			}
			this.changedObjects.clear();
		}
		if (!this.removedObjects.isEmpty()) {
			Enumeration<String> entityIds = this.removedObjects.keys();
			while (entityIds.hasMoreElements()) {
				String entityId = entityIds.nextElement();
				EntitySecondLevelCacheManager sc = removedObjects.get(entityId);
				sc.removeEntitySharedInfo(entityId, true);
				i++;
			}
			this.removedObjects.clear();
		}
		if (!this.cachesToClear.isEmpty()) {
			for (QueryCacheManager cache : this.cachesToClear) {
				cache.clear();
			}
			this.cachesToClear.clear();
		}
		return i;
	}

	private int rollbackChangedObject() {
		int i = 0;
		if (!this.changedObjects.isEmpty()) {
			for (String entityId : this.changedObjects) {
				this.em.getFirstLevelCache().removeFirstLevelCacheObject(
						entityId, false);
				i++;
			}
			this.changedObjects.clear();
		}
		if (!this.removedObjects.isEmpty()) {
			this.removedObjects.clear();
		}
		this.cachesToClear.clear();
		return i;
	}

	/**
	 * TODO tu można byłoby zaimplementować ustawienie poziomu transakcji.
	 */
	public void begin() {
		this.em.checkEntityManagerIsOpen();
		isActive = true;
		summaryUpdateCounts = 0;
	}

	public void commit() {
		this.em.checkEntityManagerIsOpen();
		checkTransactionIsActive();
		if (isRollbackOnly) {
			throw new JDBCTransactionException(
					"Transakcja zostala oznaczona jako tylko do wycofania");
		}
		try {
			if (em.getFlushMode().equals(FlushModeType.COMMIT))
				em.flush();
			em.getConnection().commit();
			int updateCounts = setChangedObjectAsCommited();
			logger.trace("\n|\n\\--> Zatwierdzilem {} zmiany encji\n",
					updateCounts);
		} catch (SQLException e) {
			throw new JDBCTransactionException(e);
		}
		isActive = false;
	}

	public boolean getRollbackOnly() {
		return isRollbackOnly;
	}

	public boolean isActive() {
		return isActive;
	}

	public void rollback() {
		this.em.checkEntityManagerIsOpen();
		checkTransactionIsActive();

		try {
			em.getConnection().rollback();
			int rollbackCounts = rollbackChangedObject();
			logger.trace("\n|\n\\--> Wycofalem {} zmiany encji\n",
					rollbackCounts);
		} catch (SQLException e) {
			throw new JDBCTransactionException(e);
		}
	}

	public void setRollbackOnly() {
		this.em.checkEntityManagerIsOpen();
		checkTransactionIsActive();

		isRollbackOnly = true;
	}

	/**
	 * Podczas trwania transakcji sumowana jest liczba zmian. Suma ta jest
	 * zerowana podczas uruchomienia transakcji za pomocą metody
	 * {@link #begin()} .
	 * 
	 * @return liczba zmienionych wierszy w calej transakcji
	 */
	public int getSummaryUpdateCounts() {
		return summaryUpdateCounts;
	}

	/**
	 * Metoda dodaje do {@link #summaryUpdateCounts} liczbe zmienionych wierszy
	 * co daje nam koncowy obraz w postaci liczby wszystkich zmian dokonanych na
	 * encjach.
	 * 
	 * @param liczba
	 *            zmienionych wierszy
	 */
	void sumUpdateCounts(int updateCounts) {
		this.summaryUpdateCounts = this.summaryUpdateCounts + updateCounts;
	}

	void addChangedObject(String entityId) {
		this.changedObjects.add(entityId);
	}

	void addRemovedObject(String entityId,
			EntitySecondLevelCacheManager secondCache) {
		this.removedObjects.put(entityId, secondCache);
	}

	void addCacheToClear(QueryCacheManager cacheToClear) {
		this.cachesToClear.add(cacheToClear);
	}

}
