package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.util.Vector;

import pl.slawas.scjdbc4jpa.jdbc.where.BindParam;

public class QueryClause implements _IClause, Serializable {

	private static final long serialVersionUID = -7161942965900209875L;

	/**
	 * Podstawowa klauzula zapytania {@code SELECT ... FROM ...} bez klauzul
	 * warunków zapytania, grupowania i sortowania.
	 */
	private final String baseClause;

	private StringBuffer queryClause;

	private boolean isExtendedClause = false;

	private final Vector<BindParam> bindedParams = new Vector<BindParam>();

	/**
	 * Konstruktor obiektu klauzuli zapytania.
	 * 
	 * @param baseClause
	 *            podstawowa klauzula zapytania {@code SELECT ... FROM ...} bez
	 *            klauzul warunków zapytania, grupowania i sortowania.
	 */
	public QueryClause(String baseClause) {
		super();
		this.baseClause = baseClause;
		this.queryClause = new StringBuffer(this.baseClause);
	}

	/**
	 * Metoda dodająca rozszerzone elementy zapytania takie jak klauzle warunków
	 * zapytania, grupowania i sortowania.
	 * <p>
	 * <font color="#dd0000"><b>Uwaga!</b></font> metoda czyści wszelkie
	 * rozszerzone klauzule warunków zapytania, grupowania i sortowania do
	 * zapytania podstawowego ({@link #baseClause}), i dodaje tą, która jest
	 * aktulanie w parametrze. Jeżeli chcesz dodać kolejną klauzulę rozszeroną
	 * bez konsekwencji obcięcia zapytania do klauzuli podstawowej użyj metody
	 * {@link #addNextExtendClause(String)}.
	 * </p>
	 * 
	 * @param extendClause
	 *            klauzua warunków zapytania, grupowania i sortowania.
	 * 
	 * @return własnie tworzone zapytanie
	 */
	public QueryClause addExtendClause(String extendClause) {
		if (!isExtendedClause) {
			isExtendedClause = true;
		} else {
			this.queryClause = new StringBuffer(this.baseClause);
		}
		queryClause.append(extendClause);
		return this;
	}

	/**
	 * Metoda dodająca rozszerzone elementy zapytania takie jak klauzle warunków
	 * zapytania, grupowania i sortowania.
	 * <p>
	 * <font color="#dd0000"><b>Uwaga!</b></font> metoda zachowuje wszelkie
	 * dotychczas dodane rozszerzone klauzule warunków zapytania, grupowania i
	 * sortowania, i dodaje tą, która jest aktulanie w parametrze. Jeżeli chcesz
	 * dodać kolejną klauzulę rozszeroną bez konsekwencji zachowania poprzednich
	 * klauzul (chcesz aby klauzula, zanim zostanie dodane rozszerzenie, została
	 * obcięta do zapytania podstawowego ({@link #baseClause})) użyj metody
	 * {@link #addExtendClause(String)}.
	 * </p>
	 * 
	 * @param extendClause
	 *            klauzua warunków zapytania, grupowania i sortowania.
	 * 
	 * @return własnie tworzone zapytanie
	 */
	public QueryClause addNextExtendClause(String extendClause) {
		if (!isExtendedClause) {
			isExtendedClause = true;
		}
		queryClause.append(extendClause);
		return this;
	}

	/**
	 * @return the queryClause
	 */
	public String getClause() {
		return queryClause.toString();
	}

	/**
	 * Metoda czyszcząca wszystkie prametry do bindowania
	 */
	public void clearBindedParams() {
		bindedParams.clear();
	}

	/**
	 * @return the bindedParams
	 */
	public Vector<BindParam> getBindedParams() {
		return bindedParams;
	}

	/**
	 * Dodawanie parametru bindowania
	 * 
	 * @param bindParam
	 * @return właśnie tworzona klauzula zapytania
	 */
	public QueryClause addBindParam(BindParam bindParam) {
		this.bindedParams.add(bindParam);
		return this;
	}

	@Override
	public String toString() {
		StringBuffer binds = new StringBuffer();
		if (this.bindedParams != null && !this.bindedParams.isEmpty()) {
			for (BindParam bind : this.bindedParams) {
				binds.append("\n\t" + bind.toString());
			}
		} else {
			binds.append("\n\tnull");
		}
		return "QueryClause [" + "\n queryClause:" + "\n "
				+ queryClause.toString() + "\n bindedParams:"
				+ binds.toString() + "\n]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((bindedParams == null) ? 0 : bindedParams.hashCode());
		result = prime
				* result
				+ ((queryClause == null) ? 0 : queryClause.toString()
						.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		QueryClause other = (QueryClause) obj;
		if (bindedParams == null) {
			if (other.bindedParams != null)
				return false;
		} else if (!bindedParams.equals(other.bindedParams))
			return false;
		if (queryClause == null) {
			if (other.queryClause != null)
				return false;
		} else if (!queryClause.toString().equals(other.queryClause.toString()))
			return false;
		return true;
	}

	/**
	 * @return the isExtendedClause
	 */
	boolean isExtendedClause() {
		return isExtendedClause;
	}

	/**
	 * @return the baseClause
	 */
	public String getBaseClause() {
		return baseClause;
	}

}
