package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityValueParser;

/**
 * 
 * ResultRow obiekt wiersza tabeli dla analizy rezultatu zapytania sql
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class ResultRow implements Serializable {

	private static final long serialVersionUID = 676091196707625533L;

	final static private Logger logger = LoggerFactory
			.getLogger(ResultRow.class);

	private final JDBCTableDesc tableDescription;

	private final int rowNumber;

	private final boolean entityAvailable;

	private final Hashtable<String, Object> columns;

	/**
	 * Konstruktor wiersza wyniku
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rowNumber
	 *            numer wiersza
	 * @param columns
	 *            kolumny z wartościami
	 * @param entityAvailable
	 *            flaga czy ma zostac wygenerowana encja podczas odwołania sie
	 *            do metody {@link #getEntity()} albo
	 *            {@link #refreshEntity(Object)}. Dla niektorych wyników zapytan
	 *            nie możliwym jest wygenerowanie encji poniważ może to być
	 *            zapytanie zliczajace wiersze, albo zawierające same funkcje
	 *            itp. - wtedy parametr powinien przyjać wartość {@code true}
	 */
	public ResultRow(JDBCTableDesc tableDescription, int rowNumber,
			Hashtable<String, Object> columns, boolean entityAvailable) {
		this.tableDescription = tableDescription;
		this.columns = columns;
		this.rowNumber = rowNumber;
		this.entityAvailable = entityAvailable;

	}

	/**
	 * Pobranie obiektu encji na podstawie wartości kolumn wiersza wyniku.
	 * 
	 * @return obiekt encji
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws SQLException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 */
	public Object getEntity() throws InstantiationException,
			IllegalAccessException, NoSuchMethodException,
			InvocationTargetException, SQLException {

		if (this.entityAvailable) {

			Class<?> clazz = this.tableDescription.getEntityClass();
			if (logger.isTraceEnabled()) {
				logger.trace("Klasa encji {}...", clazz.getName());
				logger.trace("Analiza wiersza nr {}", this.getRowNumber());
				logger.trace("Bede tworzyc obiekty encji dla klasy {}", clazz);
			}

			Object entity = null;
			entity = clazz.newInstance();

			Enumeration<String> columnKeys = this.tableDescription
					.getColumnKeys();
			while (columnKeys.hasMoreElements()) {
				/*
				 * przepisywanie danych z kolumny rezultatu do pola encji -
				 * START
				 */
				String entityFieldName = columnKeys.nextElement();
				JDBCColumnDesc col = this.tableDescription
						.getColumn(entityFieldName);
				SimpleSQLType type = SimpleSQLType.get(col.getColumnType());

				if (entity != null && columns.get(entityFieldName) != null) {
					JDBCEntityValueParser.write2EntityObject(entity,
							columns.get(entityFieldName), col, type,
							entityFieldName);
				}

				/*
				 * przepisywanie danych z kolumny rezultatu do pola encji -
				 * KONIEC
				 */
			}

			return entity;
		}
		throw new IllegalAccessError("Encja nie jest dostepna. "
				+ "Flaga entityIsAvailable:[false], a zatem wynika to miedzy "
				+ "innymi z parametrow utworzonego zapytania.");
	}

	/**
	 * Metoda przepisujaca wynik zapuyania (wiersza) do istniejącej encji. Mozna
	 * ja wykorzystać do "odświeżenia" obiektu.
	 * 
	 * @param entity
	 *            aktulana encje
	 * @return odświeżony obiekt encji
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws InvocationTargetException
	 * @throws SQLException
	 */
	public Object refreshEntity(Object entity) throws InstantiationException,
			IllegalAccessException, NoSuchMethodException,
			InvocationTargetException, SQLException {

		if (this.entityAvailable) {

			Enumeration<String> columnKeys = this.tableDescription
					.getColumnKeys();
			while (columnKeys.hasMoreElements()) {
				/*
				 * przepisywanie danych z kolumny rezultatu do pola encji -
				 * START
				 */
				String entityFieldName = columnKeys.nextElement();
				JDBCColumnDesc col = this.tableDescription
						.getColumn(entityFieldName);
				SimpleSQLType type = SimpleSQLType.get(col.getColumnType());

				if (entity != null && columns.get(entityFieldName) != null) {
					JDBCEntityValueParser.write2EntityObject(entity,
							columns.get(entityFieldName), col, type,
							entityFieldName);
				}

				/*
				 * przepisywanie danych z kolumny rezultatu do pola encji -
				 * KONIEC
				 */
			}

			return entity;
		}
		throw new IllegalAccessError("Encja nie jest dostepna. "
				+ "Flaga entityIsAvailable:[false], a zatem wynika to miedzy "
				+ "innymi z parametrow utworzonego zapytania.");
	}

	/**
	 * @return opis tabeli
	 */
	public JDBCTableDesc getTableDescription() {
		return tableDescription;
	}

	public Enumeration<String> getColumnsKeys() {
		return (columns != null ? columns.keys() : null);
	}

	public Object getColumnValue(String key) {
		return (columns != null ? columns.get(key) : null);
	}

	/**
	 * @return liczba kolumn
	 */
	public int length() {
		return (columns != null ? columns.size() : 0);
	}

	/**
	 * @return numer wiersza
	 */
	public int getRowNumber() {
		return rowNumber;
	}

	/**
	 * @return the entityAvailable
	 */
	public boolean isEntityAvailable() {
		return entityAvailable;
	}

}
