package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.paging.PagingParamsReadOnly;
import pl.slawas.paging.ResultMessage;
import pl.slawas.scjdbc4jpa.config.JdbcConfig;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;

/**
 * 
 * ResultRowMapper
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.12 $
 * 
 */
public class ResultRowMapper implements Serializable {

	private static final long serialVersionUID = 67609119670762533L;

	final static private Logger logger = LoggerFactory
			.getLogger(ResultRowMapper.class);

	private final JDBCTableDesc tableDescription;

	private final SortedMap<Integer, ResultRow> rows;

	private final boolean createEntityObject;

	/** typ rezultatu */
	private final ResultType resultType;

	private boolean hasMoreResultRows = false;

	private Integer lastRowPosition = LoadJdbcConfiguration.FIRST_ROW_POSITION;

	private Integer startRowPosition = LoadJdbcConfiguration.FIRST_ROW_POSITION;

	private final PagingParams pagingParams;

	/**
	 * Lista użytych na kolumnach funkcji SQL-owych. Pierwszy parametr to nazwa
	 * aliasu uzywanego w zapytaniu, drugi to definicja użytej funkcji.
	 */
	private final Hashtable<String, _IFunction> functions;

	/**
	 * Konstruktor dla mapperów, dla których niniejszy obiekt bedzie
	 * rozszerzeniem. Przydatny np. do testowania poprawności implementacji
	 * metod tego obiektu.
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rows
	 *            lista wyników zapytania
	 * @param createEntityObject
	 *            czy ma być tworzony obiekt encji (jeżeli {@code true}, to
	 *            obiekt encji bedzie utworzony i dostępny w obiekcie
	 *            {@link ResultRow}). Obiekt może nie być tworzony w przypadku
	 *            kiesdy zapytanie składa się tylko z funkcji
	 *            {@link SelectQuery#setUseFunctionsOnly(boolean)}
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 */
	protected ResultRowMapper(JDBCTableDesc tableDescription,
			SortedMap<Integer, ResultRow> rows, boolean createEntityObject,
			Integer startPosition, PagingParams pagingParams,
			Hashtable<String, _IFunction> functions) {

		this.tableDescription = tableDescription;
		this.createEntityObject = createEntityObject;
		this.functions = functions;
		this.resultType = ResultType.RESTRICTED;

		/* USTAWIANIE PARAMETRÓW STRONICOWANIA - START */
		/* sprawdzenie parametrów wejsciowych */
		if (pagingParams == null) {
			pagingParams = LoadJdbcConfiguration.getPagingParams(null);
		}
		/* ustalenie pozycji startowej */
		if (startPosition == null) {
			startPosition = LoadJdbcConfiguration.FIRST_ROW_POSITION;
		}
		this.startRowPosition = startPosition;
		logger.trace("Startowa pozycja wiersza: {}", this.startRowPosition);

		/* ustalenie maksymalnej liczby wierszy */
		int resultMaxRowSize = pagingParams.getMaxCount()
				- pagingParams.getOffset();
		logger.trace("Maksymalna liczba wierszy: {}", resultMaxRowSize);
		switch (this.resultType) {
		case RESTRICTED:
			break;
		case UNLIMITED:
			resultMaxRowSize = ResultType.UNLIMITED_RESULT_SIZE;
			pagingParams.setMaxCount(resultMaxRowSize);
			break;
		default:
			break;
		}
		/* zapisanie parametrów do obiektu tylko do odczytu */
		pagingParams.setPage(pagingParams.getMinimalPage());
		this.pagingParams = new PagingParamsReadOnly(pagingParams);
		/* USTAWIANIE PARAMETRÓW STRONICOWANIA - KONIEC */

		int endPosition = this.startRowPosition + resultMaxRowSize
				- LoadJdbcConfiguration.FIRST_ROW_POSITION;
		this.hasMoreResultRows = (endPosition < rows.size());
		if (this.hasMoreResultRows) {
			this.lastRowPosition = endPosition;
		} else {
			this.lastRowPosition = rows.size();
		}
		logger.trace("Ostatnia pozycja wiersza: {}", this.lastRowPosition);
		this.rows = rows.subMap(startRowPosition, this.lastRowPosition + 1);

	}

	/**
	 * Właściwa implementacja konstruktoa, za pomocą którego pobierane są
	 * wiersze jakie zostaną znalezione za pomocą zapytania SQL z przedziału,
	 * gdzie podstwą zawarcia wiersza na liście jest jego numer. Wiersze są
	 * numerowane począwszy od wartości {@link #FIRST_ROW_POSITION}.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font>Tylko do użytku wewnętrznego.
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * @see ResultType
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param createEntityObject
	 *            czy ma być tworzony obiekt encji (jeżeli {@code true}, to
	 *            obiekt encji bedzie utworzony i dostępny w obiekcie
	 *            {@link ResultRow}). Obiekt może nie być tworzony w przypadku
	 *            kiesdy zapytanie składa się tylko z funkcji
	 *            {@link SelectQuery#setUseFunctionsOnly(boolean)}
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 * @param columnsToUse
	 *            niedomyślna lista kolumn użyta w zapytaniu. Niedomyślna
	 *            oznacza, że w zapytaniu biorą new wszystkie kolumny
	 *            odpowiedające polom encji.
	 * @param resultType
	 *            typ rezultatu, za pomocą którego definiuje się czy mają zostać
	 *            użyte restrykcje związane maksymalną liczbą stron wyników w
	 *            zwracanym rezultacie zapytania.
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalArgumentException
	 */
	protected ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			boolean createEntityObject, Integer startPosition,
			Integer endPosition, PagingParams pagingParams,
			Hashtable<String, _IFunction> functions, String[] columnsToUse,
			ResultType resultType) throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException, NoSuchMethodException,
			IllegalArgumentException, InvocationTargetException {

		this.tableDescription = tableDescription;
		this.rows = new TreeMap<Integer, ResultRow>();
		this.createEntityObject = createEntityObject;
		this.functions = functions;
		this.resultType = resultType;

		/* USTAWIANIE PARAMETRÓW STRONICOWANIA - START */
		/* sprawdzenie parametrów wejsciowych */
		if (pagingParams == null) {
			pagingParams = LoadJdbcConfiguration.getPagingParams(null);
		}
		/* ustalenie pozycji startowej */
		boolean rsNext = false;
		if (startPosition == null) {
			startPosition = LoadJdbcConfiguration.FIRST_ROW_POSITION;
			rsNext = rs.next();
		} else {
			rsNext = rs.absolute(startPosition);
		}
		this.startRowPosition = startPosition;
		logger.debug("Pozycja startu {}...", this.startRowPosition);

		/* ustalenie maksymalnej liczby wierszy */
		int resultMaxRowSize = pagingParams.getMaxCount()
				- pagingParams.getOffset();
		switch (this.resultType) {
		case RESTRICTED:
			break;
		case UNLIMITED:
			resultMaxRowSize = ResultType.UNLIMITED_RESULT_SIZE;
			pagingParams.setMaxCount(resultMaxRowSize);
			break;
		default:
			break;
		}
		logger.trace("Maksymalna liczba wierszy: {}", resultMaxRowSize);
		/* zapisanie parametrów do obiektu tylko do odczytu */
		PagingParams tmpPagingParams = pagingParams.copy();
		tmpPagingParams.setPage(pagingParams.getMinimalPage());
		this.pagingParams = new PagingParamsReadOnly(tmpPagingParams);
		/* USTAWIANIE PARAMETRÓW STRONICOWANIA - KONIEC */

		Class<?> clazz = this.tableDescription.getEntityClass();
		logger.debug("Klasa encji {}...", clazz.getName());
		if (logger.isDebugEnabled() && createEntityObject) {
			logger.debug("Beda dostepne obiekty encji dla klasy {}", clazz);
		}
		if (!rsNext) {
			return;
		}

		int rowcounter = LoadJdbcConfiguration.FIRST_ROW_POSITION;
		while (rsNext) {
			Hashtable<String, Object> columns = new Hashtable<String, Object>();

			String[] sortedLinks = (columnsToUse != null
					&& columnsToUse.length > 0 ? columnsToUse
					: (createEntityObject ? this.tableDescription
							.getSortedDbColumnNames()
							: ArrayUtils.EMPTY_STRING_ARRAY));

			for (int i = 0; i < sortedLinks.length; i++) {
				JDBCColumnDesc col = this.tableDescription
						.getColumnByDbColumnName(sortedLinks[i]);
				/*
				 * przepisywanie danych z kolumny rezultatu do kolumny wiersza -
				 * START
				 */
				String entityFieldName = col.getEntityFieldName();
				SimpleSQLType type = SimpleSQLType.get(col.getColumnType());
				if (type != null) {
					Object colVal = type.getValue(
							rs,
							(StringUtils.isNotBlank(col.getAlias()) ? col
									.getAlias() : col.getName()));
					if (colVal != null) {
						columns.put(entityFieldName, colVal);
					}
				} else {
					if (logger.isTraceEnabled())
						logger.warn(
								"Nie udalo sie zidentyfikowac typu SQL dla kolumny {} ({})",
								new Object[] { col.getName(),
										col.getColumnType() });
				}

				/*
				 * przepisywanie danych z kolumny rezultatu do kolumny wiersza -
				 * KONIEC
				 */
			}

			if (this.functions != null && !this.functions.isEmpty()) {
				/* Obsługa funkcji SQL-owych - START */
				Enumeration<_IFunction> functionListElements = functions
						.elements();
				while (functionListElements.hasMoreElements()) {
					_IFunction function = functionListElements.nextElement();
					SimpleSQLType type = SimpleSQLType.get(function
							.getReturnType());
					String aliasFunctionName = function.getAlias();
					if (type != null) {
						Object colVal = type.getValue(rs, aliasFunctionName);
						if (colVal != null) {
							columns.put(aliasFunctionName, colVal);
						}
					} else {
						if (logger.isTraceEnabled())
							logger.warn(
									"Nie udalo sie zidentyfikowac typu SQL dla kolumny reprezentujacej funkcyje {} ({})",
									new Object[] { aliasFunctionName,
											function.getReturnType() });
					}
				}
				/* Obsługa funkcji SQL-owych - KONIEC */
			}

			this.lastRowPosition = (startPosition + rowcounter) - 1;
			if (logger.isTraceEnabled()) {
				String firstKey = "n/a";
				String firstValue = "n/a";
				if (!columns.isEmpty()) {
					firstKey = columns.keys().nextElement();
					firstValue = columns.get(firstKey).toString();
				}
				logger.trace("Wiersz nr {}: [{} = {}]", new Object[] {
						this.lastRowPosition, firstKey, firstValue });
			}

			this.rows.put(rowcounter, new ResultRow(tableDescription,
					rowcounter, columns, createEntityObject));
			rsNext = rs.next();
			rowcounter++;
			if (endPosition != null && rowcounter > endPosition) {
				break;
			}
			if (rowcounter > resultMaxRowSize) {
				this.hasMoreResultRows = true;
				if (logger.isTraceEnabled())
					logger.warn(
							"Przekroczono maksymalna liczbe zwracanych wynikow: {}",
							resultMaxRowSize);
				break;
			}
		}

	}

	/**
	 * Konstruktor za pomocą którego pobierane są wiersze jakie zostaną
	 * znalezione za pomocą zapytania SQL z przedziału, gdzie podstwą zawarcia
	 * wiersza na liście jest jego numer. Wiersze są numerowane począwszy od
	 * wartości {@link #FIRST_ROW_POSITION}.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font> Liczba wierszy wyniku podlega
	 * restrykcjom związanym z maksymalną liczbą wierszy w wyniku
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}.
	 * </p>
	 * 
	 * Gdy użyjemy tego konstruktora zmieni się znaczenie i działanie metod:
	 * <ol>
	 * <li>
	 * {@link #getRows()}: metoda nie bedzie zwracać wszystkich wierszy tylko
	 * listę wierszy o numerach większych lub równych parametrowi
	 * 'startPosition' i mniejszych lub równych parametrowi 'endPosition'</li>
	 * <li>
	 * {@link #getRows(Integer, Integer)}: dla tej metody początkowy zbór
	 * wyników bedzie już wcześniej ograniczony do listy wierszy o numerach
	 * większych lub równych parametrowi 'startPosition' i mniejszych lub
	 * równych parametrowi 'endPosition'</li>
	 * </ol>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param createEntityObject
	 *            czy ma być tworzony obiekt encji (jeżeli {@code true}, to
	 *            obiekt encji bedzie utworzony i dostępny w obiekcie
	 *            {@link ResultRow}). Obiekt może nie być tworzony w przypadku
	 *            kiesdy zapytanie składa się tylko z funkcji
	 *            {@link SelectQuery#setUseFunctionsOnly(boolean)}
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 * @param columnsToUse
	 *            niedomyślna lista kolumn użyta w zapytaniu. Niedomyślna
	 *            oznacza, że w zapytaniu biorą new wszystkie kolumny
	 *            odpowiedające polom encji.
	 * 
	 * 
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalArgumentException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			boolean createEntityObject, Integer startPosition,
			Integer endPosition, PagingParams pagingParams,
			Hashtable<String, _IFunction> functions, String[] columnsToUse)
			throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException, NoSuchMethodException,
			IllegalArgumentException, InvocationTargetException {
		this(tableDescription, rs, createEntityObject, startPosition,
				endPosition, pagingParams, functions, columnsToUse,
				ResultType.RESTRICTED);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wiersze jakie zostaną
	 * znalezione za pomocą zapytania SQL z przedziału, gdzie podstwą zawarcia
	 * wiersza na liście jest jego numer, bez genarcji obiektów encji.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Obiekty encji nie zostaną
	 * wygenerowane, ponieważ pole mapper-a {@link #createEntityObject} zostanie
	 * ustawione na wartość {@code false}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Użycie tego konstruktora
	 * implikuje brak obsługi wstawionych do klauzuli zapytania funkcji
	 * SQL-owych
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * @see #JDBCRowMapper(JDBCTableDesc, ResultSet, boolean, Integer, Integer)
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * 
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalArgumentException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			Integer startPosition, Integer endPosition,
			PagingParams pagingParams) throws SQLException,
			InstantiationException, IllegalAccessException, SecurityException,
			NoSuchMethodException, IllegalArgumentException,
			InvocationTargetException {
		this(tableDescription, rs, false, startPosition, endPosition,
				pagingParams, null, null);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wszystkie wiersze, jakie
	 * zostaną znalezione za pomocą zapytania SQL.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Liczba wierszy, czyli rozmiar
	 * wyniku, zostanie ograniczony do maksymalnej możliwej liczby wierszy
	 * zdefiniowanej przez pomnożenie parametru
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES} i
	 * {@link JdbcConfig#DEFAULT_PAGE_SIZE}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Użycie tego konstruktora
	 * implikuje brak obsługi wstawionych do klauzuli zapytania funkcji
	 * SQL-owych
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param createEntityObject
	 *            czy ma być tworzony obiekt encji (jeżeli {@code true}, to
	 *            obiekt encji bedzie utworzony i dostępny w obiekcie
	 *            {@link ResultRow}). Obiekt może nie być tworzony w przypadku
	 *            kiesdy zapytanie składa się tylko z funkcji
	 *            {@link SelectQuery#setUseFunctionsOnly(boolean)}
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			boolean createEntityObject) throws SQLException,
			InstantiationException, IllegalAccessException, SecurityException,
			IllegalArgumentException, NoSuchMethodException,
			InvocationTargetException {
		this(tableDescription, rs, createEntityObject, null, null, null, null,
				null);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wszystkie wiersze, jakie
	 * zostaną znalezione za pomocą zapytania SQL, bez genarcji obiektów encji.
	 * 
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Liczba wierszy, czyli rozmiar
	 * wyniku, zostanie ograniczony do maksymalnej możliwej liczby wierszy
	 * zdefiniowanej przez pomnożenie parametru
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES} i
	 * {@link JdbcConfig#DEFAULT_PAGE_SIZE}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Obiekty encji nie zostaną
	 * wygenerowane, ponieważ pole mapper-a {@link #createEntityObject} zostanie
	 * ustawione na wartość {@code false}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Użycie tego konstruktora
	 * implikuje brak obsługi wstawionych do klauzuli zapytania funkcji
	 * SQL-owych
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * @see #JDBCRowMapper(JDBCTableDesc, ResultSet, boolean)
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs)
			throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException,
			IllegalArgumentException, NoSuchMethodException,
			InvocationTargetException {
		this(tableDescription, rs, false, null, null, null, null, null);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wiersze jakie zostaną
	 * znalezione za pomocą zapytania SQL z przedziału, gdzie podstwą zawarcia
	 * wiersza na liście jest jego numer, bez genarcji obiektów encji, z obsługą
	 * funkocji SQL-owych.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Obiekty encji nie zostaną
	 * wygenerowane, ponieważ pole mapper-a {@link #createEntityObject} zostanie
	 * ustawione na wartość {@code false}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Użycie tego konstruktora
	 * implikuje brak obsługi wstawionych do klauzuli zapytania funkcji
	 * SQL-owych
	 * </p>
	 * 
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * @see #JDBCRowMapper(JDBCTableDesc, ResultSet, boolean, Integer, Integer)
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 * @param columnsToUse
	 *            niedomyślna lista kolumn użyta w zapytaniu. Niedomyślna
	 *            oznacza, że w zapytaniu biorą new wszystkie kolumny
	 *            odpowiedające polom encji.
	 * 
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalArgumentException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			Integer startPosition, Integer endPosition,
			PagingParams pagingParams, Hashtable<String, _IFunction> functions,
			String[] columnsToUse) throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException, NoSuchMethodException,
			IllegalArgumentException, InvocationTargetException {
		this(tableDescription, rs, false, startPosition, endPosition,
				pagingParams, functions, columnsToUse);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wszystkie wiersze, jakie
	 * zostaną znalezione za pomocą zapytania SQL z obsługą funkcji SQL-owych.
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Liczba wierszy, czyli rozmiar
	 * wyniku, zostanie ograniczony do maksymalnej możliwej liczby wierszy
	 * zdefiniowanej przez pomnożenie parametru
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES} i
	 * {@link JdbcConfig#DEFAULT_PAGE_SIZE}.
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param createEntityObject
	 *            czy ma być tworzony obiekt encji (jeżeli {@code true}, to
	 *            obiekt encji bedzie utworzony i dostępny w obiekcie
	 *            {@link ResultRow}). Obiekt może nie być tworzony w przypadku
	 *            kiesdy zapytanie składa się tylko z funkcji
	 *            {@link SelectQuery#setUseFunctionsOnly(boolean)}
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 * @param columnsToUse
	 *            niedomyślna lista kolumn użyta w zapytaniu. Niedomyślna
	 *            oznacza, że w zapytaniu biorą new wszystkie kolumny
	 *            odpowiedające polom encji.
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			boolean createEntityObject,
			Hashtable<String, _IFunction> functions, String[] columnsToUse)
			throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException,
			IllegalArgumentException, NoSuchMethodException,
			InvocationTargetException {
		this(tableDescription, rs, createEntityObject, null, null, null,
				functions, columnsToUse);
	}

	/**
	 * Konstruktor za pomocą którego pobierane są wszystkie wiersze, jakie
	 * zostaną znalezione za pomocą zapytania SQL, bez genarcji obiektów encji,
	 * z obsługą funkocji SQL-owych.
	 * 
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Liczba wierszy, czyli rozmiar
	 * wyniku, zostanie ograniczony do maksymalnej możliwej liczby wierszy
	 * zdefiniowanej przez pomnożenie parametru
	 * {@link LoadJdbcConfiguration#RESULT_MAX_PAGES} i
	 * {@link JdbcConfig#DEFAULT_PAGE_SIZE}.
	 * </p>
	 * <p>
	 * <font color="#aa0000">Uwaga!</font></br> Obiekty encji nie zostaną
	 * wygenerowane, ponieważ pole mapper-a {@link #createEntityObject} zostanie
	 * ustawione na wartość {@code false}.
	 * </p>
	 * 
	 * @see LoadJdbcConfiguration#RESULT_MAX_PAGES
	 * @see #JDBCRowMapper(JDBCTableDesc, ResultSet, boolean)
	 * 
	 * @param tableDescription
	 *            opis tabeli
	 * @param rs
	 *            wynik zapytania
	 * @param functions
	 *            tablica użytych w zapytaniu funkcji. Wyniki użytuch funkcji
	 *            dostepne będą w obiekcie {@link ResultRow}) (za pomocą metod
	 *            {@link ResultRow#getColumnsKeys()} oraz
	 *            {@link ResultRow#getColumnValue(String)}).
	 * @param columnsToUse
	 *            niedomyślna lista kolumn użyta w zapytaniu. Niedomyślna
	 *            oznacza, że w zapytaniu biorą new wszystkie kolumny
	 *            odpowiedające polom encji.
	 * 
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public ResultRowMapper(JDBCTableDesc tableDescription, ResultSet rs,
			Hashtable<String, _IFunction> functions, String[] columnsToUse)
			throws SQLException, InstantiationException,
			IllegalAccessException, SecurityException,
			IllegalArgumentException, NoSuchMethodException,
			InvocationTargetException {
		this(tableDescription, rs, false, null, null, null, functions,
				columnsToUse);
	}

	public boolean isEmpty() {
		return rows.isEmpty();
	}

	public int size() {
		return this.pagingParams.getOffset() + rows.size();
	}

	/**
	 * @return definicja tabeli
	 */
	public JDBCTableDesc getTableDescription() {
		return tableDescription;
	}

	/**
	 * @return lista definicji kolumn
	 */
	public List<JDBCColumnDesc> getColumns() {
		List<JDBCColumnDesc> columns = new ArrayList<JDBCColumnDesc>();
		Enumeration<String> columnKeys = this.tableDescription.getColumnKeys();
		while (columnKeys.hasMoreElements()) {
			columns.add(this.tableDescription.getColumn(columnKeys
					.nextElement()));
		}
		return columns;
	}

	/**
	 * @return cała lista wierszy
	 */
	public Collection<ResultRow> getRows() {
		return rows.values();
	}

	/**
	 * Pobieranie przedziału wierszy, gdzie podstwą zawarcia wiersza na liście
	 * jest jego numer. Wiersze są numerowane począwszy od wartości
	 * {@link #FIRST_ROW_POSITION}.
	 * 
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @return lista wierszy o numerach większych lub równych parametrowi
	 *         'startPosition' (wraz z 'startPosition') i mniejszych lub równych
	 *         parametrowi 'endPosition' (wraz z 'endPosition')
	 */
	public Collection<ResultRow> getRows(Integer startPosition,
			Integer endPosition) {
		startPosition = setStartPosition(startPosition);
		endPosition = setEndPosition(endPosition);
		SortedMap<Integer, ResultRow> subMap = rows.subMap(startPosition
				- this.pagingParams.getOffset(), endPosition
				- this.pagingParams.getOffset() + 1);
		return subMap.values();
	}

	/**
	 * Pobieranie wartości listy dla wiersza o konkretnym numerze
	 * 
	 * @param position
	 *            numer wiersza
	 * @return obiekt wiersza zajmującego daną pozycję
	 */
	public ResultRow getRow(Integer position) {
		return rows.get(position - this.pagingParams.getOffset());
	}

	/**
	 * @return informacja o tym czy obiekty encji zostały utworzone
	 */
	public boolean isCreatedEntityObject() {
		return createEntityObject;
	}

	/**
	 * @return the resultSize
	 */
	public int getResultSize() {
		return this.pagingParams.getOffset() + rows.size();
	}

	/**
	 * Otworzenie obiektu rezultatu (wyniku) zapytania
	 * 
	 * @param queryInfo
	 *            informacje o zapytaniu
	 * @param startPosition
	 *            pozycja pierwszego wiersza
	 * @param endPosition
	 *            pozycja ostatniego wiersza
	 * @return obiekt rezultatu (wyniku) zapytania
	 */
	public QueryResult prepareQueryResult(SelectQuery.QueryCacheInfo queryInfo,
			Integer startPosition, Integer endPosition) {

		logger.trace("Tworze obiekt rezultatu");
		int resultSize = this.getResultSize();
		logger.trace("Rozmiar rezultatu: {}", resultSize);
		if (resultSize > 0) {
			logger.trace("startPosition: {}, endPosition: {}", new Object[] {
					startPosition, endPosition });
			logger.trace("this.startRowPosition: {}, this.lastRowPosition: {}",
					new Object[] { this.lastRowPosition, this.lastRowPosition });
			int start = setStartPosition(startPosition);
			int end = setEndPosition(endPosition);
			logger.trace("start: {}, end: {}", new Object[] { start, end });
			if (resultSize == ((end - start) + 1)) {
				logger.trace("Pobieramm caly zbior rezultatu: ", this.getRows()
						.size());
				QueryResult out = new QueryResult(queryInfo, start, end,
						this.startRowPosition, resultSize,
						this.lastRowPosition, this.getRows(),
						(hasMoreResultRows ? ResultMessage.FRAGMENT
								: ResultMessage.ALL), this.pagingParams);
				out.setHasMoreResultRows(hasMoreResultRows);
				return out;
			} else {
				logger.trace("Rozmiar calego zbioru rezultatu: {}", this
						.getRows().size());
				logger.trace(
						"Pobieramm fragment zbioru rezultatu (od {} do {}): {}",
						new Object[] { start, end,
								this.getRows(start, end).size() });
				QueryResult out = new QueryResult(queryInfo, start, end,
						this.startRowPosition, resultSize,
						this.lastRowPosition, this.getRows(start, end),
						ResultMessage.FRAGMENT, this.pagingParams);
				out.setHasMoreResultRows(hasMoreResultRows);
				return out;
			}
		} else {
			return new QueryResult(queryInfo, 0, 0, 0, 0, 0, null,
					ResultMessage.NO_DATA_FOUND, this.pagingParams);
		}

	}

	/**
	 * @param endPosition
	 * @return
	 */
	private int setEndPosition(Integer endPosition) {
		int end = (endPosition != null && endPosition < this.lastRowPosition ? endPosition
				: this.lastRowPosition);
		return end;
	}

	/**
	 * @param startPosition
	 * @return
	 */
	private int setStartPosition(Integer startPosition) {
		int start = (startPosition != null
				&& startPosition > this.startRowPosition ? startPosition
				: this.startRowPosition);
		return start;
	}

	/**
	 * Nieraz bywa, że liczba wyników przekracza zdefiniowaną granicę
	 * maksymalnej liczby wierszy {@link #RESULT_ROW_MAX_SIZE}. Gdy to nastąpi
	 * jest generowane ostrzeżenie do loga, a jednocześnie ustawiana jest flaga
	 * {@link #hasMoreResultRows}, której wartość można pobrać za pomocą tej
	 * metody.
	 * 
	 * @return {@code true} jeżeli maksymalna liczba wyników została
	 *         przekroczona.
	 */
	public boolean hasMoreResultRows() {
		return hasMoreResultRows;
	}

	/**
	 * @return zwraca pozycje ostatniego wiersza, może to być pozycja startowa
	 *         do pobrania kolejnej paczki wyników gdy zostanie przekroczona
	 *         granica maksymalnej liczby wierszy w wyniku zapytania.
	 */
	public Integer getLastRowPosition() {
		return lastRowPosition;
	}

	/**
	 * @return the resultType
	 */
	public ResultType getResultType() {
		return resultType;
	}

	/**
	 * @return the pagingParams
	 */
	public PagingParams getPagingParams() {
		return pagingParams;
	}

}
