package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.sql.DataSource;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache._IObjectCache;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheManager;
import pl.slawas.scjdbc4jpa.jdbc.cache.managers.QueryCacheType;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.CreateSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ExecuteSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.IdIsNotEmbeddableException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectIsNotEntityException;
import pl.slawas.scjdbc4jpa.jdbc.sort.SortCondition;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;
import pl.slawas.scjdbc4jpa.jdbc.where.BindParam;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;

/**
 * 
 * SelectQuery - przygotowanie i egzekucja zapytania SQL
 * 
 * @see ResultRowMapper
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.13 $
 * 
 */
public class SelectQuery implements Serializable {

	private static final long serialVersionUID = -8293633552304794460L;

	final static private Logger logger = LoggerFactory
			.getLogger(SelectQuery.class);

	private String[] columnsToUse = null;

	private String comment;

	private SelectAfterClause afterSelectClause;

	protected final JDBCTableDesc table;

	protected PagingParams pagingParams;

	private WhereClause whereClause;

	private SortClause sortClause;

	private GroupByClause groupByClause;

	private boolean generateEntityObject = true;

	private boolean useFunctionsOnly = false;

	private boolean queryClauseDefinitionIsChanged = false;

	protected QueryClause queryClause;

	protected QueryCacheManager queryCacheManager;

	/**
	 * Lista użytych na kolumnach funkcji SQL-owych. Pierwszy parametr to nazwa
	 * aliasu uzywanego w zapytaniu, drugi to definicja użytej funkcji.
	 */
	private final Hashtable<String, _IFunction> functions = new Hashtable<String, _IFunction>();

	/**
	 * Zostanie utworzony nowy obiekt managera kesza zapytan. Nazwa regionu
	 * użytego kesza zostanie odziedziczona po opisie tabeli (
	 * {@link JDBCTableDesc#getCacheRegionName()}) i dodany zostanie suffix
	 * {@link QueryCacheType#OVER} zgodnie z zasadami w implementacji metody
	 * {@link QueryCacheManager#createQueryRegionName(JDBCTableDesc, QueryCacheType)}
	 * .
	 * 
	 * @param cacheUsage
	 *            informacja o tym czy ma być używany kesz. Jako, że jest to
	 *            inicjalizacjia obiektu, to parametr powinien przyjmować jedną
	 *            z dwóch wartości: {@link CacheUsage#TO_USE} oraz
	 *            {@link CacheUsage#NONE}
	 * @param table
	 *            opis tablicy dla której bedzie budowane zapytanie
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 */
	public SelectQuery(CacheUsage cacheUsage, JDBCTableDesc table,
			PagingParams pagingParams) {
		this(new QueryCacheManager(cacheUsage, table, QueryCacheType.OVER),
				table, pagingParams);
	}

	/**
	 * 
	 * @param queryCacheManager
	 *            manager kesza zapytań
	 * @param table
	 *            opis tablicy dla której bedzie budowane zapytanie
	 * @param pagingParams
	 *            Parametry stronicowania. Jeżeli zostanie podany jako wartość
	 *            {@code null}, to wtedy zostanie przyjęty domyślny rozmiar
	 *            strony zdefiniowany w stałej {@link Page#DEFAULT_PAGE_SIZE}
	 *            oraz maksymalna liczba stron
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}
	 */
	protected SelectQuery(QueryCacheManager queryCacheManager,
			JDBCTableDesc table, PagingParams pagingParams) {
		this.queryCacheManager = queryCacheManager;
		this.table = table;
		this.pagingParams = (pagingParams == null ? LoadJdbcConfiguration
				.getPagingParams(null) : pagingParams);
	}

	/**
	 * @return the columnsToUse lista nazw pól kolumn (
	 *         {@link JDBCColumnDesc#getName()}), które mają być użyte w
	 *         zapytaniu
	 */
	public String[] getColumnsToUse() {
		return columnsToUse;
	}

	/**
	 * Ustawienie listy kolumn jakie mają zostać użyte w zapytaniu. Lista to
	 * bazodanowe nazwy kolumn.
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @see JDBCColumnDesc#getName()
	 * @see JDBCTableDesc#getDbColumnNames()
	 * 
	 * @param columnsToUse
	 *            lista nazw pól kolumn ({@link JDBCColumnDesc#getName()})
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setColumnsToUse(String[] columnsToUse) {
		this.columnsToUse = columnsToUse;
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * Ustawienie listy kolumn jakie mają zostać użyte w zapytaniu. Lista to
	 * bazodanowe nazwy kolumn.
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @see JDBCColumnDesc#getName()
	 * @see JDBCTableDesc#getDbColumnNames()
	 * 
	 * @param columnsToUseList
	 *            lista nazw kolumn ({@link JDBCColumnDesc#getName()}).
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setColumnsToUse(Set<String> columnsToUseList) {
		columnsToUse = new String[columnsToUseList.size()];
		int i = 0;
		for (String usedColumnKey : columnsToUseList) {
			columnsToUse[i] = usedColumnKey;
			i++;
		}
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * @return the comment
	 */
	public String getComment() {
		return comment;
	}

	/**
	 * Ustawianie komnetarza zapytania. Za pomocą komntarzy możemy również
	 * dodawać wskazówki dla optymalizatora przygotowującego p[lan zapytania.
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param comment
	 *            komentarz
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setComment(String comment) {
		this.comment = comment;
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * @return the afterSelectClause
	 */
	public SelectAfterClause getAfterSelectClause() {
		return afterSelectClause;
	}

	/**
	 * Ustawianie klauzuli występującej po słowie kluczowym {@code SELECT}.
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @see SelectAfterClause
	 * 
	 * @param afterSelectClause
	 *            obiekt klauzuli
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setAfterSelectClause(SelectAfterClause afterSelectClause) {
		this.afterSelectClause = afterSelectClause;
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * @return the where
	 */
	public WhereClause getWhereClause() {
		return whereClause;
	}

	/**
	 * Ustawienie klauzuli 'WHERE' w zapytaniu (warunków zapytania).
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param where
	 *            obiekt klauzuli 'WHERE'
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setWhereClause(WhereClause where) {
		this.whereClause = where;
		return this;
	}

	/**
	 * @return opis tabeli, na podstawie którego bedzie tworzone zapytanie.
	 */
	public JDBCTableDesc getTable() {
		return table;
	}

	/**
	 * Metoda tworząca klauzulę zapytania. Jej wywołanie spowoduje utworzenie
	 * klauzuli, która jest tworzona na podstawie parametrów:
	 * <ol>
	 * <li>{@link #getAfterSelectClause()}</li>
	 * <li>{@link #getComment()}</li>
	 * <li>{@link #getColumnsToUse()}</li>
	 * <li>{@link #getWhereClause()}</li>
	 * <li>{@link #getGroupByClause()}</li>
	 * <li>{@link #getSortClause()}</li>
	 * <li>{@link #functions}</li>
	 * <li>{@link #isUseFunctionsOnly()}</li>
	 * </ol>
	 * 
	 * @return obiekt własnie tworzonego zapytania
	 * @throws ObjectIsNotEntityException
	 * @throws JDBCEntityParserException
	 * @throws IdIsNotEmbeddableException
	 * @throws CreateSQLQueryException
	 */
	public SelectQuery create() throws ObjectIsNotEntityException,
			JDBCEntityParserException, IdIsNotEmbeddableException,
			CreateSQLQueryException {

		if (this.queryClause == null || this.queryClauseDefinitionIsChanged) {
			/** tworzenie nowego zapyta nia bazowego - START */
			StringBuffer queryClause = new StringBuffer("SELECT ");

			if (this.afterSelectClause != null) {
				queryClause.append(this.afterSelectClause.toString() + " ");
			}

			if (StringUtils.isNotBlank(this.comment)
					&& this.comment.startsWith("/*")
					&& this.comment.endsWith("*/")) {
				queryClause.append(this.comment + " ");
			}
			boolean funcitonsAreDefined = false;
			/* dodaje funkcje do klauzuli */
			if (functions != null && !functions.isEmpty()) {
				Enumeration<_IFunction> functionList = functions.elements();
				int functionCounter = 0;
				while (functionList.hasMoreElements()) {
					_IFunction function = functionList.nextElement();
					String columnClause = "";
					columnClause = (functionCounter != 0 ? "," : " ")
							+ function.getClause() + " as "
							+ function.getAlias();
					queryClause.append("\n  " + columnClause);
					logger.trace("Dodalem klauzule funkcji: {}", columnClause);
					functionCounter++;
				}
				funcitonsAreDefined = (functionCounter != 0);
			}

			/* dodaje kolumny do klauzuli */
			String[] sortedLinks = (columnsToUse != null
					&& columnsToUse.length > 0 ? columnsToUse
					: (funcitonsAreDefined && this.isUseFunctionsOnly() ? ArrayUtils.EMPTY_STRING_ARRAY
							: table.getSortedDbColumnNames()));
			for (int i = 0; i < sortedLinks.length; i++) {
				JDBCColumnDesc column = table
						.getColumnByDbColumnName(sortedLinks[i]);
				String columnClause = "";
				String prefix = (i != 0 || funcitonsAreDefined ? "," : " ");
				switch (column.getSource()) {
				case FOREIGN:
					columnClause = prefix
							+ column.getName()
							+ (StringUtils.isNotBlank(column.getAlias()) ? " as "
									+ column.getAlias()
									: "");
					queryClause.append("\n  " + columnClause);
					break;
				case MAIN:
					columnClause = prefix
							+ (StringUtils.isNotBlank(column.getTable()
									.getAlias()) ? column.getTable().getAlias()
									+ "." : "")
							+ column.getName()
							+ (StringUtils.isNotBlank(column.getAlias()) ? " as "
									+ column.getAlias()
									: "");
					queryClause.append("\n  " + columnClause);
					break;
				}
				logger.trace("Dodalem klauzule kolumny: {}", columnClause);
			}

			/* dodaje tabele */
			queryClause.append("\n FROM "
					+ table.getName()
					+ (StringUtils.isNotBlank(table.getAlias()) ? " "
							+ table.getAlias() : ""));

			List<JDBCKeyDesc> foreignKeys = table.getSortedForeignKeys();
			if (!foreignKeys.isEmpty())
				for (JDBCKeyDesc foreignKey : foreignKeys) {
					JDBCTableDesc joinedTable = table
							.getJoinedTable(foreignKey);
					StringBuffer onClause = new StringBuffer();
					/* dodaje JOINY */
					for (int position = 0; position < foreignKey.getColumns()
							.size(); position++) {
						JDBCColumnDesc mainColumn = foreignKey
								.getColumn(position);
						JDBCColumnDesc dbReferencedColumn = joinedTable
								.getPrimaryKey().getColumn(position);
						String onClauseStr = "";
						switch (foreignKey.getType()) {
						case FK:
							onClauseStr = table.getAlias() + "."
									+ mainColumn.getName() + " = "
									+ joinedTable.getAlias() + "."
									+ dbReferencedColumn.getName();
							break;
						case FFK:
							onClauseStr = mainColumn.getName() + " = "
									+ joinedTable.getAlias() + "."
									+ dbReferencedColumn.getName();
							break;
						default:
							break;
						}
						onClause.append((position != 0 ? " AND " : "")
								+ onClauseStr);
						logger.trace("Dodalem klauzule on: {}", onClauseStr);
					}

					queryClause.append("\n  LEFT JOIN " + joinedTable.getName()
							+ " " + joinedTable.getAlias() + " ON ("
							+ onClause.toString() + ")");
				}
			this.queryClause = new QueryClause(queryClause.toString());
			this.queryClauseDefinitionIsChanged = false;
			/** tworzenie nowego zapyta nia bazowego - KONIEC */
		}

		if (this.queryClause == null)
			throw new CreateSQLQueryException(
					"Niepowodzenie utworzenia zapytania podstawowego");

		/* dodanie klauzuli WHERE */
		Vector<BindParam> bindParams = null;
		boolean extendClauseIsAdded = false;
		if (whereClause != null) {
			this.queryClause.addExtendClause("\n " + whereClause.getClause());
			bindParams = whereClause.getBindParams();
			extendClauseIsAdded = true;
		}

		/* dodanie klauzuli GROUP BY */
		if (groupByClause != null) {
			if (extendClauseIsAdded) {
				this.queryClause.addNextExtendClause("\n "
						+ groupByClause.getClause());
			} else {
				this.queryClause.addExtendClause("\n "
						+ groupByClause.getClause());
				extendClauseIsAdded = true;
			}
		}

		/* dodanie klauzuli ORDER BY */
		if (sortClause != null) {
			if (extendClauseIsAdded) {
				this.queryClause.addNextExtendClause("\n "
						+ sortClause.getClause());
			} else {
				this.queryClause
						.addExtendClause("\n " + sortClause.getClause());
				extendClauseIsAdded = true;
			}
		}

		if (bindParams != null && !bindParams.isEmpty()) {
			this.queryClause.clearBindedParams();
			for (BindParam bindParam : bindParams) {
				this.queryClause.addBindParam(bindParam);
			}
		}
		return this;
	}

	/**
	 * Metoda właściwa wykonania zapytania do bazy danych
	 * 
	 * @param conn
	 *            aktuwne połączenie bazy danych
	 * @param cacheUsage
	 *            flaga użycia kesza
	 * @param queryId
	 *            identyfikator zapytania
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param resultType
	 *            typ rezultatu {@link ResultType}. Jeżeli parametr przyjmuje
	 *            wartość {@link ResultType#UNLIMITED}, to parametr użycia kesza
	 *            nie ma znaczenia, bo taki wynik nie bedzie keszowany.
	 * @return mapowany wynik zapytania
	 * @throws ExecuteSQLQueryException
	 */
	protected ResultRowMapper executeQuery(Connection conn,
			CacheUsage cacheUsage, String queryId, Integer startPosition,
			Integer endPosition, ResultType resultType)
			throws ExecuteSQLQueryException {

		ResultRowMapper rowMapper = null;
		logger.trace("Wykonywanie zapytania z dyrektywa uzycia kesza: {}",
				cacheUsage);
		logger.debug("KESZ DOSTEPNY?: {}", this.cacheIsAllowed());

		/* czy rezultat ma byc dodany kesz-a */
		boolean addResultRowMapperToCache = false;
		/* czy ma byc wykonane zapytanie */
		boolean executeQuery = true;

		switch (resultType) {
		case UNLIMITED:
			rowMapper = executeQuery(conn, startPosition, endPosition,
					ResultType.UNLIMITED);
			break;
		case RESTRICTED:
		default:
			/* zapytanie z resytykcjami co do rozmiaru tezultatu */
			switch (cacheUsage) {
			case NONE:
				break;
			case REFRESH:
				addResultRowMapperToCache = this.cacheIsAllowed();
				cacheUsage = CacheUsage.TO_USE;
				break;
			case TO_USE:
				if (this.cacheIsAllowed()) {
					try {
						rowMapper = (ResultRowMapper) this.queryCacheManager
								.getCache().get(queryId);
						if (rowMapper != null) {
							logger.trace(
									"Wynik zapytania pobrano z kesza (Region: {})",
									this.queryCacheManager.getCacheRegionName());
							executeQuery = false;
						} else {
							addResultRowMapperToCache = true;
						}
					} catch (ObjectCacheErrorException e) {
						logger.warn("{}", e.getMessage());
					}
				}
				break;
			default:
				break;
			}
			if (executeQuery) {
				rowMapper = executeQuery(conn, startPosition, endPosition,
						ResultType.RESTRICTED);
			}

			/* dodanie obiektu mapowania wierszy do kesz-a */
			if (rowMapper.getResultSize() > 0 && addResultRowMapperToCache) {
				try {
					this.queryCacheManager.getCache().put(queryId, rowMapper);
					logger.trace(
							"Wynik zapytania dodano do kesza (Region: {})",
							new Object[] { this.queryCacheManager
									.getCacheRegionName() });
				} catch (ObjectCacheErrorException e) {
					logger.warn("{}", e.getMessage());
				}
			}
			break;
		}

		return rowMapper;

	}

	protected boolean cacheIsAllowed() {
		return this.queryCacheManager != null
				&& this.queryCacheManager.cacheIsAllowed();
	}

	/**
	 * Metoda wykonania zapytania do bazy danych
	 * 
	 * @param conn
	 *            aktywne połączenie do bazy danych
	 * @param startPosition
	 *            pozycja piuerwszego wiersza
	 * @param endPosition
	 *            pozycja ostatniego wiersza
	 * @param resultType
	 *            od tego parametru zalezy jaki obiekt zostanie utworzony:
	 *            {@link ResultRowMapper} czy {@link DumpRowMapper};
	 * @return mapowany wynik zapytania
	 * @throws ExecuteSQLQueryException
	 */
	private ResultRowMapper executeQuery(Connection conn,
			Integer startPosition, Integer endPosition, ResultType resultType)
			throws ExecuteSQLQueryException {

		ResultSet rs = null;
		Statement stmt = null;
		PreparedStatement pstmt = null;
		ResultRowMapper rowMapper = null;

		logger.trace("Zapytanie do bazy danych: {}", queryClause.toString());
		try {

			if (queryClause.getBindedParams().isEmpty()) {
				stmt = conn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				rs = stmt.executeQuery(queryClause.getClause());
				switch (resultType) {
				case UNLIMITED:
					rowMapper = new DumpRowMapper(table, rs,
							this.isGenerateEntityObject(), startPosition,
							endPosition, this.pagingParams, functions,
							columnsToUse);
					break;
				case RESTRICTED:
					rowMapper = new ResultRowMapper(table, rs,
							this.isGenerateEntityObject(), startPosition,
							endPosition, this.pagingParams, functions,
							columnsToUse);
					break;
				}
			} else {
				pstmt = conn.prepareStatement(queryClause.getClause(),
						ResultSet.TYPE_SCROLL_INSENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				for (BindParam bindParam : queryClause.getBindedParams()) {
					bindParam.getType().bindValue(pstmt,
							bindParam.getParameterIndex(),
							bindParam.getParamValue());
				}
				rs = pstmt.executeQuery();
				switch (resultType) {
				case UNLIMITED:
					rowMapper = new DumpRowMapper(table, rs,
							this.isGenerateEntityObject(), startPosition,
							endPosition, this.pagingParams, functions,
							columnsToUse);
					break;
				case RESTRICTED:
					rowMapper = new ResultRowMapper(table, rs,
							this.isGenerateEntityObject(), startPosition,
							endPosition, this.pagingParams, functions,
							columnsToUse);
					break;
				}
			}
		} catch (SQLException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (SecurityException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (IllegalArgumentException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (InstantiationException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (IllegalAccessException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (NoSuchMethodException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (InvocationTargetException e) {
			throw new ExecuteSQLQueryException(e);
		} finally {
			if (pstmt != null)
				try {
					pstmt.close();
				} catch (SQLException e) {
				}
			if (stmt != null) {
				try {
					stmt.close();
				} catch (SQLException e) {
				}
			}
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException e) {
				}
			}
		}
		return rowMapper;
	}

	/**
	 * Uruchomienie (realizacja) zapytania na podstawie obiektu połaczenia do
	 * źródła danych. Metoda nie zamyka połączenia po wykonaniu zapytania.
	 * 
	 * @param conn
	 *            obiekt połączenia
	 * @return wynik zapytania
	 * @throws ExecuteSQLQueryException
	 */
	public QueryResult execute(Connection conn, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		return execute(conn, null, null, cacheUsage);
	}

	/**
	 * Uruchomienie (realizacja) zapytania na podstawie obiektu połaczenia do
	 * źródła danych. Metoda nie zamyka połączenia po wykonaniu zapytania.
	 * 
	 * @param conn
	 *            obiekt połączenia
	 * @param startPosition
	 *            początkowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od pierwszego wiersza.
	 * @param endPosition
	 *            końcowy numer wiersza. Jeżeli podamy wartość {@code null},
	 *            lista bedzie tworzona od ostatniego wiersza.
	 * @param cacheUsage
	 *            informacja w jaki sposób czy ma być użyty kesz wyników zapytań
	 *            {@link CacheUsage}
	 * @return wynik zapytania (lista wierszy od 'startPosition' do
	 *         'endPosition')
	 * @throws ExecuteSQLQueryException
	 */
	public QueryResult execute(Connection conn, Integer startPosition,
			Integer endPosition, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		if (this.queryClause == null
				|| StringUtils.isBlank(this.queryClause.getClause()))
			throw new ExecuteSQLQueryException(
					"Klauzula zapytania jest null. "
							+ "Sprawdz czy wczesniej uruchomiles z sukcesem metode SelectQuery.create().");

		String queryId = this.getQueryId();
		ResultRowMapper rowMapper = executeQuery(conn, cacheUsage, queryId,
				startPosition, endPosition, ResultType.RESTRICTED);

		/* Utworzenie obiektu wyniku zapytania */
		return rowMapper.prepareQueryResult(
				createQueryCacheInfo(cacheUsage, queryId), startPosition,
				endPosition);

	}

	/**
	 * Utworzenie obiektu informacji o zapytaniu
	 * 
	 * @param cacheUsage
	 *            użycie kesz-a
	 * @param queryId
	 *            identyfikator zapytania
	 * @return obiekt informacji o zapytaniu
	 */
	protected QueryCacheInfo createQueryCacheInfo(CacheUsage cacheUsage,
			String queryId) {
		return new QueryCacheInfo(queryId, cacheUsage,
				this.queryCacheManager.getCache());
	}

	/**
	 * Uruchomienie (realizacja) zapytania na podstawie obiektu połaczenia do
	 * źródła danych. Metoda nie zamyka połączenia po wykonaniu zapytania.
	 * Metoda wykorzystywana do pobierania jednowierszowych wyników zapytania.
	 * 
	 * @param conn
	 *            połączenie ze źródła danych
	 * @param cacheUsage
	 *            informacja w jaki sposób czy ma być użyty kesz wyników zapytań
	 *            {@link CacheUsage}
	 * @return pierwszy wiersz wyników zapytania
	 * @throws ExecuteSQLQueryException
	 */
	public ResultRow uniqeExecute(Connection conn, CacheUsage cacheUsage)
			throws ExecuteSQLQueryException {
		Collection<ResultRow> result = execute(conn, cacheUsage).getResult();
		if (result != null && !result.isEmpty()) {
			return result.iterator().next();
		}
		return null;
	}

	/**
	 * Metoda zwracająca klauzule zapytania (treść zapytania). Jeżeli wcześniej
	 * nie została wywołana metoda {@link #create()} i obiekt klauzuli jest
	 * pusty to zostanie ona wykonana teraz.
	 * 
	 * @return klauzula zapytania
	 * @throws CreateSQLQueryException
	 * @throws IdIsNotEmbeddableException
	 * @throws JDBCEntityParserException
	 * @throws ObjectIsNotEntityException
	 */
	public QueryClause getQueryClause() throws ObjectIsNotEntityException,
			JDBCEntityParserException, IdIsNotEmbeddableException,
			CreateSQLQueryException {
		if (this.queryClause == null
				|| StringUtils.isBlank(queryClause.getClause()))
			this.create();
		return queryClause;
	}

	/**
	 * @return the sortClause
	 */
	public SortClause getSortClause() {
		return sortClause;
	}

	/**
	 * Ustawienie klauzuli 'ORDER BY' w zapytaniu (warunków sortowania).
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param sortClause
	 *            obiekt klauzuli 'ORDER BY'
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setSortClause(SortClause sortClause) {
		this.sortClause = sortClause;
		return this;
	}

	/**
	 * @return the groupByClause
	 */
	public GroupByClause getGroupByClause() {
		return groupByClause;
	}

	/**
	 * Ustawienie klauzuli 'GROUP BY' w zapytaniu (warunków grupowania).
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na tworzone
	 * zapytanie tylko i wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param groupByClause
	 *            obiekt klauzuli 'GROUP BY'
	 * @return obiekt własnie tworzonego zapytania
	 */
	public SelectQuery setGroupByClause(GroupByClause groupByClause) {
		this.groupByClause = groupByClause;
		return this;
	}

	/**
	 * Zwraca wartość flagi decydującej czy w wyniku zapytania ma być
	 * wygenerowany obiekt encji, z którym jest związany dany opis tabeli (
	 * {@link #table}).
	 * 
	 * @return the generateEntityObject
	 */
	public boolean isGenerateEntityObject() {
		return generateEntityObject;
	}

	/**
	 * Ustawianie flagi {@link #generateEntityObject}, która mówi zapytaniu czy
	 * podczas wykonywania zapytania (metody: {@link #execute(Connection)} oraz
	 * {@link #execute(DataSource)}) czy mapper wierszy ma próbować utworzyć
	 * encje obiektów z którymi jest związany dany opis tabeli ({@link #table}.
	 * <p>
	 * Flaga jest ustawiina domyślnie na wartość {@code true}
	 * </p>
	 * <p>
	 * <b>Uwaga!</b> ustawienie odpowiedniej wartości ma wpływ na zwracany wynik
	 * podczas wykonywania zapytania, zatem ustawianie flagi ma tylko i
	 * wyłącznie sens przed wywołaniem metod {@link #execute(Connection)} lub
	 * {@link #execute(DataSource)}.
	 * </p>
	 * 
	 * @param generateEntityObject
	 *            the generateEntityObject to set
	 */
	public void setGenerateEntityObject(boolean generateEntityObject) {
		this.generateEntityObject = generateEntityObject;
		this.queryClauseDefinitionIsChanged = true;
	}

	/**
	 * Dodawanie funkcji do klauzuli zapytania
	 * <p>
	 * <b>Uwaga!</b> dodawanie funkcji ma wpływ na tworzone zapytanie tylko i
	 * wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param function
	 *            funkcja SQL-owa
	 * @return obiekt własnie tworzonego zapytania
	 * @throws CreateSQLQueryException
	 */
	public SelectQuery addFunction(_IFunction function)
			throws CreateSQLQueryException {
		String functionKey = function.getAlias();
		/* sprawdzam czy moge dodać funkcje o takim aliasie */
		/* sprawdzam konflikty kluczy */
		if (table.columnExists(functionKey))
			throw new CreateSQLQueryException("Alias funkcji '" + functionKey
					+ "' jest w konflikcie z kluczem istniejacej kolumny");

		/* sprawdzam konflikty aliasów */
		Enumeration<String> columnKeys = table.getColumnKeys();
		while (columnKeys.hasMoreElements()) {
			JDBCColumnDesc columnDesc = table.getColumn(columnKeys
					.nextElement());
			if (columnDesc.getAlias().equalsIgnoreCase(functionKey)) {
				throw new CreateSQLQueryException("Alias funkcji '"
						+ functionKey
						+ "' jest w konflikcie z aliasem istniejacej kolumny");
			}
		}

		/* OK moge dodac funkcje */
		functions.put(functionKey, function);
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * Usuwanie funkcji z klauzuli zapytania
	 * <p>
	 * <b>Uwaga!</b> dodawanie funkcji ma wpływ na tworzone zapytanie tylko i
	 * wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param functionKey
	 *            alias dodanej wcześniej funkcji
	 * @return obiekt własnie tworzonego zapytania
	 * @throws CreateSQLQueryException
	 */
	public SelectQuery removeFunction(String functionKey)
			throws CreateSQLQueryException {
		functions.remove(functionKey);
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * Usuwanie funkcji z klauzuli zapytania
	 * <p>
	 * <b>Uwaga!</b> dodawanie funkcji ma wpływ na tworzone zapytanie tylko i
	 * wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @param function
	 *            wcześniej dodana funkcja
	 * @return obiekt własnie tworzonego zapytania
	 * @throws CreateSQLQueryException
	 */
	public SelectQuery removeFunction(_IFunction function)
			throws CreateSQLQueryException {
		String functionKey = function.getAlias();
		functions.remove(functionKey);
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * Czyszczenie listy funkcji z klauzuli zapytania
	 * <p>
	 * <b>Uwaga!</b> dodawanie funkcji ma wpływ na tworzone zapytanie tylko i
	 * wyłącznie przed wywołaniem metody {@link #create()}
	 * </p>
	 * 
	 * @return obiekt własnie tworzonego zapytania
	 * @throws CreateSQLQueryException
	 */
	public SelectQuery clearFunctions() throws CreateSQLQueryException {
		functions.clear();
		this.queryClauseDefinitionIsChanged = true;
		return this;
	}

	/**
	 * Zwraca informację czy podczas tworzenia klauzuli zapytania mają zostać
	 * użyte tylko i wyłącznie zadeklarowane funkcje SQL-owe. Jeżeli ta flaga
	 * jest ustawiona na {@code false} i lista kolumn do użycia (
	 * {@link #columnsToUse}) jest pusta, to wtedy domyślnie dodawane są
	 * wszystkie kolumny zawarte w opisie tabeli ({@link #table}).
	 * 
	 * @return the useFunctionsOnly
	 */
	public boolean isUseFunctionsOnly() {
		return useFunctionsOnly;
	}

	/**
	 * Ustawia flagę decydującą o tym czy klauzula ma zawierać tylko
	 * zadeklarowane funkcje gdy lista kolumn do użycia ({@link #columnsToUse})
	 * jest pusta. Gdy flaga zostanie ustawiona na wartość {@code true} to pole
	 * {@link #generateEntityObject} zostanie automatycznie ustawione na wartość
	 * {@code false}.
	 * 
	 * @param useFunctionsOnly
	 *            the useFunctionsOnly to set
	 */
	public void setUseFunctionsOnly(boolean useFunctionsOnly) {
		if (useFunctionsOnly)
			this.generateEntityObject = false;

		this.useFunctionsOnly = useFunctionsOnly;
		this.queryClauseDefinitionIsChanged = true;
	}

	/**
	 * Metoda tworząca nazwę pola (kolumny), która wykorzystywana jest w
	 * zapytaniu SQL. Metoda pomocna do wykorzystania np. podczas obsługi
	 * funkcji czy też budowania klauzuli sortowania i warunków zapytania.
	 * 
	 * @see ConditionClauseSupport
	 * @see SortCondition
	 * 
	 * @param column
	 *            opis kolumny
	 * @return nazwa pola
	 */
	public static String getFieldName(JDBCColumnDesc column) {
		String fieldName;
		if (column == null)
			throw new IllegalArgumentException(
					"Definicja kolumny nie moze byc pusta");
		switch (column.getSource()) {
		case MAIN:
			fieldName = (StringUtils.isNotBlank(column.getTable().getAlias()) ? column
					.getTable().getAlias() + "."
					: "")
					+ column.getName();
			break;
		default:
			fieldName = column.getName();
			break;
		}
		return fieldName;
	}

	public String getQueryId() {
		byte[] hash;
		final int prime = 31;
		int pagingCode = 1;
		pagingCode = prime * pagingCode + this.pagingParams.getMaxCount();
		pagingCode = prime * pagingCode + this.pagingParams.getMaxPageSize();
		pagingCode = prime * pagingCode + this.pagingParams.getOffset();
		pagingCode = prime * pagingCode + this.pagingParams.getPageSize();
		StringBuffer idCodeBase = new StringBuffer(
				Integer.toBinaryString(pagingCode));
		MessageDigest digest;
		try {
			digest = java.security.MessageDigest.getInstance("MD5");
			idCodeBase.append(this.queryClause.toString());
			digest.update(idCodeBase.toString().getBytes());
			hash = digest.digest();
		} catch (NoSuchAlgorithmException e) {
			idCodeBase.append(Integer.toBinaryString(queryClause.hashCode()));
			hash = idCodeBase.toString().getBytes();
		}

		return new String(hash);
	}

	/**
	 * 
	 * QueryCacheInfo obiekt z informacjami o uzyciu kesza do keszowania wyników
	 * zapytania.
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.13 $
	 * 
	 */
	public static class QueryCacheInfo implements Serializable {

		private static final long serialVersionUID = -7125338842450591769L;

		private final String queryId;

		private final CacheUsage cacheUsage;

		private transient final _IObjectCache cache;

		/**
		 * @param queryId
		 * @param cacheUsage
		 * @param cache
		 */
		private QueryCacheInfo(String queryId, CacheUsage cacheUsage,
				_IObjectCache cache) {
			super();
			this.queryId = queryId;
			this.cacheUsage = cacheUsage;
			this.cache = cache;
		}

		/**
		 * @return the queryId
		 */
		public String getQueryId() {
			return queryId;
		}

		/**
		 * @return the cacheUsage
		 */
		public CacheUsage getCacheUsage() {
			return cacheUsage;
		}

		/**
		 * @return the cache
		 */
		public _IObjectCache getCache() {
			return cache;
		}

		public boolean cacheIsAllowed() {
			return (this.cache != null);
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result
					+ ((cache == null) ? 0 : cache.getRegionName().hashCode());
			result = prime * result
					+ ((cacheUsage == null) ? 0 : cacheUsage.hashCode());
			result = prime * result
					+ ((queryId == null) ? 0 : queryId.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			QueryCacheInfo other = (QueryCacheInfo) obj;
			if (cache == null) {
				if (other.cache != null)
					return false;
			} else if (!cache.getRegionName().equals(
					other.cache.getRegionName()))
				return false;
			if (cacheUsage == null) {
				if (other.cacheUsage != null)
					return false;
			} else if (!cacheUsage.equals(other.cacheUsage))
				return false;
			if (queryId == null) {
				if (other.queryId != null)
					return false;
			} else if (!queryId.equals(other.queryId))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "QueryCacheInfo ["
					+ "cacheUsage="
					+ cacheUsage
					+ ","
					+ (cache != null ? "cacheRegionName="
							+ cache.getRegionName() + "," : "") + "queryId="
					+ queryId + "]";
		}

	}

	/**
	 * @param queryCacheManager
	 *            the queryCacheManager to set
	 */
	public void setQueryCacheManager(QueryCacheManager queryCacheManager) {
		this.queryCacheManager = queryCacheManager;
	}

	/**
	 * @param pagingParams
	 *            the {@link #pagingParams} to set
	 */
	public void setPagingParams(PagingParams pagingParams) {
		this.pagingParams.copy(pagingParams);
	}

}
