package pl.slawas.scjdbc4jpa.jdbc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.BindParamException;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityValueParser;
import pl.slawas.scjdbc4jpa.jdbc.where.BindParam;
import pl.slawas.scjdbc4jpa.jdbc.where.BindParams;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSimple;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionOperator;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionParam;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionType;

/**
 * 
 * WhereClause klauzula warunków zapytania SQL
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class WhereClause implements _IClause, Serializable {

	private static final long serialVersionUID = -7307968594304256779L;

	final static private Logger logger = LoggerFactory
			.getLogger(WhereClause.class);

	private final Vector<ConditionClauseSupport> conditionsList;

	private final ConditionClauseSupport[] conditionsArray;

	private final StringBuffer clause;

	private final Vector<BindParam> bindParams;

	/**
	 * Konstruktor podstwowej klauzuli warunków wyszukiwania dla klucza
	 * (głownego lub obcego) tabeli, czyli np. dla klucza głownego
	 * {@code WHERE id = ?}.
	 * 
	 * @param pk
	 *            definicja klucza tabeli (klucz głowny lub obcy)
	 * @param id
	 *            wartość klucza
	 */
	public WhereClause(JDBCKeyDesc pk, Object id) {
		this(getKeyCondition(pk, id));
	}

	/**
	 * Konstruktor podstwowej klauzuli warunków wyszukiwania dla klucza
	 * (głownego lub obcego) tabeli, czyli np. dla klucza głownego
	 * {@code WHERE id = ?} z pustymi wartościami bindowanymi.
	 * 
	 * @param pk
	 *            definicja klucza tabeli (klucz głowny lub obcy)
	 * 
	 * @param pk
	 */
	public WhereClause(JDBCKeyDesc pk) {
		this(getKeyCondition(pk));
	}

	public WhereClause(Vector<ConditionClauseSupport> conditions) {
		if (conditions == null || conditions.isEmpty())
			throw new IllegalArgumentException(
					"Lista warunkow wyszukiwania nie moze byc pusta");

		this.conditionsList = conditions;
		this.conditionsArray = new ConditionClauseSupport[this.conditionsList.size()];
		int i = 0;
		for (ConditionClauseSupport cond : this.conditionsList) {
			this.conditionsArray[i] = cond;
			i++;
		}
		this.clause = createClause();
		this.bindParams = createBindParams();
	}

	public WhereClause(ConditionClauseSupport[] conditions) {
		if (conditions == null || conditions.length == 0)
			throw new IllegalArgumentException(
					"Lista warunkow wyszukiwania nie moze byc pusta");

		this.conditionsArray = conditions;
		this.conditionsList = new Vector<ConditionClauseSupport>(
				this.conditionsArray.length);
		for (ConditionClauseSupport cond : this.conditionsArray) {
			this.conditionsList.add(cond);
		}
		this.clause = createClause();
		this.bindParams = createBindParams();
	}

	/**
	 * Tworzenie listy bindowanych wartości dla tworzonego zapytania
	 * 
	 * @param pstmt
	 *            tworzone zapytanie
	 * @return Lista bindowanych parametrów
	 */
	private Vector<BindParam> createBindParams() {
		int parameterIndex = 1;
		Vector<BindParam> bindParams = new Vector<BindParam>();
		ConditionParam[] params = new ConditionParam[this.conditionsArray.length];
		for (int i = 0; i < this.conditionsArray.length; i++) {
			params[i] = this.conditionsArray[i].getParam();
		}
		parameterIndex = bindValues(parameterIndex, bindParams, params);
		return bindParams;
	}

	/**
	 * Tworzenie klauzuli {@code WHERE}
	 * 
	 * @return utworzona klauzula
	 */
	private StringBuffer createClause() {
		StringBuffer out = new StringBuffer("WHERE");
		for (int i = 0; i < this.conditionsArray.length; i++) {
			String subClause = this.conditionsArray[i].getClause();
			if (i == 0) {
				if (subClause.startsWith(ConditionOperator.AND.getOperator())) {
					subClause = subClause.substring(ConditionOperator.AND
							.getOperator().length());
				} else {
					subClause = subClause.substring(ConditionOperator.OR
							.getOperator().length());
				}
			} else {
				out.append(" ");
			}
			out.append(subClause);
		}
		return out;
	}

	/**
	 * @return the conditionsList
	 */
	public Vector<ConditionClauseSupport> getConditionsList() {
		return conditionsList;
	}

	/**
	 * @return the conditionsArray
	 */
	public ConditionClauseSupport[] getConditionsArray() {
		return conditionsArray;
	}

	/**
	 * Metoda pobierania klauzuli wyszukiwania
	 * 
	 * @return klauzula sortowania
	 */
	public String getClause() {
		return this.clause.toString();
	}

	/**
	 * Bindowanych wartości nowych wartości do tej samej klauzuli {@code WHERE}
	 * zapytania.
	 * 
	 * @param params
	 *            lista nowych bindowanych wartości
	 * @return Lista bindowanych parametrów
	 */
	public Vector<BindParam> setBindParams(BindParam[] params) {
		for (BindParam param : params) {
			BindParam exists = this.bindParams
					.get(param.getParameterIndex() - 1);
			if (exists.getType().equals(param.getType())) {
				exists.setParamValue(param.getParamValue());
			} else {
				throw new BindParamException("Parametry na tej samej pozycji ["
						+ param.getParameterIndex()
						+ "] nie maja zgodnego typu. " + "Jest "
						+ param.getType() + ", a powinien byc "
						+ exists.getType());
			}
		}
		return this.bindParams;
	}

	/**
	 * Pobieranie bindowanych wartości do tworzonego zapytania
	 * 
	 * @return Lista bindowanych parametrów
	 */
	public Vector<BindParam> getBindParams() {
		return this.bindParams;
	}

	/**
	 * Metoda realizująca bindoeanie wartości wywoływana rekurencyjnie dla
	 * parametrów typu {@link ConditionType#BOOL}
	 * 
	 * @param parameterIndex
	 *            indeks parametrów bindowanych
	 * @param bindParams
	 *            lista bindowanych parametrów
	 * @param params
	 *            parametry zapytania
	 * @return nowa wartość indeksu parametrów bindowanych
	 */
	private int bindValues(int parameterIndex, Vector<BindParam> bindParams,
			ConditionParam[] params) {
		for (int i = 0; i < params.length; i++) {
			ConditionParam param = params[i];
			switch (param.getType()) {
			case NULL:
				break;
			case SIMPLE:
			case SIMPLEfx:
			case LIKE:
			case IN:
				if (param.isBindedValue()) {
					parameterIndex = bindParam(parameterIndex, bindParams,
							param.getFieldType(), param.getValue());
				}
				break;
			case RANGE:
			case RANGEfx:
				if (param.isBindedValue()) {
					parameterIndex = bindParam(parameterIndex, bindParams,
							param.getFieldType(), param.getLowerBound());
					parameterIndex = bindParam(parameterIndex, bindParams,
							param.getFieldType(), param.getUpperBound());
				}
				break;
			case BOOL:
				parameterIndex = bindValues(parameterIndex, bindParams,
						params[i].getParams());
				break;
			}

		}
		return parameterIndex;
	}

	/**
	 * Metoda realizująca podstawowe bindowanie wartości
	 * 
	 * @param parameterIndex
	 *            indeks parametrów bindowanych
	 * @param bindParams
	 *            lista bindowanych parametrów
	 * @param type
	 *            typ SQL-owy {@link SimpleSQLType}
	 * @param paramValue
	 *            wartość bindowanego parametru
	 * @return nowa wartość indeksu parametrów bindowanych
	 */
	private int bindParam(int parameterIndex, Vector<BindParam> bindParams,
			SimpleSQLType type, Object paramValue) {
		if (paramValue != null && paramValue.getClass().isArray()) {
			for (Object val : (Object[]) paramValue) {
				bindParams.add(new BindParam(parameterIndex, type, val));
				parameterIndex++;
			}
		} else {
			bindParams.add(new BindParam(parameterIndex, type, paramValue));
			parameterIndex++;
		}
		return parameterIndex;
	}

	/**
	 * Tworzenie klauzuli wyszukiwania po kluczu głownym encji
	 * 
	 * @param key
	 *            definicja klucza
	 * @param keyValue
	 *            wartość klucza
	 * @return klauzula warunku wyszukiwania po kluczu głownym.
	 */
	public static ConditionClauseSupport[] getKeyCondition(JDBCKeyDesc key,
			Object keyValue) {
		ConditionClauseSupport[] conditionClauses = new ConditionClauseSupport[key
				.getColumns().size()];

		try {
			Hashtable<JDBCColumnDesc, Object> keyMap = JDBCEntityValueParser
					.readKeyObject(key, keyValue);
			Enumeration<Integer> positions = key.getColumns().keys();
			while (positions.hasMoreElements()) {
				Integer position = positions.nextElement();
				JDBCColumnDesc column = key.getColumn(position);
				conditionClauses[position] = new ConditionClauseSimple(
						ConditionOperator.AND, column, keyMap.get(column), true);

				if (logger.isTraceEnabled())
					logger.trace(
							"Kolumna [{}]:\n {}\n Value: {}",
							new Object[] { position, column.toString(),
									keyMap.get(column) });
			}
		} catch (IllegalArgumentException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (SecurityException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (IllegalAccessException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (InvocationTargetException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		}
		return conditionClauses;
	}

	/**
	 * Tworzenie klauzuli wyszukiwania po kluczu głownym encji z pustymi
	 * warunkami.
	 * 
	 * @param key
	 *            definicja klucza
	 * @return klauzula warunku wyszukiwania po kluczu głownym.
	 */
	public static ConditionClauseSupport[] getKeyCondition(JDBCKeyDesc key) {
		ConditionClauseSupport[] conditionClauses = new ConditionClauseSupport[key
				.getColumns().size()];

		try {
			Enumeration<Integer> positions = key.getColumns().keys();
			while (positions.hasMoreElements()) {
				Integer position = positions.nextElement();
				JDBCColumnDesc column = key.getColumn(position);
				conditionClauses[position] = new ConditionClauseSimple(
						ConditionOperator.AND, column, null, true);

				if (logger.isTraceEnabled())
					logger.trace(
							"Kolumna [{}]:\n {}\n Value: {}",
							new Object[] { position, column.toString(), "NULL" });
			}
		} catch (IllegalArgumentException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (SecurityException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		}
		return conditionClauses;
	}

	/**
	 * Tworzenie obiektu bindowanego parametru
	 * 
	 * @param parameterIndex
	 *            wartość indeksu jaka ma być przypisana bindowanej wartości
	 * @param column
	 *            definicja kolumny, której dotyczy warunek zapytania
	 * @param paramValue
	 *            bindowana wartość
	 * @return obiekt bindowanego parametru
	 */
	public static BindParam createBindParam(int parameterIndex,
			JDBCColumnDesc column, Object paramValue) {
		return new BindParam(parameterIndex, SimpleSQLType.get(column
				.getColumnType()), paramValue);
	}

	/**
	 * Tworzenie obiektu z listą bindowanych parametrów na podstawie definicji
	 * klucza.
	 * 
	 * @param parameterIndex
	 *            wartość indeksu jaka ma być przypisana bindowanej wartości
	 * @param key
	 *            definicja klucza, której dotyczy warunek zapytania
	 * @param paramValue
	 *            bindowana wartość
	 * @return klucz moze być złożony stąd zwracany jest obiekt z listą
	 *         paramerów bindowanych oraz z ostatnia pozycja indeksy parametrów
	 */
	public static BindParams createBindParams(int parameterIndex,
			JDBCKeyDesc key, Object paramValue) {

		BindParam[] params = new BindParam[key.getColumns().size()];
		int startParameterIndex = parameterIndex;

		try {
			Hashtable<JDBCColumnDesc, Object> keyMap = JDBCEntityValueParser
					.readKeyObject(key, paramValue);
			Enumeration<Integer> positions = key.getColumns().keys();
			while (positions.hasMoreElements()) {
				Integer position = positions.nextElement();
				JDBCColumnDesc column = key.getColumn(position);
				int parameterIndexPosition = startParameterIndex + position;
				params[position] = new BindParam(parameterIndexPosition,
						SimpleSQLType.get(column.getColumnType()),
						keyMap.get(column));

				if (logger.isTraceEnabled())
					logger.trace(
							"Kolumna [position: {}, parameterIndexPosition: {}]:\n {}\n Value: {}",
							new Object[] { position, parameterIndexPosition,
									column.toString(), keyMap.get(column) });

				if (positions.hasMoreElements()) {
					parameterIndex++;
				}
			}
			return new BindParams(parameterIndex, params);
		} catch (IllegalArgumentException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (SecurityException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (IllegalAccessException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		} catch (InvocationTargetException e) {
			throw new BindParamException(
					"Problem z pobraniem wartosci klucza.", e);
		}

	}

}
