package pl.slawas.scjdbc4jpa.jdbc.beans;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Vector;

import pl.slawas.support.EnumerationSupport;

/**
 * 
 * JdbcColumnDesc - opis kolumny
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
public class JDBCColumnDesc implements Serializable {

	private static final long serialVersionUID = -2896154145674440482L;

	/**
	 * Maksymalna długość indentyfikatora (aliasu) kolumny.
	 */
	public static final int ALIAS_MAX_LENGTH = 30;

	private String alias;

	private final JDBCTableDesc table;

	private final String entityFieldName;

	private final Enumeration<Class<?>> entityFieldTypes;

	private final String name;

	private final Class<?> columnType;

	private JDBCColumnSourceType source = JDBCColumnSourceType.MAIN;

	private JDBCKeyDesc sourceKey;

	private final Enumeration<Method> setters;

	private final Enumeration<Method> getters;

	/**
	 * @param table
	 *            definicja tabeli, z której pochodzi kolumna.
	 * @param entityFieldType
	 *            typ pola encji
	 * @param columnType
	 *            typ kolumny w bazie danych
	 * @param entityFieldName
	 *            nazwa pola encji
	 * @param name
	 *            nazwa kolumny w bazie danych
	 * @param setter
	 *            metoda ustawiajaca wartosc w encji
	 * @param getter
	 *            metoda pobierajaca wartosc z encji
	 */
	public JDBCColumnDesc(JDBCTableDesc table, Class<?> entityFieldType,
			Class<?> columnType, String entityFieldName, String name,
			Method setter, Method getter) {
		super();
		this.table = table;
		this.entityFieldName = entityFieldName;
		this.name = name;
		this.columnType = columnType;
		this.entityFieldTypes = (new EntityFieldTypesEnumerator())
				.add(entityFieldType);
		this.setters = (new EntityMethodsEnumerator()).add(setter);
		this.getters = (new EntityMethodsEnumerator()).add(getter);
		;
	}

	/**
	 * @param table
	 *            definicja tabeli, z której pochodzi kolumna.
	 * @param entityFieldType
	 *            typ pola encji
	 * @param columnType
	 *            typ kolumny w bazie danych
	 * @param entityFieldName
	 *            nazwa pola encji
	 * @param name
	 *            nazwa kolumny w bazie danych
	 * @param source
	 *            typ źródła pochodzenia kolumny
	 * @param setter
	 *            metoda ustawiajaca wartosc w encji
	 * @param getter
	 *            metoda pobierajaca wartosc z encji
	 */
	public JDBCColumnDesc(JDBCTableDesc table, Class<?> entityFieldType,
			Class<?> columnType, String entityFieldName, String name,
			JDBCColumnSourceType source, Method setter, Method getter) {
		this(table, entityFieldType, columnType, entityFieldName, name, setter,
				getter);
		this.source = source;
	}

	/**
	 * @return the alias
	 */
	public String getAlias() {
		return alias;
	}

	/**
	 * @param alias
	 *            the alias to set
	 */
	public void setAlias(String alias) {
		if (alias.length() > ALIAS_MAX_LENGTH)
			alias = alias.substring(0, 3) + "_"
					+ Integer.toString(alias.hashCode());
		this.alias = alias;
	}

	/**
	 * @return the entityFieldName
	 */
	public String getEntityFieldName() {
		return entityFieldName;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return the source
	 */
	public JDBCColumnSourceType getSource() {
		return source;
	}

	/**
	 * @param source
	 *            the source to set
	 */
	public void setSource(JDBCColumnSourceType source) {
		this.source = source;
	}

	/**
	 * @return the columnType
	 */
	public Class<?> getColumnType() {
		return columnType;
	}

	@Override
	public String toString() {
		StringBuffer out = new StringBuffer("JdbcColumnDesc [");
		out.append("\n\tname = " + name);
		out.append("\n\tcolumnType = " + columnType);
		out.append("\n\talias = " + alias);
		out.append("\n\tsource = " + source);
		out.append("\n\ttable = " + table.toString());
		out.append("\n\tentityFieldName = " + entityFieldName);
		Enumeration<Class<?>> types = this.getEntityFieldTypes();
		int i = 0;
		out.append("\n\tentityFieldTypes:");
		while (types.hasMoreElements()) {
			out.append("\n\t  [" + i + "] " + types.nextElement());
			i++;
		}
		Enumeration<Method> setters = this.getSetters();
		i = 0;
		out.append("\n\tsetters:");
		while (setters.hasMoreElements()) {
			out.append("\n\t  [" + i + "] " + setters.nextElement().toString());
			i++;
		}
		Enumeration<Method> getters = this.getGetters();
		i = 0;
		out.append("\n\tgetters:");
		while (getters.hasMoreElements()) {
			out.append("\n\t  [" + i + "] " + getters.nextElement().toString());
			i++;
		}
		out.append("\n]");
		return out.toString();
	}

	/**
	 * @return the entityFieldTypes
	 */
	public Enumeration<Class<?>> getEntityFieldTypes() {
		return ((EntityFieldTypesEnumerator) this.entityFieldTypes)
				.getNewInstance();
	}

	/**
	 * 
	 * @param entityFieldType
	 */
	public void addEntityFieldType(Class<?> entityFieldType) {
		((EntityFieldTypesEnumerator) this.entityFieldTypes)
				.add(entityFieldType);
	}

	/**
	 * @return the sourceKey
	 */
	public JDBCKeyDesc getSourceKey() {
		return sourceKey;
	}

	/**
	 * @param sourceKey
	 *            the sourceKey to set
	 */
	public void setSourceKey(JDBCKeyDesc sourceKey) {
		this.sourceKey = sourceKey;
	}

	/**
	 * @return the table, definicja tabeli, z której pochodzi kolumna
	 */
	public JDBCTableDesc getTable() {
		return table;
	}

	/**
	 * @return the setters
	 */
	public Enumeration<Method> getSetters() {
		return ((EntityMethodsEnumerator) this.setters).getNewInstance();
	}

	/**
	 * @return the getters
	 */
	public Enumeration<Method> getGetters() {
		return ((EntityMethodsEnumerator) this.getters).getNewInstance();
	}

	/**
	 * dodawanie kolejnego settera (wykorzystywane dla kluczy obcych, kluczy
	 * głownych złozonych itp.
	 */
	public void addSetter(Method setter) {
		((EntityMethodsEnumerator) this.setters).add(setter);
	}

	/**
	 * dodawanie kolejnego gettera (wykorzystywane dla kluczy obcych, kluczy
	 * głownych złozonych itp.
	 */
	public void addGetter(Method getter) {
		((EntityMethodsEnumerator) this.getters).add(getter);
	}

	/**
	 * 
	 * EntityFieldTypesEnumerator - prywatna implementacja enumeratora typów
	 * encji
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.4 $
	 * 
	 */
	private class EntityFieldTypesEnumerator extends
			EnumerationSupport<Class<?>> {

		private static final long serialVersionUID = -8465097628182639472L;

		private EntityFieldTypesEnumerator() {
			super();
		}

		private EntityFieldTypesEnumerator(Vector<Class<?>> elements) {
			super(elements);
		}

		@Override
		public EnumerationSupport<Class<?>> getNewInstance() {
			return new EntityFieldTypesEnumerator(elements);
		}

	}

	/**
	 * 
	 * EntityMethodsEnumerator - prywatna implementacja enumeratora metod
	 * ustawiających i pobierającuych wartrości kolumny z encji.
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.4 $
	 * 
	 */
	private class EntityMethodsEnumerator extends EnumerationSupport<Method> {

		private static final long serialVersionUID = -3052246291705669109L;

		private EntityMethodsEnumerator() {
			super();
		}

		private EntityMethodsEnumerator(Vector<Method> elements) {
			super(elements);
		}

		@Override
		public EnumerationSupport<Method> getNewInstance() {
			return new EntityMethodsEnumerator(elements);
		}

	}

}
