package pl.slawas.scjdbc4jpa.jdbc.beans;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.helpers.Strings;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityParserLevel;

/**
 * 
 * JdbcTableDesc - obiekt pomocniczy opisujący tabele
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.6 $
 * 
 */
public class JDBCTableDesc implements Serializable {

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCTableDesc.class);

	private static final long serialVersionUID = -5990909663752677918L;

	private int joinedTableCounter = 0;

	private int columnCounter = 0;

	private String cacheRegionName;

	private final Class<?> entityClass;

	private final String name;

	private String alias;

	private JDBCKeyDesc primaryKey;

	/**
	 * Poziom oipisu tablicy
	 */
	private final JDBCEntityParserLevel level;

	/**
	 * Lista kolumn. Pierwszy parametr to nazwa pola w encji, drugi to opis
	 * kolumny
	 */
	private final Hashtable<String, JDBCColumnDesc> columns = new Hashtable<String, JDBCColumnDesc>();

	/**
	 * Linkowanie nazw kolumn bazy danych do nazw pol w encji. Pierwszy parametr
	 * to nazwa kolumny w bazie danych, drugi to nazwa pola w encji
	 */
	private final Hashtable<String, String> dbColumnLinks = new Hashtable<String, String>();

	/**
	 * Lista tabel połączonych. Pierwszy parametr to obiekt definiujący klucz
	 * obcym w związku, drugi to obiekt opisujący tabele powiązaną
	 */
	private final Hashtable<JDBCKeyDesc, JDBCTableDesc> joinedTables = new Hashtable<JDBCKeyDesc, JDBCTableDesc>();

	private final Hashtable<String, JDBCKeyDesc> foreignKeyList = new Hashtable<String, JDBCKeyDesc>();

	/**
	 * 
	 * @param level
	 * @param entityClass
	 * @param tableName
	 * @param alias
	 */
	public JDBCTableDesc(JDBCEntityParserLevel level, Class<?> entityClass,
			String tableName, String alias) {
		this(level, entityClass, tableName);
		this.alias = alias;
	}

	/**
	 * 
	 * @param level
	 * @param entityClass
	 * @param tableName
	 */
	public JDBCTableDesc(JDBCEntityParserLevel level, Class<?> entityClass,
			String tableName) {
		super();
		this.level = level;
		this.name = tableName;
		this.entityClass = entityClass;
	}

	/**
	 * @return nazwa tabeli w bazie danych
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return alias tabeli używany w zapytaniach złożonych
	 */
	public String getAlias() {
		return alias;
	}

	/**
	 * Nieposortowana lista kluczy kolumn (nazw pol encji). Za pomocą tej listy
	 * można wyciągnąc informacje o kolumnach.
	 * 
	 * @see #getColumn(String)
	 * 
	 * @return lista kluczy kolumn (kluczem jest nazwa pola encji)
	 */
	public Enumeration<String> getColumnKeys() {
		return columns.keys();
	}

	/**
	 * Pobieranie opisu kolumny na podstawie klucza przechowywania opisów
	 * kolumn, czyli nazwy pola w encji.
	 * 
	 * @see #getColumnKeys()
	 * 
	 * @param entityFieldName
	 *            nazwa pola encji, klucz
	 * @return opis kolumny
	 */
	public JDBCColumnDesc getColumn(String entityFieldName) {
		return columns.get(entityFieldName);
	}

	/**
	 * Nieposortowana lista nazw bazodanowych dla kolumn. Za pomocą tej listy
	 * można wyciągnąc informacje o kolumnach.
	 * 
	 * @see JDBCTableDesc#getColumnByDbColumnName(String)
	 * 
	 * @return lista nazw kolumn bazodanowych wykorzystywanych w zapytaniu.
	 *         Kluczem przechowywania listy kolumn w definicji tabeli jhest
	 *         nazwa encji, ale neraz potrzebny jest dostęp po nazwie
	 *         bazodanowej kolumny. Stąd powstała specjalna struktuira
	 *         {@link #dbColumnLinks} linkująca nazwy bazodanowe kolumn z
	 *         kluczem przechowyaniam czylki nazwą pola encji.
	 */
	public Enumeration<String> getDbColumnNames() {
		return dbColumnLinks.keys();
	}

	/**
	 * Pobieranie opisu kolumny na podstawie nazwy bazodanowej kolumny.
	 * 
	 * @see #getSortedDbColumnNames()
	 * 
	 * @param dbColumnName
	 *            nazwa bazodanowa kolumny
	 * @return opis kolumny
	 */
	public JDBCColumnDesc getColumnByDbColumnName(String dbColumnName) {
		String entityFieldName = dbColumnLinks.get(dbColumnName);
		return columns.get(entityFieldName);
	}

	/**
	 * @return liczba kolumn
	 */
	public int getColumnsCount() {
		return columns.size();
	}

	/**
	 * Metoda dodawania kolumny do definicji tablicy
	 * 
	 * @param column
	 *            definicja kolumny
	 * @return tworzony obiekt opisu tabeli
	 * @throws JDBCEntityParserException
	 */
	public JDBCTableDesc addColumn(JDBCColumnDesc column)
			throws JDBCEntityParserException {

		if (column.getSource() == null)
			throw new JDBCEntityParserException(
					"Brak definicji pochodzenia kolumny");

		columnCounter++;
		if (StringUtils.isBlank(column.getAlias())) {
			column.setAlias("col_" + columnCounter);
		} else {
			switch (column.getSource()) {
			case MAIN:
				column.setAlias(column.getAlias() + "_" + columnCounter);
				break;
			default:
				break;
			}
		}

		if (columns.containsKey(column.getEntityFieldName())) {
			JDBCColumnDesc oldDescription = columns.get(column
					.getEntityFieldName());
			dbColumnLinks.remove(oldDescription.getName());
		}
		columns.put(column.getEntityFieldName(), column);
		dbColumnLinks.put(column.getName(), column.getEntityFieldName());
		return this;
	}

	/**
	 * @return the entityName
	 */
	public String getEntityName() {
		return entityClass.getSimpleName();
	}

	/**
	 * @return lista kluczy obcych
	 */
	public Enumeration<JDBCKeyDesc> getForeignKeys() {
		return joinedTables.keys();
	}

	/**
	 */

	/**
	 * Pobieranie klucza obcego
	 * 
	 * @param keyName
	 *            nazwa klucza
	 * @return klucz obcy o podanej nazwie
	 */
	public JDBCKeyDesc getForeignKey(String keyName) {
		return foreignKeyList.get(keyName);
	}

	/**
	 * @return informacje o tablicy powiązanej
	 */
	public JDBCTableDesc getJoinedTable(JDBCKeyDesc foreignKey) {
		return joinedTables.get(foreignKey);
	}

	public final static int MAX_TABLE_ALIAS_LENGTH = 3;

	/**
	 * Metoda dodawania powiązań z tabelami
	 * 
	 * @param foreignKey
	 *            klucz obcy
	 * @param joinedTable
	 *            połączona tabela
	 * @param setter
	 *            metoda za pomocą, której ustawiany jest obiekt obcy w obiekcie
	 *            głównym
	 * @param getter
	 *            metoda za pomocą, której pobierany jest obiekt obcy z obiektu
	 *            głównego
	 * 
	 * @return tworzony obiekt opisu tabeli
	 * @throws JDBCEntityParserException
	 */
	public JDBCTableDesc addJoinedTable(JDBCKeyDesc foreignKey,
			JDBCTableDesc joinedTable, Method setter, Method getter)
			throws JDBCEntityParserException {

		StringBuffer logOut;
		int step = 0;

		logger.trace("Dodaje powiazanie z encja {}",
				joinedTable.getEntityName());

		joinedTableCounter++;
		if (StringUtils.isBlank(alias)) {
			this.alias = this.name.substring(
					0,
					(this.name.length() < MAX_TABLE_ALIAS_LENGTH ? this.name
							.length() : MAX_TABLE_ALIAS_LENGTH)).toLowerCase();
		}

		String oldJoinedTableAlias = joinedTable.getAlias();
		if (StringUtils.isBlank(oldJoinedTableAlias)) {
			joinedTable.setAlias(foreignKey.getName() + "_"
					+ joinedTableCounter);
		} else {
			joinedTable
					.setAlias(oldJoinedTableAlias + "_" + joinedTableCounter);
		}
		joinedTables.put(foreignKey, joinedTable);
		foreignKeyList.put(foreignKey.getName(), foreignKey);

		/* teraz musze dodac ewentulane tablice powiazane z tablica powiązaną */
		Enumeration<JDBCKeyDesc> joinedTableForeignKeys = joinedTable
				.getForeignKeys();
		while (joinedTableForeignKeys.hasMoreElements()) {
			step = 1;
			logOut = new StringBuffer();

			JDBCKeyDesc joinedTableForeignKey = joinedTableForeignKeys
					.nextElement();
			if (logger.isTraceEnabled())
				logOut.append("\n\n Podlaczam klucz obcy z tabeli "
						+ joinedTable.toString() + "\n Analiza klucza '"
						+ joinedTableForeignKey.getName() + "': START {");

			/* krok 1: pobieram tablice obcą */
			JDBCTableDesc tableForeign = joinedTable
					.getJoinedTable(joinedTableForeignKey);
			if (logger.isTraceEnabled())
				step = logStep(logOut, step, "Tabela klucza obcego",
						tableForeign.toString());

			/* krok 2: ustawiam nowy alias dla podłączonej tabeli */
			String newAlias4TableForeign = joinedTable.getAlias() + "_"
					+ tableForeign.getAlias();
			tableForeign.setAlias(newAlias4TableForeign);
			if (logger.isTraceEnabled())
				step = logStep(logOut, step, "Nowy alias tabeli",
						tableForeign.getAlias());

			/** bede tworzyc nowy klucz */
			/* krok 3: ustawiam nowa nazwe dla nowego klucza */
			String newForeignKeyName = foreignKey.getName() + "."
					+ joinedTableForeignKey.getName();
			if (logger.isTraceEnabled())
				step = logStep(logOut, step, "Nowa nazwa klucza obcego",
						newForeignKeyName);

			/* krok 4: tworze nowy klucz */
			JDBCKeyDesc newForeignKey = new JDBCKeyDesc(
					joinedTable.getEntityClass(), newForeignKeyName,
					JDBCKeyType.FFK);
			if (logger.isTraceEnabled())
				step = logStep(logOut, step, "Nowy klucz obcy",
						newForeignKey.toString());

			for (int position = 0; position < joinedTableForeignKey
					.getColumns().size(); position++) {
				JDBCColumnDesc foreignColumn = joinedTableForeignKey
						.getColumn(position);

				String newForeignColumnName = joinedTable.getAlias() + "."
						+ foreignColumn.getName();

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "(" + position
							+ ") Nowa nazwa kolumny klucza obcego",
							newForeignColumnName);

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "(" + position
							+ ") Pobrana kolumna klucza obcego",
							foreignColumn.toString());

				String newForeignColumnAlias = tableForeign.getAlias() + "_"
						+ foreignColumn.getAlias();

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "(" + position
							+ ") Nowy alias kolumny klucza obcego",
							newForeignColumnAlias);

				String newForeignColumnEntityName = foreignKey.getName() + "."
						+ foreignColumn.getEntityFieldName();

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "(" + position
							+ ") Nowa nazwa encji dla kolumny klucza obcego",
							newForeignColumnEntityName);

				JDBCColumnDesc newForeignColumn = new JDBCColumnDesc(
						foreignColumn.getTable(), joinedTable.getEntityClass(),
						foreignColumn.getColumnType(),
						newForeignColumnEntityName, newForeignColumnName,
						JDBCColumnSourceType.FOREIGN, setter, getter);
				newForeignColumn.setAlias(newForeignColumnAlias);
				Enumeration<Class<?>> newForeignColumnTypes = foreignColumn
						.getEntityFieldTypes();
				while (newForeignColumnTypes.hasMoreElements()) {
					newForeignColumn.addEntityFieldType(newForeignColumnTypes
							.nextElement());
				}
				/* pobieram gettery klucza obcego */
				Enumeration<Method> fkGetters = foreignColumn.getGetters();
				while (fkGetters.hasMoreElements()) {
					newForeignColumn.addGetter(fkGetters.nextElement());
				}
				/* pobieram settery klucza obcego */
				Enumeration<Method> fkSetters = foreignColumn.getSetters();
				while (fkSetters.hasMoreElements()) {
					newForeignColumn.addSetter(fkSetters.nextElement());
				}

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "(" + position
							+ ") Nowa kolumna klucza obcego",
							newForeignColumn.toString());

				newForeignKey.addColumn(position, newForeignColumn);
			}

			/* dodaje tabele */
			joinedTables.put(newForeignKey, tableForeign);
			if (logger.isTraceEnabled())
				logOut.append("\n    --------------------------------------------"
						+ "\n } Analiza klucza '"
						+ joinedTableForeignKey.getName() + "': KONIEC\n");
			logger.trace(logOut.toString());
		}

		/* dobra dodaje kolumny do zapytania */
		Enumeration<String> joinedColsKeys = joinedTable.getColumnKeys();
		String joinedTableAlias = joinedTable.getAlias();
		String foreignKeyName = foreignKey.getName();

		while (joinedColsKeys.hasMoreElements()) {
			/* krok 1 - pobranie kolumny tablicy połączonej */
			JDBCColumnDesc joinedColumn = joinedTable.getColumn(joinedColsKeys
					.nextElement());

			String newJoinedColumnName;
			String newJoinedColumnAlias;
			String newJoinedColumnEntityFieldName;
			logOut = new StringBuffer();
			step = 1;

			if (logger.isTraceEnabled()) {
				logOut.append("\n\n Podlaczam kolumny z tabeli "
						+ joinedTable.toString() + "\n START {");
				step = logStep(logOut, step, "Pobrana kolumna",
						joinedColumn.toString());
			}

			/* krok 2 - ustalenie nowej nazwy */
			switch (joinedColumn.getSource()) {
			case MAIN:
				newJoinedColumnName = joinedTableAlias + "."
						+ joinedColumn.getName();
				break;
			case FOREIGN:
				newJoinedColumnName = joinedTableAlias + "_"
						+ joinedColumn.getName();
				break;
			default:
				newJoinedColumnName = null;
			}
			if (logger.isTraceEnabled())
				step = logStep(logOut, step, "Nowa nazwa kolumny",
						newJoinedColumnName);

			if (newJoinedColumnName != null) {
				/* krok 2 - ustalenie nowego aliasu */
				newJoinedColumnAlias = joinedTableAlias + "_"
						+ joinedColumn.getAlias();
				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "Nowy alias kolumny",
							newJoinedColumnAlias);

				/* krok 3 - ustalenie nowej nazwy dla pola encji */
				newJoinedColumnEntityFieldName = foreignKeyName + "."
						+ joinedColumn.getEntityFieldName();
				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "Nowa nazwa pola encji",
							newJoinedColumnEntityFieldName);

				/* krok 4 - dolaczenie nowej kolumny obcej */
				JDBCColumnDesc foreignColumn = new JDBCColumnDesc(
						joinedColumn.getTable(), foreignKey.getKeyType(),
						joinedColumn.getColumnType(),
						newJoinedColumnEntityFieldName, newJoinedColumnName,
						JDBCColumnSourceType.FOREIGN, setter, getter);
				foreignColumn.setSourceKey(foreignKey);
				foreignColumn.setAlias(newJoinedColumnAlias);
				Enumeration<Class<?>> joinedColumnTypes = joinedColumn
						.getEntityFieldTypes();
				while (joinedColumnTypes.hasMoreElements()) {
					foreignColumn.addEntityFieldType(joinedColumnTypes
							.nextElement());
				}
				/* pobieram gettery kolumny obcej */
				Enumeration<Method> fkGetters = joinedColumn.getGetters();
				if (!fkGetters.hasMoreElements())
					logger.warn("Kolumna {} nie ma gettersow?",
							joinedColumn.getName());
				while (fkGetters.hasMoreElements()) {
					foreignColumn.addGetter(fkGetters.nextElement());
				}
				/* pobieram settery kolumny obcej */
				Enumeration<Method> fkSetters = joinedColumn.getSetters();
				while (fkSetters.hasMoreElements()) {
					foreignColumn.addSetter(fkSetters.nextElement());
				}

				if (logger.isTraceEnabled())
					step = logStep(logOut, step, "Dodawana kolumna obca",
							foreignColumn.toString());

				this.addColumn(foreignColumn);

				if (logger.isTraceEnabled())
					logOut.append("\n    --------------------------------------------"
							+ "\n } KONIEC\n");

				logger.trace(logOut.toString());

			} else {
				logger.warn("Nie udalo sie ustalic nowej nazwy dla kolumny {}",
						joinedColumn.toString());
			}
		}

		return this;
	}

	private static int logStep(StringBuffer logOut, int step, String label,
			String value) {
		logOut.append("\n    --------------------------------------------"
				+ "\n[" + step + "] " + label + ": " + value + "");
		return step + 1;
	}

	/**
	 * @return the primaryKey
	 */
	public JDBCKeyDesc getPrimaryKey() {
		return primaryKey;
	}

	/**
	 * @param primaryKey
	 *            the primaryKey to set
	 */
	public void setPrimaryKey(JDBCKeyDesc primaryKey) {
		this.primaryKey = primaryKey;
	}

	/**
	 * @param alias
	 *            the alias to set
	 */
	public void setAlias(String alias) {
		this.alias = alias;
	}

	/**
	 * @return the entityClass
	 */
	public Class<?> getEntityClass() {
		return entityClass;
	}

	@Override
	public String toString() {
		return "JDBCTableDesc [\n\talias=" + alias + "\n\tentityClass="
				+ entityClass + "\n\tname=" + name + "]";
	}

	/**
	 * Zwraca tablice posortowanych nazw kolumn bazodanowych.
	 * 
	 * @see #getDbColumnNames()
	 * 
	 * @return posortowana lista
	 */
	public String[] getSortedDbColumnNames() {

		/* dodaje kolumny do klauzuli */
		Enumeration<String> links = this.getDbColumnNames();
		String[] sortedLinks = new String[columns.size()];
		int iCounter = 0;
		while (links.hasMoreElements()) {
			sortedLinks[iCounter] = links.nextElement();
			iCounter++;
		}
		try {
			Strings.sort(sortedLinks);
		} catch (Exception e) {
			logger.warn("Problem sortowania listy", e);
		}

		return sortedLinks;

	}

	/**
	 * Zwraca posortowaną listę kluczy dla przechowywanych kolumn (lista nazw
	 * pól encji).
	 * 
	 * @see #getColumnKeys()
	 * 
	 * @return posortowana lista
	 */
	public String[] getSortedColumnKeys() {

		Enumeration<String> columnKeys = this.getColumnKeys();
		String[] sortedColumns = new String[columns.size()];
		int iCounter = 0;
		while (columnKeys.hasMoreElements()) {
			sortedColumns[iCounter] = columnKeys.nextElement();
			iCounter++;
		}
		try {
			Strings.sort(sortedColumns);
		} catch (Exception e) {
			logger.warn("Problem sortowania listy", e);
		}

		return sortedColumns;

	}

	/**
	 * Zwraca posortowaną po nazwie klucza, listę kluczy obcych
	 * 
	 * @see #getForeignKeys()
	 * 
	 * @return lista kluczy obcych
	 */
	public List<JDBCKeyDesc> getSortedForeignKeys() {

		Enumeration<JDBCKeyDesc> foreignKeys = this.getForeignKeys();
		List<JDBCKeyDesc> sortedForeignKeys = new ArrayList<JDBCKeyDesc>();
		while (foreignKeys.hasMoreElements()) {
			sortedForeignKeys.add(foreignKeys.nextElement());
		}
		Collections.sort(sortedForeignKeys, new JDBCKeyDesc.SortByName());
		return sortedForeignKeys;
	}

	/**
	 * Sprawdza czy istnieje w opisie tabeli kolumna po podanym kluczu
	 * (przypominam, że kluczami przechowywanych kolumn jest nazwa pola encji,
	 * którą reprezentuje kolumna).
	 * 
	 * @param entityFieldName
	 *            nazwa pola encji, nazwa klucza przechowywania kolumn.
	 * @return informacja o tym czy definicja kolumny o podanym kluczu istnieje
	 *         w definicji tabeli. Zwraca wartość {@code true} jeżeli tak, albo
	 *         {@code false} jeżeli nie
	 */
	public boolean columnExists(String entityFieldName) {
		return columns.containsKey(entityFieldName);
	}

	/**
	 * @see JDBCEntityParserLevel
	 * 
	 * @return poziom utworzenia opisu tabeli
	 */
	public JDBCEntityParserLevel getLevel() {
		return level;
	}

	/**
	 * @return the cacheRegionName
	 */
	public String getCacheRegionName() {
		return (StringUtils.isBlank(cacheRegionName) ? this.getEntityName()
				: cacheRegionName);
	}

	/**
	 * @param cacheRegionName
	 *            the cacheRegionName to set
	 */
	public void setCacheRegionName(String cacheRegionName) {
		this.cacheRegionName = cacheRegionName;
	}

}
