package pl.slawas.scjdbc4jpa.jdbc.cache;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import net.sf.ehcache.Cache;
import net.sf.ehcache.Element;
import net.sf.ehcache.config.CacheConfiguration;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;

/**
 * 
 * EHCache - klasa kesza, wzorowana na keszu hibernate-a
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class EhCache implements Serializable, _IObjectCache {

	private static final long serialVersionUID = 600977337387514385L;

	private static final Logger logger = LoggerFactory.getLogger(EhCache.class);

	private final Cache ehCache;

	public EhCache(net.sf.ehcache.Cache ehCache) {
		this.ehCache = ehCache;
	}

	public Object get(Object key) throws ObjectCacheErrorException {
		try {
			logger.trace("[{}] key: {}",
					new Object[] { ehCache.getName(), key });
			if (key == null) {
				return null;
			} else {
				Element element = ehCache.get(key);
				if (element == null) {
					return null;
				} else {
					return element.getObjectValue();
				}
			}
		} catch (net.sf.ehcache.CacheException e) {
			throw new ObjectCacheErrorException(e);
		}
	}

	public Object read(Object key) throws ObjectCacheErrorException {
		return get(key);
	}

	public void update(Object key, Object value)
			throws ObjectCacheErrorException {
		put(key, value);
	}

	public void put(Object key, Object value) throws ObjectCacheErrorException {
		try {
			Element element = new Element(key, value);
			ehCache.put(element);
		} catch (IllegalArgumentException e) {
			throw new ObjectCacheErrorException(e);
		} catch (IllegalStateException e) {
			throw new ObjectCacheErrorException(e);
		} catch (net.sf.ehcache.CacheException e) {
			throw new ObjectCacheErrorException(e);
		}

	}

	public void remove(Object key) throws ObjectCacheErrorException {
		try {
			ehCache.remove(key);
		} catch (ClassCastException e) {
			throw new ObjectCacheErrorException(e);
		} catch (IllegalStateException e) {
			throw new ObjectCacheErrorException(e);
		} catch (net.sf.ehcache.CacheException e) {
			throw new ObjectCacheErrorException(e);
		}
	}

	public void clear() throws ObjectCacheErrorException {
		try {
			ehCache.removeAll();
		} catch (IllegalStateException e) {
			throw new ObjectCacheErrorException(e);
		} catch (net.sf.ehcache.CacheException e) {
			throw new ObjectCacheErrorException(e);
		}
	}

	public void destroy() throws ObjectCacheErrorException {
		try {
			ehCache.getCacheManager().removeCache(ehCache.getName());
		} catch (IllegalStateException e) {
			throw new ObjectCacheErrorException(e);
		} catch (net.sf.ehcache.CacheException e) {
			throw new ObjectCacheErrorException(e);
		}
	}

	public String getRegionName() {
		return ehCache.getName();
	}

	public long getSizeInMemory() {
		try {
			return ehCache.calculateInMemorySize();
		} catch (Throwable t) {
			return -1;
		}
	}

	public long getElementCountInMemory() throws ObjectCacheErrorException {
		try {
			return ehCache.getMemoryStoreSize();
		} catch (net.sf.ehcache.CacheException ce) {
			throw new ObjectCacheErrorException(ce);
		}
	}

	public long getElementCountOnDisk() {
		return ehCache.getDiskStoreSize();
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public Map toMap() throws ObjectCacheErrorException {
		try {
			Map result = new HashMap();
			Iterator iter = ehCache.getKeys().iterator();
			while (iter.hasNext()) {
				Object key = iter.next();
				result.put(key, ehCache.get(key).getObjectValue());
			}
			return result;
		} catch (Exception e) {
			throw new ObjectCacheErrorException(e);
		}
	}

	public String toString() {
		return "EHCache(" + getRegionName() + ')';
	}

	public _IObjectCacheStatistics getStatistics() {
		return new EhCacheStatistics(ehCache);
	}

	public long getTimeToLiveSeconds() {
		return ehCache.getCacheConfiguration().getTimeToLiveSeconds();
	}

	public void setTimeToLiveSeconds(long timeToLiveSeconds) {
		CacheConfiguration config = this.ehCache.getCacheConfiguration();
		config.setTimeToLiveSeconds(timeToLiveSeconds);
		config.setTimeToIdleSeconds(timeToLiveSeconds / 2);
		config.setDiskExpiryThreadIntervalSeconds(timeToLiveSeconds);
	}

}
