package pl.slawas.scjdbc4jpa.jdbc.cache;

import java.io.Serializable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.config.JpaConfig;
import pl.slawas.scjdbc4jpa.config.JpaProperties;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;

/**
 * 
 * ObjectWithCacheSupport - abstrakcja obiektów wykorzystujących pamięć
 * podręczną.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.2 $
 * 
 */
@SuppressWarnings("serial")
public abstract class ObjectWithCacheSupport implements Serializable,
		_IObjectWithCache {

	final static private Logger logger = LoggerFactory
			.getLogger(ObjectWithCacheSupport.class);

	/** nazwa regionu używanego cache */
	private final String cacheRegionName;

	private final transient _IObjectCache cache;

	private final CacheSharing cacheSharing;
	/** obiekt dostawcy pamięci podręcznej */
	private final CacheProviderEnum provider;

	protected CacheUsage cacheUsage;

	private long cacheTimeToLive = CacheProviderEnum.cacheTimeToLive;

	/**
	 * @param cacheUsage
	 *            informacja o tym czy ma być używana pamięć podręczna. Jako, że
	 *            jest to inicjalizacja obiektu, to parametr powinien przyjmować
	 *            jedną z dwóch wartości: {@link CacheUsage#TO_USE} oraz
	 *            {@link CacheUsage#NONE}
	 * @param cacheRegionName
	 *            nazwa regionu pamięci podręcznej
	 * @param cacheSharing
	 *            informacja o współdzielenia cache'a ({@link CacheSharing}.
	 * @param useDefaultRegion
	 *            informacja o tym czy ma być użyty domyślny region. Jeżeli
	 *            podamy ten parametr jako wartość {@code null}, to wtedy
	 *            informacja ta zostanie pobrana z wartości
	 *            {@link CacheProviderEnum#CACHE_USE_DEFAULT_REGION}.
	 */
	protected ObjectWithCacheSupport(CacheUsage cacheUsage,
			String cacheRegionName, CacheSharing cacheSharing,
			Boolean useDefaultRegion) {

		/** obiekt tworzonego cache'a */
		_IObjectCache preparedCache = null;
		/** nazwa regionu tworzonego cache'a */
		String preparedCacheRegionName = null;
		this.cacheUsage = cacheUsage;
		CacheProviderEnum lPprovider = null;

		if (cacheUsage != null && !cacheUsage.equals(CacheUsage.NONE)) {
			try {
				/* ustawienie parametrów cache'a */
				lPprovider = CacheProviderEnum
						.valueOf(CacheProviderEnum.CACHE_PROVIDER);
				if (lPprovider.isAllowed()) {
					logger.trace("[ {} ] Zdefiniowano uzycie cache'a '{}'.",
							new Object[] { getClass().getSimpleName(),
									lPprovider.toString() });
					_IObjectCacheProvider cacheManager = lPprovider
							.getProvider();
					preparedCacheRegionName = JpaProperties.CACHE_DEFAULT_REGION_NAME;

					if ((useDefaultRegion == null ? CacheProviderEnum.CACHE_USE_DEFAULT_REGION
							: useDefaultRegion)) {
						logger.trace(
								"[ {} ] Zdefiniowano uzycie domyslnego regionu '{}'.",
								new Object[] { getClass().getSimpleName(),
										JpaProperties.CACHE_DEFAULT_REGION_NAME });
					} else {
						preparedCacheRegionName = cacheRegionName;
						logger.trace(
								"[ {} ] Zdefiniowano uzycie regionu '{}'.",
								new Object[] { getClass().getSimpleName(),
										preparedCacheRegionName });
					}
					preparedCache = cacheManager
							.getCache(preparedCacheRegionName);
					if (CacheProviderEnum.CACHE_USE_DEFAULT_REGION) {
						preparedCache.setTimeToLiveSeconds(cacheTimeToLive);
					} else {
						String originalKey = new String(preparedCacheRegionName);
						String cacheTimetToLiveSeconds = JpaConfig
								.getInstance()
								.get(originalKey
										+ "."
										+ JpaProperties.CACHE_TIME_TO_LIVE_SUFFIX);
						if (cacheTimetToLiveSeconds != null) {
							preparedCache.setTimeToLiveSeconds(Integer
									.valueOf(cacheTimetToLiveSeconds));
						} else {
							preparedCache.setTimeToLiveSeconds(cacheTimeToLive);
						}
					}
				} else {
					if (logger.isTraceEnabled())
						logger.warn("[ {} ] Nie zdefiniowano uzycia kesz-a.",
								new Object[] { getClass().getSimpleName() });
					preparedCache = null;
				}
			} catch (IllegalArgumentException e) {
				logger.error("[ {} ] Zla nazwa provider'a cache'a: {}.",
						new Object[] { getClass().getSimpleName(),
								CacheProviderEnum.CACHE_PROVIDER });
				logger.warn("[ {} ] Definicja uzycia cache'a pominięta.",
						new Object[] { getClass().getSimpleName() });
				preparedCache = null;
			}
		}
		this.cache = preparedCache;
		this.cacheRegionName = preparedCacheRegionName;
		this.cacheSharing = cacheSharing;
		this.provider = lPprovider;
	}

	public boolean cacheIsAllowed() {
		return (this.cache != null && !cacheUsage.equals(CacheUsage.NONE));
	}

	public _IObjectCache getCache() {
		return this.cache;
	}

	public String getCacheRegionName() {
		return this.cacheRegionName;
	}

	public long getCacheTimeToLive() {
		return this.cacheTimeToLive;
	}

	public void setCacheTimeToLive(long cacheTimeToLive) {
		if (this.cacheIsAllowed()) {
			logger.debug(
					"[{}] Ustawiam niedomyśny cacheTimeToLive: {}",
					new Object[] { getClass().getSimpleName(), cacheTimeToLive });
			this.cacheTimeToLive = cacheTimeToLive;
			this.cache.setTimeToLiveSeconds(cacheTimeToLive);
		}

	}

	public boolean cacheIsShared() {
		return this.cacheSharing.isShared();
	}

	public void close() {
		if (cacheIsAllowed() && !cacheIsShared()) {
			try {
				this.cache.destroy();
				logger.trace("Zniszczylem cache {}", this.cache.getRegionName());
			} catch (ObjectCacheErrorException ignore) {
				logger.warn("Ooops... nie udalo sie niszczenie cache'a '"
						+ this.cache.getRegionName() + "'", ignore);
			}
		}

	}

	@Override
	protected void finalize() throws Throwable {
		try {
			this.close();
		} finally {
			super.finalize();
		}
	}

	/**
	 * @return the {@link #cacheUsage}
	 */
	public CacheUsage getCacheUsage() {
		return cacheUsage;
	}

	/**
	 * @param cacheUsage
	 *            the {@link #cacheUsage} to set
	 */
	public void setCacheUsage(CacheUsage cacheUsage) {
		this.cacheUsage = cacheUsage;
	}

	@Override
	public CacheProviderEnum getCacheProvider() {
		return this.provider;
	}

}
