package pl.slawas.scjdbc4jpa.jdbc.cache.managers;

import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.SelectQuery;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache._IObjectCache;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.EntityNotJetSavedIdDbException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;

/**
 * 
 * EntityCacheUtils wsparcie narzędziowe dla obróbki danych rezydujących w keszu
 * związanych z encjami.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.1 $
 * 
 */
public class EntityCacheUtils {

	final static private Logger logger = LoggerFactory
			.getLogger(EntityCacheUtils.class);

	/**
	 * Sprawdzenie czy można użyć obiektu z kesz-a
	 * 
	 * @param cachedObject
	 *            obiekt w keszu
	 * @return jeżeli można użyc element z kesz-a to zostanie zwrócona wartość
	 *         {@code true}, w przeciwnym wypadku {@code false}.
	 */
	public static boolean cachedObjectCanBeUsed(Object cachedObject) {
		if (cachedObject == null)
			return false;
		CacheUsage status = ((_ICachedEntityInfo) cachedObject)
				.getObjectCacheStatus();
		if (status.equals(CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE))
			throw new EntityNotJetSavedIdDbException();
		return status.equals(CacheUsage.TO_USE);
	}

	/**
	 * Pobietranie (tworzenie) identyfikatora encji.
	 * 
	 * @param primaryKey
	 *            obiekt klucza głownego encji
	 * @param entityClassName
	 *            nazwa klasy encji
	 * @return utworzony identyfikator
	 */
	public static final String getEntityId(Object primaryKey,
			String entityClassName) {
		String entityId;
		entityId = entityClassName + "[pk@"
				+ Integer.toHexString(primaryKey.hashCode()) + "]";
		return entityId;
	}

	/**
	 * Usuwanie skeszowanych zapytań
	 * 
	 * @param cleanerName
	 *            nazwa czyścicela (potrzebne do rejestracji informacji w logu)
	 * @param cachedQueryInfoList
	 *            lista informacji o zapytaniach, które mają zostać usuniete z
	 *            kesza.
	 * @return liczba faktycznie usunietych z kesza zapytań
	 */
	public static int clearCachedQuery(String cleanerName,
			Set<SelectQuery.QueryCacheInfo> cachedQueryInfoList) {
		int counter = 0;
		for (SelectQuery.QueryCacheInfo info : cachedQueryInfoList) {
			if (info.cacheIsAllowed()) {
				_IObjectCache usedCache = info.getCache();
				try {
					if (logger.isTraceEnabled()) {
						logger.trace(
								"[{}] Usuwam zapytanie o id = '{}' z kesza '{}'",
								new Object[] { cleanerName, info.getQueryId(),
										usedCache.getRegionName() });
					}
					usedCache.remove(info.getQueryId());
					counter++;
				} catch (ObjectCacheErrorException e) {
					if (logger.isTraceEnabled()) {
						logger.error("[" + cleanerName
								+ "] Blad usuwania obiektu z kesz-a '"
								+ usedCache.getRegionName() + "'", e);
					}
				}

			}
		}
		cachedQueryInfoList.clear();
		return counter;
	}

	/**
	 * Usuwanie skeszowanych zapytań
	 * 
	 * @param cleanerName
	 *            nazwa czyścicela (potrzebne do rejestracji informacji w logu)
	 * @param cachedQueryInfoList
	 *            lista informacji o zapytaniach, które mają zostać usuniete z
	 *            kesza.
	 * @param withoutQuery
	 *            zapytanie, które ma być pominięte podczas usuwania (to
	 *            zapytanie zostanie w keszu)
	 * @return liczba faktycznie usunietych z kesza zapytań
	 */
	public static int clearCachedQuery(String cleanerName,
			Set<SelectQuery.QueryCacheInfo> cachedQueryInfoList,
			SelectQuery.QueryCacheInfo withoutQuery) {

		if (withoutQuery == null)
			return clearCachedQuery(cleanerName, cachedQueryInfoList);

		int counter = 0;
		for (SelectQuery.QueryCacheInfo info : cachedQueryInfoList) {
			if (info.cacheIsAllowed()) {
				_IObjectCache usedCache = info.getCache();
				try {
					if (!info.getQueryId().equals(withoutQuery.getQueryId())) {
						if (logger.isTraceEnabled()) {
							logger.trace(
									"[{}] Usuwam zapytanie o id = '{}' z kesza '{}'",
									new Object[] { cleanerName,
											info.getQueryId(),
											usedCache.getRegionName() });
						}
						usedCache.remove(info.getQueryId());
						counter++;
					} else if (logger.isTraceEnabled()) {
						logger.trace(
								"[{}] Pomijam usuwanie zapytania o id = '{}' w keszu '{}'",
								new Object[] { cleanerName, info.getQueryId(),
										usedCache.getRegionName() });
					}
				} catch (ObjectCacheErrorException e) {
					if (logger.isTraceEnabled()) {
						logger.error("[" + cleanerName
								+ "] Blad usuwania obiektu z kesz-a '"
								+ usedCache.getRegionName() + "'", e);
					}
				}

			}
		}
		cachedQueryInfoList.clear();
		cachedQueryInfoList.add(withoutQuery);
		return counter;
	}

}
