package pl.slawas.scjdbc4jpa.jdbc.cache.managers;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityMetaData;
import pl.slawas.scjdbc4jpa.jdbc.SelectQuery;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheSharing;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache.ObjectWithCacheSupport;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityParserLevel;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityValueParser;

/**
 * 
 * EntityFirstLevelCacheManager zarządzjąca keszem podręcznym z encjami.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.1 $
 * 
 */
public class EntityFirstLevelCacheManager extends ObjectWithCacheSupport {

	private static final long serialVersionUID = 4899440999809248586L;

	final static private Logger logger = LoggerFactory
			.getLogger(EntityFirstLevelCacheManager.class);

	/**
	 * @param cacheRegionName
	 *            nazwa regionu kesza
	 */
	public EntityFirstLevelCacheManager(String cacheRegionName) {
		super(CacheUsage.TO_USE, cacheRegionName, CacheSharing.NONE, false);
	}

	/**
	 * Pobranie obiektu z informacjami o encji z kesza
	 * 
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @return keszowany obiekt z danymi encji albo wartość {@code null}, gdy
	 *         danego obiektu o zadanym 'entityId' nie ma w keszu.
	 */
	public FirstLevelCacheObject getFirstLevelCacheObject(String entityId) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return null;

		try {
			Object cachedObject = getCache().get(entityId);
			if (cachedObject != null) {
				return (FirstLevelCacheObject) cachedObject;
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad pobierania obiektu z kesz-a '"
						+ getCache().getRegionName() + "'", e);
			}
		}
		return null;
	}

	/**
	 * Usuwanie obiektu z informacjami o encji z kesza
	 * 
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @param isPermanentRemove
	 *            flaga informująca czy usuwanie keszowanego obiektu jest
	 *            związane z jej ostatecznym usunięciem encji.
	 */
	public void removeFirstLevelCacheObject(String entityId,
			boolean isPermanentRemove) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return;

		try {
			FirstLevelCacheObject cachedObject = (FirstLevelCacheObject) getCache()
					.get(entityId);
			if (cachedObject != null) {
				if (isPermanentRemove) {
					cachedObject.removeEntitySharedInfo(entityId);
				}
				getCache().remove(entityId);
				if (logger.isTraceEnabled()) {
					logger.trace("[{}] Usunalem obiekt z kesz-a", new Object[] {
							getCacheRegionName(), entityId });
				}
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad usuwania obiektu z kesz-a '"
						+ getCache().getRegionName() + "'", e);
			}
		}
	}

	/**
	 * Odświeżanie wartości encji w keszu (w obiekcie z informacjami o encji)
	 * 
	 * @param <T>
	 * @param entityManagerName
	 *            nazwa entity managera (potrzebne do celów informacyjnych)
	 * @param cachedObject
	 *            keszowany obiekt
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @param queryCacheInfo
	 *            informacje o keszoawanym wyniku zapytania
	 * @param metadata
	 *            metadane dotyczące encji
	 * @param newEntityValue
	 *            nowa wartość odświeżanej encji
	 * @param newCacheUsageStatus
	 *            status użycia w keszu
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws SQLException
	 * @throws InstantiationException
	 */
	public <T> void refreshEntityInFirstLevelCache(String entityManagerName,
			FirstLevelCacheObject cachedObject, String entityId,
			SelectQuery.QueryCacheInfo queryCacheInfo,
			JDBCEntityMetaData metadata, T newEntityValue,
			CacheUsage newCacheUsageStatus) throws IllegalAccessException,
			InvocationTargetException, NoSuchMethodException, SQLException,
			InstantiationException {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return;

		if (cachedObject == null) {
			/* dodanie obiektu do kesza - START */
			cachedObject = addEntity2FirstLevelCache(entityManagerName,
					entityId, metadata, newEntityValue, newCacheUsageStatus);
			/* dodanie obiektu do kesza - KONIEC */
		} else {
			/* odświeżenie obiektu w keszu - START */
			cachedObject.refreshEntity(newEntityValue);
			if (cachedObject.getObjectCacheStatus().equals(CacheUsage.REFRESH)) {
				cachedObject.clearCachedQueryInCache(queryCacheInfo);
				cachedObject.setObjectCacheStatus(CacheUsage.TO_USE);
			}
			/* odświeżenie obiektu w keszu - KONIEC */
		}
		if (queryCacheInfo != null)
			cachedObject.addCachedQueryInfo(queryCacheInfo);
	}

	/**
	 * Utworzenie obiektu z informacjami o encji i dodanie go do kesza
	 * pierwszego poziomu.
	 * 
	 * @param <T>
	 * @param entityManagerName
	 *            nazwa entity managera (potrzebne do celów informacyjnych)
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @param metadata
	 *            metadane dotyczące encji
	 * @param newEntityValue
	 *            nowa wartość encji
	 * @param cacheUsageStatus
	 *            status użycia w keszu
	 * @return dodany keszowany obiekt
	 */
	public <T> FirstLevelCacheObject addEntity2FirstLevelCache(
			String entityManagerName, String entityId,
			JDBCEntityMetaData metadata, T newEntityValue,
			CacheUsage cacheUsageStatus) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return null;

		FirstLevelCacheObject cachedObject;
		cachedObject = new FirstLevelCacheObject(entityManagerName, metadata,
				entityId, newEntityValue, cacheUsageStatus);
		try {
			getCache().put(entityId, cachedObject);
			if (logger.isTraceEnabled()) {
				logger.trace("[{}] Dodalem obiekt do kesz-a", new Object[] {
						getCacheRegionName(), entityId });
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad dodania obiektu z kesz-a '"
						+ getCacheRegionName() + "'", e);
			}
		}
		return cachedObject;
	}

	/**
	 * Ustawienie statusu obiektu w keszu do {@link CacheUsage#REFRESH}
	 * 
	 * @param entityId
	 *            identyfikator obiektu w keszu (
	 *            {@link #getEntityId(Object, String)})
	 */
	public void setToRefreshFirstLevelCacheObject(String entityId) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return;

		try {
			FirstLevelCacheObject cachedObject = (FirstLevelCacheObject) getCache()
					.get(entityId);
			if (cachedObject != null) {
				cachedObject.setObjectCacheStatus(CacheUsage.REFRESH);
				if (logger.isTraceEnabled()) {
					logger.trace(
							"[{}] Obiekt oznaczony jako 'do aktualizacji' w kesz-u",
							new Object[] { getCacheRegionName(), entityId });
				}
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad aktuakizacji obiektu w kesz-u '"
						+ getCache().getRegionName() + "'", e);
			}
		}
	}

	/**
	 * 
	 * FirstLevelCacheObject obiekt przechowywania informacji o encji w keszu
	 * pierwszego poziomu.
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.1 $
	 * 
	 */
	public static class FirstLevelCacheObject implements Serializable,
			_ICachedEntityInfo, _ICachedQueryInfoManager {

		private static final long serialVersionUID = -5358072358648428212L;

		private final String entityManagerName;

		private final JDBCEntityMetaData metaData;

		private final EntitySecondLevelCacheManager secondLevelCacheManager;

		private Object entity;

		private final String entityId;

		private boolean isFlushedAfterDML = true;

		private final Object lockObjectCacheStatus = new Object();

		/**
		 * lokalny status dla encji właśnie wstawionych/zmienianych (używanych
		 * podczas aktywnej transakcji)
		 */
		private CacheUsage lObjectCacheStatus;

		/**
		 * lokalna lista zapytań dla encji własnie wstawionych/zmienianych
		 * (używanych podczas aktywnej transakcji)
		 */
		private Set<SelectQuery.QueryCacheInfo> lCachedQueryInfoList = new HashSet<SelectQuery.QueryCacheInfo>();

		/**
		 * @param entityManagerName
		 *            nazwa entity managera (potrzebne do rejestracji informacji
		 *            w logu)
		 * @param metaData
		 *            metadane dotyczące encji
		 * @param entityId
		 *            identyfikator encji
		 *            {@link JDBCEntityMetaData#getEntityId(Object, String)}).
		 *            Jeżeli bedzie {@code null}, to współdzielona informacja o
		 *            encji nie będzie dostępna, ze wszystkimi jej
		 *            konsekwencjami.
		 * @param entity
		 *            obiekt encji
		 * @param objectCacheStatus
		 *            status użycia encji w keszu
		 */
		private <T> FirstLevelCacheObject(String entityManagerName,
				JDBCEntityMetaData metaData, String entityId, T entity,
				CacheUsage objectCacheStatus) {
			super();
			if (metaData == null)
				throw new IllegalArgumentException(
						"Nie zadeklarowac pustej wartosci metadanych");
			if (entity == null)
				throw new IllegalArgumentException(
						"Nie zadeklarowac pustej wartosci encji");
			if (entityId == null)
				throw new IllegalArgumentException(
						"Identyfikator encji nie moze byc pusty");
			if (objectCacheStatus == null)
				throw new IllegalArgumentException(
						"Status uzycia ecncji w keszu nie moze byc pusty");
			if (entityManagerName == null)
				throw new IllegalArgumentException(
						"Nazwa manager-a encji nie moze byc pusta");

			this.lObjectCacheStatus = objectCacheStatus;
			this.entityId = entityId;
			this.metaData = metaData;
			this.entity = entity;
			this.entityManagerName = entityManagerName;
			this.secondLevelCacheManager = this.metaData
					.getEntitySecondLevelCacheManager();
			if (canUseSecondLevelCacheObject()) {
				/* sprawdzam czy jest informacja wspóldzielona w keszu 2 poziomu */
				EntitySecondLevelCacheManager.SecondLevelCacheObject sharedInfo = this.secondLevelCacheManager
						.getEntitySharedInfo(entityId);
				if (sharedInfo == null) {
					/* nie ma zatem dodaje */
					this.secondLevelCacheManager.addEntitySharedInfo(entityId,
							objectCacheStatus);
				}
			}

		}

		private boolean canUseSecondLevelCacheObject() {
			/*
			 * nie moge używać kesza współdzielonego, dla encji która jeszcze
			 * nie jest w bazie, albo jest w trakcie transakcji
			 */
			return !this.lObjectCacheStatus
					.equals(CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE)
					&& this.secondLevelCacheManager.cacheIsAllowed();
		}

		/**
		 * Usuwanie informacji o encji z kesza współdzielonego.
		 * 
		 * @param entityId
		 *            identyfikator encji
		 *            {@link JDBCEntityMetaData#getEntityId(Object, String)}).
		 */
		public void removeEntitySharedInfo(String entityId) {
			if (entityId == null)
				throw new IllegalArgumentException(
						"Nie mozna usuwac z kesz-a informacji o encji gdy nieznany jest jej identyfikator");
			if (this.secondLevelCacheManager.cacheIsAllowed())
				this.secondLevelCacheManager.removeEntitySharedInfo(entityId,
						true);
		}

		/**
		 * @return the entity
		 */
		public Object getEntity() {
			return entity;
		}

		/**
		 * @param entity
		 *            the entity to set
		 * @throws InvocationTargetException
		 * @throws IllegalAccessException
		 * @throws IllegalArgumentException
		 * @throws InstantiationException
		 * @throws SQLException
		 * @throws NoSuchMethodException
		 */
		private synchronized Object refreshEntity(Object entity)
				throws IllegalArgumentException, IllegalAccessException,
				InvocationTargetException, NoSuchMethodException, SQLException,
				InstantiationException {

			if (entity == null) {
				this.entity = entity;
				return this.entity;
			}

			if (this.entity == null) {
				this.entity = entity;
				return this.entity;
			}
			JDBCEntityValueParser.copyEntity(
					metaData.getTableDefinition(JDBCEntityParserLevel.THIRD),
					this.entity, entity);
			return this.entity;
		}

		/**
		 * @return the objectCacheStatus - informacja o tym czy obiekt może być
		 *         użyty, czy ma być odświeżony. Zwracana jest jedną z czterech
		 *         wartości: {@link CacheUsage#TO_USE},
		 *         {@link CacheUsage#REFRESH} ,
		 *         {@link CacheUsage#TO_USE_IN_FIRST_LEVEL_CACHE} oraz
		 *         {@link CacheUsage#TO_REMOVE}.
		 */
		public CacheUsage getObjectCacheStatus() {
			if (canUseSecondLevelCacheObject()) {
				EntitySecondLevelCacheManager.SecondLevelCacheObject sharedInfo = this.secondLevelCacheManager
						.getEntitySharedInfo(entityId);
				if (sharedInfo != null) {
					return sharedInfo.getObjectCacheStatus();
				} else {
					/*
					 * jezeli 2 poziom jest dostepny, to wspóldzielona
					 * informacja powinna się tam znajdować, czyba, że ktoś
					 * usunął daną encję, zatem nie pozostaje nam nic innego jak
					 * oznaczyć element jako "do usunięcia"
					 */
					return CacheUsage.TO_REMOVE;
				}
			} else {
				/* 2 poziom nie jest dostepny - uzywam lokalnego statusu */
				return this.lObjectCacheStatus;
			}
		}

		/**
		 * @param objectCacheStatus
		 *            ustawianie informacji o tym czy obiekt może być użyty, czy
		 *            ma być odświeżony. Parametr może przyjmować jedną z trzech
		 *            wartości: {@link CacheUsage#TO_USE},
		 *            {@link CacheUsage#REFRESH}
		 */
		public void setObjectCacheStatus(
				CacheUsage objectCacheStatus) {
			if (objectCacheStatus == null)
				throw new IllegalArgumentException(
						"Nie mozna ustawiac statusu do wartosci pustej");
			if (objectCacheStatus
					.equals(CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE))
				throw new IllegalArgumentException(
						"Nie zmieniac statusu obiektu na "
								+ objectCacheStatus
								+ ". Wpierw go usun z kesza pierwszego poziomu, a potem dodaj na nowo.");

			synchronized (this.lockObjectCacheStatus) {
				if (canUseSecondLevelCacheObject()) {
					EntitySecondLevelCacheManager.SecondLevelCacheObject sharedInfo = this.secondLevelCacheManager
							.getEntitySharedInfo(entityId);
					if (sharedInfo != null) {
						sharedInfo.setObjectCacheStatus(objectCacheStatus);
						sharedInfo.clearCachedQueryInCache();
					} else {
						/*
						 * jezeli 2 poziom jest dostepny, to wspóldzielona
						 * informacja powinna się tam znajdować, czyba, że ktoś
						 * usunął daną encję, zatem nie pozostaje nam nic innego
						 * jak oznaczyć element jako "do usunięcia"
						 */
						objectCacheStatus = CacheUsage.TO_REMOVE;
					}

				} else {
					switch (objectCacheStatus) {
					case REFRESH:
					case TO_USE:
						/*
						 * sprawdzam czy jest informacja wspóldzielona w keszu 2
						 * poziomu
						 */
						EntitySecondLevelCacheManager.SecondLevelCacheObject sharedInfo = this.secondLevelCacheManager
								.getEntitySharedInfo(entityId);
						if (sharedInfo == null) {
							/* nie ma, zatem dodaje */
							sharedInfo = this.secondLevelCacheManager
									.addEntitySharedInfo(entityId,
											objectCacheStatus);
						}
						sharedInfo.setObjectCacheStatus(objectCacheStatus);
						sharedInfo.clearCachedQueryInCache();
						break;
					default:
						break;
					}
				}
				this.lObjectCacheStatus = objectCacheStatus;
			}
		}

		public void addCachedQueryInfo(SelectQuery.QueryCacheInfo queryInfo) {
			if (queryInfo == null)
				throw new IllegalArgumentException(
						"Nie mozna ustawiac pustej informacji o zapytaniu");

			if (canUseSecondLevelCacheObject()) {
				EntitySecondLevelCacheManager.SecondLevelCacheObject sharedInfo = this.secondLevelCacheManager
						.getEntitySharedInfo(entityId);
				if (sharedInfo != null) {
					sharedInfo.addCachedQueryInfo(queryInfo);
				} else {
					/*
					 * jezeli 2 poziom jest dostepny, to wspóldzielona
					 * informacja powinna się tam znajdować, czyba, że ktoś
					 * usunął daną encję, zatem nie pozostaje nam nic innego jak
					 * oznaczyć element jako "do usunięcia"
					 */
					lObjectCacheStatus = CacheUsage.TO_REMOVE;
				}
			} else {
				this.lCachedQueryInfoList.add(queryInfo);
			}
		}

		/**
		 * Usuwanie zapytań z współdzielonego kesza dla obiektu w keszu
		 * pierwszego poziomu, który ma status {@link CacheUsage#REFRESH}
		 */
		public void clearCachedQueryInCache() {
			switch (this.lObjectCacheStatus) {
			case REFRESH:
			case TO_REMOVE:
				EntityCacheUtils.clearCachedQuery(entityManagerName
						+ ".FirstLevelCacheObject", this.lCachedQueryInfoList);
				break;
			default:
				break;
			}
		}

		/**
		 * Usuwanie zapytań z współdzielonego kesza dla obiektu w keszu
		 * pierwszego poziomu, który ma status {@link CacheUsage#REFRESH} z
		 * pominięciem zapytania, którego identyfikator został podany.
		 * 
		 * @param withoutQueryId
		 *            identyfikator zapytania, które ma nie być ususnięte z
		 *            współdzielonego kesz-a, bo np. właśnie zostało wykonane, i
		 *            obiekt w keszu pierwszego poziomu zostanie na jego
		 *            podstawie odświeżony.
		 */
		public void clearCachedQueryInCache(
				SelectQuery.QueryCacheInfo withoutQueryId) {
			switch (this.lObjectCacheStatus) {
			case REFRESH:
			case TO_REMOVE:
				EntityCacheUtils.clearCachedQuery(entityManagerName
						+ ".FirstLevelCacheObject", this.lCachedQueryInfoList,
						withoutQueryId);
				break;
			default:
				break;
			}
		}

		@Override
		protected void finalize() throws Throwable {
			try {
				this.clearCachedQueryInCache();
			} catch (Exception ignore) {
			} finally {
				super.finalize();
			}
		}

		/**
		 * @return the entityId
		 */
		public String getEntityId() {
			return entityId;
		}

		/**
		 * @return the isFlushedAfterDML
		 */
		public boolean isFlushedAfterDML() {
			return isFlushedAfterDML;
		}

		/**
		 * ustwia flage, że obiekt encji jeszcze nie zostal wyslany do bazy
		 * danych.
		 */
		public void setIsNotFlushedAfterDML() {
			this.isFlushedAfterDML = false;
			EntityCacheUtils.clearCachedQuery(entityManagerName
					+ ".FirstLevelCacheObject", this.lCachedQueryInfoList);
		}

		/**
		 * ustwia flage, że obiekt encji zostal wyslany do bazy danych.
		 */
		public void setIsFlushedAfterDML() {
			this.isFlushedAfterDML = true;
		}

		/**
		 * @return the secondLevelCacheManager
		 */
		public EntitySecondLevelCacheManager getSecondLevelCacheManager() {
			return secondLevelCacheManager;
		}

	}

}
