package pl.slawas.scjdbc4jpa.jdbc.cache.managers;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.SelectQuery;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheSharing;
import pl.slawas.scjdbc4jpa.jdbc.cache.CacheUsage;
import pl.slawas.scjdbc4jpa.jdbc.cache.ObjectWithCacheSupport;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectCacheErrorException;

/**
 * 
 * EntitySecondLevelCacheManager zarządzjąca keszem współdzielonych informacji o
 * encjach
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.1 $
 * 
 */
public class EntitySecondLevelCacheManager extends ObjectWithCacheSupport {

	private static final long serialVersionUID = 4899440999809248586L;

	final static Logger logger = LoggerFactory
			.getLogger(EntitySecondLevelCacheManager.class);

	/**
	 * 
	 * 
	 * @param cacheUsage
	 *            informacja o tym czy ma być używany kesz. Jako, że jest to
	 *            inicjalizacjia obiektu, to parametr powinien przyjmować jedną
	 *            z dwóch wartości: {@link CacheUsage#TO_USE} oraz
	 *            {@link CacheUsage#NONE}
	 * @param tableDefinition
	 *            definicja tabeli, po to by na jej podstawie ustalić nazwę
	 *            regionu dl kesz-a
	 */
	public EntitySecondLevelCacheManager(CacheUsage cacheUsage,
			JDBCTableDesc tableDefinition) {
		super(cacheUsage, tableDefinition.getCacheRegionName(),
				CacheSharing.SHARED, false);
	}

	/**
	 * Pobranie współdzielonej informacji o encji z kesza
	 * 
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @return keszowany obiekt współdzielonej informacji o encji albo wartość
	 *         {@code null}, gdy danego obiektu o zadanym 'entityId' nie ma w
	 *         keszu.
	 */
	SecondLevelCacheObject getEntitySharedInfo(String entityId) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return null;

		try {
			Object cachedObject = getCache().get(entityId);
			if (cachedObject != null) {
				return (SecondLevelCacheObject) cachedObject;
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad pobierania obiektu z kesz-a '"
						+ getCache().getRegionName() + "'", e);
			}
		}
		return null;
	}

	/**
	 * Dodanie (utworzenie) współdzielonej informacji o encji do kesza
	 * 
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @param objectCacheStatus
	 *            status dla wstawianego obiektu. Może przyjmować następujące
	 *            wartości: {@link CacheUsage#TO_REMOVE},
	 *            {@link CacheUsage#TO_USE}, {@link CacheUsage#REFRESH}
	 * @return dodany keszowany obiekt współdzielonej informacji o encji
	 */
	SecondLevelCacheObject addEntitySharedInfo(String entityId,
			CacheUsage objectCacheStatus) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return null;

		SecondLevelCacheObject cachedObject = new SecondLevelCacheObject(
				getCache().getRegionName(), objectCacheStatus);
		try {
			getCache().put(entityId, cachedObject);
			if (logger.isTraceEnabled()) {
				logger.trace("[{}] Dodalem obiekt do kesz-a", new Object[] {
						getCache().getRegionName(), entityId });
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad dodania obiektu do kesz-a '"
						+ getCache().getRegionName() + "'", e);
			}
		}
		return cachedObject;
	}

	/**
	 * Usuwanie współdzielonej informacji o encji z kesza
	 * 
	 * @param entityId
	 *            identyfikator encji {@link #getEntityId(Object, String)})
	 * @param cleanCachedQuery
	 *            flaga informująca czy mają być również wyczyszczone skeszowane
	 *            wyniki zapytań
	 */
	public void removeEntitySharedInfo(String entityId, boolean cleanCachedQuery) {
		/* jak kesz nie jest dostepny to nic tu po nas */
		if (!this.cacheIsAllowed())
			return;

		try {
			Object cachedObject = getCache().get(entityId);
			if (cachedObject != null) {
				if (cleanCachedQuery) {
					((SecondLevelCacheObject) cachedObject)
							.setObjectCacheStatus(CacheUsage.REFRESH);
					((SecondLevelCacheObject) cachedObject)
							.clearCachedQueryInCache();
				}
				getCache().remove(entityId);
				if (logger.isTraceEnabled()) {
					logger.trace("[{}] Usunalem obiekt z kesz-a", new Object[] {
							getCache().getRegionName(), entityId });
				}
			}
		} catch (ObjectCacheErrorException e) {
			if (logger.isTraceEnabled()) {
				logger.error("Blad usuwania obiektu z kesz-a '"
						+ getCache().getRegionName() + "'", e);
			}
		}
	}

	/**
	 * 
	 * SecondLevelCacheObject obiekt przechowywany w keszu współdzielonych
	 * informacji o encjach
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.1 $
	 * 
	 */
	static class SecondLevelCacheObject implements Serializable,
			_ICachedEntityInfo, _ICachedQueryInfoManager {

		private static final long serialVersionUID = -7431510412202997245L;

		private Set<SelectQuery.QueryCacheInfo> cachedQueryInfoList = new HashSet<SelectQuery.QueryCacheInfo>();

		private CacheUsage objectCacheStatus;

		private final String regionName;

		private final Object lockObjectCacheStatus = new Object();

		private SecondLevelCacheObject(String regionName,
				CacheUsage objectCacheStatus) {
			if (objectCacheStatus == null)
				throw new IllegalArgumentException(
						"Status uzycia ecncji w keszu nie moze byc pusty");
			if (objectCacheStatus
					.equals(CacheUsage.TO_USE_IN_FIRST_LEVEL_CACHE))
				throw new IllegalArgumentException(
						"Nie mozna tworzyc obiektu kesza drugiego poziomu ze statusem "
								+ objectCacheStatus);
			if (regionName == null)
				throw new IllegalArgumentException(
						"Nazwa regionu nie moze byc pusta");

			this.regionName = regionName;
			this.objectCacheStatus = objectCacheStatus;

		}

		public CacheUsage getObjectCacheStatus() {
			return objectCacheStatus;
		}

		public void setObjectCacheStatus(CacheUsage objectCacheStatus) {
			if (objectCacheStatus == null)
				throw new IllegalArgumentException(
						"Nie mozna ustawiac statusu do wartosci pustej");
			synchronized (this.lockObjectCacheStatus) {
				if (logger.isTraceEnabled()) {
					logger.trace("[{}] Obiekt ma nowy status w kesz-u: {}",
							new Object[] { regionName, objectCacheStatus });
				}
				this.objectCacheStatus = objectCacheStatus;
				clearCachedQueryInCache();
			}
		}

		public synchronized void addCachedQueryInfo(
				SelectQuery.QueryCacheInfo queryInfo) {
			if (queryInfo == null)
				throw new IllegalArgumentException(
						"Nie mozna ustawiac pustej informacji o zapytaniu");
			this.cachedQueryInfoList.add(queryInfo);
		}

		public synchronized void clearCachedQueryInCache() {
			switch (this.objectCacheStatus) {
			case REFRESH:
			case TO_REMOVE:
				EntityCacheUtils.clearCachedQuery(regionName
						+ ".SecondLevelCacheObject", this.cachedQueryInfoList);
				break;
			default:
				break;
			}
		}

		public synchronized void clearCachedQueryInCache(
				SelectQuery.QueryCacheInfo withoutQuery) {
			switch (this.objectCacheStatus) {
			case REFRESH:
			case TO_REMOVE:
				EntityCacheUtils.clearCachedQuery(regionName
						+ ".SecondLevelCacheObject", this.cachedQueryInfoList,
						withoutQuery);
				break;
			default:
				break;
			}
		}

		@Override
		protected void finalize() throws Throwable {
			try {
				this.clearCachedQueryInCache();
			} catch (Exception ignore) {
			} finally {
				super.finalize();
			}
		}

	}

}
