package pl.slawas.scjdbc4jpa.jdbc.dao;

import java.io.Serializable;
import java.lang.reflect.ParameterizedType;
import java.util.List;

import javax.persistence.EntityTransaction;
import javax.persistence.PersistenceException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.paging.Page;
import pl.slawas.paging.PagingParams;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.core.ScObjectFactory;
import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityManager;
import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityManagerFactory;
import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityMetaData;
import pl.slawas.scjdbc4jpa.jdbc.JDBCQuery;
import pl.slawas.scjdbc4jpa.jdbc.JDBCTransaction;
import pl.slawas.scjdbc4jpa.jdbc.PagedResult;
import pl.slawas.scjdbc4jpa.jdbc.SortClause;
import pl.slawas.scjdbc4jpa.jdbc.dml.DataModificationType;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCDaoTemplateException;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;

/**
 * 
 * JDBCDaoTemplate - abstrakcja wspierająca implementacje warstwy DAO.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.6 $
 * 
 * @param <Obj>
 *            klasa encji
 */
@SuppressWarnings("serial")
public abstract class JdbcDAOSupport<Obj> implements Serializable,
		_IJdbcDAO<Obj> {

	final protected transient Logger logger = LoggerFactory
			.getLogger(getClass());

	private final Class<Obj> persistentClass;

	private ScObjectFactory aof;

	private JDBCEntityManager em;

	private JDBCEntityMetaData metaData = null;

	@SuppressWarnings("unchecked")
	public JdbcDAOSupport() {
		super();
		this.persistentClass = (Class<Obj>) ((ParameterizedType) this
				.getClass().getGenericSuperclass()).getActualTypeArguments()[0];

	}

	public Class<Obj> getPersistentClass() {
		return persistentClass;
	}

	private int endBulkTransaction(boolean autoCommit) {
		if (autoCommit && this.em.getTransaction().isActive()) {
			this.em.getTransaction().commit();
			return ((JDBCTransaction) em.getTransaction())
					.getSummaryUpdateCounts();
		}
		return 0;
	}

	private int flushAndCommitAnalyze(int i, boolean autoCommit) {
		if ((i != 0) && (i % em.getBatchSize() == 0)) {
			this.em.flush();
		}
		int updateCounts = 0;
		if (autoCommit && (i != 0) && (i % em.getBatchCommitSize() == 0)) {
			this.em.getTransaction().commit();
			updateCounts = ((JDBCTransaction) em.getTransaction())
					.getSummaryUpdateCounts();
			this.em.getTransaction().begin();
		}
		return updateCounts;
	}

	/**
	 * Metoda wykorzystywana tylko i wyłącznie przez metody przetwarzające listy
	 * encji do sprawdzenie czy transakcja jest aktywna w przypadku, gdy ich
	 * parametr 'autoCommit' jest ustawiony na {@code true}. Jeżeli transakcja
	 * nie jest aktywna następuje jej aktywacja, a jeżeli <b>NIE</b> jest
	 * aktywna, zapisywane jest ostrzeżenie o tym, że process łączy się do
	 * aktywnej transakcji. To ostrzeżenie jest bardzo ważne ponieważ, podczas
	 * auto zatwierdzania paczek danych (zobacz metody
	 * {@link #flushAndCommitAnalyze(int, boolean)} oraz
	 * {@link #endBulkTransaction(boolean)}) nastąpi zatwierdzenie tej aktywnej
	 * transakcji, być może wbrew zamierzeniom programisty. Ponadto może
	 * nastąpić przekłamanie informacji o liczbie przetworzonych danych.
	 * 
	 * @see #insertList(List, boolean)
	 * @see #updateList(List, boolean)
	 * @see #removeList(List, boolean)
	 */
	private void joinTransaction() {
		if (!this.em.getTransaction().isActive()) {
			if (logger.isDebugEnabled())
				logger.debug("\n\n Join to new transaction!!!\n\n");
			this.em.getTransaction().begin();
		} else {
			logger.warn("Lacze sie do juz aktywnej transakcji");
		}
	}

	/**
	 * Sprawdzenie czy manager encji jest w kontekście. Metoda wykorzystywana w
	 * implementacji metod zmiany danych takich jak {@link #insert(Object)},
	 * {@link #update(Object)} itp.
	 */
	protected void checkEm() {
		if (this.em == null || !this.em.isOpen()) {
			throw new JDBCDaoTemplateException(
					"EntityManager nie jest w kontekscie");
		}
	}

	/**
	 * Metoda wyzwalacza po operacji wstawienia do bazy danych
	 * 
	 * @param entityObject
	 *            wstawiana encja.
	 */
	protected void afterInsert(Obj entityObject) {
	}

	/**
	 * Metoda wyzwalacza po operacji usuwania z bazy danych
	 * 
	 * @param entityObject
	 *            usuwana encja.
	 */
	protected void afterRemove(Obj entityObject) {
	}

	/**
	 * Metoda wyzwalacza po operacji aktualizacji w bazie danych
	 * 
	 * @param entityObject
	 *            aktualizowana encja.
	 */
	protected void afterUpdate(Obj entityObject) {
	}

	/**
	 * Metoda wyzwalacza przed operacją wstawienia do bazy danych. Możliwość
	 * modyfikacji danych, przed wykonaniem wstawiania.
	 * 
	 * @param entityObject
	 *            wstawiana encja
	 * @return zmodyfikowana encja
	 */
	protected Obj beforeInsert(Obj entityObject) {
		return entityObject;
	}

	/**
	 * Metoda wyzwalacza przed operacją usuwania z bazy danych. Możliwość
	 * modyfikacji danych, przed wykonaniem usuwania.
	 * 
	 * @param entityObject
	 *            usuwana encja
	 * @return zmodyfikowana encja
	 */
	protected Obj beforeRemove(Obj entityObject) {
		return entityObject;
	}

	/**
	 * Metoda wyzwalacza przed operacją aktualizacji w bazie danych. Możliwość
	 * modyfikacji danych, przed wykonaniem aktualizacji.
	 * 
	 * @param entityObject
	 *            aktualizowana encja
	 * @return zmodyfikowana encja
	 */
	protected Obj beforeUpdate(Obj entityObject) {
		return entityObject;
	}

	public Obj find(Object id) {
		return getEntityManager().find(getPersistentClass(), id);
	}

	public PagedResult find(Page page, ConditionClauseSupport[] conditions,
			SortClause sort) {
		return getEntityManager().find(getPersistentClass(), page, conditions,
				sort);
	}

	public PagedResult dump(Page page, ConditionClauseSupport[] conditions,
			SortClause sort) {
		return getEntityManager().dump(getPersistentClass(), page, conditions,
				sort);
	}

	public PagedResult findByIdList(Object[] idList, SortClause sort) {
		JDBCEntityManager localEM = getEntityManager();
		if (idList.length > localEM.getDefaultPagingParams().getMaxCount()) {
			throw new JDBCDaoTemplateException(
					"Liczba warunkow nie moze przekraczac maksymalnej mozliwej liczby zwracanych wierszy: "
							+ localEM.getDefaultPagingParams().getMaxCount());
		}

		JDBCQuery query = getEntityManager().createQuery(persistentClass);
		ConditionClauseSupport condition = JdbcDAOUtils.createINConditionForIdList(
				getMetaData(), idList);

		query.addCondition(condition);
		query.setSortClause(sort);
		return query.execute();
	}

	public JDBCEntityManager getEntityManager() {
		if (this.em == null || !this.em.isOpen()) {
			this.em = (JDBCEntityManager) this.aof.getEntityManager();
		}
		return this.em;
	}

	public void setEntityManager(JDBCEntityManager em) {
		this.em = em;
	}

	public void insert(Obj entityObject) {
		checkEm();
		try {
			entityObject = this.beforeInsert(entityObject);
			this.em.persist(entityObject);
			this.afterInsert(entityObject);
			logger.trace("{}: inserted.", entityObject.getClass()
					.getSimpleName());
		} catch (PersistenceException e) {
			throw new JDBCDaoTemplateException("Blad wstawienia wiersza", e);
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad wstawienia wiersza", e);
		}
	}

	public int insertList(List<Obj> entityObjects, boolean autoCommit) {
		checkEm();
		Page pageDefinition = this.em.getFirstPageDefinition();
		PagingParams pagingParams = LoadJdbcConfiguration
				.getPagingParams(pageDefinition.getSize());
		if (entityObjects.size() > pagingParams.getMaxCount()) {
			throw new JDBCDaoTemplateException(
					"Liczba wstawianych encji nie przekraczac maksymalnej mozliwej liczby zwracanych wierszy: "
							+ pagingParams.getMaxCount());
		}

		try {
			int i = 0;
			int updateCounts = 0;
			if (autoCommit)
				this.joinTransaction();
			for (Obj entityObject : entityObjects) {
				updateCounts = updateCounts
						+ flushAndCommitAnalyze(i, autoCommit);
				entityObject = this.beforeInsert(entityObject);
				this.em.persist(entityObject);
				this.afterInsert(entityObject);
				i++;
			}
			updateCounts = updateCounts + endBulkTransaction(autoCommit);
			if (updateCounts < 0)
				updateCounts = 0;
			if (logger.isTraceEnabled())
				logger.trace("" + persistentClass.getSimpleName() + ": "
						+ Integer.toString(updateCounts) + " rows inserted.");
			return updateCounts;
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad wstawienia wiersza", e);
		}
	}

	/**
	 * TODO Zaimplementować metodę masowego przetwarzania danych
	 * <p>
	 * Masowe przetwarzanie danych. Metoda kontrowersyjna w użyciu, ale może być
	 * przydatna. W zależności od implementacji należy pamiętać, o tym że:
	 * </p>
	 * <ol>
	 * <li><font color="#aa0000">Uwaga!</font> Podczas masowego przetwarzania
	 * danych <b>NIE</b> ma obsługi metoda wyzwalaczy
	 * {@link #beforeInsert(Object)}, {@link #afterInsert(Object)},
	 * {@link #beforeUpdate(Object)}, {@link #afterUpdate(Object)},
	 * {@link #beforeRemove(Object)}, {@link #afterRemove(Object)}. Dane powinny
	 * być już odpowiednio przygotowane (wstawianie, aktualizacja), a jeżeli
	 * nastąpi usuwanie, to trzeba odpowiednio wcześniej wykonać zadania, które
	 * zawarte zostały w wyzwalaczach.</li>
	 * <li>zobacz uwagi {@link JDBCEntityManager#persist(List)}</li>
	 * <li>zobacz uwagi {@link JDBCEntityManager#merge(List)}</li>
	 * <li>zobacz uwagi {@link JDBCEntityManager#remove(List)}</li>
	 * <li>
	 * <font color="#aa0000">Ważnym jest, że metoda zawiera automatyczne
	 * zatwierdzanie transakcji (zatwierdzanie zostanie automatycznie robione co
	 * {@link JDBCEntityManager#getBatchCommitSize()}) oraz sama zarządza
	 * transakcją! Nie należy poprzedzać wywołania tej metody metodą
	 * {@link EntityTransaction#begin()} oraz umieszczać
	 * {@link EntityTransaction#commit()} za, bo z pewnością wyleci
	 * wyjątek.</font></li>
	 * </ol>
	 * 
	 * 
	 * @param entityObjects
	 *            lista encji
	 * @param type
	 *            typ masowego przetwarzania
	 * @return liczba przetworzonych wierszy
	 */
	@Deprecated
	public int massiveDataModification(List<Obj> entityObjects,
			DataModificationType type) {
		checkEm();
		throw new JDBCDaoTemplateException(
				"Metoda jeszcze nie zaimplementowana");
	}

	public void update(Obj entityObject) {
		checkEm();
		try {
			entityObject = this.beforeUpdate(entityObject);
			this.em.merge(entityObject);
			this.afterUpdate(entityObject);
			if (logger.isTraceEnabled())
				logger.trace("" + entityObject.getClass().getSimpleName()
						+ ": changed.");
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad aktualizacji wiersza", e);
		}

	}

	public int updateList(List<Obj> entityObjects, boolean autoCommit) {
		checkEm();
		Page pageDefinition = this.em.getFirstPageDefinition();
		PagingParams pagingParams = LoadJdbcConfiguration
				.getPagingParams(pageDefinition.getSize());
		if (entityObjects.size() > pagingParams.getMaxCount()) {
			throw new JDBCDaoTemplateException(
					"Liczba zmienianych encji nie przekraczac maksymalnej mozliwej liczby zwracanych wierszy: "
							+ pagingParams.getMaxCount());
		}

		try {
			int i = 0;
			int updateCounts = 0;
			if (autoCommit)
				this.joinTransaction();
			for (Obj entityObject : entityObjects) {
				updateCounts = updateCounts
						+ flushAndCommitAnalyze(i, autoCommit);
				entityObject = this.beforeUpdate(entityObject);
				this.em.merge(entityObject);
				this.afterUpdate(entityObject);
				i++;
			}
			updateCounts = updateCounts + endBulkTransaction(autoCommit);
			if (updateCounts < 0)
				updateCounts = 0;
			if (logger.isTraceEnabled())
				logger.trace("" + persistentClass.getSimpleName() + ": "
						+ Integer.toString(updateCounts) + " rows updated.");
			return updateCounts;
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad aktualizacji wiersza", e);
		}
	}

	public void remove(Obj entityObject) {
		checkEm();
		try {
			entityObject = this.beforeRemove(entityObject);
			this.em.remove(entityObject);
			this.afterRemove(entityObject);
			if (logger.isTraceEnabled())
				logger.trace("" + entityObject.getClass().getSimpleName()
						+ ": deleted.");
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad usuwania wiersza", e);
		}
	}

	public int removeList(List<Obj> entityObjects, boolean autoCommit) {
		checkEm();
		Page pageDefinition = this.em.getFirstPageDefinition();
		PagingParams pagingParams = LoadJdbcConfiguration
				.getPagingParams(pageDefinition.getSize());
		if (entityObjects.size() > pagingParams.getMaxCount()) {
			throw new JDBCDaoTemplateException(
					"Liczba usuwanych encji nie przekraczac maksymalnej mozliwej liczby zwracanych wierszy: "
							+ pagingParams.getMaxCount());
		}

		try {
			int i = 0;
			int updateCounts = 0;
			if (autoCommit)
				this.joinTransaction();
			for (Obj entityObject : entityObjects) {
				updateCounts = updateCounts
						+ flushAndCommitAnalyze(i, autoCommit);
				entityObject = this.beforeRemove(entityObject);
				this.em.remove(entityObject);
				this.afterRemove(entityObject);
				i++;
			}
			updateCounts = updateCounts + endBulkTransaction(autoCommit);
			if (updateCounts < 0)
				updateCounts = 0;
			if (logger.isTraceEnabled())
				logger.trace("" + persistentClass.getSimpleName() + ": "
						+ Integer.toString(updateCounts) + " rows deleted.");
			return updateCounts;
		} catch (Exception e) {
			throw new JDBCDaoTemplateException("Blad usuwania wiersza", e);
		}
	}

	public void setScObjectFactory(ScObjectFactory aof) {
		if (aof.getEntityManagerFactory() instanceof JDBCEntityManagerFactory) {
			this.aof = aof;
		} else
			throw new UnsupportedOperationException(
					"Metoda nie wspierana przez ten obiekt "
							+ "dla fabryki obiektów innej niż pl.slawas.scjdbc4jpa.jdbc.dao.JDBCEntityManagerFactory");
	}

	@Override
	protected void finalize() throws Throwable {
		try {
			if (this.em != null && this.em.isOpen()) {
				this.em.close();
			}
		} finally {
			super.finalize();
		}
	}

	public JDBCEntityMetaData getMetaData() {
		if (this.metaData == null) {
			this.metaData = ((JDBCEntityManagerFactory) aof
					.getEntityManagerFactory())
					.getMetaData(getPersistentClass().getName());
		}
		return metaData;
	}

}
