package pl.slawas.scjdbc4jpa.jdbc.dao;

import java.util.List;

import javax.persistence.EntityTransaction;

import pl.slawas.paging.Page;
import pl.slawas.scjdbc4jpa.config.LoadJdbcConfiguration;
import pl.slawas.scjdbc4jpa.core.ScObjectFactory;
import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityManager;
import pl.slawas.scjdbc4jpa.jdbc.JDBCEntityMetaData;
import pl.slawas.scjdbc4jpa.jdbc.PagedResult;
import pl.slawas.scjdbc4jpa.jdbc.SortClause;
import pl.slawas.scjdbc4jpa.jdbc.where.ConditionClauseSupport;

public interface _IJdbcDAO<Obj> {

	/**
	 * @return the persistentClass
	 */
	public Class<Obj> getPersistentClass();

	/**
	 * Wyszukiwanie encji na podstawie jej identyfikatora.
	 * 
	 * @see JDBCEntityManager#find(Class, Object)
	 * 
	 * @param id
	 *            identyfikator encji
	 * @return znaleziona encja
	 */
	public Obj find(Object id);

	/**
	 * Wyszukiwanie encji na podstawie zadanych kryteriów.
	 * 
	 * @see JDBCEntityManager#find(Class, Page, ConditionClauseSupport[], SortClause)
	 * 
	 * @param page
	 *            strona wyników. Jeżeli podamy wartość {@code null}, to
	 *            zostanie użyta pierwsza predefniowana strona (
	 *            {@link JDBCEntityManager#getFirstPageDefinition()})
	 * @param conditions
	 *            lista warunków wyszukiwania
	 * @param sort
	 *            warunki sortowania
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult find(Page page, ConditionClauseSupport[] conditions,
			SortClause sort);

	/**
	 * Eksport danych. Wyszukiwanie encji na podstawie zadanych kryteriów z
	 * pominięciem użycia keszy oraz poza restrykcjami związanymi z maksymalną
	 * możliwą liczbą wierszy w zwracanym wyniku.
	 * 
	 * @see JDBCEntityManager#dump(Class, Page, ConditionClauseSupport[], SortClause)
	 * 
	 * @param page
	 *            strona wyników. Jeżeli podamy wartość {@code null}, to
	 *            zostanie użyta pierwsza predefniowana strona (
	 *            {@link JDBCEntityManager#getFirstPageDefinition()})
	 * @param conditions
	 *            lista warunków wyszukiwania
	 * @param sort
	 *            warunki sortowania
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult dump(Page page, ConditionClauseSupport[] conditions,
			SortClause sort);

	/**
	 * Wyszukiwanie encji na podstawie listy identyfikatorów encji, poprzez
	 * wykorzystanie SQL-owej klauzuli {@code IN}.
	 * 
	 * @param idList
	 *            lista identyfikatorów. Długość listy ograniczona jest do
	 *            ogólnych warunków maksymalnej liczby wierszy jaką może zwrócić
	 *            stronicowany wynik (zoabcz
	 *            {@link LoadJdbcConfiguration#RESULT_MAX_PAGES}).
	 * @param sort
	 *            warunki sortowania
	 * @return stronicowany wynik zapytania
	 */
	public PagedResult findByIdList(Object[] idList, SortClause sort);

	/**
	 * Pobieranie/<b>tworzenie (w przypadku gdy nie jest ustawiony)</b>
	 * manager-a encji.
	 * 
	 * @return manager encji
	 */
	public JDBCEntityManager getEntityManager();

	/**
	 * Ustawianie manager-a encji.
	 * 
	 * @param em
	 *            manager encji
	 */
	public void setEntityManager(JDBCEntityManager em);

	/**
	 * Wstawianie encji do bazy danych.
	 * 
	 * @see #beforeInsert(Object)
	 * @see #afterInsert(Object)
	 * @param entityObject
	 *            wstawiana encja
	 */
	public void insert(Obj entityObject);

	/**
	 * Wstawianie listy encji do bazy danych.
	 * 
	 * @see #beforeInsert(Object)
	 * @see #afterInsert(Object)
	 * 
	 * @param entityObjects
	 *            lista encji do wstawienia
	 * @param autoCommit
	 *            wartości {@code true} albo {@code false}. Jeżeli będzie
	 *            ustawiona wartość {@code true} to wtedy zostanie automatycznie
	 *            utworzona transakcja (jeżeli nie jest aktywna) oraz będzie
	 *            robione zatwierdzenie co
	 *            {@link JDBCEntityManager#getBatchCommitSize()}.
	 *            <p>
	 *            <font color="#aa0000">Ważnym jest, że gdy parametr przyjmuje
	 *            wartość {@code true}, aby nie poprzedzać wywołanie danej
	 *            metody metodą {@link EntityTransaction#begin()} oraz nie
	 *            umieszczać {@link EntityTransaction#commit()} za, bo z
	 *            pewnością wyleci wyjątek.</font>
	 *            </p>
	 * @return liczba dodanych wierszy
	 */
	public int insertList(List<Obj> entityObjects, boolean autoCommit);

	/**
	 * Aktualizacja encji w bazie danych.
	 * 
	 * @see #beforeUpdate(Object)
	 * @see #afterUpdate(Object)
	 * @param entityObject
	 *            aktualizowana encja
	 */
	public void update(Obj entityObject);

	/**
	 * Aktualizacja listy encji w bazie danych.
	 * 
	 * @see #beforeUpdate(Object)
	 * @see #afterUpdate(Object)
	 * 
	 * @param entityObjects
	 *            lista encji do aktualizacji
	 * @param autoCommit
	 *            wartości {@code true} albo {@code false}. Jeżeli będzie
	 *            ustawiona wartość {@code true} to wtedy zostanie automatycznie
	 *            robione zatwierdzenie co
	 *            {@link JDBCEntityManager#getBatchCommitSize()}.
	 *            <p>
	 *            <font color="#aa0000">Ważnym jest, że gdy parametr przyjmuje
	 *            wartość {@code true}, aby nie poprzedzać wywołanie danej
	 *            metody metodą {@link EntityTransaction#begin()} oraz nie
	 *            umieszczać {@link EntityTransaction#commit()} za, bo z
	 *            pewnością wyleci wyjątek.</font>
	 *            </p>
	 * @return liczba zmienionych wierszy
	 */
	public int updateList(List<Obj> entityObjects, boolean autoCommit);

	/**
	 * Usuwanie encji z bazy danych.
	 * 
	 * @see #beforeRemove(Object)
	 * @see #afterRemove(Object)
	 * @param entityObject
	 *            usuwana encja
	 */
	public void remove(Obj entityObject);

	/**
	 * Usuwanie listy encji z bazy danych.
	 * 
	 * @see #beforeRemove(Object)
	 * @see #afterRemove(Object)
	 * 
	 * @param entityObjects
	 *            lista encji do usunięcia
	 * @param autoCommit
	 *            wartości {@code true} albo {@code false}. Jeżeli będzie
	 *            ustawiona wartość {@code true} to wtedy zostanie automatycznie
	 *            robione zatwierdzenie co
	 *            {@link JDBCEntityManager#getBatchCommitSize()}.
	 *            <p>
	 *            <font color="#aa0000">Ważnym jest, że gdy parametr przyjmuje
	 *            wartość {@code true}, aby nie poprzedzać wywołanie danej
	 *            metody metodą {@link EntityTransaction#begin()} oraz nie
	 *            umieszczać {@link EntityTransaction#commit()} za, bo z
	 *            pewnością wyleci wyjątek.</font>
	 *            </p>
	 * @return liczba usuniętych wierszy
	 */
	public int removeList(List<Obj> entityObjects, boolean autoCommit);

	public void setScObjectFactory(ScObjectFactory aof);

	/**
	 * Pobieranie metadanych z fabryki manager-a encji na temat obsługiwanej
	 * (podanej w parametrarze implementacji &lt;Obj>) encji.
	 * 
	 * @return the metaData
	 */
	public JDBCEntityMetaData getMetaData();

}