package pl.slawas.scjdbc4jpa.jdbc.ddl;

import java.io.Serializable;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Enumeration;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnSourceType;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;

/**
 * 
 * TableCreator creator prostych tabel na podstawie definicji tabeli. Kreator
 * nie tworzy sekwencji, jezeli w encji zostaly zadeklarowane generatory. Do ich
 * utworzenia trzeba {@link SequenceCreator}. Kreator również nie zarządza
 * związkami, trzeba je ręcznie utworzyć z poziomu bazy danych. Klasa ma
 * wspierać tylko proste konstrukcje a ich poprawność należy zawsze zweryfikować
 * z poziomu bazy danych.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.8 $
 * 
 */
public class TableCreator implements Serializable {

	private static final long serialVersionUID = -2060989233114114637L;

	final static private Logger logger = LoggerFactory
			.getLogger(TableCreator.class);

	private final Dialect dialect;

	/**
	 * @param dialect
	 *            dialekt {@link Dialect}
	 */
	public TableCreator(Dialect dialect) {
		this.dialect = dialect;
	}

	public void createDataStore(Connection conn, JDBCTableDesc table)
			throws SQLException {

		logger.debug("Tworzenie tabeli {}", table.getName());
		CallableStatement call = conn
				.prepareCall("DROP TABLE "
						+ table.getName()
						+ (dialect.equals(Dialect.ORACLE) ? " CASCADE CONSTRAINTS PURGE"
								: ""));
		try {
			call.execute();
		} catch (SQLException e) {
			logger.error(
					"Operacja usuniecia tabeli zakonczona niepowodzeniem.\n Przyczyna: ",
					e.getMessage());
		}
		StringBuffer createClause = new StringBuffer("CREATE TABLE "
				+ table.getName() + "(");
		Enumeration<String> keys = table.getColumnKeys();
		int i = 0;
		while (keys.hasMoreElements()) {
			JDBCColumnDesc column = table.getColumn(keys.nextElement());
			if (column.getSource().equals(JDBCColumnSourceType.MAIN)) {
				SimpleSQLType type = SimpleSQLType.get(column.getColumnType());
				createClause.append((i != 0 ? "\n ," : "\n  ")
						+ column.getName() + " "
						+ type.getCreationType(dialect));
			}
			i++;
		}
		JDBCKeyDesc pk = table.getPrimaryKey();
		if (pk != null) {
			Enumeration<Integer> pkKeyPositions = pk.getColumns().keys();
			createClause.append((i != 0 ? "\n ," : "\n  ") + "PRIMARY KEY (");
			int pkPos = 0;
			while (pkKeyPositions.hasMoreElements()) {
				JDBCColumnDesc column = pk.getColumns().get(
						pkKeyPositions.nextElement());
				createClause
						.append((pkPos != 0 ? ", " : "") + column.getName());
				pkPos++;
			}
			createClause.append(")");
		}
		createClause.append("\n)");
		logger.debug("\nOperacja DDL tworzenia tabeli: \n----\n{}\n----",
				createClause.toString());
		call = conn.prepareCall(createClause.toString());
		call.execute();
		call.close();
		logger.debug("Zostala utworzona tabela : {}", table);
	}
}
