package pl.slawas.scjdbc4jpa.jdbc.dml;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.sql.DataSource;

import oracle.jdbc.OraclePreparedStatement;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.QueryClause;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.BindParamException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.CreateDmlQueryException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ExecuteSQLQueryException;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityParserLevel;
import pl.slawas.scjdbc4jpa.jdbc.utils.JDBCEntityValueParser;
import pl.slawas.scjdbc4jpa.jdbc.where.BindParam;

@SuppressWarnings("serial")
public abstract class DmlQuerySupport implements Serializable, _IDmlQuery {

	final static private Logger logger = LoggerFactory
			.getLogger(DmlQuerySupport.class);

	protected JDBCTableDesc table;

	protected QueryClause queryClause;

	protected Hashtable<String, Vector<BindParam>> bindParamsMapper;

	protected PreparedStatement pstmt = null;

	protected Connection conn = null;

	protected boolean isMyPrivateConnection = false;

	protected final DataModificationType type;

	/**
	 * Konstruktor operacji DML. Nalezy pamietać aby opis tabeli był
	 * zdefiniowany na poziomie {@link JDBCEntityParserLevel#FIRST}.
	 * 
	 * @param table
	 *            opis tabeli
	 * @param type
	 *            typ modyfikacji danych
	 */
	public DmlQuerySupport(JDBCTableDesc table, DataModificationType type) {
		this.table = table;
		JDBCEntityParserLevel level = this.table.getLevel();
		if (level == null || !level.equals(JDBCEntityParserLevel.FIRST)) {
			throw new CreateDmlQueryException(
					"Aby utworzyc zapytanie DML opis tabeli musi byc na poziomie JDBCEntityParserLevel#FIRST.");
		}
		StringBuffer jdbcSQL = init();
		queryClause = new QueryClause(jdbcSQL.toString());
		Enumeration<Vector<BindParam>> bindParamsMapperElements = bindParamsMapper
				.elements();
		while (bindParamsMapperElements.hasMoreElements()) {
			for (BindParam param : bindParamsMapperElements.nextElement()) {
				queryClause.addBindParam(param);
			}
		}
		this.type = type;
	}

	/**
	 * Metoda budująca klauzulę zapytania DML
	 * 
	 * @return klauzula zapytania.
	 */
	protected abstract StringBuffer init();

	public void setParam(Object param) {
		try {
			Hashtable<String, Object> fields = JDBCEntityValueParser
					.readEntityObjectAsSimpleMap(table, param);
			Enumeration<String> bindMapperKeys = this.bindParamsMapper.keys();
			while (bindMapperKeys.hasMoreElements()) {
				String entityFieldName = bindMapperKeys.nextElement();
				Vector<BindParam> vector = this.bindParamsMapper
						.get(entityFieldName);
				for (BindParam item : vector) {
					item.setParamValue(fields.get(entityFieldName));
				}
			}
		} catch (IllegalArgumentException e) {
			throw new BindParamException("[" + getClass().getSimpleName()
					+ "] Problem bindowania zmiennych", e);
		} catch (IllegalAccessException e) {
			throw new BindParamException("[" + getClass().getSimpleName()
					+ "] Problem bindowania zmiennych", e);
		} catch (InvocationTargetException e) {
			throw new BindParamException("[" + getClass().getSimpleName()
					+ "] Problem bindowania zmiennych", e);
		}

	}

	/**
	 * @return the queryClause
	 */
	public QueryClause getQueryClause() {
		return queryClause;
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + " [" + queryClause.toString() + "]";
	}

	public int executeUpdate(Object param) throws ExecuteSQLQueryException {
		this.setParam(param);
		return this.executeUpdate();
	}

	public int executeUpdate() throws ExecuteSQLQueryException {
		return executeUpdate(false);
	}

	public int addBatch(Object param) throws ExecuteSQLQueryException {
		this.setParam(param);
		return this.executeUpdate(true);
	}

	public int addBatch() throws ExecuteSQLQueryException {
		return this.executeUpdate(true);
	}

	protected int executeUpdate(boolean addToBatch)
			throws ExecuteSQLQueryException {
		if (this.queryClause == null
				|| StringUtils.isBlank(this.queryClause.getClause()))
			throw new ExecuteSQLQueryException("Klauzula zapytania jest null. "
					+ "Zle zaimplentowana abstrakcja operacji DML.");
		if (this.pstmt == null)
			throw new ExecuteSQLQueryException(
					"Zapytanie nie zostalo przygotowane. "
							+ "Przed wywolaniem tej metody musisz uruchomic "
							+ "jedną z implementacji metody pl.slawas.jdbc.dml._IDmlQuery#prepareStatement().");

		int rs = 0;
		try {
			if (!queryClause.getBindedParams().isEmpty()) {
				for (BindParam bindParam : queryClause.getBindedParams()) {
					bindParam.getType().bindValue(pstmt,
							bindParam.getParameterIndex(),
							bindParam.getParamValue());
				}
			}
			if (addToBatch) {
				if (this.pstmt instanceof OraclePreparedStatement) {
					logger.trace("[OraclePreparedStatement] Add batch: {}",
							queryClause.toString());
					((OraclePreparedStatement) this.pstmt).executeUpdate();
				} else {
					logger.trace("[PreparedStatement] Add batch: {}",
							queryClause.toString());
					this.pstmt.addBatch();
				}
			} else {
				logger.trace("Zapytanie do bazy danych: {}",
						queryClause.toString());
				rs = pstmt.executeUpdate();
			}
		} catch (SQLException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (SecurityException e) {
			throw new ExecuteSQLQueryException(e);
		} catch (IllegalArgumentException e) {
			throw new ExecuteSQLQueryException(e);
		}
		return rs;
	}

	public void close() {
		try {
			if (this.pstmt != null) {
				pstmt.close();
			}
			if (this.conn != null && isMyPrivateConnection
					&& !this.conn.isClosed()) {
				this.conn.close();
			}
			this.conn = null;
			this.pstmt = null;
		} catch (SQLException e) {
		}
	}

	public _IDmlQuery prepareStatement(DataSource dataSource)
			throws SQLException {
		if (this.queryClause == null
				|| StringUtils.isBlank(this.queryClause.getClause()))
			throw new IllegalArgumentException("Klauzula zapytania jest null. "
					+ "Zle zaimplentowana abstrakcja operacji DML.");

		if (this.conn != null && isMyPrivateConnection && !this.conn.isClosed()) {
			this.close();
		}
		this.conn = dataSource.getConnection();
		if (this.conn != null) {
			isMyPrivateConnection = true;
			this.pstmt = conn.prepareStatement(queryClause.getClause());
		} else {
			throw new SQLException(
					"Nie udalo sie pobrac polaczenia ze zrodla danych");
		}
		return this;
	}

	public _IDmlQuery prepareStatement(Connection conn) throws SQLException {
		if (this.queryClause == null
				|| StringUtils.isBlank(this.queryClause.getClause()))
			throw new IllegalArgumentException("Klauzula zapytania jest null. "
					+ "Zle zaimplentowana abstrakcja operacji DML.");

		this.conn = conn;
		if (this.conn != null) {
			isMyPrivateConnection = false;
			this.pstmt = conn.prepareStatement(queryClause.getClause());
		} else {
			throw new SQLException("Obiekt polaczenia nie moze byc pusty");
		}
		return this;
	}

	@Override
	protected void finalize() throws Throwable {
		try {
			this.close();
		} finally {
			super.finalize();
		}
	}

	public int executeBatch() throws ExecuteSQLQueryException {
		if (this.pstmt == null)
			throw new ExecuteSQLQueryException(
					"Zapytanie nie zostalo przygotowane. "
							+ "Przed wywolaniem tej metody musisz uruchomic "
							+ "jedną z implementacji metody pl.slawas.jdbc.dml._IDmlQuery#prepareStatement().");
		try {
			if (pstmt instanceof OraclePreparedStatement) {
				logger.trace("[OraclePreparedStatement] Send batch.");
				return ((OraclePreparedStatement) pstmt).sendBatch();
			} else {
				int[] result = this.pstmt.executeBatch();
				int sum = 0;
				for (int i : result) {
					sum = sum + i;
				}
				return sum;
			}
		} catch (SQLException e) {
			throw new ExecuteSQLQueryException(e);
		}
	}

	public void setBatchSize(int batchSize) throws ExecuteSQLQueryException {
		if (this.pstmt == null)
			throw new ExecuteSQLQueryException(
					"Zapytanie nie zostalo przygotowane. "
							+ "Przed wywolaniem tej metody musisz uruchomic "
							+ "jedną z implementacji metody pl.slawas.jdbc.dml._IDmlQuery#prepareStatement().");
		try {
			if (pstmt instanceof OraclePreparedStatement) {
				((OraclePreparedStatement) pstmt).setExecuteBatch(batchSize);
			} else
				throw new UnsupportedOperationException(
						"Opreacja wspierana tylko i wylacznie dla polaczen Oracle");
		} catch (SQLException e) {
			throw new ExecuteSQLQueryException(e);
		}
	}

	public boolean isOracleStatement() throws ExecuteSQLQueryException {
		if (this.pstmt == null)
			throw new ExecuteSQLQueryException(
					"Zapytanie nie zostalo przygotowane. "
							+ "Przed wywolaniem tej metody musisz uruchomic "
							+ "jedną z implementacji metody pl.slawas.jdbc.dml._IDmlQuery#prepareStatement().");
		return (pstmt instanceof OraclePreparedStatement);
	}

	/**
	 * @return the type
	 */
	public DataModificationType getType() {
		return type;
	}
}