package pl.slawas.scjdbc4jpa.jdbc.sql.functions;

import java.lang.reflect.InvocationTargetException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.ResultRowMapper;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionImplementationException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionNotFoundException;

/**
 * 
 * Function odwołanie się do implementacji funkcji w zależności od dialektu
 * serwera bazy danych.
 * 
 * @see #Concat
 * @see #LPad
 * @see #LTrim
 * @see #NVL
 * @see #Reverse
 * @see #RPad
 * @see #RTrim
 * @see #ToChar
 * @see #ToDate
 * @see #ToNumber
 * @see #ToUpperCase
 * @see #ToLowerCase
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
public enum Function {

	/**
	 * funkcja łączenia łańcuchów znakowych
	 */
	Concat,
	/**
	 * funkcja dopełnienia z lewej strony łańcuchów znakowych
	 */
	LPad,
	/**
	 * funkcja obcinania z lewej strony łańcuchów znakowych
	 */
	LTrim,
	/**
	 * funkcja obsługi wartości pustej (zastąpienie wartości pustej
	 * zadeklarowaną wartościa)
	 */
	NVL,
	/**
	 * funkcja obsługi odwrócenia wartości
	 */
	Reverse,
	/**
	 * funkcja dopełniania z prawej strony łańcuchów znakowych
	 */
	RPad,
	/**
	 * funkcja obcinania z prawej strony łańcuchów znakowych
	 */
	RTrim,
	/**
	 * funkcja zamieniania obiektów do łańcucha znakowego
	 */
	ToChar,
	/**
	 * funkcja zamieniania obiektów (łańcuchów znakowych) do daty
	 */
	ToDate,
	/**
	 * funkcja zamieniania obiektów (łańcuchów znakowych) do liczby
	 */
	ToNumber,
	/**
	 * funkcja zamieniania obiektów (łańcuchów znakowych) do obiektów (łańcuchów
	 * znakowych) pisanych wielkimi literami
	 */
	ToUpperCase,
	/**
	 * funkcja zamieniania obiektów (łańcuchów znakowych) do obiektów (łańcuchów
	 * znakowych) pisanych małymi literami
	 */
	ToLowerCase,
	/**
	 * Funkcja obcinająca wartości typu date oraz numeryczne
	 */
	Trunc,
	/**
	 * Funkcja 'brak funkcji'. Moze wydawać się to bez sensu, ale podczas
	 * budowania zapytań, zupełnie niezależnych od zbioru encji (tzw. zapytań
	 * natywnych), potrzebujemy zdefiniować typ dla kolumny występującej w
	 * zapytaniu. Wtedy używamy takiej 'sztucznej' funkcji, w celu rozpoznania
	 * tej kolumny przez {@link ResultRowMapper}-a.
	 */
	None;

	final protected static Logger logger = LoggerFactory
			.getLogger(Function.class);

	/**
	 * Tworzenie obiektu funkcji w zależności od dialektu serwera bazy danych.
	 * <p>
	 * <font color="#dd0000"><b>Uwaga!</b></font> Dla funkcji {@link #NVL} oraz
	 * {@link #Reverse} wymagane jest podanie typu zwracanej wartości funkcji, w
	 * związku z tym, podczas wywołania tej metody dla tych funkcji zostanie
	 * domyślnie ustawiony typ {@link String} oraz wygenerowane zostanie
	 * ostrzężenie do loga.
	 * </p>
	 * 
	 * @see _IFunction
	 * @see FunctionSupport
	 * 
	 * @param dialect
	 *            dialekt serwera bazy danych
	 * @param alias
	 *            alias dla kolumny wyniku funkcji
	 * @param arguments
	 *            argumenty funkcji
	 * @return obiekt funkcji
	 * @throws SQLFunctionNotFoundException
	 *             brak implementacji funkcji dla danego dialektu
	 * @throws SQLFunctionImplementationException
	 *             błedy argumentów inicjalizacji funkcji dla danego dialektu
	 */
	public _IFunction get(Dialect dialect, String alias, String[] arguments)
			throws SQLFunctionNotFoundException,
			SQLFunctionImplementationException {
		String functionClassName = this.getDeclaringClass().getPackage()
				.getName()
				+ "."
				+ dialect.toString().toLowerCase()
				+ "."
				+ this.toString();
		try {
			Class<?> clazz = this.getDeclaringClass().getClassLoader()
					.loadClass(functionClassName);
			Class<?>[] params = null;
			switch (this) {
			case NVL:
			case None:
			case Reverse:
			case Trunc:
				logger.warn("[get(Dialect, String, String[])] Ustawiam domyślny typ zwracanego obiektu na java.lang.String dla implementacji funkcji '"
						+ this.toString() + "'");
				params = new Class<?>[] { Class.class, String.class,
						(new String[] {}).getClass() };
				return (_IFunction) clazz.getConstructor(params).newInstance(
						new Object[] { String.class, alias, arguments });
			default:
				params = new Class<?>[] { String.class,
						(new String[] {}).getClass() };
				return (_IFunction) clazz.getConstructor(params).newInstance(
						new Object[] { alias, arguments });
			}
		} catch (ClassNotFoundException e) {
			throw new SQLFunctionNotFoundException("[" + dialect.toString()
					+ "]", e);
		} catch (IllegalArgumentException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (SecurityException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (InstantiationException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (IllegalAccessException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (InvocationTargetException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (NoSuchMethodException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		}
	}

	/**
	 * Tworzenie obiektu funkcji w zależności od dialektu serwera bazy danych.
	 * <p>
	 * <font color="#dd0000"><b>Uwaga!</b></font> Tylko dla funkcji {@link #NVL}
	 * oraz {@link #Reverse} wymagane jest podanie typu zwracanej wartości
	 * funkcji, w związku z tym, podczas wywołania tej metody dla innych funkcji
	 * zostanie pominiety argument 'returnType' oraz wygenerowane zostanie
	 * ostrzężenie do loga.
	 * </p>
	 * 
	 * @see _IFunction
	 * @see FunctionSupport
	 * 
	 * @param dialect
	 *            dialekt serwera bazy danych
	 * @param returnType
	 *            typ zwracanej przez funkcję wartości
	 * @param alias
	 *            alias dla kolumny wyniku funkcji
	 * @param arguments
	 *            argumenty funkcji
	 * @return obiekt funkcji
	 * @throws SQLFunctionNotFoundException
	 *             brak implementacji funkcji dla danego dialektu
	 * @throws SQLFunctionImplementationException
	 *             błedy argumentów inicjalizacji funkcji dla danego dialektu
	 */
	public _IFunction get(Dialect dialect, Class<?> returnType, String alias,
			String[] arguments) throws SQLFunctionNotFoundException,
			SQLFunctionImplementationException {
		String functionClassName = this.getDeclaringClass().getPackage()
				.getName()
				+ "."
				+ dialect.toString().toLowerCase()
				+ "."
				+ this.toString();
		try {
			Class<?> clazz = this.getDeclaringClass().getClassLoader()
					.loadClass(functionClassName);
			Class<?>[] params = null;
			switch (this) {
			case NVL:
			case None:
			case Reverse:
			case Trunc:
				params = new Class<?>[] { Class.class, String.class,
						(new String[] {}).getClass() };
				return (_IFunction) clazz.getConstructor(params).newInstance(
						new Object[] { returnType, alias, arguments });
			default:
				logger.warn("[get(Dialect, Class<?>, String, String[])] Pomijam argument 'returnType' dla implementacji funkcji '"
						+ this.toString() + "'");
				params = new Class<?>[] { String.class,
						(new String[] {}).getClass() };
				return (_IFunction) clazz.getConstructor(params).newInstance(
						new Object[] { alias, arguments });
			}
		} catch (ClassNotFoundException e) {
			throw new SQLFunctionNotFoundException("[" + dialect.toString()
					+ "]", e);
		} catch (IllegalArgumentException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (SecurityException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (InstantiationException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (IllegalAccessException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (InvocationTargetException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		} catch (NoSuchMethodException e) {
			throw new SQLFunctionImplementationException("["
					+ dialect.toString() + "]", e);
		}
	}

}
