package pl.slawas.scjdbc4jpa.jdbc.types;

import java.math.BigDecimal;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Date;
import java.sql.NClob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;

import pl.slawas.scjdbc4jpa.dialect.Dialect;

/**
 * 
 * SimpleSQLType - typu SQL-owe - lista i podstawowa obsługa
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.5 $
 * 
 */
public enum SimpleSQLType {

	STRING // java.lang.String
	, INTEGER // java.lang.Integer
	, INT // int
	, LONG // java.lang.Long or long
	, SHORT // java.lang.Short or short
	, FLOAT // java.lang.Float or float
	, DOUBLE // java.lang.Double or double
	, BIGDECIMAL // java.lang.BigDecimal
	, BYTE // java.lang.Byte or byte
	, BOOLEAN // java.lang.Boolean or boolean
	, DATE // java.util.Date or java.sql.Date
	, CALENDAR // java.util.Calendar
	, CLOB // java.sql.Clob
	, NCLOB // java.sql.NClob
	, BLOB // java.sql.Blob
	, TIME // java.sql.Time
	, TIMESTAMP // java.sql.Timestamp
	;

	/**
	 * Pobieranie obiektu {@link SimpleSQLType} na podstawie typu (klasy)
	 * zdefiniowanego dla kolumny w zapytaniu.
	 */
	public static SimpleSQLType get(Class<?> objectClass) {

		if (objectClass.equals(String.class)) {
			return STRING;
		}
		if (objectClass.equals(Integer.class)) {
			return INTEGER; // java.lang.Integer
		}
		if (objectClass.equals(int.class)) {
			return INT; // int
		}
		if (objectClass.equals(Long.class) || objectClass.equals(long.class)) {
			return LONG; // java.lang.Long or long
		}
		if (objectClass.equals(Short.class) || objectClass.equals(short.class)) {
			return SHORT; // java.lang.Short or short
		}
		if (objectClass.equals(Float.class) || objectClass.equals(float.class)) {
			return FLOAT; // java.lang.Float or float
		}
		if (objectClass.equals(Double.class)
				|| objectClass.equals(double.class)) {
			return DOUBLE; // java.lang.Double or double
		}
		if (objectClass.equals(BigDecimal.class)) {
			return BIGDECIMAL; // java.lang.BigDecimal
		}
		if (objectClass.equals(Byte.class) || objectClass.equals(byte.class)) {
			return BYTE; // java.lang.Byte or byte
		}
		if (objectClass.equals(Boolean.class)
				|| objectClass.equals(boolean.class)) {
			return BOOLEAN; // java.lang.Boolean or boolean
		}
		if (objectClass.equals(java.util.Date.class)
				|| objectClass.equals(java.sql.Date.class)) {
			return DATE; // java.util.Date or java.sql.Date
		}
		if (objectClass.equals(Calendar.class)) {
			return CALENDAR; // java.util.Calendar
		}
		if (ArrayUtils.contains(objectClass.getInterfaces(), Clob.class)
				|| objectClass.equals(Clob.class)) {
			return CLOB; // java.sql.Clob
		}
		if (ArrayUtils.contains(objectClass.getInterfaces(), NClob.class)
				|| objectClass.equals(NClob.class)) {
			return NCLOB; // java.sql.NClob
		}
		if (ArrayUtils.contains(objectClass.getInterfaces(), Blob.class)
				|| objectClass.equals(Blob.class)) {
			return BLOB; // java.sql.Blob
		}
		if (objectClass.equals(Time.class)) {
			return TIME; // java.sql.Time
		}
		if (objectClass.equals(Timestamp.class)) {
			return TIMESTAMP; // java.sql.Timestamp
		}

		return null;
	}

	/**
	 * Pobieranie wartości dla danego typu z resultSet-a (wybniku zapytania) na
	 * podstawie nazwy kolumny (albo jej aliasu, jezeli został zdefiniwany -
	 * klauzula 'AS &lt;alias>') w zapytaniu.
	 * 
	 * @param rs
	 *            wynik zapytania
	 * @param columnLabel
	 *            nazwa albo alias kolumny w zapytaniu
	 * @return wartość okreslonego typu dla kolumny o podanej nazwie/aliasie
	 * @throws SQLException
	 */
	public Object getValue(ResultSet rs, String columnLabel)
			throws SQLException {
		switch (this) {
		case STRING:
			return rs.getString(columnLabel); // java.lang.String
		case INTEGER:
		case INT:
			return rs.getInt(columnLabel); // java.lang.Integer
		case LONG:
			return rs.getLong(columnLabel); // java.lang.Long or long
		case SHORT:
			return rs.getShort(columnLabel); // java.lang.Short or short
		case FLOAT:
			return rs.getFloat(columnLabel); // java.lang.Float or float
		case DOUBLE:
			return rs.getDouble(columnLabel); // java.lang.Double or double
		case BIGDECIMAL:
			return rs.getBigDecimal(columnLabel); // java.lang.BigDecimal
		case BYTE:
			return rs.getByte(columnLabel); // java.lang.Byte or byte
		case BOOLEAN:
			return rs.getString(columnLabel); // java.lang.Boolean or boolean
		case DATE:
			return rs.getDate(columnLabel); // java.util.Date or java.sql.Date
		case CALENDAR:
			Date date = rs.getDate(columnLabel);
			if (date != null) {
				Calendar calendar = Calendar.getInstance();
				calendar.setTime(date);
				return calendar; // java.util.Calendar
			} else {
				return null;
			}
		case CLOB:
			return rs.getClob(columnLabel); // java.sql.Clob
		case NCLOB:
			return rs.getNClob(columnLabel); // java.sql.NClob
		case BLOB:
			return rs.getBlob(columnLabel); // java.sql.Blob
		case TIME:
			return rs.getTime(columnLabel); // java.sql.Time
		case TIMESTAMP:
			return rs.getTimestamp(columnLabel); // java.sql.Timestamp
		default:
			break;
		}

		return rs.getObject(columnLabel);

	}

	/**
	 * Pobranie wartości stałej {@link Types} odpowiadającej danemu typowi SQL
	 * 
	 * @return wartość stałej {@link Types} np. {@link #STRING} odpowiada
	 *         {@link Types#VARCHAR}
	 */
	public int getSqlType() {
		switch (this) {
		case STRING:
			return Types.VARCHAR; // java.lang.String
		case INTEGER:
		case INT:
			return Types.INTEGER; // java.lang.Integer
		case LONG:
			return Types.NUMERIC; // java.lang.Long or long
		case SHORT:
			return Types.NUMERIC; // java.lang.Short or short
		case FLOAT:
			return Types.NUMERIC; // java.lang.Float or float
		case DOUBLE:
			return Types.NUMERIC; // java.lang.Double or double
		case BIGDECIMAL:
			return Types.NUMERIC; // java.lang.BigDecimal
		case BYTE:
			return Types.NUMERIC; // java.lang.Byte or byte
		case BOOLEAN:
			return Types.VARCHAR; // java.lang.Boolean or boolean
		case DATE:
			return Types.DATE; // java.util.Date or java.sql.Date
		case CALENDAR:
			return Types.DATE; // java.util.Calendar
		case CLOB:
			return Types.CLOB; // java.sql.Clob
		case NCLOB:
			return Types.NCLOB; // java.sql.NClob
		case BLOB:
			return Types.BLOB; // java.sql.Blob
		case TIME:
			return Types.TIME; // java.sql.Time
		case TIMESTAMP:
			return Types.TIMESTAMP; // java.sql.Timestamp
		default:
			break;
		}

		return Types.VARCHAR;

	}

	/**
	 * Pobieranie wartości dla danego typu z resultSet-a (wybniku zapytania) na
	 * podstawie pozycji kolumny w zapytaniu.
	 * 
	 * @param rs
	 *            wynik zapytania
	 * @param position
	 *            pozycja kolumny w zapytaniu
	 * @return wartość okreslonego typu dla kolumny na danej pozycji
	 * @throws SQLException
	 */
	public Object getValue(ResultSet rs, int position) throws SQLException {

		switch (this) {
		case STRING:
			return rs.getString(position); // java.lang.String
		case INTEGER:
		case INT:
			return rs.getInt(position); // java.lang.Integer
		case LONG:
			return rs.getLong(position); // java.lang.Long or long
		case SHORT:
			return rs.getShort(position); // java.lang.Short or short
		case FLOAT:
			return rs.getFloat(position); // java.lang.Float or float
		case DOUBLE:
			return rs.getDouble(position); // java.lang.Double or double
		case BIGDECIMAL:
			return rs.getBigDecimal(position); // java.lang.BigDecimal
		case BYTE:
			return rs.getByte(position); // java.lang.Byte or byte
		case BOOLEAN:
			return rs.getString(position); // java.lang.Boolean or boolean
		case DATE:
			return rs.getDate(position); // java.util.Date or java.sql.Date
		case CALENDAR:
			Date date = rs.getDate(position);
			Calendar calendar = Calendar.getInstance();
			calendar.setTime(date);
			return calendar; // java.util.Calendar
		case CLOB:
			return rs.getClob(position); // java.sql.Clob
		case NCLOB:
			return rs.getNClob(position); // java.sql.NClob
		case BLOB:
			return rs.getBlob(position); // java.sql.Blob
		case TIME:
			return rs.getTime(position); // java.sql.Time
		case TIMESTAMP:
			return rs.getTimestamp(position); // java.sql.Timestamp
		default:
			break;
		}

		return rs.getObject(position);

	}

	/**
	 * Bindowanie wartości do zapytania
	 * 
	 * @param ps
	 *            preparowanie zapytanie
	 * @param parameterIndex
	 *            pozycja bindowanej wartości (Zaczynamy od {@code 1}).
	 * @param value
	 *            bindowana wartość
	 * @throws SQLException
	 */
	public void bindValue(PreparedStatement ps, int parameterIndex, Object value)
			throws SQLException {

		if (null == value) {
			ps.setNull(parameterIndex, this.getSqlType());
			return;
		}

		switch (this) {
		case STRING:
			// java.lang.String
			ps.setString(parameterIndex, (String) value);
			return;
		case INTEGER:
		case INT:
			// java.lang.Integer
			ps.setInt(parameterIndex, (Integer) value);
			return;
		case LONG:
			// java.lang.Long or long
			ps.setLong(parameterIndex, (Long) value);
			return;
		case SHORT:
			// java.lang.Short or short
			ps.setShort(parameterIndex, (Short) value);
			return;
		case FLOAT:
			// java.lang.Float or float
			ps.setFloat(parameterIndex, (Float) value);
			return;
		case DOUBLE:
			// java.lang.Double or double
			ps.setDouble(parameterIndex, (Double) value);
			return;
		case BIGDECIMAL:
			// java.lang.BigDecimal
			ps.setBigDecimal(parameterIndex, (BigDecimal) value);
			return;
		case BYTE:
			// java.lang.Byte or byte
			ps.setByte(parameterIndex, (Byte) value);
			return;
		case BOOLEAN:
			ps.setString(parameterIndex, (String) value);
			return;
		case DATE:
			ps.setDate(parameterIndex,
					new Date(((java.util.Date) value).getTime()));
			return;
		case CALENDAR:
			Calendar calendar = (Calendar) value;
			ps.setDate(parameterIndex, new Date((calendar.getTime()).getTime()));
			return;
		case CLOB:
		case NCLOB:
		case BLOB:
			throw new IllegalArgumentException(
					"Brak wsparcia dla wyszukiwania po polach typu LOB");
		case TIME:
			// java.sql.Time
			ps.setTime(parameterIndex, (Time) value);
			return;
		case TIMESTAMP:
			// java.sql.Timestamp
			ps.setTimestamp(parameterIndex, (Timestamp) value);
			return;
		default:
			break;
		}

		ps.setObject(parameterIndex, value);

	}

	/**
	 * Bindowanie wartości do zapytania
	 * 
	 * @param ps
	 *            preparowanie zapytanie
	 * @param parameterIndex
	 *            pozycja bindowanej wartości (Zaczynamy od {@code 1}).
	 * @param value
	 *            wartość w postaci string-a. Ważne by dla typu
	 *            {@link #TIMESTAMP} wartość spełniała format
	 *            {@code "yyyy-mm-dd hh:mm:ss[.f...]"}, a dla typu {@link #TIME}
	 *            format wartości musi być następujący: {@code "hh:mm:ss"}. Dla
	 *            typów daty wartość powinna spełniać format podany w parametrze
	 *            {@code dateFormat}.
	 * @param dateFormat
	 *            format na potrzeby sparsowania obiektu daty. Jeżeli podczas
	 *            konwersji daty, format daty będzie {@code null}, to nastąpi
	 *            próba sparsowania wartości do postaci {@link Long}, czyli
	 *            zakładamy, że wtedy czas jest podany w postaci liczby
	 *            milisekund.
	 * @throws SQLException
	 * @throws ParseException
	 */
	public void bindValue(PreparedStatement ps, int parameterIndex,
			String value, SimpleDateFormat dateFormat) throws SQLException,
			ParseException {

		Object parsedValue = this.parseValue(value, dateFormat);
		this.bindValue(ps, parameterIndex, parsedValue);
	}

	/**
	 * Transformacja wartości obiektu do łańcucha znakowego.
	 * <p>
	 * <font color="#dd2222">Uwaga!</font> Wartość typu daty jest konwertowana
	 * do wartości typu {@link Long} (czas w milisekundach) a następnie do
	 * 'string-a'.
	 * </p>
	 * 
	 * @param value
	 *            wartość
	 * @return repreezentacja znakowa wartości
	 */
	public String getStringValue(Object value) {
		if (value == null) {
			return "NULL";
		}
		if (value instanceof String) {
			return (String) value;
		}

		switch (this) {
		case INTEGER:
		case INT:
			return Integer.toString((Integer) value);
		case LONG:
			return Long.toString((Long) value);
		case SHORT:
			return Short.toString((Short) value);
		case FLOAT:
			return Float.toString((Float) value);
		case DOUBLE:
			return Double.toString((Double) value);
		case BIGDECIMAL:
			return ((BigDecimal) value).toString();
		case BYTE:
			return Byte.toString((Byte) value);
		case BOOLEAN:
			return (String) value;
		case DATE:
			if (value instanceof java.util.Date) {
				java.util.Date d = (java.util.Date) value;
				return Long.toString(d.getTime());
			} else {
				return Long.toString(((Date) value).getTime());
			}
		case CALENDAR:
			Calendar calendar = (Calendar) value;
			Date d = new Date((calendar.getTime()).getTime());
			return Long.toString(d.getTime());
		case CLOB:
		case NCLOB:
		case BLOB:
			return value.toString();
		case TIME:
			return ((Time) value).toString();
		case TIMESTAMP:
			return ((Timestamp) value).toString();
		default:
			break;
		}

		return value.toString();

	}

	/**
	 * Parsowanie wartości obiektu w postacji łańcucha znakowego.
	 * 
	 * @param value
	 *            wartość w postaci string-a. Ważne by dla typu
	 *            {@link #TIMESTAMP} wartość spełniała format
	 *            {@code "yyyy-mm-dd hh:mm:ss[.f...]"}, a dla typu {@link #TIME}
	 *            format wartości musi być następujący: {@code "hh:mm:ss"}. Dla
	 *            typów daty wartość powinna spełniać format podany w parametrze
	 *            {@code dateFormat}.
	 * @param dateFormat
	 *            format na potrzeby sparsowania obiektu daty. Jeżeli podczas
	 *            konwersji daty, format daty będzie {@code null}, to nastąpi
	 *            próba sparsowania wartości do postaci {@link Long}, czyli
	 *            zakładamy, że wtedy czas jest podany w postaci liczby
	 *            milisekund.
	 * @return sparsowana wartość obiektu
	 * @throws ParseException
	 */
	public Object parseValue(String value, SimpleDateFormat dateFormat)
			throws ParseException {
		if (StringUtils.isBlank(value)) {
			return null;
		}

		switch (this) {
		case INTEGER:
		case INT:
			// zrobie obsluge wyjątku dla liczb "calkowitych", z prowadzącym '0'
			// miejsca dziesiętnego.
			String[] numberElements = value.split("\\.");
			return Integer.parseInt(numberElements[0]);
		case LONG:
			return Long.parseLong(value);
		case SHORT:
			return Short.parseShort(value);
		case FLOAT:
			return Float.parseFloat(value);
		case DOUBLE:
			return Double.parseDouble(value);
		case BIGDECIMAL:
			return BigDecimal.valueOf(Double.parseDouble(value));
		case BYTE:
			return Byte.parseByte(value);
		case BOOLEAN:
			return value;
		case DATE:
			if (dateFormat != null) {
				return dateFormat.parse(value);
			} else {
				return new java.util.Date(Long.parseLong(value));
			}
		case CALENDAR:
			Calendar calendar = Calendar.getInstance();
			if (dateFormat != null) {
				calendar.setTime(dateFormat.parse(value));
			} else {
				calendar.setTime(new java.util.Date(Long.parseLong(value)));
			}
			return calendar;
		case CLOB:
		case NCLOB:
		case BLOB:
			throw new IllegalArgumentException(
					"Brak wsparcia dla parsowania pol typu LOB");
		case TIME:
			return Time.valueOf(value);
		case TIMESTAMP:
			return Timestamp.valueOf(value);
		default:
			break;
		}

		return value.toString();

	}

	/**
	 * Pobieranie opisu typu na potrzeby utworzenia klauzuli tworzenia tablicy i
	 * innych obiektów obsługiwanej bazy danych
	 * 
	 * @param dialect
	 *            dialekt bazy danych
	 * @return opis typu np. {@code VARCHAR2(4000)}
	 */
	public String getCreationType(Dialect dialect) {
		switch (dialect) {
		case ORACLE:
			switch (this) {
			case STRING:
				return "VARCHAR2(4000)";
			case INTEGER:
			case BYTE:
			case INT:
				return "INTEGER";
			case LONG:
			case SHORT:
			case FLOAT:
			case DOUBLE:
			case BIGDECIMAL:
				return "NUMBER";
			case BOOLEAN:
				return "VARCHAR2(5)";
			case TIME:
			case DATE:
			case CALENDAR:
				return "DATE";
			case CLOB:
			case NCLOB:
				return "CLOB";
			case BLOB:
				return "BLOB";
			case TIMESTAMP:
				return "TIMESTAMP";
			default:
				break;
			}
			break;
		default:
			switch (this) {
			case STRING:
				return "VARCHAR(255)";
			case INTEGER:
			case BYTE:
			case INT:
				return "INTEGER";
			case LONG:
			case SHORT:
			case FLOAT:
			case DOUBLE:
			case BIGDECIMAL:
				return "NUMBER";
			case BOOLEAN:
				return "VARCHAR(5)";
			case TIME:
			case DATE:
			case CALENDAR:
				return "DATE";
			case CLOB:
			case NCLOB:
				return "CLOB";
			case BLOB:
				return "BLOB";
			case TIMESTAMP:
				return "TIMESTAMP";
			default:
				break;
			}
			break;
		}
		return null;

	}

	/**
	 * Pobranie klasy z pakietu {@code java.*} odpowiadającej danemu typowi SQL
	 * 
	 * @return klasa z pakietu {@code java.*}
	 */
	public Class<?> getObjectClass() {
		switch (this) {
		case STRING: // java.lang.String
			return String.class;
		case INTEGER: // java.lang.Integer
		case INT: // int
			return Integer.class;
		case LONG: // java.lang.Long or long
			return Long.class;
		case SHORT: // java.lang.Short or short
			return Short.class;
		case FLOAT: // java.lang.Float or float
			return Float.class;
		case DOUBLE: // java.lang.Double or double
			return Double.class;
		case BIGDECIMAL: // java.lang.BigDecimal
			return BigDecimal.class;
		case BYTE: // java.lang.Byte or byte
			return Byte.class;
		case BOOLEAN: // java.lang.Boolean or boolean
			return Boolean.class;
		case DATE: // java.util.Date or java.sql.Date
			return java.sql.Date.class;
		case CALENDAR: // java.util.Calendar
			return java.util.Calendar.class;
		case CLOB: // java.sql.Clob
			return java.sql.Clob.class;
		case NCLOB: // java.sql.NClob
			return java.sql.NClob.class;
		case BLOB: // java.sql.Blob
			return java.sql.Blob.class;
		case TIME: // java.sql.Time
			return java.sql.Time.class;
		case TIMESTAMP: // java.sql.Timestamp
			return java.sql.Timestamp.class;
		default:
			return Object.class;
		}
	}

}
