package pl.slawas.scjdbc4jpa.jdbc.utils;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.helpers.PrimitiveType;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityRuntimeException;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;

/**
 * 
 * JDBCEntityValueParser klasa narzedziowa zawierająca metody przeksztalcenia
 * encji do struktur definicji JDBC.
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class JDBCEntityValueParser {

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCEntityValueParser.class);

	/**
	 * Przeczytanie wartości pól obiektu encji. Metoda wykorzystywana do
	 * operacji wstawiania i modyfikacji danych w bazie danych.
	 * 
	 * @param table
	 *            definicja tabeli związana z obiektem encji
	 * @param entity
	 *            obiekt encji
	 * @return mapa definicji kolumn tabeli z odpowiednimi wartościami
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<JDBCColumnDesc, Object> readEntityObject(
			JDBCTableDesc table, Object entity)
			throws IllegalArgumentException, IllegalAccessException,
			InvocationTargetException {

		Hashtable<JDBCColumnDesc, Object> values = new Hashtable<JDBCColumnDesc, Object>();
		if (entity == null)
			return values;
		entity = checkEntityValue(table, entity);

		Enumeration<String> columnNames = table.getColumnKeys();
		while (columnNames.hasMoreElements()) {
			String entityFieldName = columnNames.nextElement();
			JDBCColumnDesc column = table.getColumn(entityFieldName);
			write2column(column, entity, 0, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól obiektu encji i zwrócenie prostej mapy ['nazwa
	 * pola encji', 'wartość'].
	 * 
	 * @param table
	 *            definicja tabeli związana z obiektem encji
	 * @param entity
	 *            obiekt encji
	 * @return prosta mapa ['nazwa pola encji', 'wartość']
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<String, Object> readEntityObjectAsSimpleMap(
			JDBCTableDesc table, Object entity)
			throws IllegalArgumentException, IllegalAccessException,
			InvocationTargetException {

		Hashtable<String, Object> values = new Hashtable<String, Object>();
		if (entity == null)
			return values;
		entity = checkEntityValue(table, entity);

		Enumeration<String> columnNames = table.getColumnKeys();
		while (columnNames.hasMoreElements()) {
			String entityFieldName = columnNames.nextElement();
			JDBCColumnDesc column = table.getColumn(entityFieldName);
			write2field(column, entity, 0, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól obiektu encji. Metoda wykorzystywana do
	 * przeczytwania wartości kluczy.
	 * 
	 * @param key
	 *            definicja klucza związanego z obiektem encji
	 * @param entity
	 *            obiekt encji
	 * @return mapa definicji kolumn klucza z odpowiednimi wartościami
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<JDBCColumnDesc, Object> readEntityObject(
			JDBCKeyDesc key, Object entity) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {

		Hashtable<JDBCColumnDesc, Object> values = new Hashtable<JDBCColumnDesc, Object>();
		if (entity == null)
			return values;
		entity = checkEntityValue(key.getColumns().get(0).getTable(), entity);

		Enumeration<Integer> columnPositions = key.getColumns().keys();
		while (columnPositions.hasMoreElements()) {
			Integer position = columnPositions.nextElement();
			JDBCColumnDesc column = key.getColumn(position);
			write2column(column, entity, 0, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól obiektu encji i zwrócenie prostej mapy ['nazwa
	 * pola encji', 'wartość']. Metoda wykorzystywana do przeczytwania wartości
	 * kluczy.
	 * 
	 * @param key
	 *            definicja klucza związanego z obiektem encji
	 * @param entity
	 *            obiekt encji
	 * @return prosta mapa ['nazwa pola encji', 'wartość']
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<String, Object> readEntityObjectAsSimpleMap(
			JDBCKeyDesc key, Object entity) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {

		Hashtable<String, Object> values = new Hashtable<String, Object>();
		if (entity == null)
			return values;
		entity = checkEntityValue(key.getColumns().get(0).getTable(), entity);

		Enumeration<Integer> columnPositions = key.getColumns().keys();
		while (columnPositions.hasMoreElements()) {
			Integer position = columnPositions.nextElement();
			JDBCColumnDesc column = key.getColumn(position);
			write2field(column, entity, 0, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól obiektu encji i zwrócenie prostej mapy
	 * ['pozycja kolumny w kluczu', 'wartość']. Metoda wykorzystywana do
	 * przeczytwania wartości kluczy.
	 * 
	 * @param key
	 *            definicja klucza związanego z obiektem encji
	 * @param entity
	 *            obiekt encji
	 * @return prosta mapa ['pozycja kolumny w kluczu', 'wartość']
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<Integer, Object> readEntityObjectAsPositionMap(
			JDBCKeyDesc key, Object entity) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {

		Hashtable<Integer, Object> values = new Hashtable<Integer, Object>();
		if (entity == null)
			return values;
		entity = checkEntityValue(key.getColumns().get(0).getTable(), entity);

		Enumeration<Integer> columnPositions = key.getColumns().keys();
		while (columnPositions.hasMoreElements()) {
			Integer position = columnPositions.nextElement();
			JDBCColumnDesc column = key.getColumn(position);
			write2position(position, column, entity, 0, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól klucza obiektu encji i zwrócenie prostej mapy
	 * ['pozycja kolumny w kluczu', 'wartość']. Metoda wykorzystywana do
	 * przeczytwania wartości kluczy.
	 * 
	 * @param key
	 *            definicja klucza związanego z obiektem encji
	 * @param keyValue
	 *            obiekt klucza encji
	 * @return prosta mapa ['pozycja kolumny w kluczu', 'wartość']
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<Integer, Object> readKeyObjectAsPositionMap(
			JDBCKeyDesc key, Object keyValue) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {

		Hashtable<Integer, Object> values = new Hashtable<Integer, Object>();
		if (keyValue == null)
			return values;
		keyValue = checkKeyValue(key, keyValue);

		Enumeration<Integer> columnPositions = key.getColumns().keys();
		while (columnPositions.hasMoreElements()) {
			Integer position = columnPositions.nextElement();
			JDBCColumnDesc column = key.getColumn(position);
			write2position(position, column, keyValue, 1, values);
		}

		return values;
	}

	/**
	 * Przeczytanie wartości pól klucza obiektu encji i zwrócenie mapowania
	 * definicji kolumny do wartości pola. Metoda wykorzystywana do
	 * przeczytwania wartości kluczy.
	 * 
	 * @param key
	 *            definicja klucza związanego z obiektem encji
	 * @param keyValue
	 *            obiekt klucza encji
	 * @return mapa definicji kolumny do wartości pola
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static Hashtable<JDBCColumnDesc, Object> readKeyObject(
			JDBCKeyDesc key, Object keyValue) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {

		Hashtable<JDBCColumnDesc, Object> values = new Hashtable<JDBCColumnDesc, Object>();
		if (keyValue == null)
			return values;
		keyValue = checkKeyValue(key, keyValue);

		Enumeration<Integer> columnPositions = key.getColumns().keys();
		while (columnPositions.hasMoreElements()) {
			Integer position = columnPositions.nextElement();
			JDBCColumnDesc column = key.getColumn(position);
			write2column(column, keyValue, 1, values);
		}

		return values;
	}

	/**
	 * Metoda pomocnicza analizująca obiekt encji i wiążąca odpowiednie wartośći
	 * z definicjami kolumn.
	 * 
	 * @param column
	 *            definicja kolumny
	 * @param entity
	 *            obiekt encji
	 * @param startRecursiveLevel
	 *            poziom zagnieżdżenia, od którego ma być analizowany obiekt
	 *            encji. Dla samej encji parametr przyjmuje wartość {@code 0},
	 *            dla klucza {@code 1}
	 * @param values
	 *            aktualizowana mapa definicji kolumn z odpowiednimi wartościami
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void write2column(JDBCColumnDesc column, Object entity,
			int startRecursiveLevel, Hashtable<JDBCColumnDesc, Object> values)
			throws IllegalAccessException, InvocationTargetException {

		Object value = null;
		value = readFieldValue(column, entity, startRecursiveLevel);
		if (value != null) {
			values.put(column, value);
		}
	}

	/**
	 * Metoda pomocnicza analizująca obiekt encji i wiążąca odpowiednie wartośći
	 * z nazwami pól encji.
	 * 
	 * @param column
	 *            definicja kolumny
	 * @param entity
	 *            obiekt encji
	 * @param startRecursiveLevel
	 *            poziom zagnieżdżenia, od którego ma być analizowany obiekt
	 *            encji. Dla samej encji parametr przyjmuje wartość {@code 0},
	 *            dla klucza {@code 1}
	 * @param values
	 *            aktualizowana mapa definicji kolumn z odpowiednimi wartościami
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void write2field(JDBCColumnDesc column, Object entity,
			int startRecursiveLevel, Hashtable<String, Object> values)
			throws IllegalAccessException, InvocationTargetException {

		Object value = null;
		value = readFieldValue(column, entity, startRecursiveLevel);
		if (value != null) {
			values.put(column.getEntityFieldName(), value);
		}
	}

	/**
	 * Metoda pomocnicza analizująca obiekt encji i wiążąca odpowiednie wartośći
	 * z pozycją w zapytaniu SQL.
	 * 
	 * @param position
	 *            pozycja kolumny
	 * @param column
	 *            definicja kolumny
	 * @param entity
	 *            obiekt encji
	 * @param startRecursiveLevel
	 *            poziom zagnieżdżenia, od którego ma być analizowany obiekt
	 *            encji. Dla samej encji parametr przyjmuje wartość {@code 0},
	 *            dla klucza {@code 1}
	 * @param values
	 *            aktualizowana mapa definicji kolumn z odpowiednimi wartościami
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void write2position(Integer position, JDBCColumnDesc column,
			Object entity, int startRecursiveLevel,
			Hashtable<Integer, Object> values) throws IllegalAccessException,
			InvocationTargetException {

		Object value = null;
		value = readFieldValue(column, entity, startRecursiveLevel);
		if (value != null) {
			values.put(position, value);
		}
	}

	/**
	 * Metoda pomocnicza czytająca wartość pola na podstawie defincji kolumny
	 * 
	 * @param column
	 *            definicja kolumny
	 * @param entity
	 *            obiekt encji
	 * @param startRecursiveLevel
	 *            poziom zagnieżdżenia, od którego ma być analizowany obiekt
	 *            encji. Dla samej encji parametr przyjmuje wartość {@code 0},
	 *            dla klucza {@code 1}
	 * @return wartość pola
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static Object readFieldValue(JDBCColumnDesc column, Object entity,
			int startRecursiveLevel) throws IllegalAccessException,
			InvocationTargetException {
		Object value;
		Enumeration<Method> entityFieldGetters = column.getGetters();
		String entityFieldName = column.getEntityFieldName();
		String[] entityFieldElements = entityFieldName.split("\\.");
		Object complexField = entity;
		for (int recursiveLevel = 0; recursiveLevel < entityFieldElements.length; recursiveLevel++) {
			if (complexField == null)
				break;

			Method getter = entityFieldGetters.nextElement();
			if (recursiveLevel >= startRecursiveLevel) {
				complexField = getter.invoke(complexField);
			}
		}
		value = complexField;
		return value;
	}

	/**
	 * Metoda przepisująca wartości kolumn do obiektu encji
	 * 
	 * 
	 * @param entity
	 *            obiekt encji (musi być nie pusty!)
	 * @param value
	 *            wartość pola jaka ma być przepisana do pola zdefiniowanego
	 *            przez opis kolumny
	 * @param col
	 *            opis kolumny
	 * @param type
	 *            SQL-owy typ polumny, której wartość ma być przepisana do pola
	 *            encji
	 * @param entityFieldName
	 *            nazwa poola encji, które ma być uzupełnione
	 * @throws NoSuchMethodException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws SQLException
	 * @throws InstantiationException
	 */
	public static void write2EntityObject(Object entity, Object value,
			JDBCColumnDesc col, SimpleSQLType type, String entityFieldName)
			throws NoSuchMethodException, IllegalAccessException,
			InvocationTargetException, SQLException, InstantiationException {
		if (entity == null) {
			return;
		}

		/* tworzenie pola obiektu encji - START */
		String[] entityFieldElements = entityFieldName.split("\\.");
		Enumeration<Class<?>> entityFieldTypes = col.getEntityFieldTypes();
		Enumeration<Method> entityFieldSetters = col.getSetters();

		Enumeration<Method> entityFieldGetters = col.getGetters();
		if (entityFieldElements.length == 1) {
			/* pole proste - START */
			Class<?> entityFieldType = entityFieldTypes.nextElement();
			setSimpleField(entity, type, value, entityFieldType,
					entityFieldSetters.nextElement());
			/* pole proste - KONIEC */
		} else {
			/* pole z obiektem złożonym - START */
			int recursiveLevel = 0;
			Object complexField = entity;

			while (entityFieldTypes.hasMoreElements()) {
				Class<?> entityFieldType = entityFieldTypes.nextElement();
				String complexFieldName = entityFieldElements[recursiveLevel];
				logger.trace("\n[Rekursja {}] Info:\n {}\n {}\n {}",
						new Object[] { recursiveLevel, complexField.getClass(),
								entityFieldType, complexFieldName });

				Method setter = entityFieldSetters.nextElement();
				Method getter = entityFieldGetters.nextElement();
				Object fieldValue = (Object) getter.invoke(complexField);

				if (fieldValue == null && entityFieldTypes.hasMoreElements()) {
					fieldValue = entityFieldType.newInstance();
				}
				if (entityFieldTypes.hasMoreElements()) {
					setter.invoke(complexField, fieldValue);
				} else {
					setSimpleField(complexField, type, value, entityFieldType,
							setter);
				}
				complexField = fieldValue;
				recursiveLevel++;
			}

			/* pole z obiektem złożonym - KONIEC */
		}
		/* tworzenie obiektu encji - KONIEC */
	}

	/**
	 * Ustawienie wartości pola obiektu java.
	 * 
	 * @param entity
	 *            obiekt właściciela pola (obiekt, w którym zdefiniowane jest
	 *            pole)
	 * @param type
	 *            SQL-owy typ kolumny
	 * @param val
	 *            wartość kolumny
	 * @param entityFieldType
	 *            typ pola
	 * @param method
	 *            metoda ustawiająca wartość kolumny w polu obiektu java
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws SQLException
	 */
	private static void setSimpleField(Object entity, SimpleSQLType type,
			Object val, Class<?> entityFieldType, Method method)
			throws IllegalAccessException, InvocationTargetException,
			SQLException {

		if (val == null) {
			/** obsluga wartosci null */
			method.invoke(entity, new Object[] { null });
		} else if (val.getClass().equals(entityFieldType)) {
			/* typy kolumny w bazie i encji takie same - START */
			method.invoke(entity, val);
			/* typy kolumny w bazie i encji takie same - KONIEC */
		} else {
			logger.trace("Wystepuje roznica typow: entity={}, val={}",
					new Object[] { entityFieldType.getName(),
							val.getClass().getName() });
			/* typy kolumny w bazie i encji są różne - START */
			if (entityFieldType.isPrimitive()) {
				/* obsługa typów prymitywnych - START */
				/* typy prymitywne (zobacz {@link PrimitiveType}) */
				primitiveTypeAnalyze(entity, val, entityFieldType, method);
				/* obsługa typów prymitywnych - KONIEC */
			} else if (entityFieldType.isArray()) {
				/* obsługa macierzy typów podstawowych - START */
				arrayTypeAnalyze(entity, type, val, entityFieldType, method);
				/* obsługa macierzy typów podstawowych - KONIEC */
			} else {
				boolean isResolved = false;

				switch (type) {
				case DATE:
					/* obsługa daty nie SQL - START */
					Date noJavaSqlDate = (Date) val;
					method.invoke(entity, noJavaSqlDate);
					/* obsługa daty nie SQL - KONIEC */
					isResolved = true;
					break;
				case CALENDAR:
					if (val instanceof Calendar) {
						Calendar gCalendar = (Calendar) val;
						method.invoke(entity, gCalendar);
						isResolved = true;
					} else if (val instanceof Date) {
						/* obsługa daty - START */
						Calendar calendar = Calendar.getInstance();
						calendar.setTime((Date) val);
						method.invoke(entity, calendar);
						/* obsługa daty - KONIEC */
						isResolved = true;
					}
					break;
				case CLOB:
					Clob clob = (Clob) val;
					if (entityFieldType.equals(String.class)) {
						/* obsługa kmonwersji CLOB do String-a - START */
						String strVal = clob.getSubString(1,
								(int) clob.length());
						method.invoke(entity, strVal);
						/* obsługa kmonwersji CLOB do String-a - KONIEC */
					} else {
						method.invoke(entity, clob);
					}
					isResolved = true;
					break;
				case BLOB:
					Blob blob = (Blob) val;
					method.invoke(entity, blob);
					isResolved = true;
					break;
				default:
					break;
				}

				if (!isResolved && logger.isTraceEnabled()) {
					logger.warn(
							"Nie obsluguje roznic typow: entity={}, val={}",
							new Object[] { entityFieldType.getName(),
									val.getClass().getName() });
				}
			}
			/* typy kolumny w bazie i encji są różne - KONIEC */
		}
	}

	/**
	 * Analiza i przekształcenie do obiektu encji typu macierzowego.
	 * 
	 * @param entity
	 *            encja
	 * @param type
	 *            typ kolumny w bazie danych (SQL)
	 * @param val
	 *            wartość kolumny
	 * @param entityFieldType
	 *            typ pola encji
	 * @param method
	 *            metoda ustwienia wartości pola w encji
	 * @throws SQLException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void arrayTypeAnalyze(Object entity, SimpleSQLType type,
			Object val, Class<?> entityFieldType, Method method)
			throws SQLException, IllegalAccessException,
			InvocationTargetException {
		Class<?> componentType = entityFieldType.getComponentType();
		if (type.equals(SimpleSQLType.BLOB)
				&& (componentType.equals(Byte.class) || componentType
						.equals(byte.class))) {
			Blob blob = (Blob) val;
			byte[] primitiveArray = blob.getBytes(1, (int) blob.length());
			if (componentType.equals(byte.class)) {
				val = primitiveArray;
				method.invoke(entity, val);
			} else {
				Byte[] byteArray = new Byte[primitiveArray.length];
				for (int i = 0; i < primitiveArray.length; i++) {
					byteArray[i] = primitiveArray[i];
				}
				val = byteArray;
				method.invoke(entity, val);
			}
		} else if (logger.isTraceEnabled()) {
			logger.warn("Nie obsluguje roznic typow: entity={}, val={}",
					new Object[] { entityFieldType.getName(),
							val.getClass().getName() });
		}
	}

	/**
	 * Analiza i przekształcenie do obiektu encji typu prostego.
	 * 
	 * @param entity
	 *            encja
	 * @param val
	 *            wartość kolumny
	 * @param entityFieldType
	 *            typ pola encji
	 * @param method
	 *            metoda ustwienia wartości pola w encji
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void primitiveTypeAnalyze(Object entity, Object val,
			Class<?> entityFieldType, Method method)
			throws IllegalAccessException, InvocationTargetException {
		PrimitiveType p = PrimitiveType.get(entityFieldType.getName());
		if (p != null) {
			switch (p) {
			case INT:
				int valInt = ((Integer) val).intValue();
				method.invoke(entity, valInt);
				break;
			case BOOLEAN:
				boolean booleanValue = Boolean.valueOf(val.toString())
						.booleanValue();
				method.invoke(entity, booleanValue);
				break;
			case BYTE:
				byte byteValue = ((Byte) val).byteValue();
				method.invoke(entity, byteValue);
				break;
			case CHAR:
				char valChar = (char) Integer.valueOf(val.toString())
						.intValue();
				method.invoke(entity, valChar);
				break;
			case DOUBLE:
				double doubleValue = ((Double) val).doubleValue();
				method.invoke(entity, doubleValue);
				break;
			case FLOAT:
				float floatValue = ((Float) val).floatValue();
				method.invoke(entity, floatValue);
				break;
			case LONG:
				long longValue = ((Long) val).longValue();
				method.invoke(entity, longValue);
				break;
			case SHORT:
				short shortValue = ((Short) val).shortValue();
				method.invoke(entity, shortValue);
				break;
			default:
				if (logger.isTraceEnabled()) {
					logger.warn("Nie obsługuje typu {}",
							entityFieldType.getName());
				}
				break;
			}
		} else if (logger.isTraceEnabled()) {
			logger.warn("Nie obsluguje roznic typow: entity={}, val={}",
					new Object[] { entityFieldType.getName(),
							val.getClass().getName() });
		}
	}

	/**
	 * Sprawdzenie poprawności obiektu wartości poprzez kastowania go na typ
	 * klucza.
	 * 
	 * @param key
	 *            klucz
	 * @param keyValue
	 *            obiekt wartości
	 * @return skastowany obiekt wartości
	 */
	private static Object checkKeyValue(JDBCKeyDesc key, Object keyValue) {
		Class<?> keyClass = key.getKeyType();
		logger.trace(
				"Buduje warunek zapytania klucza glownego dla pola typu {} dla wartosci typu {}",
				new Object[] { keyClass, keyValue.getClass() });
		try {
			return keyClass.cast(keyValue);
		} catch (java.lang.ClassCastException e) {
			throw new JDBCEntityRuntimeException("Nie mozna rzutowac "
					+ keyValue.getClass() + " do " + keyClass, e);
		}
	}

	/**
	 * Sprawdzenie poprawności obiektu wartości poprzez kastowania go na typ
	 * tabeli zawarty w opisie tabeli.
	 * 
	 * @param table
	 *            opis tabeli
	 * @param entity
	 *            obiekt encji z wartościami
	 * @return skastowany obiekt wartości
	 */
	private static Object checkEntityValue(JDBCTableDesc table, Object entity) {
		Class<?> entityClass = table.getEntityClass();
		logger.trace(
				"Buduje warunek zapytania klucza glownego dla pola typu {}",
				new Object[] { entityClass });
		try {
			return entityClass.cast(entity);
		} catch (java.lang.ClassCastException e) {
			throw new JDBCEntityRuntimeException("Nie mozna rzutowac "
					+ entity.getClass() + " do " + entityClass, e);
		}
	}

	/**
	 * Kopiowanie wartosci encji pomiedzy roznymi obiektami encji
	 * 
	 * @param table
	 *            definicja tabeli
	 * @param target
	 *            encja docelowa (do której mają trafić kopiowane wartości)
	 * @param source
	 *            encja źródłowa (z której będą pobierane kopiowane wartości)
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws NoSuchMethodException
	 * @throws SQLException
	 * @throws InstantiationException
	 */
	public static void copyEntity(JDBCTableDesc table, Object target,
			Object source) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException,
			NoSuchMethodException, SQLException, InstantiationException {

		if (target == null || source == null) {
			throw new IllegalArgumentException(
					"Obiekty zrodla kopiowania oraz obiektu docelowego nie moga byc puste");
		}

		if (!target.getClass().equals(source.getClass())) {
			throw new IllegalArgumentException(
					"Kopiowane obiekty musza byc tej samej klasy");
		}

		if (!table.getEntityClass().equals(source.getClass())) {
			throw new IllegalArgumentException(
					"Zla definicja tabeli. Tabela definiuje encje o klasie "
							+ table.getEntityClass().getName()
							+ " natomiast zrodlo jest klasy "
							+ source.getClass().getName());
		}

		/** przekształcam nowe wartosci encji do listy pól */
		Hashtable<String, Object> entitiFelds = JDBCEntityValueParser
				.readEntityObjectAsSimpleMap(table, source);

		Enumeration<String> columnKeys = table.getColumnKeys();
		while (columnKeys.hasMoreElements()) {
			/*
			 * przepisywanie danych z listy pól do pola encji - START
			 */
			String entityFieldName = columnKeys.nextElement();
			JDBCColumnDesc col = table.getColumn(entityFieldName);
			SimpleSQLType type = SimpleSQLType.get(col.getColumnType());

			JDBCEntityValueParser.write2EntityObject(target,
					entitiFelds.get(entityFieldName), col, type,
					entityFieldName);
			/*
			 * przepisywanie danych z listy pól do pola encji - KONIEC
			 */
		}

		return;
	}

}
