package pl.slawas.scjdbc4jpa.jdbc.utils;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.sql.Blob;
import java.sql.Clob;
import java.util.Enumeration;

import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.EmbeddedId;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.JoinColumns;
import javax.persistence.Lob;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnSourceType;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyDesc;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCKeyType;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCTableDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.IdIsNotEmbeddableException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.JDBCEntityParserException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.ObjectIsNotEntityException;

/**
 * 
 * JdbcPersistenceEntityParser - parsowanie encji z annotacjami
 * javax.persistence.*
 * 
 * @see javax.persistence.Column;
 * @see javax.persistence.Embeddable;
 * @see javax.persistence.EmbeddedId;
 * @see javax.persistence.Entity;
 * @see javax.persistence.Id;
 * @see javax.persistence.JoinColumn;
 * @see javax.persistence.JoinColumns;
 * @see javax.persistence.ManyToOne;
 * @see javax.persistence.Table;
 * @see javax.persistence.Transient;
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.6 $
 * 
 */
public class JDBCPersistenceEntityParser {

	final static private Logger logger = LoggerFactory
			.getLogger(JDBCPersistenceEntityParser.class);

	/**
	 * Parsowanie encji do obiektu opisu tabeli pojedynczego wiersza.
	 * Implentacja posiada ograniczenia i jeżeli ktoś chciałby ją wykorzystać,
	 * musi pamiętac, że pomijane są pola z annotację
	 * {@link javax.persistence.OneToMany}, ze wzgledu na to, że powiązania
	 * takie zwracają listy (wiele wierszy) i zresztą takie pola encji nie
	 * należą do tabeli.
	 * 
	 * @param entityClass
	 *            klasa encji
	 * @param level
	 *            poziom zagnieżdżenia parsowania {@link JDBCEntityParserLevel},
	 *            jeżeli podamy {@code null} automatycznie zostanie przypisany
	 *            {@link JDBCEntityParserLevel#FIRST}
	 * @return obiekt opisu tabeli
	 * @throws ObjectIsNotEntityException
	 * @throws JDBCEntityParserException
	 * @throws IdIsNotEmbeddableException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 */
	public static JDBCTableDesc getTable(Class<?> entityClass,
			JDBCEntityParserLevel level) throws ObjectIsNotEntityException,
			JDBCEntityParserException, IdIsNotEmbeddableException,
			SecurityException, NoSuchMethodException {
		if (level == null)
			level = JDBCEntityParserLevel.FIRST;

		return getTable(1, entityClass, level);
	}

	/**
	 * Metoda właściwa parsowania encji do obiektu opisu tabeli i jeżeli trzeba
	 * jest wywoływana rekurencyjnie dla budowania tablic powiązanych.
	 * Implementacja pomija pola z annotację {@link javax.persistence.OneToMany}
	 * .
	 * 
	 * @see #getTable(Class, JDBCEntityParserLevel)
	 * 
	 * @param iterator
	 *            nymer rekurencji
	 * @param entityClass
	 *            klasa encji
	 * @param level
	 *            poziom zagnieżdżenia parsowania {@link JDBCEntityParserLevel},
	 * @return obiekt opisu tabeli
	 * @throws ObjectIsNotEntityException
	 * @throws JDBCEntityParserException
	 * @throws IdIsNotEmbeddableException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 */
	private static JDBCTableDesc getTable(int iterator, Class<?> entityClass,
			JDBCEntityParserLevel level) throws ObjectIsNotEntityException,
			JDBCEntityParserException, IdIsNotEmbeddableException,
			SecurityException, NoSuchMethodException {

		if (entityClass.getAnnotation(Entity.class) == null) {
			throw new ObjectIsNotEntityException(
					"Obiekt "
							+ entityClass.getName()
							+ " nie jest encja  (brak annotacji javax.persistence.Entity)");
		}

		/* utworzenie obiektu opisu tabeli */
		String tableName = entityClass.getSimpleName();
		Table tableAnnotation = entityClass.getAnnotation(Table.class);
		if (tableAnnotation != null) {
			tableName = (StringUtils.isNotBlank(tableAnnotation.schema()) ? tableAnnotation
					.schema() + "."
					: "")
					+ (StringUtils.isNotBlank(tableAnnotation.name()) ? tableAnnotation
							.name() : tableName);
		}
		JDBCTableDesc table = new JDBCTableDesc(level, entityClass, tableName);
		logger.debug("[{}] Tworze opis tabeli {}",
				new Object[] { entityClass.getSimpleName(), tableName });

		/* uzupełnienie informacji tabeli o kolumny i związki */
		Field[] entityFields = entityClass.getDeclaredFields();
		for (Field entityField : entityFields) {
			logger.debug("Analiza pola: {}...", entityField.getName());
			AnalizedField analizedField = skipField(entityClass, entityField);
			if (analizedField.isToSkip()) {
				logger.debug("Pomijam w analizie pole: {}",
						entityField.getName());
				continue;
			}
			String dbColumnName = entityField.getName();
			String entityFieldName = dbColumnName;
			Method setter = analizedField.getSetter();
			Method getter = analizedField.getGetter();

			if (entityField.isAnnotationPresent(Id.class)) {
				/* klucz prosty - START */
				if (entityField.isAnnotationPresent(Column.class)) {
					dbColumnName = ((Column) entityField
							.getAnnotation(Column.class)).name();
				}

				JDBCColumnDesc column = new JDBCColumnDesc(table,
						entityField.getType(), entityField.getType(),
						entityFieldName, dbColumnName,
						JDBCColumnSourceType.MAIN, setter, getter);
				table.addColumn(column);

				JDBCKeyDesc pk = new JDBCKeyDesc(entityField.getType(),
						entityFieldName, JDBCKeyType.PK);
				pk.addColumn(0, column);
				table.setPrimaryKey(pk);
				/* klucz prosty - KONIEC */
			} else if (entityField.isAnnotationPresent(EmbeddedId.class)) {
				/* klucz złozony - START */
				Class<?> fieldClass = entityField.getType();
				if (fieldClass.getAnnotation(Embeddable.class) == null) {
					throw new IdIsNotEmbeddableException(
							"Obiekt klucza  "
									+ fieldClass.getName()
									+ " nie jest kluczem złozonym (brak annotacji javax.persistence.Embeddable)");
				}
				JDBCKeyDesc pk = new JDBCKeyDesc(fieldClass, entityFieldName,
						JDBCKeyType.PK);
				Field[] keyFields = fieldClass.getDeclaredFields();
				int i = 0;
				for (Field keyField : keyFields) {
					logger.debug("Analiza pola klucza zlozonego: {}...",
							entityField.getName());
					AnalizedField analizedKeyField = skipField(fieldClass,
							keyField);
					if (analizedKeyField.isToSkip()) {
						logger.debug("Pomijam w analizie pole: {}",
								entityField.getName());
						continue;
					}
					String keyFieldName = entityFieldName + "."
							+ keyField.getName();
					String keyDbColumnName = keyField.getName();
					if (keyField.isAnnotationPresent(Column.class)) {
						keyDbColumnName = ((Column) keyField
								.getAnnotation(Column.class)).name();
					}

					Method keySetter = analizedKeyField.getSetter();
					Method keyGetter = analizedKeyField.getGetter();

					JDBCColumnDesc column = new JDBCColumnDesc(table,
							entityField.getType(), keyField.getType(),
							keyFieldName, keyDbColumnName,
							JDBCColumnSourceType.MAIN, setter, getter);
					column.addEntityFieldType(keyField.getType());
					column.addSetter(keySetter);
					column.addGetter(keyGetter);

					pk.addColumn(i, column);
					table.addColumn(column);
					i++;
				}
				table.setPrimaryKey(pk);
				/* klucz złozony - KONIEC */
			} else if (entityField.isAnnotationPresent(Column.class)) {
				/* kolumna zwykła - START */
				dbColumnName = ((Column) entityField
						.getAnnotation(Column.class)).name();
				addSimpleColumn(table, entityField, dbColumnName,
						entityFieldName, setter, getter);
				/* kolumna zwykła - KONIEC */
			} else if (entityField.isAnnotationPresent(ManyToOne.class)) {
				/* klucz obcy - START */
				JDBCTableDesc joinedTable = null;
				JDBCKeyDesc fk = null;
				switch (level) {
				case FIRST:
					joinedTable = getTable(iterator++, entityField.getType(),
							JDBCEntityParserLevel.FIRST);
					break;
				case SECOND:
					fk = new JDBCKeyDesc(entityField.getType(),
							entityFieldName, JDBCKeyType.FK);
					joinedTable = getTable(iterator++, entityField.getType(),
							JDBCEntityParserLevel.FIRST);
					break;
				case THIRD:
					fk = new JDBCKeyDesc(entityField.getType(),
							entityFieldName, JDBCKeyType.FK);
					joinedTable = getTable(iterator++, entityField.getType(),
							JDBCEntityParserLevel.SECOND);
					break;
				}

				JDBCKeyDesc joinedTablePK = joinedTable.getPrimaryKey();

				if (entityField.isAnnotationPresent(JoinColumns.class)) {
					/* klucz obcy złożony - START */
					JoinColumn joinColumns[] = ((JoinColumns) entityField
							.getAnnotation(JoinColumns.class)).value();

					for (JoinColumn joinColumn : joinColumns) {
						dbColumnName = joinColumn.name();
						logger.debug("Kolumna powiazania: {}", dbColumnName);
						String dbReferencedColumnName = joinColumn
								.referencedColumnName();
						logger.debug("Kolumna referencji: {}",
								dbReferencedColumnName);

						/* szukam odpowiednich typów dla kolumn */
						Class<?> entityFieldType = null;
						Enumeration<Method> pkSetters = null;
						Enumeration<Method> pkGetters = null;
						String colEntityFieldName = entityFieldName;
						for (int i = 0; i < joinColumns.length; i++) {
							JDBCColumnDesc pkCol = joinedTablePK.getColumn(i);
							if (dbReferencedColumnName.equals(pkCol.getName())) {
								entityFieldType = pkCol.getColumnType();
								colEntityFieldName = colEntityFieldName + "."
										+ pkCol.getEntityFieldName();
								pkSetters = pkCol.getSetters();
								pkGetters = pkCol.getGetters();
								break;
							}
						}
						JDBCColumnDesc column = new JDBCColumnDesc(table,
								entityField.getType(), entityFieldType,
								colEntityFieldName, dbColumnName,
								JDBCColumnSourceType.MAIN, setter, getter);
						column.addEntityFieldType(joinedTablePK.getKeyType());
						column.addEntityFieldType(entityFieldType);
						if (pkGetters != null)
							while (pkGetters.hasMoreElements()) {
								column.addGetter(pkGetters.nextElement());
							}
						if (pkSetters != null)
							while (pkSetters.hasMoreElements()) {
								column.addSetter(pkSetters.nextElement());
							}

						table.addColumn(column);
						if (fk != null) {
							/*
							 * uzupełniam informacje o kluczu obcym (ustawiam
							 * kolumny na odpowiedniej pozycji)
							 */
							for (int i = 0; i < joinColumns.length; i++) {
								JDBCColumnDesc pkCol = joinedTablePK
										.getColumn(i);
								if (dbReferencedColumnName.equals(pkCol
										.getName())) {
									fk.addColumn(i, column);
									break;
								}
							}
						}
					}
					if (fk != null) {
						/* dodaje informacje o powiazaniach */
						table.addJoinedTable(fk, joinedTable, setter, getter);
					}

					/* klucz obcy złożony - KONIEC */
				} else {
					/* klucz obcy prosty - START */
					if (entityField.isAnnotationPresent(JoinColumn.class)) {
						dbColumnName = ((JoinColumn) entityField
								.getAnnotation(JoinColumn.class)).name();
					}

					/* szukam odpowiedniego typu dla kolumny */
					JDBCColumnDesc column = new JDBCColumnDesc(table,
							entityField.getType(), joinedTablePK.getKeyType(),
							entityFieldName
									+ "."
									+ joinedTablePK.getColumn(0)
											.getEntityFieldName(),
							dbColumnName, JDBCColumnSourceType.MAIN, setter,
							getter);
					column.addEntityFieldType(joinedTablePK.getKeyType());
					/* pobieram gettery klucza głownego */
					Enumeration<Method> pkGetters = joinedTablePK.getColumn(0)
							.getGetters();
					while (pkGetters.hasMoreElements()) {
						column.addGetter(pkGetters.nextElement());
					}
					/* pobieram settery klucza głownego */
					Enumeration<Method> pkSetters = joinedTablePK.getColumn(0)
							.getSetters();
					while (pkSetters.hasMoreElements()) {
						column.addSetter(pkSetters.nextElement());
					}

					table.addColumn(column);
					if (fk != null) {
						/* dodaje informacje o powiazaniach */
						fk.addColumn(0, column);
						table.addJoinedTable(fk, joinedTable, setter, getter);
					}
					/* klucz obcy prosty - KONIEC */
				}
				/* klucz obcy - KONIEC */
			} else {
				/*
				 * kolumna nie opisana annotacja javax.persistence.Column, ale
				 * też nie opisana annotacją javax.persistence.Transient - je
				 * też trzeba brać pod uwagę
				 */
				/* START */
				addSimpleColumn(table, entityField, dbColumnName,
						entityFieldName, setter, getter);
				/* KONIEC */
			}
		}

		return table;
	}

	/**
	 * Metoda analizująca czy dane pole encji ma być pominięte w iterpretacji
	 * tego pola jako kolumny
	 * 
	 * @param entityClazz
	 *            klasa encji
	 * @param entityField
	 *            pole encji
	 * @return wynik analizy pola w postaci obiektu {@link AnalizedField}.
	 *         Jezeli pole nie jest kolumna w bazie danych (pole ma być
	 *         pominięte w iterpretacji jako kolumna w bazie) metoda zwróci
	 *         wartość obiekt z @ {@link AnalizedField#toSkip} ustawionym na
	 *         {@code true}, w przeciwnym razie {@code false} oraz zostaną
	 *         ustawione odpowiednie pola {@link AnalizedField#setter} oraz
	 *         {@link AnalizedField#getter}.
	 */
	public static AnalizedField skipField(Class<?> entityClazz,
			Field entityField) {

		/* tych pól z definicji nie obsługujemy */
		if (entityField.getName().equals("serialVersionUID")
				|| entityField.isAnnotationPresent(Transient.class)
				|| entityField.isAnnotationPresent(OneToMany.class)
				|| entityField.isSynthetic() || entityField.isEnumConstant())
			return new AnalizedField(true, null, null);

		Method setter = null;
		Method getter = null;
		/* reszte musimy sprawdzić pod kątem istnienia setterów i getterów */
		try {
			String getterMethodName = "get"
					+ entityField.getName().substring(0, 1).toUpperCase()
					+ entityField.getName().substring(1);
			String setterMethodName = "set"
					+ entityField.getName().substring(0, 1).toUpperCase()
					+ entityField.getName().substring(1);
			Class<?> parTypes[] = new Class[] { entityField.getType() };
			setter = entityClazz.getMethod(setterMethodName, parTypes);
			getter = entityClazz.getMethod(getterMethodName);
		} catch (SecurityException e) {
			return new AnalizedField(true, null, null);
		} catch (NoSuchMethodException e) {
			return new AnalizedField(true, null, null);
		}

		return new AnalizedField(false, setter, getter);
	}

	/**
	 * Metoda wspierająca dodanie prostej kolumny. Implementacja zawiera
	 * dodatkową analize czy kolumna w bazie danych nie jest CLOB albo BLOB.
	 * 
	 * @param table
	 *            definicja tabeli
	 * @param entityField
	 *            pole encji
	 * @param dbColumnName
	 *            nazwa kolumny
	 * @param entityFieldName
	 *            nazwa pola encji
	 * @throws JDBCEntityParserException
	 */
	private static void addSimpleColumn(JDBCTableDesc table, Field entityField,
			String dbColumnName, String entityFieldName, Method setter,
			Method getter) throws JDBCEntityParserException {

		Class<?> columnType = entityField.getType();
		if (entityField.isAnnotationPresent(Lob.class)) {
			if (columnType.equals(String.class)) {
				columnType = Clob.class;
			} else if (columnType.isArray()) {
				Class<?> componentType = columnType.getComponentType();
				if (componentType.equals(Byte.class)
						|| componentType.equals(byte.class)) {
					columnType = Blob.class;
				} else {
					columnType = Clob.class;
				}
			}
		}
		if (columnType.equals(char.class)) {
			columnType = String.class;
		}

		JDBCColumnDesc column = new JDBCColumnDesc(table,
				entityField.getType(), columnType, entityFieldName,
				dbColumnName, JDBCColumnSourceType.MAIN, setter, getter);
		table.addColumn(column);
	}

	/**
	 * 
	 * AnalizedField wynik analizy pola przez metode
	 * {@link JDBCPersistenceEntityParser#skipField(Class, Field)}
	 * 
	 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
	 * @version $Revision: 1.6 $
	 * 
	 */
	public static class AnalizedField {

		private final boolean toSkip;

		private final Method getter;

		private final Method setter;

		/**
		 * @param toSkip
		 * @param setter
		 * @param getter
		 */
		public AnalizedField(boolean toSkip, Method setter, Method getter) {
			super();
			this.toSkip = toSkip;
			this.getter = getter;
			this.setter = setter;
		}

		/**
		 * @return the toSkip
		 */
		public boolean isToSkip() {
			return toSkip;
		}

		/**
		 * @return the getter
		 */
		public Method getGetter() {
			return getter;
		}

		/**
		 * @return the setter
		 */
		public Method getSetter() {
			return setter;
		}

	}

}
