package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.jdbc.SelectQuery;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.types.SimpleSQLType;

/**
 * 
 * ConditionClauseEmbededIn warunek zapytania z klauzulą IN z wieloma kolumnami
 * (np. dla kluczy złożonych).
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;1&quot;, &quot;'Natural'&quot;, &quot;0&quot;, &quot;'Surprise'&quot; };
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * Przykład z użyciem zmiennych bindowanych:
 * </p>
 * 
 * <pre>
 * Object[] expectedObjectParams = new Object[] { 1, &quot;Natural&quot;, 0, &quot;Surprise&quot; };
 * 
 * expectedCondition = &quot;AND (FATHER_ID, FATHER_TYPE) NOT IN ((?,?),(?,?))&quot;;
 * tested = new ConditionClauseEmbededIn(ConditionOperator.ANDNOT,
 * 		new JDBCColumnDesc[] { table.getColumn(&quot;id.fatherId&quot;),
 * 				table.getColumn(&quot;id.type&quot;) }, expectedObjectParams, true)
 * 
 * ;
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.2 $
 * 
 */
public class ConditionClauseEmbededIn extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param columns
	 *            lista kolumn, których dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseEmbededIn(ConditionOperator conditionOperator,
			JDBCColumnDesc[] columns, Object[] value, Boolean isBinded) {
		super(null, Boolean.class);
		init(conditionOperator, columns, value, isBinded);
	}

	private void init(ConditionOperator conditionOperator,
			JDBCColumnDesc[] columns, Object[] value, Boolean isBinded) {

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.BOOL);
		this.setParamOperator(ParamOperatorIn.IN);

		if (value == null || value.length == 0)
			throw new IllegalArgumentException(
					"Lista wartosci kryterium nie moze byc pusta");
		if (value.length > ConditionClauseIn.MAX_IN_VALUES)
			throw new IllegalArgumentException(
					"Lista wartosci kryterium nie moze byc "
							+ "dluzsza od zalozonej wartoci granicznej: "
							+ ConditionClauseIn.MAX_IN_VALUES);

		this.setConditionValue(value);
		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		int conditionsSize = value.length;
		if ((conditionsSize % columns.length) != 0)
			throw new IllegalArgumentException(
					"Lista wartosci musi byc wielokrotnoscia "
							+ "liczby kolumn uczestniczacych w warynku zapytania");

		StringBuffer embededFieldName = new StringBuffer();
		for (int i = 0; i < columns.length; i++) {
			embededFieldName.append((i == 0 ? "" : ", ")
					+ SelectQuery.getFieldName(columns[i % columns.length]));
		}
		this.getParam().setField(embededFieldName.toString());

		ConditionParam[] params = new ConditionParam[conditionsSize];
		for (int i = 0; i < conditionsSize; i++) {
			ConditionParam param = new ConditionParam();
			param.setOperator(ParamOperatorIn.IN);
			param.setType(ConditionType.IN);
			param.setField(SelectQuery
					.getFieldName(columns[i % columns.length]));
			SimpleSQLType type = SimpleSQLType.get(columns[i % columns.length]
					.getColumnType());
			if (type == null)
				throw new IllegalArgumentException("Niepoprawny typ pola");
			param.setFieldType(type);
			param.setValue(value[i]);

			if (isBinded != null && !isBinded)
				param.setValueAsLiteral();

			params[i] = param;
		}
		this.getParam().setParams(params);
	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		ConditionParam param = getParam();
		out.append(conditionOperator);
		out.append(" (" + param.getField() + ")");
		out.append(" " + operator);
		ConditionParam[] params = param.getParams();
		String[] fields = param.getField().split("\\,");
		if (param.isBindedValue()) {
			out.append(" (");
			for (int i = 0; i < params.length; i++) {
				boolean isFirstArg = (i % fields.length) == 0;
				boolean isLastArg = (i % fields.length) == fields.length - 1;
				out.append((0 == i ? "" : ",") + (isFirstArg ? "(" : "") + "?"
						+ (isLastArg ? ")" : ""));
			}
			out.append(")");
			return out.toString();
		}
		out.append(" (");
		for (int i = 0; i < params.length; i++) {
			if (params[i].getValue() instanceof String) {
				boolean isFirstArg = (i % fields.length) == 0;
				boolean isLastArg = (i % fields.length) == fields.length - 1;
				out.append((0 == i ? "" : ",") + (isFirstArg ? "(" : "")
						+ (String) params[i].getValue()
						+ (isLastArg ? ")" : ""));
			} else
				throw new IllegalArgumentException(
						"Jezeli uzywamy literalow to wartosc parametru musi byc obiektem klasy [Ljava.lang.String ");
		}
		out.append(")");
		return out.toString();
	}

}
