package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionImplementationException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionNotFoundException;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions.Function;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;

/**
 * 
 * ConditionClauseIn warunek zapytania z klauzulą IN.
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;AND chi.name NOT IN ('slawas','karol','tomek','ania')&quot;;
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * tested = new ConditionClauseIn(ConditionOperator.ANDNOT,
 * 		table.getColumn(&quot;name&quot;), expectedParams, false)
 * 
 * ;
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.5 $
 * 
 */
public class ConditionClauseIn extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Maksymalna liczba wartości jako argumenty klauzuli IN. Nie wolno go
	 * przektoczyć, ze względu na mechanizmy związane z pobieraniem encji (encje
	 * pobierane są klauzulą IN, i nie ważne z jakiego serwera bazy danych
	 * korzystamy, zbyt duża liczba argumentów prowadzi do katastrofy). Wartość
	 * parametru została wyznaczona na podstawie doświadzczenia i {@code 500}
	 * daje optyymalny wynik.
	 */
	public final static int MAX_IN_VALUES = 500;

	/**
	 * Domyślny dialekt dla funkcji
	 */
	public static final Dialect DEFAULT_DIALECT = Dialect.ORACLE;

	/** dialekt serwera baz danych używany podczas tworzenia klauzuli funkcji */
	private Dialect serverDialect;

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * <li>{@link ConditionParam#isCaseSensitive()} na wartość {@code true}
	 * (czyli rozróżnienie wielkości liter)</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseIn(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object[] value, Boolean isBinded) {
		super(column);
		init(conditionOperator, value, isBinded, true, null);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link ConditionParam#isCaseSensitive()} na wartość {@code true}
	 * (czyli rozróżnienie wielkości liter)</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji
	 */
	public ConditionClauseIn(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object[] value, Boolean isBinded,
			Dialect serverDialect) {
		super(column);
		init(conditionOperator, value, isBinded, true, serverDialect);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isCaseSensitive
	 *            flaga związana z tym, czy mają być rozrózniane wielkie i małe
	 *            litery podczas realizacji warunku. Wartość {@code true}
	 *            oznacza, że wielkości liter będą rozróżniane.
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji
	 */
	public ConditionClauseIn(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object[] value, boolean isCaseSensitive,
			Boolean isBinded, Dialect serverDialect) {
		super(column);
		init(conditionOperator, value, isBinded, isCaseSensitive, serverDialect);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isCaseSensitive
	 *            flaga związana z tym, czy mają być rozrózniane wielkie i małe
	 *            litery podczas realizacji warunku. Wartość {@code true}
	 *            oznacza, że wielkości liter będą rozróżniane.
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseIn(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object[] value, boolean isCaseSensitive,
			Boolean isBinded) {
		super(column);
		init(conditionOperator, value, isBinded, isCaseSensitive, null);
	}

	private void init(ConditionOperator conditionOperator, Object[] value,
			Boolean isBinded, boolean isCaseSensitive, Dialect serverDialect) {

		if (serverDialect == null)
			this.serverDialect = DEFAULT_DIALECT;
		else
			this.serverDialect = serverDialect;

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.IN);
		if (value == null || value.length == 0)
			throw new IllegalArgumentException(
					"Lista wartosci kryterium nie moze byc pusta");
		if (value.length > MAX_IN_VALUES)
			throw new IllegalArgumentException(
					"Lista wartosci kryterium nie moze byc "
							+ "dluzsza od zalozonej wartoci granicznej: "
							+ MAX_IN_VALUES);

		this.setConditionValue(value);
		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		this.setParamOperator(ParamOperatorIn.IN);
		this.getParam().setCaseSensitive(isCaseSensitive);

	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		try {
			ConditionParam param = getParam();
			out.append(conditionOperator);
			out.append(" " + getCaseSensitivity(param, param.getField()));
			out.append(" " + operator);
			Object[] values = (Object[]) param.getValue();
			if (param.isBindedValue()) {
				out.append(" (");
				for (int i = 0; i < values.length; i++) {
					out.append((0 == i ? "" : ",") + "?");
				}
				out.append(")");
				return out.toString();
			}
			out.append(" (");
			for (int i = 0; i < values.length; i++) {
				if (values[i] instanceof String) {
					out.append((0 == i ? "" : ",")
							+ getCaseSensitivity(param, (String) values[i]));
				} else
					throw new IllegalArgumentException(
							"Jezeli uzywamy literalow to wartosc parametru musi byc obiektem klasy [Ljava.lang.String ");
			}
			out.append(")");
		} catch (SQLFunctionNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (SQLFunctionImplementationException e) {
			throw new IllegalArgumentException(e);
		}
		return out.toString();
	}

	private String getCaseSensitivity(ConditionParam param, String value)
			throws SQLFunctionNotFoundException,
			SQLFunctionImplementationException {
		if (param.isCaseSensitive())
			return value;
		else {
			_IFunction concatFunction = Function.ToUpperCase.get(serverDialect,
					null, new String[] { value });
			return concatFunction.getClause();
		}
	}

	/**
	 * @return the serverDialect
	 */
	public Dialect getServerDialect() {
		return serverDialect;
	}

}
