package pl.slawas.scjdbc4jpa.jdbc.where;

import java.util.Vector;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionImplementationException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionNotFoundException;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions.Function;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;

/**
 * 
 * ConditionClauseLike warunek zapytania z klauzulą LIKE
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * </pre>
 * 
 * </p>
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;AND UPPER(name) LIKE UPPER(('%' || 'slawas'))&quot;;
 * tested = new ConditionClauseLike(ConditionOperator.AND,
 * 		table.getColumn(&quot;name&quot;), &quot;'slawas'&quot;, true, false, false, false);
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class ConditionClauseLike extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Domyślny dialekt dla funkcji
	 */
	public static final Dialect DEFAULT_DIALECT = Dialect.ORACLE;

	/** dialekt serwera baz danych używany podczas tworzenia klauzuli funkcji */
	private Dialect serverDialect;

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param inclusiveLower
	 *            flaga decydująca o tym czy z lewej strony wartość warunku
	 *            zostanie dopełniona znakiem {@code '%'}
	 * @param inclusiveUpper
	 *            flaga decydująca o tym czy z prawej strony wartość warunku
	 *            zostanie dopełniona znakiem {@code '%'}
	 * @param isCaseSensitive
	 *            flaga związana z tym, czy mają być rozrózniane wielkie i małe
	 *            litery podczas realizacji warunku. Wartość {@code true}
	 *            oznacza, że wielkości liter będą rozróżniane.
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseLike(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object value, boolean inclusiveLower,
			boolean inclusiveUpper, boolean isCaseSensitive, Boolean isBinded,
			Dialect serverDialect) {
		super(column);
		init(conditionOperator, value, isBinded, inclusiveLower,
				inclusiveUpper, isCaseSensitive, serverDialect);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param inclusiveLower
	 *            flaga decydująca o tym czy z lewej strony wartość warunku
	 *            zostanie dopełniona znakiem {@code '%'}
	 * @param inclusiveUpper
	 *            flaga decydująca o tym czy z prawej strony wartość warunku
	 *            zostanie dopełniona znakiem {@code '%'}
	 * @param isCaseSensitive
	 *            flaga związana z tym, czy mają być rozrózniane wielkie i małe
	 *            litery podczas realizacji warunku. Wartość {@code true}
	 *            oznacza, że wielkości liter będą rozróżniane.
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseLike(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object value, boolean inclusiveLower,
			boolean inclusiveUpper, boolean isCaseSensitive, Boolean isBinded) {
		super(column);
		init(conditionOperator, value, isBinded, inclusiveLower,
				inclusiveUpper, isCaseSensitive, null);
	}

	private void init(ConditionOperator conditionOperator, Object value,
			Boolean isBinded, boolean inclusiveLower, boolean inclusiveUpper,
			boolean isCaseSensitive, Dialect serverDialect) {

		if (serverDialect == null)
			this.serverDialect = DEFAULT_DIALECT;
		else
			this.serverDialect = serverDialect;

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.LIKE);
		if (value == null)
			throw new IllegalArgumentException("Wartosc nie moze byc pusta");
		this.setConditionValue(value);
		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		this.setParamOperator(ParamOperatorLike.LIKE);
		this.getParam().setInclusiveLower(inclusiveLower);
		this.getParam().setInclusiveUpper(inclusiveUpper);
		this.getParam().setCaseSensitive(isCaseSensitive);

	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		try {
			ConditionParam param = getParam();
			out.append(conditionOperator);
			out.append(" " + getCaseSensitivity(param, param.getField()));
			out.append(" " + operator);
			if (param.isBindedValue()) {
				out.append(" "
						+ getCaseSensitivity(param, getConcatFunction("?")
								.getClause()));
				return out.toString();
			}
			if (param.getValue() instanceof String) {
				out.append(" "
						+ getCaseSensitivity(param,
								getConcatFunction((String) param.getValue())
										.getClause()));
			} else
				throw new IllegalArgumentException(
						"Jezeli uzywamy literalow to wartosc parametru musi byc obiektem klasy java.lang.String ");
		} catch (SQLFunctionNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (SQLFunctionImplementationException e) {
			throw new IllegalArgumentException(e);
		}
		return out.toString();
	}

	private _IFunction getConcatFunction(String value)
			throws SQLFunctionNotFoundException,
			SQLFunctionImplementationException {
		Vector<String> args = new Vector<String>();
		if (getParam().isInclusiveLower())
			args.add("'%'");
		args.add(value);
		if (getParam().isInclusiveUpper())
			args.add("'%'");
		String[] argsArray = args.toArray(new String[args.size()]);
		_IFunction concatFunction = Function.Concat.get(serverDialect, null,
				argsArray);
		return concatFunction;
	}

	private String getCaseSensitivity(ConditionParam param, String value)
			throws SQLFunctionNotFoundException,
			SQLFunctionImplementationException {
		if (param.isCaseSensitive())
			return value;
		else {
			_IFunction concatFunction = Function.ToUpperCase.get(serverDialect,
					null, new String[] { value });
			return concatFunction.getClause();
		}
	}

	/**
	 * @return the serverDialect
	 */
	public Dialect getServerDialect() {
		return serverDialect;
	}

}
