package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;

/**
 * 
 * ConditionClauseRange warunek zaytania obsługujący sprawdzenie, czy zadana
 * kolumna posiada wartości zawarte pomiędzy górną i dolną granicą.
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * </pre>
 * 
 * </p>
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;OR nameDay NOT BETWEEN ? AND ?&quot;;
 * expectedParam = Calendar.getInstance();
 * tested = new ConditionClauseRange(ConditionOperator.ORNOT,
 * 		table.getColumn(&quot;nameDay&quot;), expectedParam, true, expectedParam, true,
 * 		true);
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * assertEquals(&quot;Blad dolnej granicy wartosci pola dla klauzuli warunku&quot;,
 * 		expectedParam, tested.getParam().getLowerBound());
 * assertEquals(&quot;Blad gornej granicy wartosci pola dla klauzuli warunku&quot;,
 * 		expectedParam, tested.getParam().getUpperBound());
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class ConditionClauseRange extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param lowerBound
	 *            wartość dolnej granicy warunku
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperBound
	 *            wartość górnej granicy warunku
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseRange(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object lowerBound, boolean inclusiveLower,
			Object upperBound, boolean inclusiveUpper, Boolean isBinded) {
		super(column);
		init(conditionOperator, lowerBound, inclusiveLower, upperBound,
				inclusiveUpper, isBinded);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param fieldName
	 *            nazwa pola (<font color="#dd0000"> trzeba znac nazwe pola
	 *            wystepującą w zapytaniu:
	 *            {@code &lt;ewentualny_alias_tabeli>.&lt;nazwa_kolumny_w_tabeli>}
	 *            )</font>
	 * @param fieldType
	 *            klasa typu pola
	 * @param lowerBound
	 *            wartość dolnej granicy warunku
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperBound
	 *            wartość górnej granicy warunku
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseRange(ConditionOperator conditionOperator,
			String fieldName, Class<?> fieldType, Object lowerBound,
			boolean inclusiveLower, Object upperBound, boolean inclusiveUpper,
			Boolean isBinded) {
		super(fieldName, fieldType);
		init(conditionOperator, lowerBound, inclusiveLower, upperBound,
				inclusiveUpper, isBinded);
	}

	private void init(ConditionOperator conditionOperator, Object lowerBound,
			boolean inclusiveLower, Object upperBound, boolean inclusiveUpper,
			Boolean isBinded) {

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.RANGE);
		if (lowerBound == null && upperBound == null)
			throw new IllegalArgumentException(
					"Obydwie granice przedzialu nie moga byc puste. "
							+ "Jezeli chcesz ograniczyc wynik tylko z jednej "
							+ "strony uzyj 'pl.slawas.jdbc.where.ConditionClauseSimple'.");

		this.getParam().setLowerBound(lowerBound);
		this.getParam().setInclusiveLower(inclusiveLower);
		this.getParam().setUpperBound(upperBound);
		this.getParam().setInclusiveUpper(inclusiveUpper);

		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		if (this.getParam().isInclusiveLower()
				&& this.getParam().isInclusiveUpper())
			this.setParamOperator(ParamOperatorRange.BETWEEN);
		else
			this.setParamOperator(ParamOperatorRange.SIMPLE);

	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		ConditionParam param = getParam();
		out.append(conditionOperator);
		switch ((ParamOperatorRange) param.getOperator()) {
		case BETWEEN:
			out.append(" " + param.getField());
			out.append(" " + operator);
			if (param.isBindedValue()) {
				out.append(" ? AND ?");
				return out.toString();
			}
			if (param.getLowerBound() instanceof String
					&& param.getUpperBound() instanceof String) {
				out.append(" " + param.getLowerBound() + " AND "
						+ param.getUpperBound());
			} else
				throw new IllegalArgumentException(
						"Jezeli uzywamy literalow to wartosci parametrow musza byc obiektami klasy java.lang.String ");
			break;
		default:
			out.append(" " + operator);
			if (param.isBindedValue()) {
				createRange(out, param, true);
				return out.toString();
			}
			if (param.getLowerBound() instanceof String
					&& param.getUpperBound() instanceof String) {
				createRange(out, param, false);
			} else
				throw new IllegalArgumentException(
						"Jezeli uzywamy literalow to wartosci parametrow musza byc obiektami klasy java.lang.String ");
			break;
		}

		return out.toString();
	}

	/**
	 * Metoda pomocnicza budująca klauzule przedziału na podstwaie operatorów
	 * parametrów prostych {@link ParamOperatorSimple}
	 * 
	 * @param out
	 *            klauzula
	 * @param param
	 *            parametr warunku
	 * @param binded
	 *            czy wartosci mają być bindowane
	 */
	private static void createRange(StringBuffer out, ConditionParam param,
			boolean binded) {
		out.append("(" + param.getField());
		if (param.isInclusiveLower()) {
			out.append(" " + ParamOperatorSimple.GEQ.getOperator() + " ");
		} else {
			out.append(" " + ParamOperatorSimple.GT.getOperator() + " ");
		}
		out.append((binded ? "?" : (String) param.getLowerBound()));
		out.append(" AND " + param.getField());
		if (param.isInclusiveUpper()) {
			out.append(" " + ParamOperatorSimple.LEQ.getOperator() + " ");
		} else {
			out.append(" " + ParamOperatorSimple.LT.getOperator() + " ");
		}
		out.append((binded ? "?" : (String) param.getUpperBound()));
		out.append(")");
	}

}
