package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionImplementationException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionNotFoundException;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions.Function;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;

/**
 * 
 * ConditionClauseRangeFx warunek zaytania obsługujący sprawdzenie, czy zadana
 * kolumna posiada wartości zawarte pomiędzy górną i dolną granicą, z tym że
 * obydwa argumenty granic zdefiniowane są jako funkcje SQL ({@link Function}).
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * </pre>
 * 
 * </p>
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;AND NOT(chi.nameDay &gt;= TO_DATE(?,'YYYY-MM-DD HH24:MI') AND chi.nameDay &lt; TO_DATE(?,'YYYY-MM-DD HH24:MI'))&quot;;
 * String expectedLowerParam = &quot;2010-10-13 12:00&quot;;
 * String expectedUpperParam = &quot;2010-10-13 12:00&quot;;
 * tested = new ConditionClauseRangeFx(ConditionOperator.ANDNOT,
 * 		table.getColumn(&quot;nameDay&quot;), Function.ToDate, new String[] { &quot;?&quot;,
 * 				&quot;'YYYY-MM-DD HH24:MI'&quot; }, expectedLowerParam, true,
 * 		Function.ToDate, new String[] { &quot;?&quot;, &quot;'YYYY-MM-DD HH24:MI'&quot; },
 * 		expectedUpperParam, false, true)
 * 
 * ;
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.3 $
 * 
 */
public class ConditionClauseRangeFx extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Domyślny dialekt dla funkcji
	 */
	public static final Dialect DEFAULT_DIALECT = Dialect.ORACLE;

	/** dialekt serwera baz danych używany podczas tworzenia klauzuli funkcji */
	private Dialect serverDialect;

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param lowerFunction
	 *            fukcja definiująca wartość dolnej granicy ({@link Function})
	 * @param lowerArguments
	 *            argumenty fukcji definiującej wartość dolnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param lowerBindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            dolnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperFunction
	 *            fukcja definiująca wartość górnej granicy ({@link Function})
	 * @param upperArguments
	 *            argumenty fukcji definiującej wartość górnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param upperBindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            górnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseRangeFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function lowerFunction,
			String[] lowerArguments, Object lowerBindArgumnet,
			boolean inclusiveLower, Function upperFunction,
			String[] upperArguments, Object upperBindArgumnet,
			boolean inclusiveUpper, Boolean isBinded) {
		super(column);
		init(conditionOperator, lowerFunction, lowerArguments,
				new Object[] { lowerBindArgumnet }, inclusiveLower,
				upperFunction, upperArguments,
				new Object[] { upperBindArgumnet }, inclusiveUpper, isBinded,
				null);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param lowerFunction
	 *            fukcja definiująca wartość dolnej granicy ({@link Function})
	 * @param lowerArguments
	 *            argumenty fukcji definiującej wartość dolnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param lowerBindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            dolnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperFunction
	 *            fukcja definiująca wartość górnej granicy ({@link Function})
	 * @param upperArguments
	 *            argumenty fukcji definiującej wartość górnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param upperBindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            górnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseRangeFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function lowerFunction,
			String[] lowerArguments, Object lowerBindArgumnet,
			boolean inclusiveLower, Function upperFunction,
			String[] upperArguments, Object upperBindArgumnet,
			boolean inclusiveUpper, Boolean isBinded, Dialect serverDialect) {
		super(column);
		init(conditionOperator, lowerFunction, lowerArguments,
				new Object[] { lowerBindArgumnet }, inclusiveLower,
				upperFunction, upperArguments,
				new Object[] { upperBindArgumnet }, inclusiveUpper, isBinded,
				serverDialect);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param lowerFunction
	 *            fukcja definiująca wartość dolnej granicy ({@link Function})
	 * @param lowerArguments
	 *            argumenty fukcji definiującej wartość dolnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param lowerBindArgumnets
	 *            dla fukcji definiującej wartość dolnej granicy. Jeżeli w
	 *            funkcji zostaną użyte zmienne bindowane, to to jest lista
	 *            wartości, które do nich zostaną przypisane
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperFunction
	 *            fukcja definiująca wartość górnej granicy ({@link Function})
	 * @param upperArguments
	 *            argumenty fukcji definiującej wartość górnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param upperBindArgumnets
	 *            dla fukcji definiującej wartość górnej granicy. Jeżeli w
	 *            funkcji zostaną użyte zmienne bindowane, to to jest lista
	 *            wartości, które do nich zostaną przypisane
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseRangeFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function lowerFunction,
			String[] lowerArguments, Object[] lowerBindArgumnets,
			boolean inclusiveLower, Function upperFunction,
			String[] upperArguments, Object[] upperBindArgumnets,
			boolean inclusiveUpper, Boolean isBinded) {
		super(column);
		init(conditionOperator, lowerFunction, lowerArguments,
				lowerBindArgumnets, inclusiveLower, upperFunction,
				upperArguments, upperBindArgumnets, inclusiveUpper, isBinded,
				null);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param lowerFunction
	 *            fukcja definiująca wartość dolnej granicy ({@link Function})
	 * @param lowerArguments
	 *            argumenty fukcji definiującej wartość dolnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param lowerBindArgumnets
	 *            dla fukcji definiującej wartość dolnej granicy. Jeżeli w
	 *            funkcji zostaną użyte zmienne bindowane, to to jest lista
	 *            wartości, które do nich zostaną przypisane
	 * @param inclusiveLower
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości dolnej granicy warunku
	 * @param upperFunction
	 *            fukcja definiująca wartość górnej granicy ({@link Function})
	 * @param upperArguments
	 *            argumenty fukcji definiującej wartość górnej granicy (
	 *            {@link _IFunction#getArguments()})
	 * @param upperBindArgumnets
	 *            dla fukcji definiującej wartość górnej granicy. Jeżeli w
	 *            funkcji zostaną użyte zmienne bindowane, to to jest lista
	 *            wartości, które do nich zostaną przypisane
	 * @param inclusiveUpper
	 *            flaga, mówiąca o tym czy wartości spełniające warunek mają
	 *            również zawierać się w wartości górnej granicy warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseRangeFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function lowerFunction,
			String[] lowerArguments, Object[] lowerBindArgumnets,
			boolean inclusiveLower, Function upperFunction,
			String[] upperArguments, Object[] upperBindArgumnets,
			boolean inclusiveUpper, Boolean isBinded, Dialect serverDialect) {
		super(column);
		init(conditionOperator, lowerFunction, lowerArguments,
				lowerBindArgumnets, inclusiveLower, upperFunction,
				upperArguments, upperBindArgumnets, inclusiveUpper, isBinded,
				serverDialect);
	}

	private void init(ConditionOperator conditionOperator,
			Function lowerFunction, String[] lowerArguments,
			Object[] lowerBindArgumnets, boolean inclusiveLower,
			Function upperFunction, String[] upperArguments,
			Object[] upperBindArgumnets, boolean inclusiveUpper,
			Boolean isBinded, Dialect serverDialect) {

		this.setFieldType(String.class);
		if (serverDialect == null)
			this.serverDialect = DEFAULT_DIALECT;
		else
			this.serverDialect = serverDialect;

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.RANGEfx);
		if (lowerArguments == null) {
			throw new IllegalArgumentException(
					"Lista argumentow dla funkcji SQL zdefiniowanej jako dolna granica, "
							+ "nie moze byc pusta");
		}
		if (upperArguments == null) {
			throw new IllegalArgumentException(
					"Lista argumentow dla funkcji SQL zdefiniowanej jako gorna granica, "
							+ "nie moze byc pusta");
		}
		if (lowerFunction == null || upperFunction == null)
			throw new IllegalArgumentException(
					"Obydwie granice przedzialu nie moga byc puste. "
							+ "Jezeli chcesz ograniczyc wynik tylko z jednej "
							+ "strony uzyj 'pl.slawas.jdbc.where.ConditionClauseSimple'.");

		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		_IFunction lowerSqlFunction;
		try {
			lowerSqlFunction = lowerFunction.get(this.serverDialect,
					new String(), lowerArguments);
		} catch (SQLFunctionNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (SQLFunctionImplementationException e) {
			throw new IllegalArgumentException(e);
		}
		this.getParam().setLowerBound(lowerBindArgumnets);
		this.getParam().setLowerFunction(lowerSqlFunction);
		this.getParam().setInclusiveLower(inclusiveLower);

		_IFunction upperSqlFunction;
		try {
			upperSqlFunction = upperFunction.get(this.serverDialect,
					new String(), upperArguments);
		} catch (SQLFunctionNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (SQLFunctionImplementationException e) {
			throw new IllegalArgumentException(e);
		}
		this.getParam().setUpperBound(lowerBindArgumnets);
		this.getParam().setUpperFunction(upperSqlFunction);
		this.getParam().setInclusiveUpper(inclusiveUpper);

		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();

		if (this.getParam().isInclusiveLower()
				&& this.getParam().isInclusiveUpper())
			this.setParamOperator(ParamOperatorRange.BETWEEN);
		else
			this.setParamOperator(ParamOperatorRange.SIMPLE);

	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		ConditionParam param = getParam();
		out.append(conditionOperator);
		switch ((ParamOperatorRange) param.getOperator()) {
		case BETWEEN:
			out.append(" " + param.getField());
			out.append(" " + operator);
			out.append(" " + param.getLowerFunction().getClause() + " AND "
					+ param.getUpperFunction().getClause());
			break;
		default:
			out.append(" " + operator);
			out.append("(" + param.getField());
			if (param.isInclusiveLower()) {
				out.append(" " + ParamOperatorSimple.GEQ.getOperator() + " ");
			} else {
				out.append(" " + ParamOperatorSimple.GT.getOperator() + " ");
			}
			out.append(param.getLowerFunction().getClause());
			out.append(" AND " + param.getField());
			if (param.isInclusiveUpper()) {
				out.append(" " + ParamOperatorSimple.LEQ.getOperator() + " ");
			} else {
				out.append(" " + ParamOperatorSimple.LT.getOperator() + " ");
			}
			out.append(param.getUpperFunction().getClause());
			out.append(")");
			break;
		}

		return out.toString();
	}

}
