package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;

/**
 * 
 * ConditionClauseSimple warunek prosty zapytania (dla operatorów logicznych
 * dwuargumentowych {@link ParamOperatorSimple})
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, };
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;AND nameDay = ?&quot;;
 * Calendar expectedParam = Calendar.getInstance();
 * tested = new ConditionClauseSimple(ConditionOperator.AND,
 * 		table.getColumn(&quot;nameDay&quot;), expectedParam, true);
 * logger.debug(tested.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested.getClause());
 * assertEquals(&quot;Blad wartosci pola dla klauzuli warunku&quot;, expectedParam, tested
 * 		.getParam().getValue());
 * </pre>
 * 
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.4 $
 * 
 */
public class ConditionClauseSimple extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/** domyślny operator logiczny porównania wartości */
	public static final ParamOperatorSimple DEFAULT_SIMPLE_OPERATOR = ParamOperatorSimple.EQ;

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link ConditionParam#getOperator()} na wartość
	 * {@link #DEFAULT_SIMPLE_OPERATOR}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 */
	public ConditionClauseSimple(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object value, Boolean isBinded) {
		super(column);
		init(conditionOperator, value, isBinded, null);
	}

	/**
	 * Konstruktor warunku zapytania
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param operator
	 *            operator logiczny dla porównania wartości
	 *            {@link ParamOperatorSimple}
	 */
	public ConditionClauseSimple(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Object value, Boolean isBinded,
			ParamOperatorSimple operator) {
		super(column);
		init(conditionOperator, value, isBinded, operator);
	}

	/**
	 * Konstruktor warunku zapytania
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param fieldName
	 *            nazwa pola (<font color="#dd0000"> trzeba znac nazwe pola
	 *            wystepującą w zapytaniu:
	 *            {@code &lt;ewentualny_alias_tabeli>.&lt;nazwa_kolumny_w_tabeli>}
	 *            )</font>
	 * @param fieldType
	 *            klasa typu pola
	 * @param value
	 *            wartość warunku
	 * @param isBinded
	 *            czy wartość ma być bindowana, czy też użyta jako literał,
	 *            jeżeli podamy wartość {@code null} to zostanie to uznane jako
	 *            informacja, że ma być użyta flaga domyślna, która jest
	 *            ustawiona na {@code true} ({@link ConditionParam}.
	 * @param operator
	 *            operator logiczny dla porównania wartości
	 *            {@link ParamOperatorSimple}
	 */
	public ConditionClauseSimple(ConditionOperator conditionOperator,
			String fieldName, Class<?> fieldType, Object value,
			Boolean isBinded, ParamOperatorSimple operator) {
		super(fieldName, fieldType);
		init(conditionOperator, value, isBinded, operator);
	}

	private void init(ConditionOperator conditionOperator, Object value,
			Boolean isBinded, ParamOperatorSimple operator) {

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.SIMPLE);
		// if (value == null)
		// throw new IllegalArgumentException("Wartosc nie moze byc pusta");
		this.setConditionValue(value);
		if (isBinded != null && !isBinded)
			this.setValueAsLiteral();
		if (operator == null)
			operator = DEFAULT_SIMPLE_OPERATOR;
		this.setParamOperator(operator);

	}

}
