package pl.slawas.scjdbc4jpa.jdbc.where;

import pl.slawas.scjdbc4jpa.dialect.Dialect;
import pl.slawas.scjdbc4jpa.jdbc.beans.JDBCColumnDesc;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionImplementationException;
import pl.slawas.scjdbc4jpa.jdbc.exceptions.SQLFunctionNotFoundException;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions.Function;
import pl.slawas.scjdbc4jpa.jdbc.sql.functions._IFunction;

/**
 * 
 * ConditionClauseSimpleFx warunek prosty zapytania (dla operatorów logicznych
 * dwuargumentowych {@link ParamOperatorSimple}), z tym że argument wartości
 * zdefiniowany jest jako funkcje SQL ({@link Function}).
 * <p>
 * <font color="#aa0000">Uwaga!</font><br>
 * Jeżeli chcemy używac literałów dla łańcuchów znakowych ({@link String}) to
 * trzeba je wprowadzać jako argumenty zawarte pomiędzy pojedynczymi apostrofami
 * np.
 * 
 * <pre>
 * String[] expectedParams = new String[] { &quot;'slawas'&quot;, &quot;'karol'&quot;, &quot;'tomek'&quot;,
 * 		&quot;'ania'&quot; };
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * Przykład:
 * </p>
 * 
 * <pre>
 * expectedCondition = &quot;AND chi.nameDay = TO_DATE(?,'YYYY-MM-DD HH24:MI')&quot;;
 * String expectedParam = &quot;2010-10-13 12:00&quot;;
 * ConditionClauseSimpleFx tested1 = new ConditionClauseSimpleFx(
 * 		ConditionOperator.AND, table.getColumn(&quot;nameDay&quot;), Function.ToDate,
 * 		new String[] { &quot;?&quot;, &quot;'YYYY-MM-DD HH24:MI'&quot; },
 * 		new Object[] { expectedParam });
 * logger.debug(tested1.getClause());
 * assertEquals(&quot;Blad wygenerowania klauzuli warunku&quot;, expectedCondition,
 * 		tested1.getClause());
 * assertEquals(&quot;Blad wartosci pola dla klauzuli warunku&quot;, expectedParam,
 * 		((Object[]) tested1.getParam().getValue())[0]);
 * </pre>
 * 
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version $Revision: 1.2 $
 * 
 */
public class ConditionClauseSimpleFx extends ConditionClauseSupport {

	private static final long serialVersionUID = 6801361643123612029L;

	/**
	 * Domyślny dialekt dla funkcji
	 */
	public static final Dialect DEFAULT_DIALECT = Dialect.ORACLE;

	/** domyślny operator logiczny porównania wartości */
	public static final ParamOperatorSimple DEFAULT_SIMPLE_OPERATOR = ParamOperatorSimple.EQ;

	/** dialekt serwera baz danych używany podczas tworzenia klauzuli funkcji */
	private Dialect serverDialect;

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link ConditionParam#getOperator()} na wartość
	 * {@link #DEFAULT_SIMPLE_OPERATOR}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            dolnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object bindArgumnet, Dialect serverDialect) {
		super(column);
		init(serverDialect, conditionOperator, function, arguments,
				new Object[] { bindArgumnet }, null);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnet
	 *            pojedyncza warość argumentu dla fukcji definiującej wartość
	 *            dolnej granicy. Jeżeli w funkcji zostaną użyte zmienna
	 *            bindowana, to ta wartość zostanie przypisana
	 * @param operator
	 *            operator logiczny dla porównania wartości
	 *            {@link ParamOperatorSimple}
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object bindArgumnet, ParamOperatorSimple operator,
			Dialect serverDialect) {
		super(column);
		init(serverDialect, conditionOperator, function, arguments,
				new Object[] { bindArgumnet }, operator);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link ConditionParam#getOperator()} na wartość
	 * {@link #DEFAULT_SIMPLE_OPERATOR}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnets
	 *            dla fukcji definiującej wartość. Jeżeli w funkcji zostaną
	 *            użyte zmienne bindowane, to to jest lista wartości, które do
	 *            nich zostaną przypisane
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object[] bindArgumnets, Dialect serverDialect) {
		super(column);
		init(serverDialect, conditionOperator, function, arguments,
				bindArgumnets, null);
	}

	/**
	 * Konstruktor warunku zapytania.
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnets
	 *            dla fukcji definiującej wartość. Jeżeli w funkcji zostaną
	 *            użyte zmienne bindowane, to to jest lista wartości, które do
	 *            nich zostaną przypisane
	 * @param operator
	 *            operator logiczny dla porównania wartości
	 *            {@link ParamOperatorSimple}
	 * @param serverDialect
	 *            dialekt serwera baz danych używany podczas tworzenia klauzuli
	 *            funkcji {@link Dialect}
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object[] bindArgumnets, ParamOperatorSimple operator,
			Dialect serverDialect) {
		super(column);
		init(serverDialect, conditionOperator, function, arguments,
				bindArgumnets, operator);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link ConditionParam#getOperator()} na wartość
	 * {@link #DEFAULT_SIMPLE_OPERATOR}</li>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnets
	 *            dla fukcji definiującej wartość. Jeżeli w funkcji zostaną
	 *            użyte zmienne bindowane, to to jest lista wartości, które do
	 *            nich
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object[] bindArgumnets) {
		super(column);
		init(null, conditionOperator, function, arguments, bindArgumnets, null);
	}

	/**
	 * Konstruktor warunku zapytania. Przy użyciu tego konstruktora domyślnie
	 * zostaną ustawione parametry:
	 * <ul>
	 * <li>{@link #serverDialect} na wartość {@link #DEFAULT_DIALECT}</li>
	 * </ul>
	 * 
	 * @param conditionOperator
	 *            oprator warunku (jedna z wartości {@link ConditionOperator})
	 * @param column
	 *            opis kolumny, której dotyczy warunek
	 * @param function
	 *            fukcja definiująca wartość ({@link Function})
	 * @param arguments
	 *            argumenty fukcji definiującej wartość (
	 *            {@link _IFunction#getArguments()})
	 * @param bindArgumnets
	 *            dla fukcji definiującej wartość. Jeżeli w funkcji zostaną
	 *            użyte zmienne bindowane, to to jest lista wartości, które do
	 *            nich
	 * @param operator
	 *            operator logiczny dla porównania wartości
	 *            {@link ParamOperatorSimple}
	 */
	public ConditionClauseSimpleFx(ConditionOperator conditionOperator,
			JDBCColumnDesc column, Function function, String[] arguments,
			Object[] bindArgumnets, ParamOperatorSimple operator) {
		super(column);
		init(null, conditionOperator, function, arguments, bindArgumnets,
				operator);
	}

	private void init(Dialect serverDialect,
			ConditionOperator conditionOperator, Function function,
			String[] arguments, Object[] bindArgumnets,
			ParamOperatorSimple operator) {

		this.setFieldType(String.class);
		if (serverDialect == null)
			this.serverDialect = DEFAULT_DIALECT;
		else
			this.serverDialect = serverDialect;

		this.setConditionOperator(conditionOperator);
		this.setConditionType(ConditionType.SIMPLEfx);
		if (arguments == null) {
			throw new IllegalArgumentException(
					"Lista argumentow funkcji SQL nie moze byc pusta");
		}
		this.setConditionValue(arguments);

		if (function == null) {
			throw new IllegalArgumentException(
					"Klauzula funkcji SQL nie moze byc pusta");
		}
		_IFunction sqlFunction;
		try {
			sqlFunction = function.get(this.serverDialect, new String(),
					arguments);
		} catch (SQLFunctionNotFoundException e) {
			throw new IllegalArgumentException(e);
		} catch (SQLFunctionImplementationException e) {
			throw new IllegalArgumentException(e);
		}

		this.getParam().setFunction(sqlFunction);
		if (bindArgumnets == null || bindArgumnets.length == 0) {
			this.setValueAsLiteral();
		} else {
			this.setConditionValue(bindArgumnets);
		}

		if (operator == null)
			operator = DEFAULT_SIMPLE_OPERATOR;
		this.setParamOperator(operator);

	}

	/**
	 * @return the serverDialect
	 */
	public Dialect getServerDialect() {
		return serverDialect;
	}

	@Override
	public String getClause() {
		/* ustalenie operatorów */
		Operators operators = this.getOperators();
		String operator = operators.getOperator();
		ConditionOperator conditionOperator = operators.getConditionOperator();

		/* tworzenie klauzuli */
		StringBuffer out = new StringBuffer();
		ConditionParam param = getParam();
		out.append(conditionOperator);
		out.append(" " + param.getField());
		out.append(" " + operator);
		out.append(" " + param.getFunction().getClause());
		return out.toString();
	}
}
