package pl.slawas.scjdbc4jpa.jpa;

import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Set;

import javax.persistence.EntityTransaction;
import javax.persistence.TransactionRequiredException;
import javax.persistence.spi.PersistenceUnitInfo;
import javax.persistence.spi.PersistenceUnitTransactionType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.support.PersistenceExceptionTranslator;
import org.springframework.orm.jpa.EntityManagerFactoryInfo;
import org.springframework.orm.jpa.EntityManagerFactoryUtils;
import org.springframework.orm.jpa.EntityManagerPlusOperations;
import org.springframework.orm.jpa.EntityManagerProxy;
import org.springframework.orm.jpa.ExtendedEntityManagerCreator;
import org.springframework.orm.jpa.JpaDialect;
import org.springframework.transaction.support.TransactionSynchronizationAdapter;
import org.springframework.transaction.support.TransactionSynchronizationManager;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;

import pl.slawas.scjdbc4jpa.core.ScEntityManager;
import pl.slawas.scjdbc4jpa.core.ScEntityManagerHolder;

public class JpaExtendedEntityManagerUtils {

	public static ScEntityManager createApplicationManagedEntityManager(
			ScEntityManager rawEntityManager, EntityManagerFactoryInfo emfInfo) {

		return createProxy(rawEntityManager, emfInfo, false);
	}

	public static ScEntityManager createContainerManagedEntityManager(
			ScEntityManager rawEntityManager, EntityManagerFactoryInfo emfInfo) {

		return createProxy(rawEntityManager, emfInfo, true);
	}

	private static ScEntityManager createProxy(
			ScEntityManager rawEntityManager, EntityManagerFactoryInfo emfInfo,
			boolean containerManaged) {

		Assert.notNull(emfInfo, "EntityManagerFactoryInfo must not be null");
		JpaDialect jpaDialect = emfInfo.getJpaDialect();
		EntityManagerPlusOperations plusOperations = null;
		if (jpaDialect != null
				&& jpaDialect.supportsEntityManagerPlusOperations()) {
			plusOperations = jpaDialect
					.getEntityManagerPlusOperations(rawEntityManager);
		}
		PersistenceUnitInfo pui = emfInfo.getPersistenceUnitInfo();
		Boolean jta = (pui != null ? pui.getTransactionType() == PersistenceUnitTransactionType.JTA
				: null);

		return createProxy(rawEntityManager, plusOperations, jpaDialect, jta,
				false);
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	private static ScEntityManager createProxy(
			ScEntityManager rawEntityManager,
			EntityManagerPlusOperations plusOperations,
			PersistenceExceptionTranslator exceptionTranslator, Boolean jta,
			boolean containerManaged) {

		Assert.notNull(rawEntityManager, "EntityManager must not be null");
		ClassLoader cl = ExtendedEntityManagerCreator.class.getClassLoader();
		Set ifcs = ClassUtils.getAllInterfacesForClassAsSet(
				rawEntityManager.getClass(), cl);
		ifcs.add(EntityManagerProxy.class);
		if (plusOperations != null) {
			ifcs.add(EntityManagerPlusOperations.class);
		}
		return (ScEntityManager) Proxy.newProxyInstance(cl, (Class[]) ifcs
				.toArray(new Class[ifcs.size()]),
				new ExtendedEntityManagerInvocationHandler(rawEntityManager,
						plusOperations, exceptionTranslator, jta,
						containerManaged));
	}

	/**
	 * InvocationHandler for extended EntityManagers as defined in the JPA spec.
	 */
	private static class ExtendedEntityManagerInvocationHandler implements
			InvocationHandler, Serializable {

		private static final long serialVersionUID = -8804046924068110880L;

		private static final Log logger = LogFactory
				.getLog(ExtendedEntityManagerInvocationHandler.class);

		private final ScEntityManager target;

		private final EntityManagerPlusOperations plusOperations;

		private final PersistenceExceptionTranslator exceptionTranslator;

		private boolean jta;

		private final boolean containerManaged;

		private ExtendedEntityManagerInvocationHandler(ScEntityManager target,
				EntityManagerPlusOperations plusOperations,
				PersistenceExceptionTranslator exceptionTranslator,
				Boolean jta, boolean containerManaged) {

			this.target = target;
			this.plusOperations = plusOperations;
			this.exceptionTranslator = exceptionTranslator;
			this.jta = (jta != null ? jta.booleanValue() : isJtaEntityManager());
			this.containerManaged = containerManaged;

		}

		private boolean isJtaEntityManager() {
			try {
				this.target.getTransaction();
				return false;
			} catch (IllegalStateException ex) {
				logger.debug("Cannot access EntityTransaction handle - assuming we're in a JTA environment");
				return true;
			}
		}

		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
			// Invocation on EntityManager interface coming in...

			if (method.getName().equals("equals")) {
				// Only consider equal when proxies are identical.
				return (proxy == args[0]);
			} else if (method.getName().equals("hashCode")) {
				// Use hashCode of EntityManager proxy.
				return hashCode();
			} else if (method.getName().equals("getTargetEntityManager")) {
				return this.target;
			} else if (method.getName().equals("isOpen")) {
				if (this.containerManaged) {
					return true;
				}
			} else if (method.getName().equals("close")) {
				if (this.containerManaged) {
					throw new IllegalStateException(
							"Invalid usage: Cannot close a container-managed EntityManager");
				}
			} else if (method.getName().equals("getTransaction")) {
				if (this.containerManaged) {
					throw new IllegalStateException(
							"Cannot execute getTransaction() on a container-managed EntityManager");
				}
			} else if (method.getName().equals("joinTransaction")) {
				doJoinTransaction(true);
				return null;
			}

			// Do automatic joining if required.
			if (this.containerManaged
					&& method.getDeclaringClass().isInterface()) {
				doJoinTransaction(false);
			}

			try {
				if (method.getDeclaringClass().equals(
						EntityManagerPlusOperations.class)) {
					return method.invoke(this.plusOperations, args);
				} else {
					return method.invoke(this.target, args);
				}
			} catch (InvocationTargetException ex) {
				throw ex.getTargetException();
			}

		}

		/**
		 * Join an existing transaction, if not already joined.
		 * 
		 * @param enforce
		 *            whether to enforce the transaction (i.e. whether failure
		 *            to join is considered fatal)
		 */
		private void doJoinTransaction(boolean enforce) {
			if (this.jta) {
				// Let's try whether we're in a JTA transaction.
				try {
					this.target.joinTransaction();
					logger.debug("Joined JTA transaction");
				} catch (TransactionRequiredException ex) {
					if (!enforce) {
						logger.debug("No JTA transaction to join: " + ex);
					} else {
						throw ex;
					}
				}
			} else {
				if (TransactionSynchronizationManager.isSynchronizationActive()) {
					if (!TransactionSynchronizationManager
							.hasResource(this.target)
							&& !this.target.getTransaction().isActive()) {
						enlistInCurrentTransaction();
					}
					logger.debug("Joined local transaction");
				} else {
					if (!enforce) {
						logger.debug("No local transaction to join");
					} else {
						throw new TransactionRequiredException(
								"No local transaction to join");
					}
				}
			}
		}

		/**
		 * Enlist this application-managed EntityManager in the current
		 * transaction.
		 */
		private void enlistInCurrentTransaction() {
			// Resource local transaction, need to acquire the
			// EntityTransaction,
			// start a transaction now and enlist a synchronization for
			// commit or rollback later.
			EntityTransaction et = this.target.getTransaction();
			et.begin();
			if (logger.isDebugEnabled()) {
				logger.debug("Starting resource local transaction on application-managed "
						+ "EntityManager [" + this.target + "]");
			}
			ScEntityManagerHolder aemh = new ScEntityManagerHolder(this.target);
			ExtendedEntityManagerSynchronization extendedEntityManagerSynchronization = new ExtendedEntityManagerSynchronization(
					aemh, this.exceptionTranslator);
			TransactionSynchronizationManager.bindResource(this.target,
					extendedEntityManagerSynchronization);
			// TODO sprawdzic do czego to sluzy
			TransactionSynchronizationManager
					.registerSynchronization(extendedEntityManagerSynchronization);
		}
	}

	/**
	 * TransactionSynchronization enlisting an extended EntityManager with a
	 * current Spring transaction.
	 */
	private static class ExtendedEntityManagerSynchronization extends
			TransactionSynchronizationAdapter {

		private final ScEntityManagerHolder ScEntityManagerHolder;

		private final PersistenceExceptionTranslator exceptionTranslator;

		private boolean holderActive = true;

		public ExtendedEntityManagerSynchronization(
				ScEntityManagerHolder emHolder,
				PersistenceExceptionTranslator exceptionTranslator) {

			this.ScEntityManagerHolder = emHolder;
			this.exceptionTranslator = exceptionTranslator;

		}

		public int getOrder() {
			return EntityManagerFactoryUtils.ENTITY_MANAGER_SYNCHRONIZATION_ORDER + 1;
		}

		public void suspend() {
			if (this.holderActive) {
				TransactionSynchronizationManager
						.unbindResource(this.ScEntityManagerHolder
								.getScEntityManager());
			}
		}

		public void resume() {
			if (this.holderActive) {
				TransactionSynchronizationManager.bindResource(
						this.ScEntityManagerHolder.getScEntityManager(),
						this.ScEntityManagerHolder);
			}
		}

		public void beforeCompletion() {
			TransactionSynchronizationManager
					.unbindResource(this.ScEntityManagerHolder
							.getScEntityManager());
			this.holderActive = false;
		}

		public void afterCommit() {
			// Trigger commit here to let exceptions propagate to the caller.
			try {
				this.ScEntityManagerHolder.getScEntityManager()
						.getTransaction().commit();
			} catch (RuntimeException ex) {
				throw convertCompletionException(ex);
			}
		}

		public void afterCompletion(int status) {
			this.ScEntityManagerHolder.setSynchronizedWithTransaction(false);
			if (status != STATUS_COMMITTED) {
				// Haven't had an afterCommit call: trigger a rollback.
				try {
					this.ScEntityManagerHolder.getScEntityManager()
							.getTransaction().rollback();
				} catch (RuntimeException ex) {
					throw convertCompletionException(ex);
				}
			}
			// Don't close the EntityManager... That's up to the user.
		}

		private RuntimeException convertCompletionException(RuntimeException ex) {
			DataAccessException daex = (this.exceptionTranslator != null) ? this.exceptionTranslator
					.translateExceptionIfPossible(ex)
					: EntityManagerFactoryUtils
							.convertJpaAccessExceptionIfPossible(ex);
			return (daex != null ? daex : ex);
		}
	}

}
