package pl.slawas.scjdbc4jpa.jpa.dao;

import java.io.Serializable;
import java.lang.reflect.ParameterizedType;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.Query;

import org.apache.log4j.Logger;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.beans.factory.config.BeanPostProcessor;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.dao.support.DaoSupport;

import pl.slawas.scjdbc4jpa.core.ScDefaults;
import pl.slawas.scjdbc4jpa.core.ScEntityManager;
import pl.slawas.scjdbc4jpa.core.ScObjectFactory;
import pl.slawas.scjdbc4jpa.jpa.exceptions.JpaDaoBeanInitializationException;
import pl.slawas.scjdbc4jpa.jpa.mapping.RowValues;
import pl.slawas.scjdbc4jpa.jpa.mapping.Table;

/**
 * @author Sławomir Cichy &lt;slawas@slawas.pl&gt;
 * @version 0.0.1
 * 
 * @param <Obj>
 *            klasa encji, której dotyczy DAO.
 */
@SuppressWarnings("serial")
public abstract class JpaDAOSupport<Obj> extends DaoSupport implements
		ScDefaults, _IJpaDAO<Obj>, ApplicationContextAware, BeanPostProcessor,
		Serializable {

	final protected Logger log = Logger.getLogger(getClass());

	private boolean inicjalized = false;

	protected ScEntityManager aem;

	private ScObjectFactory aof;

	protected Table tableInfo = null;

	private Class<Obj> persistentClass;

	private ApplicationContext actx;

	private void endBulkTransaction(boolean autoCommit) {
		if (autoCommit && this.aem.getTransaction().isActive())
			this.aem.getTransaction().commit();
	}

	private void flushAndCommitAnalyze(int i, boolean autoCommit) {
		if ((i != 0) && (i % aem.getBatchSize() == 0)) {
			// log.debug("[Transaction] flush");
			this.aem.flush();
			this.aem.clear();
		}
		if (autoCommit && (i != 0) && (i % aem.getBatchCommitSize() == 0)) {
			// log.debug("[Transaction] commit");
			this.aem.getTransaction().commit();
			this.aem.getTransaction().begin();
		}
	}

	public void joinTransaction() {
		if (!this.aem.getTransaction().isActive()) {
			if (logger.isDebugEnabled())
				log.debug("\n\n Join to new transaction!!!\n\n");
			this.aem.getTransaction().begin();
		}
	}

	protected boolean isAllowedTableInfo() {
		return (this.tableInfo != null);
	}

	@SuppressWarnings("unchecked")
	public JpaDAOSupport() {
		super();
		this.persistentClass = (Class<Obj>) ((ParameterizedType) this
				.getClass().getGenericSuperclass()).getActualTypeArguments()[0];
	}

	public Class<Obj> getPersistentClass() {
		return persistentClass;
	}

	public Obj find(Object id) {
		this.aem.setTable(this.tableInfo);
		return (Obj) this.aem.find(persistentClass, id);
	}

	@SuppressWarnings("unchecked")
	public List<Obj> findAll() throws SQLException {
		this.aem.setTable(this.tableInfo);
		return this.aem.createQuery(
				"select r from " + persistentClass.getSimpleName() + " r")
				.getResultList();
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	public List<Obj> find(String namedQuery, Map sqlParams) throws SQLException {
		this.aem.setTable(this.tableInfo);
		Query q = this.aem.createNamedQuery(namedQuery);
		if (sqlParams != null) {
			Set params = sqlParams.entrySet();
			Iterator it = params.iterator();
			while (it.hasNext()) {
				Map.Entry entry = (Map.Entry) it.next();
				String key = (String) entry.getKey();
				String value = (String) entry.getValue();
				q.setParameter(key, value);
			}
		}
		return q.getResultList();
	}

	@SuppressWarnings("unchecked")
	public List<Obj> findByNativeQuery(String nativeQuery) throws SQLException {
		this.aem.setTable(this.tableInfo);
		return this.aem.createNativeQuery(nativeQuery).getResultList();
	}

	public boolean createFullTextIndex() throws SQLException {
		this.aem.setTable(this.tableInfo);
		// TODO zrealizoawac metode createFullTextIndex()
		return true;
	}

	public List<Obj> search(String serachText, String[] indexedFieldList)
			throws SQLException {
		this.aem.setTable(this.tableInfo);
		// TODO zrealizowac metode search(String serachText, String[]
		// indexedFieldList)
		return null;
	}

	public void insert(Obj entityObject) throws SQLException {
		this.aem.setTable(this.tableInfo);
		entityObject = this.beforeInsert(entityObject);
		this.aem.persist(entityObject);
		this.afterInsert(entityObject);
		if (logger.isDebugEnabled())
			logger.debug("" + entityObject.getClass().getSimpleName()
					+ ": inserted.");
	}

	public int insertList(List<Obj> entityObjects, boolean autoCommit)
			throws SQLException {
		this.aem.setTable(this.tableInfo);
		int i = 0;
		this.joinTransaction();
		for (Obj entityObject : entityObjects) {
			flushAndCommitAnalyze(i, autoCommit);
			entityObject = this.beforeInsert(entityObject);
			this.aem.persist(entityObject);
			this.afterInsert(entityObject);
			i++;
		}
		endBulkTransaction(autoCommit);
		if (i < 0)
			i = 0;
		if (logger.isDebugEnabled())
			logger.debug("" + persistentClass.getSimpleName() + ": "
					+ Integer.toString(i) + " rows inserted.");
		return i;
	}

	public void update(Obj entityObject) throws SQLException {
		this.aem.setTable(this.tableInfo);
		entityObject = this.beforeUpdate(entityObject);
		this.aem.merge(entityObject);
		this.afterUpdate(entityObject);
		if (logger.isDebugEnabled())
			logger.debug("" + entityObject.getClass().getSimpleName()
					+ ": changed.");

	}

	public int updateList(List<Obj> entityObjects, boolean autoCommit)
			throws SQLException {
		this.aem.setTable(this.tableInfo);
		int i = 0;
		this.joinTransaction();
		for (Obj entityObject : entityObjects) {
			flushAndCommitAnalyze(i, autoCommit);
			entityObject = this.beforeUpdate(entityObject);
			this.aem.merge(entityObject);
			this.afterUpdate(entityObject);
			i++;
		}
		endBulkTransaction(autoCommit);
		if (i < 0)
			i = 0;
		if (logger.isDebugEnabled())
			logger.debug("" + persistentClass.getSimpleName() + ": "
					+ Integer.toString(i) + " rows updated.");
		return i;
	}

	public void remove(Obj entityObject) throws SQLException {
		this.aem.setTable(this.tableInfo);
		entityObject = this.beforeRemove(entityObject);
		this.aem.remove(entityObject);
		this.afterRemove(entityObject);
		if (logger.isDebugEnabled())
			logger.debug("" + entityObject.getClass().getSimpleName()
					+ ": deleted.");
	}

	public int removeList(List<Obj> entityObjects, boolean autoCommit)
			throws SQLException {
		this.aem.setTable(this.tableInfo);
		int i = 0;
		this.joinTransaction();
		for (Obj entityObject : entityObjects) {
			flushAndCommitAnalyze(i, autoCommit);
			entityObject = this.beforeRemove(entityObject);
			this.aem.remove(entityObject);
			this.afterRemove(entityObject);
			i++;
		}
		endBulkTransaction(autoCommit);
		if (i < 0)
			i = 0;
		if (logger.isDebugEnabled())
			logger.debug("" + persistentClass.getSimpleName() + ": "
					+ Integer.toString(i) + " rows deleted.");
		return i;
	}

	@Override
	protected void checkDaoConfig() throws IllegalArgumentException {
		// TODO zaimplementować sprawdzenie checkDaoConfig()

	}

	public Table getTableInfo() {
		return this.tableInfo;
	}

	public RowValues fetchValues(Obj entityObject) throws SQLException {
		return this.tableInfo.fetchValues(entityObject);
	}

	/**
	 * Wyzwalacz, akcja uruchamiana tuż przed wstawieniem wiersza do bazy
	 * danych.
	 * 
	 * @param entityObject
	 * @return
	 * @throws SQLException
	 */
	protected Obj beforeInsert(Obj entityObject) throws SQLException {
		return entityObject;
	}

	/**
	 * Wyzwalacz, akcja uruchamiana tuż przed aktualizacją wiersza w bazie
	 * danych.
	 * 
	 * @param entityObject
	 * @return
	 * @throws SQLException
	 */
	protected Obj beforeUpdate(Obj entityObject) throws SQLException {
		return entityObject;
	}

	/**
	 * Wyzwalacz, akcja uruchamiana tuż przed usunięciem wiersza z bazy danych.
	 * 
	 * @param entityObject
	 * @return
	 * @throws SQLException
	 */
	protected Obj beforeRemove(Obj entityObject) throws SQLException {
		return entityObject;
	}

	/**
	 * Wyzwalacz, akcja uruchamiana po wstawieniu wiersza do bazy danych,
	 * 
	 * @param entityObject
	 * @throws SQLException
	 */
	protected void afterInsert(Obj entityObject) throws SQLException {
	}

	/**
	 * Wyzwalacz, akcja uruchamiana po aktualizacji wiersza w bazie danych,
	 * 
	 * @param entityObject
	 * @throws SQLException
	 */
	protected void afterUpdate(Obj entityObject) throws SQLException {
	}

	/**
	 * Wyzwalacz, akcja uruchamiana po usunięciu wiersza z bazy danych,
	 * 
	 * @param entityObject
	 * @throws SQLException
	 */
	protected void afterRemove(Obj entityObject) throws SQLException {
	}

	public ScEntityManager getEntityManager() {
		if (this.aem == null || !this.aem.isOpen()) {
			this.aem = (ScEntityManager) this.aof.getEntityManager();
		}
		return aem;
	}

	public void setEntityManager(ScEntityManager aem) {
		this.aem = aem;
	}

	@Required
	public void setScObjectFactory(ScObjectFactory aof) {
		this.aof = aof;
	}

	public void setApplicationContext(ApplicationContext arg0)
			throws BeansException {
		this.actx = arg0;
	}

	public Object postProcessAfterInitialization(Object bean, String arg1)
			throws BeansException {

		ScEntityManager em = getEntityManager();
		if (!inicjalized && em.isAllowedConfiguration()
				&& em.isAllowedMapping()) {
			if (logger.isDebugEnabled())
				logger.debug("[" + getClass().getSimpleName()
						+ "] INICJALIZACJA...");

			try {
				this.tableInfo = new Table(em, persistentClass, actx, logger);
			} catch (SecurityException e) {
				throw new JpaDaoBeanInitializationException(
						"SecurityException", e);
			} catch (InstantiationException e) {
				throw new JpaDaoBeanInitializationException(
						"InstantiationException", e);
			} catch (IllegalAccessException e) {
				throw new JpaDaoBeanInitializationException(
						"IllegalAccessException", e);
			} catch (ClassNotFoundException e) {
				throw new JpaDaoBeanInitializationException(
						"ClassNotFoundException", e);
			} catch (NoSuchMethodException e) {
				throw new JpaDaoBeanInitializationException(
						"NoSuchMethodException", e);
			}

			inicjalized = true;
		}
		if (!isAllowedTableInfo()) {
			logger.warn("Table information is not allowed.");
		} else if (logger.isDebugEnabled()) {
			logger.debug("Table information is allowed.");
		}
		return bean;
	}

	public Object postProcessBeforeInitialization(Object bean, String arg1)
			throws BeansException {
		return bean;
	}

	/**
	 * @return the aem
	 */
	public ScEntityManager getScEntityManager() {
		return aem;
	}

}
